#!/usr/local/bin/perl
##############################################################################
# Export Mascot search results                                               #
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 1998-2014 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################
#    $Archive:: /www/cgi/export_dat.pl                                     $ #
#     $Author: villek $ #
#       $Date: 2014/11/11 11:52:40 $ #
#   $Revision: 1.152 $ #
# $NoKeywords::                                                            $ #
##############################################################################
 use strict;                                                                #
 use warnings;
##############################################################################

# The following (case insensitive) URL arguments are recognised:
#
# file                      - relative path to result file (required)
# sessionid                 - Mascot security sessionID
# do_export                 - 1 to export results, 0 to display form, -1 to display command line arguments
# generate_file             - 1 (and do_export=1) to never redirect if cache files need to be created
# generate_mime             - 1 to export all messages and data in multipart MIME format
# export_format             - "XML" or "CSV" or "mzIdentML" or "mzTab" or "pepXML" or "DTASelect" or MascotDAT or MGF or ...
# _sigthreshold             - probability significance threshold, default 0.05
# report                    - max number of hits to be reported, (0 = AUTO)
# _server_mudpit_switch     - if queries / entries greater than this value, switch to Mudpit scoring, default 0.001
# _ignoreionsscorebelow     - cut off on MS/MS ions scores
# show_same_sets            - 1 to display all proteins that match the same set of peptides, otherwise just one per hit
# _showsubsets              - display protein hits that are missing up to this fraction of the protein score of the main hit
# _requireboldred           - 1 to report protein hits only if they include at least one bold, red peptide match
# group_family              - 1 to group into protein families
# use_homology              - 1 to use homology threshold, 0 to use identity threshold 
# unigene                   - UniGene index species to be used to cluster hits
# _showallfromerrortolerant - 1 to display all hits from an error tolerant search, including garbage (no UI)
# _onlyerrortolerant        - 1 to display only error tolerant matches from a new style error tolerant search
# _noerrortolerant          - 1 to display no error tolerant matches from a new style error tolerant search
# _show_decoy_report        - 1 to display report for the integrated decoy search (Mascot 2.2. & later)
# percolate                 - 1 to display scores and expect values based on Percolator PEPs  
# percolate_rt              - 1 to include retention time feature when running Percolator
# _prefertaxonomy           - taxonomy ID > 0; proteins in samesets with a taxonomy that matches _prefertaxonomy will be preferred over the others
# _target_fdr               - target FDR; if set, this adjusts _sigthreshold (at identity or homology level as specified by use_homology)
# search_master             - Over-ride toggle for show_header, show_decoy, show_mods, show_params, show_format, show_masses, show_quant_header
# show_header               - 1 to include a block containing search level information
# show_decoy                - 1 to include a block containing decoy search information
# show_mods                 - 1 to include a block containing variable modifications information
# show_params               - 1 to include a block containing search parameters
# show_format               - 1 to include a block containing format parameters
# show_masses               - 1 to include a block containing residue and element masses
# show_unassigned           - 1 to include a block containing unassigned matches
# show_header               - 1 to include a block containing quantitation level information
# protein_master            - Over-ride toggle for all protein info
# peptide_master            - Over-ride toggle for all peptide info
# show_pep_dupes            - Over-ride toggle for exporting duplicate peptides
# query_master              - Over-ride toggle for all query info

#
# If the following (case insensitive) URL arguments are set to 1, the corresponding fields 
# are included in the report unless over-ridden by "master" checkbox
#
  my @columns;
  push @columns, "prot_hit_num";  # always forced to 1
  push @columns, "prot_family_member";  # slaved off prot_hit_num if group_family true
  push @columns, "prot_acc";      # always forced to 1
  push @columns, "prot_desc";
  push @columns, "prot_score";
  push @columns, "prot_thresh";
  push @columns, "prot_expect";
  push @columns, "prot_mass";
  push @columns, "prot_matches";
  push @columns, "prot_matches_sig";    # slaved off prot_matches
  push @columns, "prot_sequences";      # slaved off prot_matches
  push @columns, "prot_sequences_sig";  # slaved off prot_matches
  push @columns, "prot_cover";
  push @columns, "prot_len";
  push @columns, "prot_pi";
  push @columns, "prot_tax_str";
  push @columns, "prot_tax_id";
  push @columns, "prot_seq";
# prot_empai (no entry in header row)
# prot_quant (no entry in header row)
  push @columns, "pep_query";     # always forced to 1
  push @columns, "pep_rank";      # always forced to 1
  push @columns, "pep_isbold";    # always forced to 1
  push @columns, "pep_isunique";  # always forced to 1
  push @columns, "pep_exp_mz";    # always forced to 1
  push @columns, "pep_exp_mr";
  push @columns, "pep_exp_z";
  push @columns, "pep_calc_mr";
  push @columns, "pep_delta";
  push @columns, "pep_start";
  push @columns, "pep_end";
  push @columns, "pep_miss";
  push @columns, "pep_score";
  push @columns, "pep_homol";
  push @columns, "pep_ident";
  push @columns, "pep_expect";
  push @columns, "pep_res_before";  # slaved off pep_seq
  push @columns, "pep_seq";
  push @columns, "pep_res_after";  # slaved off pep_seq
  push @columns, "pep_frame";
  push @columns, "pep_var_mod";
  push @columns, "pep_var_mod_pos";  # slaved off pep_var_mod
  push @columns, "pep_summed_mod_pos";  # slaved off pep_var_mod
  push @columns, "pep_local_mod_pos";  # slaved off pep_var_mod
  push @columns, "pep_var_mod_conf";  # slaved off pep_var_mod
  push @columns, "pep_num_match";
  push @columns, "pep_scan_title";
# pep_quant (no entry in header row)

# 'Global' variables
  my(
    $anyPepSumMatch,           # $objResFile->anyPeptideSummaryMatches($sec_peptides)
    $delimiter,                # delimiter for CSV is "," by default
    $export_format,            # "XML" or "CSV", etc.
    @fileIn,                   # relative paths to result files, passed as required URL argument (file or fileset)
    $idForFatalError,          # $fileIn or results collection set ID for error reporting
    $fileNameRoot,             # root of result file name, e.g. F123456
    $group_family,             # 1 to group into protein families
    $ignoreIonsScoreBelow,     # cut off on MS/MS ions scores
    $mass_type,                # MONO or AVE
    $massDP,                   # number of decimal places in formatted mass values
    $minPepLen,                # peptides shorter than $minPepLen are ignored when generating Peptide summary 
    $msresFlags,               # flags used to create Mascot Parser summary object
    $msresFlags2,              # flags2 used to create Mascot Parser summary object
    $mudpitSwitch,             # if queries / entries greater than this value, switch to Mudpit scoring, default 0.001
    $myName,                   # file name of this script for links to self (not complete path)
    $NoErrorTolerant,          # 1 to display only standard matches from a new style error tolerant search
    $numHits,                  # max number of hits to be reported
    $objDatFile,               # Mascot parser mascot.dat file object
    $objMascotOptions,         # Mascot parser configuration options object
    $objMassesFile,            # Mascot parser masses file object
    $objModFile,               # Mascot parser mod_file object
    $objParams,                # Mascot parser search parameters object
    $objProtein,               # Mascot parser protein object
    $objQuantFile,             # Mascot parser ms_quant_configfile object
    $objQuantMethod,           # Mascot parser quant_method object
    $quantProtocolType,        # REPORTER, MULTIPLEX, PRECURSOR
    $objResFile,               # Mascot parser result file object
    $objSummary,               # Mascot parser result summary object (may be protein or peptide)
    $objMS2Quant,              # Mascot parser quantitation data object (or undef if file is not Reporter or Multiplex)
    $objUmodConfigFile,        # Mascot parser Unimod configuration file object
    $objUmodConfigFile2,       # For finding the IDs from PSI-MS-MOD names. When record_id is not in the .dat file, go to unimod.xml
    $objUnigeneOptions,        # Mascot parser configuration unigene object
    $OldStyleErrTolReport,     # 1 to suppress certain parts of the report as per old-style error tolerant search
    $OnlyErrorTolerant,        # 1 to display only error tolerant matches from a new style error tolerant search
    $percolate,                # 1 to display scores and expect values based on Percolator PEPs
    $percolate_rt,             # 1 to include retention time feature when running Percolator
    $preferTaxonomy,           # index > 0 that controls which protein to show at the top of a sameset
    $protColumnCount,          # Used to align peptide match data in CSV output 
    $queryColumnCount,         # Used to align peptide match data in CSV output 
    $reportType,               # 'peptide' or 'concise'
    $RequireBoldRed,           # 1 to report protein hits only if they include at least one bold, red peptide match
    $sec_peptides,             # SEC_PEPTIDES or SEC_DECOYPEPTIDES 
    $sec_summary,              # SEC_SUMMARY or SEC_DECOYSUMMARY 
    $session,                  # Mascot security session object
    $sessionID,                # Mascot security sessionID
    $shipper,                  # flag set to 'TRUE' if this is Mascot for intranet, 'FALSE' for public internet site
    $ShowAllFromErrorTolerant, # 1 to display all hits from an error tolerant search, including garbage
    $ShowDecoyReport,          # 1 to display report for the integrated decoy search (Mascot 2.2. & later)
    $ShowSameSets,             # 1 to display same set proteins
    $ShowSubSets,              # display protein hits that are missing up to this fraction of the protein score of the main hit
    $sigThreshold,             # probability significance threshold (OneInXprobRnd)
    $targetFDR,                # target FDR or unset; if set, adjusts $sigThreshold
    $thisScript,               # CGI object
    $isCGI,                    # running under CGI?
    $tmpChecked,               # temp variable used to toggle state of CHECKED
    $UniGeneFile,              # path to a UniGene index file
    $objUniGene,               # ms_unigene object if UniGene is in use
    $URI,                      # URL as far as Mascot root directory, if executed as CGI, otherwise "../"
    $use_homology,             # 1 to use homology threshold, 0 to use identity threshold
    $use_resfile_cache,        # if given and 0, don't use resfile cache file
    $use_results_cache,        # if given and 0, don't use results cache file
    $mascotresultsNeedsPercolating, # 1 if Percolator files potentially need to be created
    $fastaLen,                 # sequence length cache (both for emPAI and mustGetProteinMass)
    %elementalMasses,          # mass values for oxygen, carbon, ...
    %groupMasses,              # mass values for residues and terminii, including possible fixed modifications (group means delta + absolute)
    %taxNames,                 # cache for looking up taxonomy name string from tax ID
    %urlParams,                # keys are URL parameter names in lower case, values are names in original case
    %vmMass,                   # variable modification mass
    %vmString,                 # variable modification name
    @quantWarnings,            # contains the HTML text for any warnings returned by quant routines
    @query_title_list,         # list of query titles, HTML escaped, index is query number
    );

# set current directory to script location for FastTrack on NT
# chdir also handles a Windows drive letter change
  if ($ENV{'WINDIR'} && $0 =~ /(.*)[\\\/]/) {
    chdir($1);
  }

# delimiter for CSV can be changed
  $delimiter = ",";

 # required Perl packages
  use CGI;
  use CGI::Carp qw(fatalsToBrowser);
  use URI;
  use HTTP::Request::Common;
  use LWP::UserAgent;
  use Cwd;
  use Scalar::Util qw(looks_like_number);
  use lib "../bin";
  use msparser;
  use lib "../bin/modules";
  use MascotDat;
  use MascotResults;
  use Util::IPC qw(pipefork);
  use CachedResults::Build::Util qw(run_ms_createpip run_percolator);
  use PeptideSummary::Util qw(key_acc key_acc_str);
  use PeptideSummary::Backend qw(duplicate_group_main sort_duplicate_group);
  use MasterResults::ProteinFamily::ReportFunc qw(taxonomy_choice);
  use Util::Peptides qw(lookup_query_level_mods);
  use Util::SiteAnalysis;
  use Util::Text qw(joinand);

  $| = 1; # set autoflush on STDOUT

# create CGI object
  $thisScript = new CGI;
 
# pull in some common subroutines
  require "./common_subs.pl";
  require "./quant_subs.pl";
  
# www.pl contains the code to decrypt result file names
# its existence indicates that this script is running on the public web site
  if (-e "./www.pl") {
    do "./www.pl";
    $shipper = 'FALSE';
  } else {
    $shipper = 'TRUE';
  }

# get the name of this script for creating URL links to self
#  $myName = $thisScript->url(-relative=>1);
  ($myName) = $0 =~ /([^\\\/]*)$/;

# taint mode requires path to be set explicitly
# Unix path may be required to find decompression utilities in &common_subs::decompress()
  if ($ENV{'WINDIR'}){
    $ENV{"PATH"} = "";
  } else {
    $ENV{"PATH"} = "/usr/local/sbin:/usr/sbin:/sbin:/usr/local/bin:/usr/bin:/bin";
  }
# to bypass taint mode check on the path, comment out above lines and uncomment following line
# ($ENV{"PATH"}) = $ENV{"PATH"} =~ /(.*)/;

# create hash for URL arguments, so that we can always use lower case keys
  foreach ($thisScript->param) {
    if ($urlParams{lc($_)} && $urlParams{lc($_)} ne $_) {
      die("URL parameter names must use consistent case");
    }
    $urlParams{lc($_)} = $_;
  }

  $isCGI = 0;

  if (defined($ENV{'REMOTE_ADDRESS'}) or defined($ENV{'GATEWAY_INTERFACE'})
          or defined($ENV{'SERVER_NAME'})) {
    $isCGI = 1;
  }

# some protein options not permitted on public web site
  if ($shipper eq 'FALSE') {
    $thisScript->param(-name=>$urlParams{'prot_cover'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_len'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_pi'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_tax_str'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_tax_id'}, -value=>'');
    $thisScript->param(-name=>$urlParams{'prot_seq'}, -value=>'');
  }

# some options are not optional!
  &set_option('prot_hit_num');
  &set_option('prot_acc');
  &set_option('pep_query');
  &set_option('pep_rank');
  &set_option('pep_isbold');
  &set_option('pep_isunique');
  &set_option('pep_exp_mz');

# save export_format in variable to make script more readable
  $export_format = "XML" unless ($export_format = $thisScript->param($urlParams{'export_format'}));

# verify the result file exists, decompress if necessary
  if (my $fileset = $thisScript->param($urlParams{'fileset'})) {
    @fileIn = common_subs::filesetFilePaths($fileset);
  }

  if (@fileIn == 0) {
    my $fileIn = $thisScript->param($urlParams{'file'})    || common_subs::fatal("no file name", __LINE__, $myName, "");
    @fileIn = ($fileIn);
    $idForFatalError = $fileIn;
  } else {
    $idForFatalError = $thisScript->param($urlParams{'fileset'});
  }

# create a datfile file object
  $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objDatFile->isValid) {
    die common_subs::checkErrorHandler($objDatFile);
  }
  $objMascotOptions = $objDatFile->getMascotOptions;
  $objUnigeneOptions = $objDatFile->getUniGeneOptions;

# untaint the file path. This regex may need fine tuning
  for (@fileIn) {
    ($_) = $_ =~ /([\/\w\-.$%&()[\]:]+)/;
    if (not $_) {
      common_subs::fatal('result filename could not be untainted', __LINE__, $myName, $idForFatalError);
    }
    if (not -e common_subs::decompress($_)) {
      # could be master_results_2.pl style filepath
      my $root = $objMascotOptions->getInterFileRelPath;
      my $filepath = join('/', $root, $_);
      if (-e common_subs::decompress($filepath)) {
        $_ = $filepath;
      } else {
        die("result file does not exist ", $_);
      }
    }
  }
  
  if (defined($thisScript->param($urlParams{'_usecache'}))) {
    $use_resfile_cache = $use_results_cache = $thisScript->param($urlParams{'_usecache'}) + 0;
  } else {
    $use_resfile_cache = $use_results_cache = 1;
    if ($export_format eq 'mzIdentML') {
      # mzIdentML requires lots of query level data, which means the resfile
      # cache creates a large I/O bottleneck. mzIdentML can also require lots
      # of protein level data, which means the results cache can also become
      # a bottleneck. However, we should keep using the cache file until
      # we are actually about to print out the data.
      if ($isCGI) {
        if (($thisScript->param($urlParams{'do_export'}) || 0) == 0) {
          # use caches
        } elsif ($thisScript->param($urlParams{'do_export'}) == -1) {
          # use caches
        } elsif (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
                and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
          # use caches
        } else {
          $use_resfile_cache = 0;
          $use_results_cache = 0;
        }
      } else {
        # Speed up command line.
        $use_resfile_cache = 0;
        $use_results_cache = 0;
      }
    } else {
      # use caches
    }
  }

# see whether to use ms_mascotresfile::RESFILE_USE_CACHE flag
  my $resfileFlags;
  if (common_subs::useResfileCache($objMascotOptions, $myName)) {
    $resfileFlags = $msparser::ms_mascotresfile::RESFILE_USE_CACHE;
  } else {
    $resfileFlags = $msparser::ms_mascotresfile::RESFILE_NOFLAG;
  }
  if (not $use_resfile_cache) {
    $resfileFlags &= ~$msparser::ms_mascotresfile::RESFILE_USE_CACHE;
  }
  my $resfileCacheDirectory = $objMascotOptions->getCacheDirectory;

  if ($resfileFlags & $msparser::ms_mascotresfile::RESFILE_USE_CACHE) {
    for (@fileIn) {
      if ((msparser::ms_mascotresfile::willCreateCache(
          common_subs::decompress($_), 
          $resfileFlags, 
          $resfileCacheDirectory
      ))[0]) {
        maybeRedirectToCaching('build:resultsfile');
      }
    }
  }

  if (common_subs::useResultsCache($objMascotOptions, $myName)) {
    $msresFlags2 = $msparser::ms_peptidesummary::MSPEPSUM_USE_CACHE;
  } else {
    $msresFlags2 = $msparser::ms_peptidesummary::MSPEPSUM_NONE;
  }
  if (not $use_results_cache) {
    $msresFlags2 &= ~$msparser::ms_peptidesummary::MSPEPSUM_USE_CACHE;
  }
  
# Choose whether to use identity or homology threshold
  $use_homology = 0;
  if (defined($thisScript->param($urlParams{'use_homology'}))) {
    $use_homology = $thisScript->param($urlParams{'use_homology'}) + 0;
  }  
  if ($use_homology) {
    $msresFlags2 = $msresFlags2 | $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH;
  }  

  $preferTaxonomy = 0;
  if (defined(my $p = $thisScript->param($urlParams{'_prefertaxonomy'}))) {
    if (looks_like_number($p) and $p > 0) {
      $preferTaxonomy = int($p);
    }
  }

  my $MIME_boundary = undef;

  if ($thisScript->param($urlParams{'generate_mime'})) {
    # We use custom functions to print MIME boundaries. The CGI module
    # provides only "server push" support.

    $MIME_boundary = common_subs::multipart_init(); # sets binmode()
    common_subs::multipart_start($MIME_boundary, 'text/plain; charset=utf-8', 'keepalive');
  }

# create Mascot Parser result file object
  $objResFile = common_subs::openResultsCollectionSet(
    $objMascotOptions,
    $resfileFlags,
    $resfileCacheDirectory,
    $idForFatalError,
    defined $MIME_boundary ? 1 : 0, # keepalive?
    \@fileIn,
    $MIME_boundary,
  );

# abort on fatal error. Note that isValid() remains true if there are only warnings
  unless ($objResFile->isValid) {
    my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die($errorList, " ", $idForFatalError);
  }
  
# create a Mascot Parser searchparams object
  $objParams = new msparser::ms_searchparams($objResFile);
  
# Security
  $sessionID = "" unless ($sessionID = $thisScript->param($urlParams{sessionid}));
  $session = new msparser::ms_session($sessionID);
  if ($session->isValid) {
    unless ($session->canResultsFileBeViewed($objParams->getUSERID)) {
      if (defined $MIME_boundary) {
        common_subs::multipart_end($MIME_boundary);
        common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
      } else {
        print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
      }
      print common_subs::notAuthorised($session->getUserName);
      if (defined $MIME_boundary) {
        common_subs::multipart_final($MIME_boundary);
      }
      exit 1;
    }
    $sessionID = $session->getID();
  } else {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
    } else {
      print $thisScript->header(-type => 'text/html', -charset => 'utf-8');
    }
    &common_subs::printHeader1("Matrix Science - Not logged in");
    &common_subs::printHeader2("", "Not logged in", 1);
    &common_subs::go2login($thisScript->self_url, $session->getLastError, $session->getLastErrorString);
    &common_subs::printFooter;
    print "</BODY>\n";
    print "</HTML>\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 1;
  }

# validate a variety of mascot.dat and URL parameters
  &validateParams();

  if (@fileIn > 1) {
    if ($export_format eq 'CSV' or $export_format eq 'mzIdentML') {
      # OK
    } else {
      common_subs::fatal("export format $export_format is not supported for merged results", __LINE__, $myName, $idForFatalError, $MIME_boundary);
    }
  }

  # Required for redirecting support in maybeRedirectToCaching().
  set_option('report', $numHits);

# fixed settings for DTASelect and pepXML
  if ($export_format eq "DTASelect" or $export_format eq "pepXML") {
    $ShowSubSets = 1;
    $RequireBoldRed = 0;
    $group_family = 0;
    # Required for redirecting support in maybeRedirectToCaching().
    set_option('_showsubsets', $ShowSubSets);
    set_option('_requireboldred', $RequireBoldRed);
    set_option('group_family', $group_family);
  }

  if ($group_family) {
    $RequireBoldRed = 0;
    set_option('_requireboldred', $RequireBoldRed);
  }

# We need to check if the file has to be cached, in order to write out
# the correct form. So, we construct the flags here as they would be used
# when creating ms_mascotresfile and ms_proteinsummary/ms_peptidesummary.
  if ($reportType eq 'concise') {
    $msresFlags = $msparser::ms_mascotresults::MSRES_MAXHITS_OVERRIDES_MINPROB
                | $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS 
                | $msparser::ms_mascotresults::MSRES_SHOW_SUBSETS;
    if (!$objResFile->isErrorTolerant()
      && $objParams->getDECOY() > 0
      && $ShowDecoyReport) {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_DECOY;
    }
  } else {
    $msresFlags = $msparser::ms_mascotresults::MSRES_MAXHITS_OVERRIDES_MINPROB;
    if ($ShowSubSets > 0) {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_SHOW_SUBSETS;
    }
    if ($group_family) {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_CLUSTER_PROTEINS;
    } else {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS;
      if ($RequireBoldRed) {
        $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_REQUIRE_BOLD_RED;
      }
    }
    if ($objResFile->isErrorTolerant()) {
      if ($objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      # new-style integrated ET search
        if ($OnlyErrorTolerant) {
          $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_ERR_TOL;
        } elsif ($NoErrorTolerant) {
        # parser default
        } else {
          $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_INTEGRATED_ERR_TOL;
        }
      }
      if ($ShowAllFromErrorTolerant) {
        $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_SHOW_ALL_FROM_ERR_TOL;
      }
    }
    if ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch) {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_MUDPIT_PROTEIN_SCORE;
    }
    if (!$objResFile->isErrorTolerant()
      && $objParams->getDECOY() > 0
      && $ShowDecoyReport) {
      $msresFlags = $msresFlags | $msparser::ms_mascotresults::MSRES_DECOY;
    }
    if (MascotResults::can_percolate($objResFile)) {
      if ($percolate == 1) {
        $msresFlags2 = $msresFlags2 | $msparser::ms_peptidesummary::MSPEPSUM_PERCOLATOR;
        $objResFile->setPercolatorFeatures(
            $objMascotOptions->getPercolatorFeatures,
            '',
            $percolate_rt,
        );
        if (MascotResults::no_percolator_files($objResFile)) {
          $mascotresultsNeedsPercolating = 1;
        }
      }
    }
  }

  if ($export_format eq 'DTASelect') {
    $ShowSameSets = $msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS;
    # Required for redirecting support in maybeRedirectToCaching().
    set_option('show_same_sets', $ShowSameSets);
  }
  
# branch here if we require the form
  if (($thisScript->param($urlParams{'do_export'}) || 0) == 0) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
    } else {
      print $thisScript->header(-type => 'text/html', -charset => "utf-8", -expires => '-1d' );
    }
    &formatControls();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  } elsif ($thisScript->param($urlParams{'do_export'}) == -1) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/html; charset=utf-8');
    } else {
      print $thisScript->header(-type => 'text/html', -charset => "utf-8", -expires => '-1d' );
    }
    &display_args();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;
  }

  if ($export_format eq "MascotDAT") {
  # Branch off to download raw Mascot result file
  # (we know that @fileIn == 1)
    exit mascotDAT($thisScript, $isCGI, $fileIn[0], $objMascotOptions, $MIME_boundary);
  }
  
  if ($export_format eq "MGF") {
  # Branch off to download MGF peak list
    exit exportMGF($thisScript, $MIME_boundary);
  }

# pepXML and DTASelect only valid for MS/MS
  if ($export_format eq "pepXML" && $reportType eq 'concise') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print "pepXML format is not valid for peptide mass fingerprint results ", $idForFatalError;
      common_subs::multipart_final($MIME_boundary);
    }
    die("pepXML format is not valid for peptide mass fingerprint results ", $idForFatalError);
  }
  if ($export_format eq "DTASelect" && $reportType eq 'concise') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print "DTASelect format is not valid for peptide mass fingerprint results ", $idForFatalError;
      common_subs::multipart_final($MIME_boundary);
    }
    die("DTASelect format is not valid for peptide mass fingerprint results ", $idForFatalError);
  }

# $objQuantMethod will encapsulate any quantitation method
# report can only handle multiplex and reporter protocols
# but may need $objQuantMethod for (say) mods and isotope shifts
  if ($objParams->getQUANTITATION() 
      and lc($objParams->getQUANTITATION()) ne "none") {
    my $method = msparser::ms_quant_method->new();
    if ($objResFile->getQuantitationMethod($method)) {
      $quantProtocolType = uc($method->getProtocol->getType());
      if ($quantProtocolType eq "MULTIPLEX" or $quantProtocolType eq "REPORTER") {
        $objQuantMethod = $method;
        # default report settings to those in the quantitation method
        quant_subs::quantFormatOverRide($objQuantMethod, $thisScript, \%urlParams, \$mudpitSwitch, \$sigThreshold, \$ShowSubSets, \$RequireBoldRed);
        # read any URL arguments and create text for format controls 
        # text is not used, but need to call the routine to set or over-ride defaults
        quant_subs::quantFormatControls($objQuantMethod, $thisScript, \%urlParams, []);
        # warn if trying to use unsupported features
        quant_subs::quantWarnings($objQuantMethod, \@quantWarnings);
      }
    }
  }

  $quantProtocolType = '' if not defined $quantProtocolType;

  if (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
          and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
    if ($reportType ne 'concise' and $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_CACHE) {
      if (msparser::ms_peptidesummary::willCreateCache(
            $objResFile, 
            $msresFlags,
            $sigThreshold, 
            $numHits, 
            $UniGeneFile,
            $ignoreIonsScoreBelow,
            $minPepLen,
            "",
            $msresFlags2
      )) {
        maybeRedirectToCaching('build:peptidesummary');
      }
    }
    if ($reportType ne 'concise' and $mascotresultsNeedsPercolating) {
      maybeRedirectToCaching('build:peptidesummary');
      # If the above function returns, it means we need to percolate right here
      # and now.
      _percolate($objResFile, 0, $percolate_rt, $MIME_boundary);
    }
    if (not $targetFDR) {
      # We only show download controls here if target FDR is not specified.
      # Otherwise download controls are shown after we've opened the results
      # file.
      if ($isCGI) {
        print $thisScript->header( -charset => "utf-8", -expires => '-1d' );
        downloadControls();
        exit 0;
      }
    }
  } else {
    if ($reportType ne 'concise' and $mascotresultsNeedsPercolating) {
      _percolate($objResFile, 0, $percolate_rt, $MIME_boundary);
    }
  }
  
# if this is an old-style error tolerant search, may need to decompress the parent
  if ($objResFile->isErrorTolerant()) {
    if ($objParams->getErrTolParentFilename()) {
      &common_subs::decompress($objParams->getErrTolParentFilename());
    }
  # set value for $OldStyleErrTolReport
    if ($ShowAllFromErrorTolerant || $OnlyErrorTolerant) {
      $OldStyleErrTolReport = 1;
    } elsif (!$objResFile->doesSectionExist($msparser::ms_mascotresfile::SEC_ERRTOLSUMMARY)) {
      $OldStyleErrTolReport = 1;
    } else {
      $OldStyleErrTolReport = 0;
    }
  } else {
    $OldStyleErrTolReport = 0;
  }

# create Mascot Parser summary object of the required type
  if ($reportType eq 'concise') {
    $objSummary = new msparser::ms_proteinsummary(
      $objResFile, 
      $msresFlags, 
      $sigThreshold,
      $numHits
    );
  } else {
    $objSummary = new msparser::ms_peptidesummary(
      $objResFile, 
      $msresFlags,
      $sigThreshold, 
      $numHits, 
      $UniGeneFile,
      $ignoreIonsScoreBelow,
      $minPepLen,
      "",
      $msresFlags2
    );
    $objSummary->setSubsetsThreshold($ShowSubSets);
  }
  
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objResFile->isValid) {
    my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die($errorList, " ", $idForFatalError);
  }

# We have to wait until this point to deal with target FDR requests, because
# the ms_mascotresults object needs to exist before we can query it for
# target FDRs. It may be that we need to redirect again if a new cache file
# has to be created.
  if ($targetFDR) {
    my ($success, $closestFDR, $minProb);
    if ($use_homology) {
      ($success, $closestFDR, $minProb) = $objSummary->getThresholdForFDRAboveHomology($targetFDR/100);
    } else {
      ($success, $closestFDR, $minProb) = $objSummary->getThresholdForFDRAboveIdentity($targetFDR/100);
    }
    my $sigThreshold_adjusted;
    if ($success) {
      $minProb = sprintf("%.5g", $minProb); # same number of significant digits as in master_results_2.pl
      if ($minProb < 1e-17) {
          $minProb = 1e-17;
      }
      if ($minProb != $sigThreshold) {
        $sigThreshold_adjusted = $minProb;
      }
    }
    if (defined $sigThreshold_adjusted) {
      if (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
              and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
        if ($reportType ne 'concise' and $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_CACHE) {
          if (msparser::ms_peptidesummary::willCreateCache(
              $objResFile, 
              $msresFlags,
              $sigThreshold_adjusted,
              $numHits, 
              $UniGeneFile,
              $ignoreIonsScoreBelow,
              $minPepLen,
              "",
              $msresFlags2
          )) {
            # We need to pass the new value to nph-cache_families.pl to ensure
            # the right cache file is created.
            maybeRedirectToCaching('build:peptidesummary', _sigthreshold => $sigThreshold_adjusted);
          }
        }
        # No need to check for percolation here.
  
        if ($isCGI) {
          print $thisScript->header( -charset => "utf-8", -expires => '-1d' );
          downloadControls($success, $closestFDR, $sigThreshold_adjusted);
          exit 0;
        }
      }

      $sigThreshold = $sigThreshold_adjusted;

      # create Mascot Parser summary object of the required type
      if ($reportType eq 'concise') {
        $objSummary = new msparser::ms_proteinsummary(
          $objResFile, 
          $msresFlags, 
          $sigThreshold,
          $numHits
        );
      } else {
        $objSummary = new msparser::ms_peptidesummary(
          $objResFile, 
          $msresFlags,
          $sigThreshold, 
          $numHits, 
          $UniGeneFile,
          $ignoreIonsScoreBelow,
          $minPepLen,
          "",
          $msresFlags2
        );
        $objSummary->setSubsetsThreshold($ShowSubSets);
      }

      unless ($objResFile->isValid) {
        my (undef, $errorList) = common_subs::checkResfileErrors($objResFile);
        if (defined $MIME_boundary) {
          common_subs::multipart_end($MIME_boundary);
          common_subs::multipart_start($MIME_boundary);
          print $errorList;
          common_subs::multipart_final($MIME_boundary);
        }
        die($errorList, " ", $idForFatalError);
      }
    } else { # success or not
      if (($thisScript->param($urlParams{'generate_file'}) || 0) != 1
              and ($thisScript->param($urlParams{'generate_mime'}) || 0) != 1) {
        # We may need to show the download controls here, as we skipped them
        # before trying to reach the target FDR.
        if ($isCGI) {
          print $thisScript->header( -charset => "utf-8", -expires => '-1d' );
          downloadControls($success, $closestFDR, $minProb);
          exit 0;
        }
      }
    }
  }

  if ($objQuantMethod
      and ($thisScript->param($urlParams{'pep_quant'}) or $thisScript->param($urlParams{'prot_quant'}))) {
    $objMS2Quant = msparser::ms_ms2quantitation->new($objSummary, $objQuantMethod);

    if (not $objMS2Quant->isValid) {
      my $errorList = common_subs::checkErrorHandler($objMS2Quant);
      if (defined $MIME_boundary) {
        common_subs::multipart_end($MIME_boundary);
        common_subs::multipart_start($MIME_boundary);
        print $errorList;
        common_subs::multipart_final($MIME_boundary);
      }
      die $errorList;
    }

    if ($objMS2Quant->hasIntensityNormalisation) {
      $objMS2Quant->normaliseIntensities();
    }
    if ($objMS2Quant->hasPeptideRatioNormalisation) {
      $objMS2Quant->normalisePeptideRatios();
    }

    $objQuantMethod = $objMS2Quant->getQuantitationMethod();
  }

# create an ms_umod_configfile object from the result file unimod section
  $objUmodConfigFile = new msparser::ms_umod_configfile;
  unless ($objResFile->getUnimod($objUmodConfigFile)) {
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($sessionID);
  # old results file with no unimod section, use current unimod.xml
    $objUmodConfigFile = new msparser::ms_umod_configfile("../config/unimod.xml",
      "../html/xmlns/schema/unimod_2/unimod_2.xsd", $cs);
  }
        
# add in any modifications locally defined in any quant method
  if ($objQuantMethod) {  
    for (my $i = 0; $i < $objQuantMethod->getNumberOfModificationGroups(); $i++) {
      my $objModGroup = $objQuantMethod->getModificationGroupByNumber($i);
      for (my $j = 0; $j < $objModGroup->getNumberOfLocalDefinitions(); $j++) {
        quant_subs::includeLocalDef($objModGroup->getLocalDefinition($j), $objUmodConfigFile);
      }
    }
    for (my $i = 0; $i < $objQuantMethod->getNumberOfComponents(); $i++) {
      my $objComponent = $objQuantMethod->getComponentByNumber($i);
      for (my $j = 0; $j < $objComponent->getNumberOfModificationGroups(); $j++) {
        my $objModGroup = $objComponent->getModificationGroupByNumber($j);
        for (my $k = 0; $k < $objModGroup->getNumberOfLocalDefinitions(); $k++) {
          quant_subs::includeLocalDef($objModGroup->getLocalDefinition($k), $objUmodConfigFile);
        }
      }
    }
  }

# use this to create an ms_masses object
  $objMassesFile = new msparser::ms_masses($objUmodConfigFile); 
  
# create an ms_modfile object from the ms_umod_configfile object
# this now has all the mods used in the search
  $objModFile = new msparser::ms_modfile($objUmodConfigFile, $msparser::ms_umod_configfile::MODFILE_FLAGS_ALL);
# abort on fatal error. Note that isValid() is true if there are only warnings
  unless ($objModFile->isValid) {
    my $errorList = common_subs::checkErrorHandler($objModFile);
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $errorList;
      common_subs::multipart_final($MIME_boundary);
    }
    die $errorList;
  }
    
# msparser MASS_TYPE
  $mass_type = $msparser::MASS_TYPE_MONO;
  if (lc($objParams->getMASS()) eq "average") {
    $mass_type = $msparser::MASS_TYPE_AVE;
  }

  if ($UniGeneFile) {
    $objUniGene = msparser::ms_unigene->new($objResFile, $UniGeneFile);

    if (not $objResFile->isValid) {
      if (defined $MIME_boundary) {
        common_subs::multipart_end($MIME_boundary);
        common_subs::multipart_start($MIME_boundary);
        print "Cannot create ms_unigene object: ".$objResFile->getLastErrorString();
        common_subs::multipart_final($MIME_boundary);
      }
      die "Cannot create ms_unigene object: ".$objResFile->getLastErrorString();
    }
  }

# Create hash for masses and variable mods
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_MASSES, $i)) {
    if ($nextKey =~ /^delta(\d+)/i) {
      my $letter = $1;
      if ($letter > 9) {
        $letter = chr($1+55);
      }
      ($vmMass{$letter}, $vmString{$letter}) = 
        split(/,/, $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $nextKey), 2);
    } elsif ($nextKey =~ /^Ignore(\d+)$/i) {
    } elsif ($nextKey =~ /^FixedMod(\d+)$/i) {
    } elsif ($nextKey =~ /^FixedModResidues(\d+)$/i) {
    } elsif ($nextKey =~ /^FixedModNeutralLoss(\d+)$/i) {
    } elsif ($nextKey =~ /^NeutralLoss(\d+)$/i) {
    } elsif ($nextKey =~ /^NeutralLoss(\d+)_master$/i) {
    } elsif ($nextKey =~ /^NeutralLoss(\d+)_slave$/i) {
    } elsif ($nextKey =~ /^NeutralLoss\d_/i) {
    } elsif ($nextKey =~ /^ReqPepNeutralLoss(\d+)$/i) {
    } elsif ($nextKey =~ /^PepNeutralLoss(\d+)$/i) {
    } else {
      my $m = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $nextKey);
      if (length($nextKey) == 1 or $nextKey =~ /_term/i) {
        $groupMasses{lc($nextKey)} = $m;
      } else {
        $elementalMasses{lc($nextKey)} = $m;
      }
    }
    $i++;
  }
# result files earlier than 1.9 don't list electron mass
  $elementalMasses{electron} = 0.000549 unless $elementalMasses{electron};
  
# parse out result file name root
  ($fileNameRoot) = $fileIn[0] =~ /([^\/\\]*)\.dat$/;
  if (@fileIn > 1) {
    $fileNameRoot .= '_merged';
  }

  my $cache_every_protein = 0;

  if ($export_format eq 'pepXML') {
      $cache_every_protein = 1;
  } elsif ($export_format eq 'mzIdentML') {
      $cache_every_protein = $thisScript->param($urlParams{'query_all'});
  }

  # Taxonomy cache is always needed if _prefertaxonomy is used, but it's
  # also used on its own.
  if ($preferTaxonomy > 0 
          or $thisScript->param($urlParams{'prot_tax_id'})
          or $thisScript->param($urlParams{'prot_tax_str'})) {
    my $prot_tax_id = $thisScript->param($urlParams{'prot_tax_id'});
    my $prot_tax_str = $thisScript->param($urlParams{'prot_tax_str'});
    my $prot_ttid = ($prot_tax_id || $prot_tax_str);

    if ($preferTaxonomy > 0) {
        $prot_tax_id = 1;
    }

    common_subs::populateGetTaxonomyCache(
      $objSummary, $objResFile, $objUniGene, $sessionID, 
      $preferTaxonomy || $ShowSameSets, $ShowSubSets,
      {
        prot_ttid => $prot_ttid,
        prot_tax_id => $prot_tax_id,
        prot_tax_str => $prot_tax_str,
      },
      $cache_every_protein,
      defined $MIME_boundary ? 1 : 0, # keepalive?
      $MIME_boundary,
    );
  }

  if ($preferTaxonomy > 0) {
    common_subs::populatePreferredTaxonomyCache(
      $objResFile,
      $objSummary,
      $preferTaxonomy,
      $ShowSameSets,
      $ShowSubSets,
      $cache_every_protein,
    );
  }

  if ($thisScript->param($urlParams{'prot_desc'})) {
    common_subs::populateProteinDescriptionCache(
      $objSummary, $objResFile, $objDatFile, $objUniGene, $sessionID, 
      $ShowSameSets, $ShowSubSets, 
      defined $MIME_boundary ? 1 : 0, # keepalive?
      $cache_every_protein,
      common_subs::samesetOrderingByPreferredTaxonomy(),
      \&common_subs::preferredAccession,
      $MIME_boundary,
    );
  }
  
  do {
    my $do_fetch = 0;
    if ($export_format eq 'XML' or $export_format eq 'CSV' or $export_format eq 'mzTab') {
      # Only XML, CSV and mzTab support emPAI.
      if ($objSummary->anyEmPAI
           and ($export_format eq 'mzTab' or $thisScript->param($urlParams{'peptide_master'}))
           and $thisScript->param($urlParams{'prot_empai'})) {
        $do_fetch = 1;
      } elsif ($thisScript->param($urlParams{'prot_mass'})) {
        $do_fetch = 1;
      } elsif ($thisScript->param($urlParams{'prot_cover'})) {
        $do_fetch = 1;
      }
    } else {
      if ($thisScript->param($urlParams{'prot_mass'})) {
        $do_fetch = 1;
      } elsif ($thisScript->param($urlParams{'prot_cover'})) {
        $do_fetch = 1;
      }
    }

    if ($do_fetch) {
    # All proteins in a single NCBI nr entry have the same sequence, so no
    # need to pass \&preferredAccession here.
      my $getseqBatchMass = common_subs::queueGetSeqProteins(
        $objSummary, $objResFile, $objDatFile,
        sub { $_[0]->getProteinMass($_[1], $_[2]) }, # only proteins without mass
        $ShowSameSets, $ShowSubSets, 1,
        $cache_every_protein,
        common_subs::samesetOrderingByPreferredTaxonomy(),
      );
      my @data_keys = qw(prot_len prot_sequence);
      # This will be used by both getProteinEmPAI and mustGetProteinMass.
      my $data = common_subs::batchGetSeqQuery(
        $objResFile->getCacheDirectory(), $getseqBatchMass, $sessionID, 
        defined $MIME_boundary ? 1 : 0, # keepalive?
        ['showlen', 'showsequence'], \@data_keys,
        $MIME_boundary
      );
      if ($objUniGene) {
        # UniGene entries don't have protein sequences in any meaningful
        # sense.
        common_subs::mapUnparsedToUniGeneEntries(
          $objUniGene, $data, { prot_len => -1, prot_sequence => -1 }
        );
      }
      $fastaLen = common_subs::parseGetSeqBatch(
        $data, $getseqBatchMass, $objResFile->params, \@data_keys,
        prot_sequence => sub { $_[0] eq '' ? -1 : $_[0] },
        prot_len => sub { $_[0] eq '' ? -1 : $_[0] },
      );
    }
  };

  do {
    my $prot_len = $thisScript->param($urlParams{'prot_len'});
    my $prot_pi = $thisScript->param($urlParams{'prot_pi'});
    my $prot_seq = undef;

    if ($shipper eq 'TRUE') {
      $prot_seq = $thisScript->param($urlParams{'prot_seq'});
    }

    if ($prot_len or $prot_pi or $prot_seq) {
      populateGetSeqCache(
        $objSummary, $objResFile, $objUniGene, undef, 
        $ShowSameSets, $ShowSubSets,
        {
          prot_len => $prot_len,
          prot_pi => $prot_pi,
          prot_seq => $prot_seq,
        },
        $cache_every_protein,
        $MIME_boundary,
      );
    }
  };

  # We need to stop emitting keepalive from here on in order to avoid
  # corrupting the output data.
  $objResFile->resetKeepAlive(0, "");

  if (defined $MIME_boundary) {
    common_subs::multipart_end($MIME_boundary);
    if ($export_format eq "XML" || $export_format eq "pepXML") {
      common_subs::multipart_start($MIME_boundary, "text/xml; charset=utf-8", "$fileNameRoot.xml");
    } elsif ($export_format eq "mzIdentML") {
      common_subs::multipart_start($MIME_boundary, "text/xml; charset=utf-8", "$fileNameRoot.mzid");
    } elsif ($export_format eq "mzTab") {
      common_subs::multipart_start($MIME_boundary, "text/plain; charset=utf-8", "$fileNameRoot.mztab");
    } elsif ($export_format eq "CSV") {
      common_subs::multipart_start($MIME_boundary, "application/vnd.ms-excel; charset=utf-8", "$fileNameRoot.csv");
    } elsif ($export_format eq "DTASelect") {
      common_subs::multipart_start($MIME_boundary, "text/plain; charset=utf-8", "DTASelect.txt");
    } else {
      common_subs::multipart_start($MIME_boundary);
    }
  } elsif ($isCGI) {
    if ($export_format eq "XML" || $export_format eq "pepXML") {
      print $thisScript->header( -type => "text/xml",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.xml");
    } elsif ($export_format eq "mzIdentML") {
      print $thisScript->header( -type => "text/xml",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.mzid");
    } elsif ($export_format eq "mzTab") {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.mztab");
    } elsif ($export_format eq "CSV") {
      print $thisScript->header( -type => "application/vnd.ms-excel",
                                 -charset => "utf-8",
                                 -attachment => "$fileNameRoot.csv");
    } elsif ($export_format eq "DTASelect") {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "DTASelect.txt");
    } else {
      print $thisScript->header( -charset => "utf-8" );
    }
  }

  if ($isCGI) {
    $URI = $thisScript->url(-path_info=>1);
    ($URI) = $URI =~ /^(.*\/).+?\/$myName/;
  } else {
    $URI = "../";
  }

# print document opening tags
  if ($export_format eq "XML") {
    print "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    print "<mascot_search_results xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"";
    print " xmlns=\"http://www.matrixscience.com/xmlns/schema/mascot_search_results_2\"";
    print " xsi:schemaLocation=\"http://www.matrixscience.com/xmlns/schema/mascot_search_results_2 ";
    if ($isCGI) {
      print $URI;
    } else {
      print "http://www.matrixscience.com/";
    }
    print "xmlns/schema/mascot_search_results_2/mascot_search_results_2.xsd\"";
    print "  majorVersion=\"2\" minorVersion=\"2\">\n";
  } elsif ($export_format eq "CSV") {
    # nothing
  } elsif ($export_format eq "pepXML") {
    print "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    print "<msms_pipeline_analysis";
    print " xmlns=\"http://regis-web.systemsbiology.net/pepXML\"";
    print " xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"";
    print " xsi:schemaLocation=\"http://regis-web.systemsbiology.net/pepXML ";
    if ($isCGI) {
      print $URI;
    } else {
      print "http://www.matrixscience.com/";
    }
    print "xmlns/schema/pepXML_v18/pepXML_v18.xsd\"";
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = gmtime(time);
    my $timestamp = sprintf("%04d-%02d-%02dT%02d:%02d:%02dZ", 
      $year+1900, $mon+1, $mday, $hour, $min, $sec);
    # optional name
    print " date=\"$timestamp\"";
    print " summary_xml=\"$fileNameRoot.xml\">\n";
  } elsif ($export_format eq "mzIdentML") {
    print "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    print "<MzIdentML id=\"report\" version=\"1.1.0\"\n";
    print "             xsi:schemaLocation=\"http://psidev.info/psi/pi/mzIdentML/1.1 ";
    if ($isCGI) {
      print $URI;
    } else {
      print "http://www.matrixscience.com/";
    }
    print "xmlns/schema/mzIdentML/mzIdentML1.1.0.xsd\"\n";
    print "             xmlns=\"http://psidev.info/psi/pi/mzIdentML/1.1\"\n";
    print "             xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n";
    printf "             creationDate=\"%s\">\n", _fmtdate_iso8601(time);
  } elsif ($export_format eq "DTASelect") {
    # nothing
  } elsif ($export_format eq 'mzTab') {
    # nothing
  } else {
    my $HTMLtitle = "\n\n<HTML>\n<HEAD>\n<TITLE>$export_format Report (";
    if (@fileIn == 1) {
      if ($objParams->getCOM()) {
        $HTMLtitle .= &noXmlTag($objParams->getCOM());
      } else {
        $HTMLtitle .= $fileIn[0];
      }
    } else {
        $HTMLtitle .= sprintf('results collection set of %d files', scalar(@fileIn));
    }
    $HTMLtitle .= ")</TITLE>\n";
    print $HTMLtitle;
    print "</HEAD>\n\n";
    print "<BODY>\n";
    print "<PRE>\n";
  }

  if ($export_format eq "pepXML") {
  # Branch off for ISB pepXML
    pepXML();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;    
  }

  # We need record_id values for the UNIMOD entries in mzIdentML and
  # mzTab. If we don't have them, try unimod.xml
  if ($objUmodConfigFile->getNumberOfModifications() > 0
          and $objUmodConfigFile->getModificationByNumber(0)->haveRecordID()) {
      $objUmodConfigFile2 = $objUmodConfigFile;
  } else {
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($sessionID);
    $objUmodConfigFile2 = new msparser::ms_umod_configfile("../config/unimod.xml",
      "../html/xmlns/schema/unimod_2/unimod_2.xsd", $cs);
  }

  if ($export_format eq "mzIdentML") {
  # Branch off for PSI mzIdentML
    eval { mzIdentML() };
    my $err;
    if ($@) {
      $err = $@;
    }
    print "</MzIdentML>\n";
    if ($err) {
      common_subs::fatal($err, __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
    } elsif (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;    
  }

  if ($export_format eq "mzTab") {
  # Branch off for PSI mzTab
    eval { mzTab() };
    if ($@) {
      common_subs::fatal($@, __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
    } elsif (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;    
  }

  if ($export_format eq "DTASelect") {
  # Branch off for Scripps DTASelect
    DTASelect();
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    exit 0;    
  }

# search information
  if ($thisScript->param($urlParams{'search_master'})) {
    if ($thisScript->param($urlParams{'show_header'})) {
      outputHeader();
    }
    if ($thisScript->param($urlParams{'show_decoy'})) {
      outputDecoy();
    }
    if ($thisScript->param($urlParams{'show_mods'})) {
      if ($objParams->getMODS || $objParams->getFixedModsName(1)) {
        outputFixedMods();
      }
      if ($objParams->getIT_MODS || $objParams->getVarModsName(1)) {
        outputVarMods();
      }
    }
    if ($thisScript->param($urlParams{'show_params'})) {
      outputParams();
    }
    if ($thisScript->param($urlParams{'show_format'})) {
      outputFormat();
    }
    if ($thisScript->param($urlParams{'show_masses'})) {
      outputMasses();
    }
    if ($thisScript->param($urlParams{'show_quant_header'})) {
      outputQuantitationHeader();
    }
  }
  
  if ($thisScript->param($urlParams{'protein_master'})) {
    my $filter_duplicates = 1;
    if ($thisScript->param($urlParams{'show_pep_dupes'})) {
      $filter_duplicates = 0;
    }

    if ($export_format eq "XML") {
      print "<hits>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Protein hits\"$delimiter\"--------------------------------------------------------\"\n\n";
    }
  
  # print column headers for CSV format
    if ($export_format eq "CSV") {
      my $string = "";
      foreach my $field (@columns) {
        if ($field =~ /^pep/ && !$thisScript->param($urlParams{'peptide_master'})) {
          next;
        }
        if ($thisScript->param($urlParams{lc($field)})) {
          if ($field eq "pep_seq") {
            $string .= "pep_res_before" . $delimiter . $field . $delimiter . "pep_res_after" . $delimiter;
          } elsif ($field eq "pep_var_mod") {
            $string .= $field . $delimiter 
              . "pep_var_mod_pos" . $delimiter
              . "pep_summed_mod_pos" . $delimiter
              . "pep_local_mod_pos" . $delimiter
              # pep_var_mod_conf is not included here; it's query-level data
              ;
          } elsif ($field eq "prot_matches" && $anyPepSumMatch) {
            $string .= $field . $delimiter . "prot_matches_sig" . $delimiter . "prot_sequences" . $delimiter . "prot_sequences_sig" . $delimiter;
          } elsif ($field eq "prot_hit_num" && $group_family && $anyPepSumMatch) {
            $string .= $field . $delimiter . "prot_family_member" . $delimiter;
          } else {
            $string .= $field . $delimiter;
          }
        }
      }
      if ($string) {
        chop $string;
      }
      print $string . "\n";
    }
  
  # start of MAIN LOOP to enumerate the protein hits
    my $sameset_ordering = common_subs::samesetOrderingByPreferredTaxonomy() || {};
    my %dupe_state_cache = ();
    my $thisHit = 1;
    $objProtein = $objSummary->getHit($thisHit);
    while (defined($objProtein)) {
    # export the hit
      outputHit($objProtein, $thisHit, $$sameset_ordering{$thisHit}, $filter_duplicates, \%dupe_state_cache);
    # and get the next primary hit
      $thisHit++;
      $objProtein = $objSummary->getHit($thisHit);
    }
  # end of MAIN LOOP to enumerate the protein hits
  
    if ($export_format eq "XML") {
      print "</hits>\n";
    } 
    
  }

  if ($thisScript->param($urlParams{'show_unassigned'})) {
    outputUnassigned();
  }

  if ($thisScript->param($urlParams{'query_master'})) {
    outputQueries();
  }
  
# print document closing tags
  if ($export_format eq "XML") {
    print "</mascot_search_results>\n";
  } elsif ($export_format eq "CSV") {
    # none
  } else {
    print "</PRE>\n";
    print "</BODY>\n";
    print "</HTML>\n";
  }

  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }

# Good Night
  exit 0;

###############################################################################
# &mustGetProteinMass()
# tries very hard to get a protein mass
# $_[0] accession string
# $_[1] is an ms_mascotresults object
# $_[2] is an ms_searchparams object
# $_[2] frame number, 0 for protein, -1 for mixed
# $_[3] database number
# globals:
# my($shipper, $fastaLen);
###############################################################################

sub mustGetProteinMass {
  my ($accession, $objSummary, $objParams, $frame, $dbIdx) = @_;

  my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
  if (defined(my $mass = $$fastaLen{$key}{'prot_mass'})) {
    return $mass == -1 ? '' : $mass;
  }

  if (my $mass = $objSummary->getProteinMass($accession, $dbIdx)) {
    $$fastaLen{$key}{'prot_mass'} = $mass;
    return $mass;
  }

# if this is the public web site, give up
  if ($shipper eq 'FALSE') {
    return '';
  }

  if (defined(my $seqString = $$fastaLen{$key}{'prot_sequence'})) {
    my $mass = -1;
    if ($seqString ne -1) {
       $mass = $objSummary->getSequenceMass($seqString);
    }
    $$fastaLen{$key}{'prot_mass'} = $mass;
    return $mass == -1 ? '' : $mass;
  }

  if ($frame == -1) {$frame = 1}

  my $seqString = mustGetProteinSeq($objParams, $accession, $frame, $dbIdx);
  if ($seqString) {
    my $mass = $objSummary->getSequenceMass($seqString);
    $$fastaLen{$key}{'prot_mass'} = $mass;
    return $mass;
  } 

 # Cannot retrieve sequence
  $$fastaLen{$key}{'prot_mass'} = -1;
  return '';
}

BEGIN {
  my %getseq_cache = ();

  my %queue_key = qw(
      prot_pi     showpi
      prot_len    showlen
      prot_seq    showsequence
  );

  my %data_key = qw(
      prot_pi     prot_pi
      prot_len    prot_len
      prot_seq    prot_sequence
  );

###############################################################################
# &populateGetSeqCache()
# $_[0] is an ms_mascotresults object
# $_[1] is an ms_mascotresfile object
# $_[2] is an ms_unigene object, if UniGene is in use
# $_[3] is a session ID (or undef)
# $_[4] should be $ShowSameSets
# $_[5] should be $ShowSubSets
# $_[6] is a hash reference that defines what is to be cached
# $_[7] should be 1 if output format is pepXML
# $_[8] MIME arguments to batchGetSeqQuery()
###############################################################################
  sub populateGetSeqCache {
    my ($objSummary, $objResFile, $objUniGene, $sessionID, $ShowSameSets,
        $ShowSubSets, $what, $include_every_protein, $MIME_boundary) = @_;

    my $getseqBatchAll = common_subs::queueGetSeqProteins(
      $objSummary, $objResFile, $objDatFile, 
      undef, $ShowSameSets, $ShowSubSets, 1, $include_every_protein,
      common_subs::samesetOrderingByPreferredTaxonomy(),
    );

    my (@queue_keys, @data_keys);

    for (keys %$what) {
      next unless $queue_key{$_};
      push @queue_keys, $queue_key{$_};
      push @data_keys, $data_key{$_};
    }

    my $data = common_subs::batchGetSeqQuery(
      $objResFile->getCacheDirectory(), $getseqBatchAll, $sessionID, 
      defined $MIME_boundary ? 1 : 0, # keepalive?
      \@queue_keys, \@data_keys, $MIME_boundary
    );

    %getseq_cache = %{ 
      common_subs::parseGetSeqBatch(
        $data, $getseqBatchAll, $objResFile->params, \@data_keys,
        prot_sequence => sub { $_[0] eq '' ? -1 : $_[0] },
        prot_len => sub { $_[0] eq '' ? -1 : $_[0] },
        prot_pi => sub { $_[0] eq '' ? -1 : $_[0] },
      ) || {}
    };

    if ($objUniGene) {
      common_subs::mapToUniGeneEntries(
        $objUniGene, \%getseq_cache, { prot_pi => -1, prot_len => -1, prot_sequence => -1 }
      );
    }
  }

###############################################################################
# &getProteinpI()
# Return the protein pI (from cache if possible)
# $_[0] is an ms_searchparams object
# $_[1] accession string
# $_[2] frame number, 0 for protein, -1 for mixed
# $_[3] database number
###############################################################################
  sub getProteinpI {
    my ($objParams, $accession, $frame, $dbIdx) = @_;

# if protein contains matches in mixed frames, then return empty handed
    if ($frame == -1) {return ''}

# first try the cache
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
    # Could be undefined if ms-getseq.exe failed.
    my $pI = $getseq_cache{$key}{'prot_pi'};
    if (defined($pI)) {
      return '' if $pI eq -1;
      return $pI;
    }
# calculate pI from ms-getseq.exe
    my @retVal = common_subs::getExtInfo($accession, $frame, $objParams->getDB($dbIdx), "pI");
    if ($retVal[0]) {
      if ($retVal[1] =~ /plain\s+([\d\.]+)\s*$/) {
        return $getseq_cache{$key}{'prot_pi'} = $1;
      }
    }
    $getseq_cache{$key}{'prot_pi'} = -1;
    return '';
  }

###############################################################################
# &getProteinLen()
# Return the protein sequence length (from cache if possible)
# $_[0] is an ms_searchparams object
# $_[1] accession string
# $_[2] frame number, 0 for protein, -1 for mixed
# $_[3] database index for access
###############################################################################
  sub getProteinLen {
    my ($objParams, $accession, $frame, $dbIdx) = @_;

# if protein contains matches in mixed frames, then choose frame 1 to get approximate length
    if ($frame == -1) {$frame = 1}

# first try the cache (populated after the header has been output)
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
    my $len = $getseq_cache{$key}{'prot_len'};
    if (defined($len)) {
      return '' if $len eq -1;
      return $len;
    }

    if (defined(my $len = $$fastaLen{$key}{'prot_len'})) {
      $getseq_cache{$key}{'prot_len'} = $len;
      return $len == -1 ? '' : $len;
    }

    my @retVal = common_subs::getExtInfo($accession, $frame, $objParams->getDB($dbIdx), "length");
    if ($retVal[0]) {
      if ($retVal[1] =~ /plain\s+([\d\.]+)\s*$/) {
        return $getseq_cache{$key}{'prot_len'} = $1;
      }
    }
    $getseq_cache{$key}{'prot_len'} = -1;
    return '';
  }

###############################################################################
# &mustGetProteinSeq()
# tries very hard to get a description string
# $_[0] is an ms_searchparams object
# $_[1] is an accession
# $_[2] frame number, 0 for protein, (-1 for mixed is not allowed)
# $_[3] database number
# my($shipper);
###############################################################################
  sub mustGetProteinSeq {
    my ($objParams, $accession, $frame, $dbIdx) = @_;

# if this is the public web site, give up
    if ($shipper eq 'FALSE') {
      return '';
    }
    
# cannot get sequence for mixed frame match
    if ($frame < 0 || $frame > 6) {
      return '';
    }
    
# first try the cache
    my $key = common_subs::makeProteinCacheKey($dbIdx, $accession, $frame);
    # Could be undefined if ms-getseq.exe failed.
    my $seq = $getseq_cache{$key}{'prot_sequence'};
    if (defined($seq)) {
      return '' if $seq eq -1;
      return $seq;
    }

    my @retVal = common_subs::getExtInfo($accession, $frame, $objParams->getDB($dbIdx), "sequence");
    if ($retVal[0]) {
      $seq = common_subs::parseSequence($retVal[1], $retVal[2]);
    }
    if ($seq) {
      return $getseq_cache{$key}{'prot_sequence'} = $seq;
    }
    $getseq_cache{$key}{'prot_sequence'} = -1;
    return '';
  }

} # end %getseq_cache


###############################################################################
# &validateParams()
# validate a variety of mascot.dat and URL parameters
# globals:
# my($massDP, $minPepLen, $ignoreIonsScoreBelow, $reportType,
# $objParams, $objResFile, $thisScript, %urlParams, $sigThreshold,
# $targetFDR, $numHits, $ShowSubSets, $RequireBoldRed,
# $mudpitSwitch, $ShowAllFromErrorTolerant, $UniGeneFile, @fileIn, 
# $idForFatalError);
###############################################################################

sub validateParams {

  my $tempString;

# $massDP determines precision on peptide mass values
# default is 2 decimal places for peptides
  $massDP = 0 unless ($massDP = $objMascotOptions->getMassDecimalPlaces);
  $massDP = int($massDP);
  if ($massDP < 1 || $massDP > 6) {
    $massDP = 2;
  }
  
# peptides shorter than $minPepLen are not interesting
# default is to ignore anything shorter than 6 residues
  if (defined($thisScript->param($urlParams{'_minpeplen'}))
    && $thisScript->param($urlParams{'_minpeplen'}) >= 0) {
    $minPepLen = $thisScript->param($urlParams{'_minpeplen'}) + 0;
  } else {
    $minPepLen = 0 unless ($minPepLen = $objMascotOptions->getMinPepLenInPepSummary );
    $minPepLen = int($minPepLen); 
    if ($minPepLen < 1 || $minPepLen > 9) {
      $minPepLen = 6;
    }
  }
  
# ignoreIonsScoreBelow is a threshold on MS/MS ions scores
# a value of 0 specifies that all peptides are included (default)
# a value greater than 1 indicates an absolute score cutoff 
# a value less than 1 is treated as a probability threshold 
  if (defined($thisScript->param($urlParams{'_ignoreionsscorebelow'}))
    && $thisScript->param($urlParams{'_ignoreionsscorebelow'}) >= 0) {
    $ignoreIonsScoreBelow = $thisScript->param($urlParams{'_ignoreionsscorebelow'}) + 0;
  } else {
    $ignoreIonsScoreBelow = 0 unless ($ignoreIonsScoreBelow = $objMascotOptions->getIgnoreIonsScoreBelow);
    if ($ignoreIonsScoreBelow < 0) {
      $ignoreIonsScoreBelow = 0;
    }
  }

# if queries / entries greater than $mudpitSwitch, default 0.001, switch to Mudpit scoring
  if (defined($thisScript->param($urlParams{'_server_mudpit_switch'}))
    && $thisScript->param($urlParams{'_server_mudpit_switch'}) > 0) {
    $mudpitSwitch = $thisScript->param($urlParams{'_server_mudpit_switch'}) + 0;
  } else {
    $mudpitSwitch = 0.001;
    my $tempString;
    $tempString = 0 unless ($tempString = $objMascotOptions->getMudpitSwitch);
    if ($tempString > 0) {
      $mudpitSwitch = $tempString;
    }
  }

# If $ShowDecoyReport is true, a report for the integrated decoy search is shown
  if (defined($thisScript->param($urlParams{'_show_decoy_report'}))) {
    $ShowDecoyReport = $thisScript->param($urlParams{'_show_decoy_report'}) + 0;
  } else {
    $ShowDecoyReport = 0;
  }

# set $section parameters to point to the correct sections
  $sec_peptides = $msparser::ms_mascotresfile::SEC_PEPTIDES;
  $sec_summary = $msparser::ms_mascotresfile::SEC_SUMMARY;
  if (!$objResFile->isErrorTolerant()
    && $objParams->getDECOY() > 0
    && $ShowDecoyReport) {
    $sec_peptides = $msparser::ms_mascotresfile::SEC_DECOYPEPTIDES;
    $sec_summary = $msparser::ms_mascotresfile::SEC_DECOYSUMMARY;
  }
# $objResFile->anyPeptideSummaryMatches($sec_peptides) is used constantly
  $anyPepSumMatch = $objResFile->anyPeptideSummaryMatches($sec_peptides);

# Report type is either peptide or concise
  $reportType = 'concise';
  if ($anyPepSumMatch) {
    $reportType = 'peptide';
  }

# set $sigThreshold to probability significance threshold, default 0.05
  if (defined($thisScript->param($urlParams{'_sigthreshold'}))
    && $thisScript->param($urlParams{'_sigthreshold'}) >= 1E-18
    && $thisScript->param($urlParams{'_sigthreshold'}) < 1) {
    $sigThreshold = $thisScript->param($urlParams{'_sigthreshold'}) + 0;
  } else {
    $sigThreshold = 0 unless ($sigThreshold = $objMascotOptions->getSigThreshold);
    if ($sigThreshold < 1E-18 || $sigThreshold >= 1) {
      $sigThreshold = 0.05;
    }
  }

  if ($objParams->getDECOY() > 0) {
    if (defined($thisScript->param($urlParams{'_target_fdr'}))) {
      my $str = $thisScript->param($urlParams{'_target_fdr'});
      if ($str ne '' and 0 < $str and $str < 100) {
        $targetFDR = 0 + $str;
      }
    }
  }

# set number of hits to display & make sure it is reasonable
# if the value of REPORT is 'AUTO', $numHits is set to 0
  $numHits = 0;
  if (defined($thisScript->param($urlParams{'report'}))) {
    if ($thisScript->param($urlParams{'report'}) =~ /auto/i) {
      $numHits = 0; 
    } else {
      $numHits = int($thisScript->param($urlParams{'report'}));
    }
  } else {
    $numHits = int($objParams->getREPORT());
  }
  if ($numHits < 0){
    $numHits = 20;
  }
  
# Display same-set proteins $ShowSameSets is true
  $ShowSameSets = 0;
  if (defined($thisScript->param($urlParams{'show_same_sets'}))) {
    $ShowSameSets = $thisScript->param($urlParams{'show_same_sets'}) + 0;
  }  

# Display a sub-set protein hit if it is missing less than $ShowSubSets fraction of the protein score of the main hit
  if (defined($thisScript->param($urlParams{'_showsubsets'}))) {
    $ShowSubSets = $thisScript->param($urlParams{'_showsubsets'}) + 0;
  } elsif ($thisScript->param($urlParams{'do_export'})) {
    $ShowSubSets = 0;
  } else {
    $ShowSubSets = 0 unless ($ShowSubSets = $objMascotOptions->getShowSubsets);
  }

# If RequireBoldRed is true, only display protein hits that include at least one bold, red peptide match
  if (defined($thisScript->param($urlParams{'_requireboldred'}))) {
    $RequireBoldRed = $thisScript->param($urlParams{'_requireboldred'}) + 0;
  } elsif ($thisScript->param($urlParams{'do_export'})) {
    $RequireBoldRed = 0;
  } else {
    $RequireBoldRed = 0 unless ($RequireBoldRed = $objMascotOptions->isRequireBoldRed);
  }

# If $ShowAllFromErrorTolerant is true, all hits from an error tolerant search are shown
  if (defined(my $all = $thisScript->param($urlParams{'_showallfromerrortolerant'}))) {
    $ShowAllFromErrorTolerant = looks_like_number($all) ? $all + 0 : 0;
  } else {
    $ShowAllFromErrorTolerant = 0;
  }
# If $OnlyErrorTolerant is true, display only error tolerant matches from a new style error tolerant search
  if (defined(my $only = $thisScript->param($urlParams{'_onlyerrortolerant'}))) {
    $OnlyErrorTolerant = looks_like_number($only) ? $only + 0 : 0;
  } else {
    $OnlyErrorTolerant = 0;
  }
# If $NoErrorTolerant is true, display only standard matches from a new style error tolerant search
  if (defined(my $no = $thisScript->param($urlParams{'_noerrortolerant'}))) {
    $NoErrorTolerant = looks_like_number($no) ? $no + 0 : 0;
  } else {
    $NoErrorTolerant = 0;
  }
# pick up illegal combinations
  if ($NoErrorTolerant) {
    $ShowAllFromErrorTolerant = 0;
  }
  if ($OnlyErrorTolerant && $NoErrorTolerant) {
    die("Meaningless to specify both _onlyerrortolerant and _noerrortolerant", $idForFatalError);
  }

# $UniGeneFile is the path to the Unigene index to be used for clustering EST matches
  if (my $species = $thisScript->param($urlParams{'unigene'})) {
    if ($species eq "None") {
      $UniGeneFile = "";
    } else {
      $UniGeneFile = $objUnigeneOptions->getPropValStringByName($species)
        || die("cannot find UniGene $species in mascot.dat ", $idForFatalError);
      $UniGeneFile =~ s/\s*$//;
    }
  } else {
    $UniGeneFile = "";
  }  
  
# Whether to group protein families
  $group_family = 0;
  if (defined($thisScript->param($urlParams{'group_family'}))) {
    $group_family = $thisScript->param($urlParams{'group_family'}) + 0;
  }  

# Whether to display scores and expect values based on Percolator PEPs
  if (defined($percolate = $thisScript->param($urlParams{'percolate'}))) {
    $percolate = $thisScript->param($urlParams{'percolate'});
  } elsif ($thisScript->param($urlParams{'do_export'})) {
    $percolate = 0;
  } else {
    $percolate = $objMascotOptions->isPercolator;
  }
  $percolate += 0;
  my($PercolatorMinQueries, $PercolatorMinSequences);
  $PercolatorMinQueries = 100 unless ($PercolatorMinQueries = $objMascotOptions->getPercolatorMinQueries);
  $PercolatorMinSequences = 100 unless ($PercolatorMinSequences = $objMascotOptions->getPercolatorMinSequences);
  if ($reportType eq 'protein' 
    || $reportType eq 'concise'
    || !$objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_PEPTIDES)
    || $objResFile->getNumQueries < $PercolatorMinQueries
    || $objResFile->getNumSeqsAfterTax < $PercolatorMinSequences
    || $objResFile->isErrorTolerant
    || !$objResFile->anyPeptideSummaryMatches($msparser::ms_mascotresfile::SEC_DECOYPEPTIDES)) {
    $percolate = -1;
  }
  if ($percolate == 1) {
    if (defined($thisScript->param($urlParams{'percolate_rt'}))) {
      $percolate_rt = $thisScript->param($urlParams{'percolate_rt'});
    } elsif ($thisScript->param($urlParams{'do_export'})) {
      $percolate_rt = 0;
    } else {
      $percolate_rt = $objMascotOptions->isPercolatorUseRT;
    }
    $percolate_rt += 0;
  } else {
    $percolate_rt = 0;
  }

}

###############################################################################
# &formatControls()
# Print format form
# globals:
# my($myName, $objResFile, $objParams, $numHits, $sigThreshold, @fileIn,
# $idForFatalError,
# $ignoreIonsScoreBelow, $ShowSubSets, $thisScript, $RequireBoldRed, 
# %urlParams, $mudpitSwitch, $objMascotOptions, $targetFDR);
###############################################################################

sub formatControls {

  &common_subs::printHeader1("Matrix Science - Mascot - Export search results");

  print <<'end_of_static_HTML_text_block';

<SCRIPT LANGUAGE="JavaScript">
<!-- Begin hiding from older browsers 

  function check_slaves(checkBox, form){
    var checkState = checkBox.checked;
    var checkName = checkBox.name;
end_of_static_HTML_text_block

  if ($shipper eq 'TRUE') {
    print "    var shipper = 1;\n";
  } else {
    print "    var shipper = 0;\n";
  }

  print <<'end_of_static_HTML_text_block';
    if (checkName == "search_master") {
      if (form.show_header) {
        form.show_header.disabled = !checkState;
      }
      if (form.show_decoy) {
        form.show_decoy.disabled = !checkState;
      }
      if (form.show_mods) {
        form.show_mods.disabled = !checkState;
      }
      if (form.show_params) {
        form.show_params.disabled = !checkState;
      }
      if (form.show_format) {
        form.show_format.disabled = !checkState;
      }
      if (form.show_masses) {
        form.show_masses.disabled = !checkState;
      }
      if (form.show_quant_header) {
        form.show_quant_header.disabled = !checkState;
      }
    } else if (checkName == "protein_master") {
      if (form.prot_score) {
        form.prot_score.disabled = !checkState;
      }
      if (form.prot_thresh) {
        form.prot_thresh.disabled = !checkState;
      }
      if (form.prot_expect) {
        form.prot_expect.disabled = !checkState;
      }
      if (form.prot_desc) {
        form.prot_desc.disabled = !checkState;
      }
      if (form.prot_mass) {
        form.prot_mass.disabled = !checkState;
      }
      if (form.prot_matches) {
        form.prot_matches.disabled = !checkState;
      }
      if (form.prot_cover) {
        form.prot_cover.disabled = !checkState || !shipper;
      }
      if (form.prot_len) {
        form.prot_len.disabled = !checkState || !shipper;
      }
      if (form.prot_empai) {
        form.prot_empai.disabled = !checkState;
      }
      if (form.prot_pi) {
        form.prot_pi.disabled = !checkState || !shipper;
      }
      if (form.prot_tax_str) {
        form.prot_tax_str.disabled = !checkState || !shipper;
      }
      if (form.prot_tax_id) {
        form.prot_tax_id.disabled = !checkState || !shipper;
      }
      if (form.prot_seq) {
        form.prot_seq.disabled = !checkState || !shipper;
      }
      if (form.prot_quant) {
        form.prot_quant.disabled = !checkState;
      }
end_of_static_HTML_text_block
  unless ($export_format eq 'mzTab') {
    print <<'end_of_static_HTML_text_block';
      if (form.peptide_master) {
        form.peptide_master.checked = !checkState;
        form.peptide_master.click();
        form.peptide_master.disabled = !checkState;
        form.peptide_master.click();
      }
end_of_static_HTML_text_block
  }
  print <<'end_of_static_HTML_text_block';
    } else if (checkName == "peptide_master") {
      if (form.pep_exp_mr) {
        form.pep_exp_mr.disabled = !checkState;
      }
      if (form.pep_exp_z) {
        form.pep_exp_z.disabled = !checkState;
      }
      if (form.pep_calc_mr) {
        form.pep_calc_mr.disabled = !checkState;
      }
      if (form.pep_delta) {
        form.pep_delta.disabled = !checkState;
      }
      if (form.pep_start) {
        form.pep_start.disabled = !checkState;
      }
      if (form.pep_end) {
        form.pep_end.disabled = !checkState;
      }
      if (form.pep_miss) {
        form.pep_miss.disabled = !checkState;
      }
      if (form.pep_score) {
        form.pep_score.disabled = !checkState;
      }
      if (form.pep_homol) {
        form.pep_homol.disabled = !checkState;
      }
      if (form.pep_ident) {
        form.pep_ident.disabled = !checkState;
      }
      if (form.pep_expect) {
        form.pep_expect.disabled = !checkState;
      }
      if (form.pep_seq) {
        form.pep_seq.disabled = !checkState;
      }
      if (form.pep_frame) {
        form.pep_frame.disabled = !checkState;
      }
      if (form.pep_var_mod) {
        form.pep_var_mod.disabled = !checkState;
      }
//      if (form.show_unassigned) {
//        form.show_unassigned.disabled = !checkState;
//      }
      if (form.pep_num_match) {
        form.pep_num_match.disabled = !checkState;
      }
      if (form.pep_scan_title) {
        form.pep_scan_title.disabled = !checkState;
      }
      if (form.pep_quant) {
        form.pep_quant.disabled = !checkState;
      }
      if (form.show_pep_dupes) {
        form.show_pep_dupes.disabled = !checkState;
      }
    } else if (checkName == "query_master") {
      if (form.query_title) {
        form.query_title.disabled = !checkState;
      }
      if (form.query_qualifiers) {
        form.query_qualifiers.disabled = !checkState;
      }
      if (form.query_params) {
        form.query_params.disabled = !checkState;
      }
      if (form.query_peaks) {
        form.query_peaks.disabled = !checkState;
      }
      if (form.query_raw) {
        form.query_raw.disabled = !checkState;
      }
    }
  }
  
  function check_rbr(checkBox, form){
    var checkState = checkBox.checked;
    var checkName = checkBox.name;
    if (form._requireboldred) {
      if (checkState) {
        form._requireboldred.checked = 0;
      }
      form._requireboldred.disabled = checkState;
    }
  }
  
  function check_target_fdr(checkBox, form){
    var checkState = checkBox.checked;
    var checkName = checkBox.name;
    if (form._target_fdr) {
      form._target_fdr.disabled = checkState;
    }
  }
  

// End hiding Javascript from old browsers. -->
</SCRIPT>


end_of_static_HTML_text_block

  common_subs::printHeader2("", "Export search results", 1, $session);

  print <<"end_of_static_HTML_text_block";
  <FORM ACTION="./$myName" NAME="Re-format" METHOD="POST" ENCTYPE="multipart/form-data">
end_of_static_HTML_text_block

  if ($thisScript->param($urlParams{'fileset'})) {
    my $fileset = common_subs::htmlEscape($thisScript->param($urlParams{'fileset'}));
    printf('  <INPUT TYPE="hidden" NAME="fileset" VALUE="%s">'."\n", $fileset);
  } else {
    my $fileIn = common_subs::htmlEscape($fileIn[0]);
    printf('  <INPUT TYPE="hidden" NAME="file" VALUE="%s">'."\n", $fileIn);
  }

  print <<"end_of_static_HTML_text_block";
  <INPUT TYPE="hidden" NAME="do_export" VALUE=1>
  <INPUT TYPE="hidden" NAME="prot_hit_num" VALUE=1>
  <INPUT TYPE="hidden" NAME="prot_acc" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_query" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_rank" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_isbold" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_isunique" VALUE=1>
  <INPUT TYPE="hidden" NAME="pep_exp_mz" VALUE=1>
  <INPUT TYPE="hidden" NAME="_showallfromerrortolerant" VALUE=$ShowAllFromErrorTolerant>
  <INPUT TYPE="hidden" NAME="_onlyerrortolerant" VALUE=$OnlyErrorTolerant>
  <INPUT TYPE="hidden" NAME="_noerrortolerant" VALUE=$NoErrorTolerant>
  <INPUT TYPE="hidden" NAME="_show_decoy_report" VALUE=$ShowDecoyReport>
  <INPUT TYPE="hidden" NAME="sessionid" VALUE="$sessionID">
end_of_static_HTML_text_block

# pass through any quantitation parameters
  while (my($key, $value) = each %urlParams) {
    if ($key =~ /^_quant/i || lc($key) eq "_min_precursor_charge") {
      print "  <INPUT TYPE=\"hidden\" NAME=\"$key\" VALUE=\"" . common_subs::htmlEscape($thisScript->param($urlParams{$key})) . "\">\n";
    }
  }

  print <<"end_of_static_HTML_text_block";
  <TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>
  
    <TR>
      <TD NOWRAP><H2>Export search results</H2></TD>
      <TD ALIGN=RIGHT VALIGN=baseline><A HREF="../help/export_help.html" target=_blank>Help</A></TD>
    </TR>

    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        Export format </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <SELECT NAME=\"export_format\" onChange=\"this.form.do_export.value=0; this.form.submit(); return true\">
end_of_static_HTML_text_block

  if (@fileIn == 1) {
    if ($export_format eq "XML" || $export_format eq "") {
      print "         <OPTION SELECTED>XML</OPTION>\n";
    } else {
      print "         <OPTION>XML</OPTION>\n";
    }
  }
  if ($export_format eq "CSV") {
    print "         <OPTION SELECTED>CSV</OPTION>\n";
  } else {
    print "         <OPTION>CSV</OPTION>\n";
  }
  if (@fileIn == 1) {
    if ($reportType eq 'peptide' && $export_format eq "pepXML") {
      print "         <OPTION SELECTED>pepXML</OPTION>\n";
    } elsif ($reportType eq 'peptide') {
      print "         <OPTION>pepXML</OPTION>\n";
    }
  }
  if ($export_format eq "mzIdentML") {
    print "         <OPTION SELECTED>mzIdentML</OPTION>\n";
  } else {
    print "         <OPTION>mzIdentML</OPTION>\n";
  }
  if ($export_format eq "mzTab") {
    print "         <OPTION SELECTED>mzTab</OPTION>\n";
  } else {
    print "         <OPTION>mzTab</OPTION>\n";
  }
  if (@fileIn == 1) {
    if ($reportType eq 'peptide' && $export_format eq "DTASelect") {
      print "         <OPTION SELECTED>DTASelect</OPTION>\n";
    } elsif ($reportType eq 'peptide') {
      print "         <OPTION>DTASelect</OPTION>\n";
    }
    if ($export_format eq "MascotDAT") {
      print "         <OPTION value=\"MascotDAT\" SELECTED>Mascot DAT File</OPTION>\n";
    } else {
      print "         <OPTION value=\"MascotDAT\">Mascot DAT File</OPTION>\n";
    }
    if ($export_format eq "MGF") {
      print "         <OPTION value=\"MGF\" SELECTED>MGF Peak List</OPTION>\n";
    } else {
      print "         <OPTION value=\"MGF\">MGF Peak List</OPTION>\n";
    }
  }
  print "        </SELECT>\n";
  print "      </TD>\n";
  print "    </TR>\n";

  unless ($export_format eq "MascotDAT" || $export_format eq "MGF") {
      
    print <<"end_of_static_HTML_text_block";
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        Significance threshold p&lt; </TD> 
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <INPUT NAME="_sigthreshold" TYPE=text SIZE=8 VALUE=$sigThreshold>
  
end_of_static_HTML_text_block
  
    if ($anyPepSumMatch) {

      print "      at&nbsp;&nbsp;";
      print "<INPUT TYPE=\"radio\" VALUE=0 NAME=\"use_homology\" ID=\"use_homology_0\"";
      unless ($use_homology){
        print " CHECKED";
      }
      print "><LABEL FOR=\"use_homology_0\">identity</LABEL>";
      print "&nbsp;&nbsp;&nbsp;&nbsp;<INPUT TYPE=\"radio\" VALUE=1 NAME=\"use_homology\" ID=\"use_homology_1\"";
      if ($use_homology){
        print " CHECKED";
      }
      print "><LABEL FOR=\"use_homology_1\">homology</LABEL>\n";
      print "<INPUT TYPE=\"hidden\" NAME=\"_minpeplen\" VALUE=$minPepLen>\n";
      print "    </TD>\n";
      print "   </TR>\n";

      if ($objParams->getDECOY() > 0) {
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>\n";
        print "     Target FDR<BR>(overrides significance<BR>threshold if set) </TD>\n";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "     <SELECT NAME=\"_target_fdr\"";
        if (defined($urlParams{'percolate'}) and $thisScript->param($urlParams{'percolate'} == 1)) {
          print " DISABLED";
        }
        print ">\n";
        my @options = ();
        my $target_found = 0;
        for my $i (0 .. $objMascotOptions->getNumberOfTargetFDRPercentages()-1) {
          my $t = $objMascotOptions->getTargetFDRPercentage($i);
          my $str = "       <OPTION VALUE=\"$t\"";
          if ($targetFDR and $t == $targetFDR) {
            $str .= " SELECTED";
            $target_found = 1;
          }
          $str .= ">$t%</OPTION>\n";
          push @options, $str;
        }
        if (not $target_found) {
          if ($targetFDR) {
            unshift @options, "       <OPTION VALUE=\"$targetFDR\" SELECTED>$targetFDR%</OPTION>\n";
            $target_found = 1;
          }
        }
        print "       <OPTION VALUE=\"\"";
        if (not $target_found) {
          print " SELECTED";
        }
        print ">&lt;not set&gt;</OPTION>\n";
        print @options;
        print "     </SELECT>\n";
        print "    </TD>\n";
        print "   </TR>\n";
      }

      unless ($export_format eq "pepXML") {  
        print "   <TR>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
        print "       Ions score cut-off </TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "       <INPUT NAME=\"_ignoreionsscorebelow\" TYPE=text SIZE=8 VALUE=$ignoreIonsScoreBelow>\n";
        print "     </TD>\n";
        print "   </TR>\n";
      }

    } else {
      print "    </TD>\n";
      print "   </TR>\n";
    }

    unless ($export_format eq "pepXML") {  
      print "   <TR>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      print "       Max. number of hits </TD>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
      if ($numHits) {
        print "       <INPUT NAME=\"report\" TYPE=text SIZE=5 VALUE=$numHits>\n";
      } else {
        print "       <INPUT NAME=\"report\" TYPE=text SIZE=5 VALUE=\"AUTO\">\n";
      }
      print "     </TD>\n"; 
      print "   </TR>\n";
      
      if ($anyPepSumMatch) {
  
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF ALIGN=RIGHT>\n";
        print "     Protein scoring </TD>\n";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "      <INPUT TYPE=\"radio\" VALUE=99999999 NAME=\"_server_mudpit_switch\" ID=\"_server_mudpit_switch_99999999\"";
        unless ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch){
          print " CHECKED";
        }
        print "><LABEL FOR=\"_server_mudpit_switch_99999999\">Standard</LABEL>";
        print "&nbsp;&nbsp;&nbsp;&nbsp;<INPUT TYPE=\"radio\" VALUE=0.000000001 NAME=\"_server_mudpit_switch\" ID=\"_server_mudpit_switch_0.000000001\"";
        if ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch){
          print " CHECKED";
        }
        print "><LABEL FOR=\"_server_mudpit_switch_0.000000001\">MudPIT</LABEL>\n";
        print "     </TD>\n";
        print "   </TR>\n";
  
      }
    }
      
    unless ($export_format eq "pepXML" || $export_format eq "DTASelect") {
      if ($ShowSameSets) {
        $tmpChecked = "CHECKED";
      } else {
        $tmpChecked = "";
      }
      &checkBox("Include same-set protein hits<BR>(additional proteins that span<BR>the same set of peptides)", 
        "show_same_sets", $tmpChecked, "");
      print "   <TR>\n";
      print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
      if ($export_format eq 'mzTab') {
        print "     </TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "       <INPUT NAME=\"_showsubsets\" TYPE=hidden VALUE=$ShowSubSets>\n";
      } else {
        print "       Include sub-set protein hits<BR>(additional proteins that span<BR>a sub-set of peptides) </TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "       <INPUT NAME=\"_showsubsets\" TYPE=text SIZE=5 VALUE=$ShowSubSets>\n";
      }
      print "     </TD>\n"; 
      print "   </TR>\n";
      if ($anyPepSumMatch) {
          
        print "   <TR>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
        print "       Group protein families</TD>\n";
        print "     <TD BGCOLOR=#EEEEFF NOWRAP>\n";
        print "     <INPUT TYPE=\"checkbox\" NAME=\"group_family\" VALUE=1 ";
        if ($group_family) {
          print "CHECKED "
        }
        print "onClick=\"check_rbr(this, this.form)\"></TD>\n";
        print "   </TR>\n";
        my $disabled = "";
        if ($group_family) {
          $disabled = "disabled";
        }
        if ($RequireBoldRed) {
          &checkBox("Require bold red", "_requireboldred", "CHECKED", $disabled);
        } else {
          &checkBox("Require bold red", "_requireboldred", "", $disabled);
        }      
      }
    # UniGene
      my %tempUniGene; # use hash to eliminate duplicates
      for (my $i=1; $i <= $objParams->getNumberOfDatabases; $i++) {
        if (my $tempString = $objUnigeneOptions->getPropValStringByName($objParams->getDB($i))) {
          my @tempList = split(/\s+/, $tempString);
          foreach (@tempList) {
            $tempUniGene{$_} = 1;
          }
        }
      }
      if (%tempUniGene) {
        print "  <TR>\n";
        print "    <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>\n";
        print "UniGene index </TD>";
        print "    <TD BGCOLOR=#EEEEFF>\n";
        print "      <SELECT NAME=\"UNIGENE\">\n";
        $tempUniGene{"None"} = 1;
        my $species = "None";
        if ($thisScript->param($urlParams{'unigene'})) {
          $species = $thisScript->param($urlParams{'unigene'});
        }
        foreach (sort keys %tempUniGene) {
          print "  <OPTION";
          if ($species eq $_) {
            print " SELECTED";
          }
          print ">$_</OPTION>\n";
        }
        print "</SELECT>\n";
        print "    </TD>\n";
        print "  </TR>\n";
      }
    }
    
    if ($percolate != -1) {
      if ($percolate == 1) {
        $tmpChecked = "CHECKED";
      } else {
        $tmpChecked = "";
      }
      &checkBox("Show Percolator scores", "percolate", $tmpChecked, "", 'check_target_fdr(this, this.form)');
    }

    print <<'end_of_static_HTML_text_block';
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        Preferred Taxonomy<SUP>*</SUP> </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
end_of_static_HTML_text_block

    print DOM->dropdown({
      values => [taxonomy_choice($objResFile)],
      name => '_prefertaxonomy',
      default => $preferTaxonomy,
    })->as_string;

    print <<'end_of_static_HTML_text_block';
      </TD>
    </TR>
end_of_static_HTML_text_block

    print <<'end_of_static_HTML_text_block';
    <TR>
      <TD NOWRAP COLSPAN=2><SUP>* Occasionally requires information to be retrieved from external utilities, which can be slow</TD>
    </TR>
end_of_static_HTML_text_block
      
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML" or $export_format eq 'mzTab') {

      print "   <TR>\n";
      print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
      print "   </TR>\n";
  
      print "   <TR>\n";
      print "     <TD NOWRAP>\n";
      print "     <H3>Search Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     <INPUT TYPE=\"checkbox\" NAME=\"search_master\" VALUE=1 CHECKED onClick=\"check_slaves(this, this.form)\"></TD>\n";
      print "   </TR>\n";
  
      &checkBox("Header", "show_header", "CHECKED", "");
      if ($objParams->getDECOY() > 0 && !$ShowDecoyReport) {
        &checkBox("Decoy", "show_decoy", "CHECKED", "");
      }
      &checkBox("Modification deltas", "show_mods", "CHECKED", "");
      &checkBox("Search parameters", "show_params", "CHECKED", "");
      &checkBox("Format parameters", "show_format", "CHECKED", "");
      &checkBox("Residue masses", "show_masses", "", "");
      if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
        &checkBox("Quantitation header", "show_quant_header", "CHECKED", "");
      }
    }
  
    print "   <TR>\n";
    print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
    print "   </TR>\n";
  
    print "   <TR>\n";
    print "     <TD NOWRAP>\n";
    if ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
      print "     <H3>Optional Protein Hit Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     &nbsp;</TD>\n";
    } else {
      print "     <H3>Protein Hit Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     <INPUT TYPE=\"checkbox\" NAME=\"protein_master\" VALUE=1 CHECKED onClick=\"check_slaves(this, this.form)\"></TD>\n";
    }
    print "   </TR>\n";
  
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML" or $export_format eq 'mzTab') {
      &checkBox("Score", "prot_score", "CHECKED", "");
      unless ($anyPepSumMatch) {
        &checkBox("Significance threshold", "prot_thresh", "CHECKED", "");
        &checkBox("Expectation value", "prot_expect", "CHECKED", "");
      }
    }
    &checkBox("Description<SUP>*</SUP>", "prot_desc", "CHECKED", "");
    unless ($export_format eq "mzIdentML" or $export_format eq 'mzTab') {
      &checkBox("Mass (Da)<SUP>*</SUP>", "prot_mass", "CHECKED", "");
    }
    my $disabled = "";
    if ($shipper eq 'FALSE') {
      $disabled = "disabled";
    }
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
      &checkBox("Number of queries matched", "prot_matches", "CHECKED", "");
    }
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
      &checkBox("Percent coverage<SUP>**</SUP>", "prot_cover", "", $disabled);
    }
    unless ($export_format eq "pepXML" or $export_format eq 'mzTab') {
      &checkBox("Length in residues<SUP>**</SUP>", "prot_len", "", $disabled);
    }
    unless ($export_format eq "mzIdentML") {
      &checkBox("pI<SUP>**</SUP>", "prot_pi", "", $disabled);
    }
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect") {
      &checkBox("Taxonomy<SUP>**</SUP>", "prot_tax_str", "", $disabled);
      unless ($export_format eq 'mzTab') {
        &checkBox("Taxonomy ID<SUP>**</SUP>", "prot_tax_id", "", $disabled);
      }
    }
    unless ($export_format eq 'mzTab') {
      &checkBox("Protein sequence<SUP>**</SUP>", "prot_seq", "", $disabled);
    }
    if ($anyPepSumMatch and ($export_format eq "XML" or $export_format eq "CSV" or $export_format eq 'mzTab')) {
      my $can_have_emPAI = 0;
      if (not ($msresFlags & $msparser::ms_mascotresults::MSRES_DECOY)
            and $objResFile->getNumQueries() >= 100) {
        $can_have_emPAI = 1;
      }
      if ($can_have_emPAI) {
        &checkBox("emPAI", "prot_empai", "", $disabled);
      }
      if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
        &checkBox("Protein quantitation", "prot_quant", "", "");
      }
    }
      
    print <<'end_of_static_HTML_text_block';

    <TR>
      <TD NOWRAP COLSPAN=2><SUP>* Occasionally requires information to be retrieved from external utilities, which can be slow<BR>
      ** Always requires information to be retrieved from external utilities, which can be slow</SUP></TD>
    </TR>

end_of_static_HTML_text_block
  
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $export_format eq "mzIdentML") {
      print "   <TR>\n";
      print "     <TD NOWRAP>\n";
      print "     <H3>Peptide Match Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     <INPUT TYPE=\"checkbox\" NAME=\"peptide_master\" VALUE=1 CHECKED onClick=\"check_slaves(this, this.form)\"></TD>\n";
      print "   </TR>\n";
    
      unless ($export_format eq 'mzTab') {
        &checkBox("Experimental Mr (Da)", "pep_exp_mr", "CHECKED", "");
      }
      unless ($export_format eq 'mzTab') {
        if ($anyPepSumMatch) {
          &checkBox("Experimental charge", "pep_exp_z", "CHECKED", "");
        } else {
          &checkBox("Experimental charge", "pep_exp_z", "", "");
        }
        &checkBox("Calculated Mr (Da)", "pep_calc_mr", "CHECKED", "");
      }
      &checkBox("Mass error (Da)", "pep_delta", "CHECKED", "");
      unless ($export_format eq 'mzTab') {
        if ($anyPepSumMatch) {
          &checkBox("Start", "pep_start", "", "");
          &checkBox("End", "pep_end", "", "");
        } else {
          &checkBox("Start", "pep_start", "CHECKED", "");
          &checkBox("End", "pep_end", "CHECKED", "");
        }
        &checkBox("Number of missed cleavages", "pep_miss", "CHECKED", "");
      }
      if ($anyPepSumMatch) {
        &checkBox("Score", "pep_score", "CHECKED", "");
        &checkBox("Homology threshold", "pep_homol", "", "");
        &checkBox("Identity threshold", "pep_ident", "", "");
        unless ($export_format eq 'mzTab') { # exported as peptide score
          &checkBox("Expectation value", "pep_expect", "CHECKED", "");
        }
      }
      unless ($export_format eq 'mzTab') { # always on in mzTab
        &checkBox("Sequence", "pep_seq", "CHECKED", "");
      }
    # enabling / disabling frame checkbox is too messy 
    # because we don't want to create a summary object at this stage
      &checkBox("Frame number", "pep_frame", "", "");
      unless ($export_format eq 'mzTab') { # always on in mzTab
        &checkBox("Variable Modifications", "pep_var_mod", "CHECKED", "");
      }
      unless ($objResFile->isPMF()){
        &checkBox("Number of fragment<BR>ion matches", "pep_num_match", "", "");
        &checkBox("Query title", "pep_scan_title", "CHECKED", "");
      }
      if ($anyPepSumMatch) {
        if ($objParams->getQUANTITATION() && lc($objParams->getQUANTITATION()) ne "none") {
          &checkBox("Peptide quantitation", "pep_quant", "", "");
        }
        unless ($export_format eq 'mzTab') {
          &checkBox("Unassigned queries<BR>(peptide matches not<BR>assigned to protein hits)", "show_unassigned", "", "");
        }
      }

      &checkBox("Show duplicate peptides", "show_pep_dupes", "CHECKED", "");
  
      print "   <TR>\n";
      print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
      print "   </TR>\n";
    }
    if ($export_format eq 'mzIdentML') {
      print "   <TR>\n";
      print "     <TD NOWRAP>\n";
      print "     <H3>Optional peptide Match Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      print "     <INPUT TYPE=\"hidden\" NAME=\"peptide_master\" VALUE=1></TD>\n";
      print "   </TR>\n";

      &checkBox("Show duplicate peptides", "show_pep_dupes", "CHECKED", "");

      print "   <TR>\n";
      print "     <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>\n";
      print "   </TR>\n";
    }
    
    unless ($export_format eq "pepXML" or $export_format eq "DTASelect" or $objResFile->isPMF() or $export_format eq 'mzTab') {
      print "   <TR>\n";
      print "     <TD NOWRAP>\n";
      print "     <H3>Query Level Information</H3></TD>\n";
      print "     <TD NOWRAP VALIGN=top>\n";
      unless ($export_format eq "mzIdentML") {
        print "     <INPUT TYPE=\"checkbox\" NAME=\"query_master\" VALUE=1 onClick=\"check_slaves(this, this.form)\"></TD>\n";
      } else {
        print "     &nbsp;</TD>\n";
      }
      print "   </TR>\n";
  
      my $sDisabelled = "";
      unless ($export_format eq "mzIdentML") {
        $sDisabelled = "disabled";
        &checkBox("Query title", "query_title", "", $sDisabelled);
        &checkBox("seq(), comp(), tag(), etc.", "query_qualifiers", "", $sDisabelled);
        &checkBox("Query level<BR>search parameters", "query_params", "", $sDisabelled);
        &checkBox("MS/MS Peak lists", "query_peaks", "", $sDisabelled);
      }
      unless ($export_format eq "mzIdentML") {
        &checkBox("Raw peptide match data", "query_raw", "", $sDisabelled);
      } else {
        &checkBox("Matched Fragment Ions", "query_raw", "", $sDisabelled);
        &checkBox("Export data for all Queries", "query_all", "", $sDisabelled);
      }
    }

  }  # closure of unless ($export_format eq "MascotDAT" || $export_format eq "MGF") {

  my $submission_scriptlet_cmdline = 'this.form.do_export.value=-1; this.form.submit(); return true';
  my $submission_scriptlet_data = 'this.form.do_export.value=1; this.form.submit(); return true';

  print <<"end_of_static_HTML_text_block";

    <TR>
      <TD NOWRAP ALIGN=RIGHT COLSPAN=2>&nbsp;</TD>
    </TR>
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=LEFT>
        <INPUT TYPE="button" VALUE="Show command line arguments" onClick="$submission_scriptlet_cmdline">
      </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <INPUT TYPE="button" VALUE="Export search results" onClick="$submission_scriptlet_data">
      </TD>
    </TR>
    
  </TABLE>
  
</FORM>

end_of_static_HTML_text_block

  &common_subs::printFooter;

  print "</BODY>\n";
  print "</HTML>\n";

}

###############################################################################
# &downloadControls($FDRsuccess, $selectedFDR, $sigThreshold_adjusted)
# Print the downloading form displayed after successful caching. If the
# optional arguments $FDRsuccess, $selectedFDR and $sigThreshold_adjusted are
# given, they will be reported as the result of trying to find the target FDR.
# globals:
# my($myName, $objResFile, $objParams, $numHits, $sigThreshold, 
# $ignoreIonsScoreBelow, $ShowSubSets, $thisScript, $RequireBoldRed, 
# %urlParams, $mudpitSwitch);
###############################################################################

sub downloadControls {
  my ($FDRsuccess, $selectedFDR, $sigThreshold_adjusted) = @_;

  common_subs::printHeader1("Matrix Science - Mascot - Export search results");
  common_subs::printHeader2("", "Export search results", 1, $session);

  if (defined $FDRsuccess) {
    my $fdr = $selectedFDR > 0 ? sprintf("%.2f", 100 * $selectedFDR) : $selectedFDR;
    my $type = $use_homology ? 'at homology level' : 'at identity level';
    if ($FDRsuccess) {
      print <<"end_of_static_HTML_text_block";
  <P>Target FDR is $targetFDR% ($type).
  <BR>Closest FDR found is <STRONG>$fdr%</STRONG> at significance level p&lt;$sigThreshold_adjusted.
  </P>
end_of_static_HTML_text_block
    } else {
      print <<"end_of_static_HTML_text_block";
  <P>Target FDR is $targetFDR% ($type).
  <BR>However, target FDR could not be reached; closest FDR found is $fdr%. Significance threshold has not been adjusted and remains at p&lt;$sigThreshold.
  </P>
end_of_static_HTML_text_block
    }
  }

  print <<"end_of_static_HTML_text_block";
  <FORM ACTION="./$myName" NAME="Download" METHOD="POST" ENCTYPE="multipart/form-data">
  <INPUT TYPE="hidden" NAME="generate_file" VALUE=1>
end_of_static_HTML_text_block

# Pass through all parameters.
  while (my($key, $value) = each %urlParams) {
    print "  <INPUT TYPE=\"hidden\" NAME=\"$key\" VALUE=\"" . common_subs::htmlEscape($thisScript->param($urlParams{$key})) . "\">\n";
  }

  my $msg = 'Exporting is ready to be started.';

  if (my $id = $thisScript->param($urlParams{'_cachewarnings'})) {
      my $session = msparser::ms_session->new($id);
      my @warnings = ();

      if ($session->isValid) {
          my $num = $session->getIntParam('num_warnings');

          for my $i (0 .. $num-1) {
              push @warnings, $session->getStringParam('warning_'.$i);
          }

          $session->destroy();
      }

      if (@warnings) {
          $msg = '<p>Caching the exported data did not succeed:</p><ul><li>';
          $msg .= join('</li><li>', map { 
                  '<FONT COLOR=#FF0000>'.common_subs::htmlEscape($_).'</FONT>'
              } @warnings
          );

          $msg .= '</li></ul> <p>This means downloading may take a long time while the file is opened and parsed.</p>';

          $msg .= '<p>Exported data is ready to be downloaded.</p>';
      }
  }

  print <<"end_of_static_HTML_text_block";
  <TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>
  
    <TR>
      <TD NOWRAP ALIGN=LEFT>$msg</TD>
    </TR>
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=LEFT>
        <INPUT TYPE="submit" VALUE="Download">
      </TD>
    </TR>
    
  </TABLE>
  
</FORM>

end_of_static_HTML_text_block

  common_subs::printFooter();

  print "</BODY>\n";
  print "</HTML>\n";

}

###############################################################################
# &checkBox()
# Print row of table for a checkbox option
###############################################################################

sub checkBox {
  
  my($description, $name, $checked, $disabled, $onclick) = @_;

  if (defined $onclick) {
    $onclick = sprintf('onClick="%s"', $onclick);
  } else {
    $onclick = '';
  }
  
  print <<"end_of_static_HTML_text_block";
    <TR>
      <TD BGCOLOR=#EEEEFF NOWRAP ALIGN=RIGHT>
        $description </TD>
      <TD BGCOLOR=#EEEEFF NOWRAP>
        <INPUT $disabled TYPE="checkbox" NAME="$name" VALUE=1 $checked $onclick>
      </TD>
    </TR> 
end_of_static_HTML_text_block
  
}

sub markExportableQueries {
  my ($queryAll, $isPMF, $sameset, $exported_queryList) = @_;

  unless ($queryAll or $isPMF) {
    for my $i (1 .. $sameset->getNumPeptides()) {
      $$exported_queryList[$sameset->getPeptideQuery($i)] = 1;
    }
  }
}

###############################################################################
# &outputHit()
# export a hit block
# $_[0] protein object
# $_[1] hit number
# $_[2] sameset ordering within hit $_[1]
# $_[3] 1 if duplicate peptides should be filtered out
# $_[4] cache for duplicate_status_main()
# globals:
# my($objSummary, %urlParams, $thisScript, $export_format, $group_family);
###############################################################################

sub outputHit {

  my $objProtein = shift;
  my $hitNum = shift;
  my $sameset_ordering = shift;
  my $filter_duplicates = shift;
  my $dupe_state_cache = shift;
  
  if ($export_format eq "XML") {
    print "<hit number=\"$hitNum\">\n";
  }

  if ($group_family) {
    
  # The lead protein for the family
    my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
    outputProtein($sameset[0], $hitNum, 1, 1, $filter_duplicates, $dupe_state_cache);
  
  # Other proteins that match the same set of peptides
    if ($ShowSameSets) {
      outputProtein($_, $hitNum, 0, 1, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
    }
    
  # Other family members
    my $j = 1;
    while (my $objProtein = $objSummary->getNextFamilyProtein($hitNum, $j++)) {
      my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
      outputProtein($sameset[0], $hitNum, 1, $j, $filter_duplicates, $dupe_state_cache);

    # And their same-sets
      if ($ShowSameSets) {
        outputProtein($_, $hitNum, 0, $j, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
      }
    }
      
  # sub-sets and intersections for the entire family
    if ($ShowSubSets > 0) {
      my $j = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($hitNum, $j++, 1)) {
        my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
        outputProtein($sameset[0], $hitNum, 0, -1, $filter_duplicates, $dupe_state_cache);

      # And their same-sets
        if ($ShowSameSets) {
          outputProtein($_, $hitNum, 0, -1, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
        }
      }
    }    
    
  } else {
  # The primary protein hit
    my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, $sameset_ordering);
    outputProtein($sameset[0], $hitNum, 1, 0, $filter_duplicates, $dupe_state_cache);
  
  # Other proteins that match the same set of peptides
    if ($ShowSameSets) {
      outputProtein($_, $hitNum, 0, 0, $filter_duplicates, $dupe_state_cache) for @sameset[1 .. $#sameset];
    }
    
  # Other proteins that match a sub-set of peptides
    if ($ShowSubSets > 0) {
      my $i = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($hitNum, $i++)) {
        outputProtein($objProtein, $hitNum, 0, 0, $filter_duplicates, $dupe_state_cache);
      }
    }
  
  }
  
  if ($export_format eq "XML") {
    print "</hit>\n";
  } 

}
  
###############################################################################
# &outputProtein()
# export a protein block
# $_[0] protein object
# $_[1] hit or family number
# $_[2] flag set for anchor protein
# $_[3] family member number (0 if not family grouping, -1 if family subsets)
# $_[4] 1 if duplicate peptides should be filtered out
# $_[5] cache for duplicate_status_main()
# globals:
# my(%urlParams, $export_format, $thisScript, $fastaLen);
###############################################################################

sub outputProtein {

  my $objProtein = shift;
  my $hitNum = shift;
  my $primary = shift;
  my $member = shift;
  my $filter_duplicates = shift;
  my $dupe_state_cache = shift;
  
  my $dbNumPrefix = "";
  if ($objParams->getNumberOfDatabases > 1 && !$objProtein->isPMFMixture) {
    $dbNumPrefix = $objProtein->getDB . "::";
  }

  if ($export_format eq "XML") {
    print "<protein accession=\"" . $dbNumPrefix 
        . common_subs::preferredAccession($objProtein->getDB(), $objProtein->getAccession()) . "\"";
    if ($member > 0) {
      print " member=\"" . $member . "\"";
    }
    print ">\n";
  }
  &printProteinRow($objProtein, $hitNum, $member);
  
  if ($export_format eq "CSV") {
    print $delimiter;
  }

  if ($thisScript->param($urlParams{'peptide_master'})) {
    printPeptideRows($objProtein, $hitNum, $primary, $member, $filter_duplicates, $dupe_state_cache);
  } else {
    if ($export_format eq "CSV") {
      print "\n";
    }
  }

# quantitation information is only output for primary protein hit  
  if ($primary) {
  # and only available if we have looped through the peptides
    if ($objSummary->anyEmPAI and $thisScript->param($urlParams{'peptide_master'})) {
      if ($thisScript->param($urlParams{'prot_empai'})) {
        my $L = -1;
        if ($objSummary->getProteinMass($objProtein->getAccession(), $objProtein->getDB)) {
          # no need to provide a length
        } else {
          $L = getProteinLen($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
          $L = -1 if not $L;
        }
        my $emPAI = $objSummary->getProteinEmPAI($objProtein->getAccession(), $objProtein->getDB(), $L);

        if ($emPAI > 0) {
        # insert emPAI information
          if ($export_format eq "CSV") {
            print $delimiter . "\"emPAI\"" . $delimiter . sprintf("%.2f", $emPAI);
          } elsif ($export_format eq "XML") {
            print "<prot_empai>" . sprintf("%.2f", $emPAI) . "</prot_empai>\n";
          }
        } else {
          if ($export_format eq "CSV") {
            print $delimiter . $delimiter;
          }
        }
      }
    # insert quantitation summary information for the primary hit
      if ($thisScript->param($urlParams{'prot_quant'}) and $objQuantMethod) {
        if ($export_format eq "CSV") {
          print $delimiter . "\"Quantitation summary for protein\"";
        }
        my $proteinRatioType = lc($objMS2Quant->getQuantitationMethod()->getProteinRatioType());
        my @ratio_names = @{ $objMS2Quant->getCombinedReportRatioNames() };
        for my $name (@ratio_names) {
          my $ratio = $objMS2Quant->getProteinRatio($objProtein->getAccession(), $objProtein->getDB(), $name);
          if ($export_format eq "CSV") {
            if ($ratio->isMissing()) {
              print $delimiter
                . '"' . &noQuotes($name) . '"'
                . $delimiter
                . '"---"'
                . $delimiter
                . '"---"'
                . $delimiter
                . '"---"'
                . $delimiter
                . '"---"'
                . $delimiter
                ;
            } else {
              my ($sd, $p_value, $is_sig) = ('', '', '');
              if ($ratio->getStandardDeviation() < 0) {
                # Ignore reason why unavailable.
                $sd = 'N/A';
              } else {
                $sd = sprintf($ratio->isLogNormal ? '%.3f' : '[%.3f]', $ratio->getStandardDeviation());
              }
              if ($ratio->getNormalityPvalue() < 0) {
                # Ignore reason why unavailable.
                $p_value = 'N/A';
              } else {
                if ($ratio->isLogNormal) {
                  if ($ratio->getHypothesisPvalue() < 0) {
                    # Ignore reason why unavailable.
                    $p_value = 'N/A';
                  } else {
                    $p_value = sprintf('%.4g', $ratio->getHypothesisPvalue());
                    $is_sig = ($ratio->isLogNormal && $ratio->isSignificant() ? '*' : '');
                  }
                }
              }
              print $delimiter
                . '"' . &noQuotes($name) . '"'
                . $delimiter
                . '"' . sprintf('%.3f', $ratio->getValue()) . '"'
                . $delimiter
                . '"' . $ratio->getSampleSize() . '"'
                . $delimiter
                . '"' . $sd . '"'
                . $delimiter
                . '"' . $is_sig . '"'
                . $delimiter
                . '"' . $p_value . '"'
                ;
            }
          } elsif ($export_format eq "XML") {
            if ($ratio->isMissing()) {
              print '<quant_prot_ratio name="' 
                . &noXmlTag($name) 
                . '" ratio="---" n="---" sd="---" significant="" p-value=""'
                . " />\n"
                ;
            } else {
              my ($sd, $p_value, $is_sig) = ('', '', '');
              if ($ratio->getStandardDeviation() < 0) {
                # Ignore reason why unavailable.
                $sd = 'N/A';
              } else {
                $sd = sprintf($ratio->isLogNormal ? '%.3f' : '[%.3f]', $ratio->getStandardDeviation());
              }
              if ($ratio->getNormalityPvalue() < 0) {
                # Ignore reason why unavailable.
                $p_value = 'N/A';
              } else {
                if ($ratio->isLogNormal) {
                  if ($ratio->getHypothesisPvalue() < 0) {
                    # Ignore reason why unavailable.
                    $p_value = 'N/A';
                  } else {
                    $p_value = sprintf('%.4g', $ratio->getHypothesisPvalue());
                    $is_sig = ($ratio->isLogNormal && $ratio->isSignificant() ? '*' : '');
                  }
                }
              }
              print '<quant_prot_ratio name="' 
                . &noXmlTag($name) 
                . '" ratio="' . sprintf('%.3f', $ratio->getValue())
                . '" n="' . $ratio->getSampleSize()
                . '" sd="' . $sd
                . '" significant="' . $is_sig
                . '" p-value="' . $p_value
                . "\" />\n"
                ;
            }
          }
        }
      }
    }
  }
  
  if ($export_format eq "XML") {
    print "</protein>\n";
  } else {
    print "\n"; 
  }

}

###############################################################################
# &printPeptideRows()
# export a peptides block
# $_[0] protein object
# $_[1] hit number
# $_[2] flag set for anchor protein
# $_[3] family member number (0 if not family grouping, -1 if family subsets)
# $_[4] 1 if duplicates should be filtered out
# $_[5] cache for duplicate_status_main()
# globals:
# my($objSummary, $export_format);
###############################################################################

sub printPeptideRows {
  
  my $objProtein = shift;
  my $hitNum = shift;
  my $primary = shift;
  my $member = shift;
  my $filter_duplicates = shift;
  my $dupe_state_cache = shift;

  my $isPMF = $objResFile->isPMF();
  
  my @selected_qp = _select_nonduplicate_protein_qp(
    $dupe_state_cache, $filter_duplicates, $objResFile, $objParams,
    $objSummary, $objProtein
  );

  my $needSpacer = 0;
  my $eol = "";   # need to suppress final EOL for CSV so as to add quant info in-line
  for (@selected_qp) {
    my ($i, $queryNum, $rank, $objPeptide) = @$_;
    print $eol;
    if ($export_format eq "CSV" && $needSpacer) {
      printProteinRow($objProtein, $hitNum, $member);
      print $delimiter;
    # print $hitNum . $delimiter x $protColumnCount;
    } else {
      $needSpacer = 1;
    }
    if ($export_format eq "XML") {
      print "<peptide query=\"$queryNum\"";
      if ($objPeptide->getAnyMatch() && !$isPMF) {
        print " rank=\"" . $objPeptide->getPrettyRank() . "\"";
        if ($objProtein->getPeptideIsBold($i)) {
          print " isbold=\"1\"";
        } else {
          print " isbold=\"0\"";
        }
        if ($objSummary->isPeptideUnique($queryNum, $rank)) {
          print " isunique=\"1\"";
        } else {
          print " isunique=\"0\"";
        }
      }
      print ">\n";
    }
    my $string = fmtPeptideDetails($objPeptide, $queryNum, $i, $objProtein, 1, $isPMF);
    if ($primary 
      && $objQuantMethod
      && ($thisScript->param($urlParams{'pep_quant'})
      || $thisScript->param($urlParams{'prot_quant'}))) {
      $string .= fmtQuantDetails($objPeptide, $queryNum, $i, $objProtein, $hitNum);
    }

    chop $string;
    if ($export_format eq "XML") {
      $string =~ s/>$delimiter</>\n</g;
    }
    if ($string) {
      print $string;
    }
    if ($export_format eq "XML") {
      print "\n</peptide>\n";
    } 
    if ($export_format eq "CSV") {
      $eol = "\n";
    }
  }
}

sub _select_nonduplicate_protein_qp {
  my (
    $dupe_state_cache, $filter_duplicates, $objResFile, $objParams,
    $objSummary, $objProtein
  ) = @_;

  my @selected_qp = ();

  if ($filter_duplicates) {
    my %duplicate_groups = ();

    for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
      next if ($objProtein->getPeptideDuplicate($i) == $msparser::ms_protein::DUPE_DuplicateSameQuery);
      my $queryNum = $objProtein->getPeptideQuery($i);
      my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
      my $key = duplicate_group_main($dupe_state_cache, $objResFile, $objParams, $objSummary, $objPeptide, $quantProtocolType);
      push @{ $duplicate_groups{$key} }, $objPeptide;
    }
  
    for (values %duplicate_groups) {
      my $arr = sort_duplicate_group($_);
      my $master_peptide = $$arr[0];
      my ($queryNum, $rank) = ($master_peptide->getQuery(), $master_peptide->getRank());
      my $i = $objProtein->getPepNumber($queryNum, $rank);
      push @selected_qp, [$i, $queryNum, $rank, $master_peptide];
    }

    @selected_qp = sort { $$a[0] <=> $$b[0] } @selected_qp;
  } else {
    for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
      next if ($objProtein->getPeptideDuplicate($i) == $msparser::ms_protein::DUPE_DuplicateSameQuery);
      my ($queryNum, $rank) = ($objProtein->getPeptideQuery($i), $objProtein->getPeptideP($i));
      my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
      push @selected_qp, [$i, $queryNum, $rank, $objPeptide];
    }
  }

  return @selected_qp;
}

###############################################################################
# &outputUnassigned()
# export an unassigned block
# $_[0] 
# globals:
# my($objSummary, $idForFatalError, $export_format);
###############################################################################

sub outputUnassigned {
  
  if ($export_format eq "XML") {
    print "<unassigned>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Peptide matches not assigned to protein hits\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  $objSummary->createUnassignedList($msparser::ms_mascotresults::QUERY);
  if ($objSummary->getNumberOfUnassigned() > 0) {
    my $isPMF = $objResFile->isPMF();

  # enumerate the peptides in the unassigned list
    for (my $i = 1; $i <= $objSummary->getNumberOfUnassigned(); $i++) {
      my $objPeptide = $objSummary->getUnassigned($i);
      if (defined($objPeptide)) {
        if ($export_format eq "CSV") {
          print $delimiter x $protColumnCount;
        }
        if ($export_format eq "XML") {
          print "<u_peptide query=\"" . $objPeptide->getQuery() . "\"";
          if ($objPeptide->getAnyMatch() && !$isPMF) {
            print " rank=\"" . $objPeptide->getPrettyRank() . "\"";
            if ($objSummary->getUnassignedIsBold($i)) {
              print " isbold=\"1\"";
            } else {
              print " isbold=\"0\"";
            }
            print " isunique=\"0\"";
          }
          print ">\n";
        }
        my $string = fmtPeptideDetails($objPeptide, $objPeptide->getQuery(), $i, "", 1, $isPMF);
        chop $string;
        if ($export_format eq "XML") {
          $string =~ s/>$delimiter</>\n</g;
        }
        if ($string) {
          print $string . "\n";
        }
        if ($export_format eq "XML") {
          print "</u_peptide>\n";
        } 
      } else {
        die("Undefined peptide object ", $idForFatalError);
      }
    }
  }
  
  if ($export_format eq "XML") {
    print "</unassigned>\n";
  } 

}

###############################################################################
# &printProteinRow()
# output fields for a protein match
# $_[0] protein object
# $_[1] hit number
# $_[2] family member number (0 if not family grouping, -1 if family subsets)
# globals:
# my($objSummary, $fileIn, %urlParams, $objParams, $export_format, 
# $thisScript);
###############################################################################

sub printProteinRow {
  
  my $objProtein = shift;
  my $hitNum = shift;
  my $member = shift;
  
  $protColumnCount = 0;

  my $dbNumPrefix = "";
  if ($objParams->getNumberOfDatabases > 1 && !$objProtein->isPMFMixture) {
    $dbNumPrefix = $objProtein->getDB . "::";
  }

  my $preferred_acc = common_subs::preferredAccession($objProtein->getDB(), $objProtein->getAccession);

  my $string = "";
  if ($export_format eq "CSV") {
    $string .=  $hitNum . $delimiter;
    if ($member > 0) {
      $string .=  $member . $delimiter;
    } elsif ($member < 0) {
      $string .=  $delimiter;
    }
    $string .= "\"" . &noQuotes($dbNumPrefix . $preferred_acc) . "\"$delimiter";
  }
  $protColumnCount++;
  if ($member != 0) {
    $protColumnCount++;
  }
  $protColumnCount++;
  if ($thisScript->param($urlParams{'prot_desc'})) {
    $string .= &formatElement($export_format, "", 1, "prot_desc", 
      common_subs::mustGetProteinDescription($objSummary, $objParams, $shipper, $preferred_acc, $objProtein->getDB)) . $delimiter;
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_score'})) {
    $string .= &formatElement($export_format, "", 0, "prot_score", 
      sprintf("%.0f", $objProtein->getScore())) . $delimiter;
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_thresh'})) {
    $string .= &formatElement($export_format, "", 0, "prot_thresh", 
      $objSummary->getProteinThreshold(1 / $sigThreshold)) . $delimiter;
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_expect'})) {
    $string .= &formatElement($export_format, "", 0, "prot_expect", 
      sprintf("%.2g", 1 / $objSummary->getProbOfProteinBeingRandomMatch($objProtein->getScore()))) . $delimiter;
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_mass'})) {
    if (mustGetProteinMass($objProtein->getAccession(), $objSummary, $objParams, $objProtein->getFrame(), $objProtein->getDB)) {
      $string .= &formatElement($export_format, "", 0, "prot_mass", 
        sprintf("%.0f", mustGetProteinMass($objProtein->getAccession(), $objSummary, $objParams, $objProtein->getFrame(), $objProtein->getDB))) . $delimiter;
    } else {
      if ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    }
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_matches'})) {
    $string .= &formatElement($export_format, "", 0, "prot_matches", 
      $objProtein->getNumDisplayPeptides(0)) . $delimiter;
    $protColumnCount++;
    if ($anyPepSumMatch) {
      $string .= &formatElement($export_format, "", 0, "prot_matches_sig", 
        $objProtein->getNumDisplayPeptides(1)) . $delimiter;
      $protColumnCount++;
      $string .= &formatElement($export_format, "", 0, "prot_sequences", 
        $objProtein->getNumDistinctPeptides(0, $msparser::ms_protein::DPF_SEQUENCE)) . $delimiter;
      $protColumnCount++;
      $string .= &formatElement($export_format, "", 0, "prot_sequences_sig", 
        $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE)) . $delimiter;
      $protColumnCount++;
    }
  }
  if ($thisScript->param($urlParams{'prot_cover'}) || $thisScript->param($urlParams{'prot_len'})) {
    my $length = getProteinLen($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
    if ($thisScript->param($urlParams{'prot_cover'})) {
      my $coverage = "";
      if ($length) {
      # accurate
        $coverage = $objProtein->getCoverage() * 100 / $length;
      } else {
        my $protMass = mustGetProteinMass($objProtein->getAccession(), $objSummary, $objParams, $objProtein->getFrame(), $objProtein->getDB);
        if ($protMass) {
        # approximate
          $coverage = $objProtein->getCoverage() * 100 * 110 / $protMass;
        }
      }
      if ($coverage) {
        $string .= &formatElement($export_format, "", 0, "prot_cover", 
          sprintf("%.1f", $coverage)) . $delimiter;
      } else {
        if ($export_format eq "CSV") {        
          $string .= $delimiter;
        }
      }
      $protColumnCount++;
    }
    if ($thisScript->param($urlParams{'prot_len'})) {
      if ($length) {
        $string .= &formatElement($export_format, "", 0, "prot_len", $length) . $delimiter;
      } else {
        if ($export_format eq "CSV") {        
          $string .= $delimiter;
        }
      }
      $protColumnCount++;
    }
  }
  if ($thisScript->param($urlParams{'prot_pi'})) {
    my $pI = getProteinpI($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
    if ($pI) {
      $string .= &formatElement($export_format, "", 0, "prot_pi", $pI) . $delimiter;
    } else {
      if ($export_format eq "CSV") {        
        $string .= $delimiter;
      }
    }
    $protColumnCount++;
  }
  if ($thisScript->param($urlParams{'prot_tax_str'}) || $thisScript->param($urlParams{'prot_tax_id'})) {
    my ($taxID, $species) = common_subs::getProteinTaxonomy($objSummary, $objParams, $preferred_acc, $objProtein->getDB);
    if ($thisScript->param($urlParams{'prot_tax_str'})) {
      if ($species) {
        $string .= &formatElement($export_format, "", 1, "prot_tax_str", $species) . $delimiter;
      } else {
        if ($export_format eq "CSV") {        
          $string .= $delimiter;
        }
      }
      $protColumnCount++;
    }
    if ($thisScript->param($urlParams{'prot_tax_id'})) {
      if ($taxID) {
        $string .= &formatElement($export_format, "", 0, "prot_tax_id", $taxID) . $delimiter;
      } else {
        if ($export_format eq "CSV") {        
          $string .= $delimiter;
        }
      }
      $protColumnCount++;
    }
  }
  if ($thisScript->param($urlParams{'prot_seq'})) {
    my $seq = mustGetProteinSeq($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
    if ($seq) {
      $string .= &formatElement($export_format, "", 0, "prot_seq", $seq) . $delimiter;
    } else {
      if ($export_format eq "CSV") {        
        $string .= $delimiter;
      }
    }
    $protColumnCount++;
  }
  
  chop $string;
  if ($export_format eq "XML") {
    $string =~ s/>$delimiter</>\n</g;
  }
  if ($string) {
    print $string;
    if ($export_format eq "XML") {
      print "\n";
    }
  }

}

###############################################################################
# &fmtPeptideDetails()
# returns fields for a single peptide match
# $_[0] peptide object
# $_[1] query number
# $_[2] peptide number
# $_[3] protein object (empty for unassigned list)
# $_[4] 1 to begin line with query number
# $_[5] if the results file is PMF
# $_[6] is the site analysis data (or undef if missing)
# globals:
# my($objSummary, %urlParams, $objResFile, $massDP, $thisScript, 
# $export_format, $sigThreshold);
###############################################################################

sub fmtPeptideDetails {
  
  my ($objPeptide, $queryNum, $i, $objProtein, $printQueryNum, $isPMF, $siteAnalysisData) = @_;

  my $string = "";
  if ($export_format eq "CSV") {
    if ($printQueryNum) {
      $string .= $queryNum . $delimiter;
    }
    if ($objPeptide->getAnyMatch() && !$isPMF) {
      $string .= $objPeptide->getPrettyRank() . $delimiter;
      if ($printQueryNum) {
        if ($objProtein) {
          if ($objProtein->getPeptideIsBold($i)) {
            $string .= "1" . $delimiter;
          } else {
            $string .= "0" . $delimiter;
          }
          if ($objSummary->isPeptideUnique($queryNum, $objPeptide->getRank)) {
            $string .= "1" . $delimiter;
          } else {
            $string .= "0" . $delimiter;
          }
        } elsif ($objSummary->getUnassignedIsBold($i)) {
          $string .= "1" . $delimiter;
          $string .= "0" . $delimiter;
        } else {
          $string .= "0" . $delimiter;
          $string .= "0" . $delimiter;
        }
      } else {
        $string .= $delimiter . $delimiter;
      }
    } else {
      $string .= $delimiter . $delimiter . $delimiter;
    }
  }
  $string .= &formatElement($export_format, "", 0, "pep_exp_mz",
    sprintf("%." . $massDP . "f", $objPeptide->getObserved())) . $delimiter;
  if ($thisScript->param($urlParams{'pep_exp_mr'})) {
    my $mrExpt = $objResFile->getObservedMrValue($queryNum);
    if ($objPeptide->getAnyMatch) {
      $mrExpt = $objPeptide->getMrExperimental;    # only works if there's a match
    }
    $string .= &formatElement($export_format, "", 0, "pep_exp_mr",
      sprintf("%." . $massDP . "f", $mrExpt)) . $delimiter;
  }
  if ($thisScript->param($urlParams{'pep_exp_z'})) {
    $string .= &formatElement($export_format, "", 0, "pep_exp_z", $objPeptide->getCharge()) . $delimiter;
  }
  if ($thisScript->param($urlParams{'pep_calc_mr'})) {
    if ($objPeptide->getAnyMatch()) {
      $string .= &formatElement($export_format, "", 0, "pep_calc_mr", 
        sprintf("%." . $massDP . "f", $objPeptide->getMrCalc())) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_delta'})) {
    if ($objPeptide->getAnyMatch()) {
      $string .= &formatElement($export_format, "", 0, "pep_delta", 
        sprintf("%." . $massDP . "f", $objPeptide->getDelta())) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_start'})) {
    if ($objProtein) {
      $string .= &formatElement($export_format, "", 0, "pep_start", 
        $objProtein->getPeptideStart($i)) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_end'})) {
    if ($objProtein) {
      $string .= &formatElement($export_format, "", 0, "pep_end", 
        $objProtein->getPeptideEnd($i)) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_miss'})) {
    if ($objPeptide->getAnyMatch()) {
      $string .= &formatElement($export_format, "", 0, "pep_miss", 
        $objPeptide->getMissedCleavages()) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_score'})) {
    if ($objPeptide->getAnyMatch()) {
      $string .= &formatElement($export_format, "", 0, "pep_score", 
        sprintf("%.2f", $objPeptide->getIonsScore)) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_homol'})) {
    if ($objPeptide->getAnyMatch() &&
      $objSummary->getHomologyThreshold($queryNum, 1 / $sigThreshold)) {
      if ($objResFile->isErrorTolerant() && $objSummary->getPeptide($queryNum, 1)->getIsFromErrorTolerant()) {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      } else {
        $string .= &formatElement($export_format, "", 0, "pep_homol", 
          $objSummary->getHomologyThreshold($queryNum, 1 / $sigThreshold)) . $delimiter;
      }
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_ident'})) {
    if ($objPeptide->getAnyMatch() &&
      $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold) &&
      $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold) > -1) {
      if ($objResFile->isErrorTolerant() && $objSummary->getPeptide($queryNum, 1)->getIsFromErrorTolerant()) {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      } else {
        $string .= &formatElement($export_format, "", 0, "pep_ident", 
          $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold)) . $delimiter;
      }
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_expect'})) {
    if ($objPeptide->getAnyMatch()) {
      if ($objResFile->isErrorTolerant() && $objSummary->getPeptide($queryNum, 1)->getIsFromErrorTolerant()) {
        if ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      } else {
        $string .= &formatElement($export_format, "", 0, "pep_expect", 
          sprintf("%.2g", $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $queryNum))) . $delimiter;
      }
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_seq'})) {
    if ($objPeptide->getAnyMatch()) {
      if ($objProtein) {
        my $res = $objProtein->getPeptideResidueBefore($i);
        $res =~ s/@/-/;
        $res =~ s/\?//;
        if ($res) {
          $string .= &formatElement($export_format, "", 0, "pep_res_before", $res) . $delimiter;
        } elsif ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $string .= &formatElement($export_format, "", 0, "pep_seq", 
        $objPeptide->getPeptideStr()) . $delimiter;
      if ($objProtein) {
        my $res = $objProtein->getPeptideResidueAfter($i);
        $res =~ s/@/-/;
        $res =~ s/\?//;
        if ($res) {
          $string .= &formatElement($export_format, "", 0, "pep_res_after", $res) . $delimiter;
        } elsif ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
    } elsif ($export_format eq "CSV") {
      $string .= "$delimiter$delimiter$delimiter";
    }
  }
  if ($thisScript->param($urlParams{'pep_frame'})) {
    if ($objProtein && $objProtein->getPeptideFrame($i) > 0) {
      $string .= &formatElement($export_format, "", 0, "pep_frame", $objProtein->getPeptideFrame($i)) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_var_mod'})) {
    if ($objPeptide->getAnyMatch()) {
      $string .= &formatElement($export_format, "", 1, "pep_var_mod", 
        $objSummary->getReadableVarMods($queryNum, $objPeptide->getRank())) . $delimiter;
    # need to re-format string from 00000012000 to 0.000001200.0
    # to prevent Excel from re-formatting it to 12000
      my $tmpvar = $objPeptide->getVarModsStr();
      my $tmpsum = $objPeptide->getSummedModsStr();
      my $tmplocal = $objPeptide->getLocalModsStr();
      for ($tmpvar, $tmpsum, $tmplocal) {
        if (/[1-9A-Za-z]/) {
          s/(.)(.*)(.)/$1\.$2\.$3/;
        } else {
          $_ = "";
        }
      }
      $string .= &formatElement($export_format, "", 0, "pep_var_mod_pos", $tmpvar) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "pep_summed_mod_pos", $tmpsum) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "pep_local_mod_pos", $tmplocal) . $delimiter;

      if ($siteAnalysisData) {
        if (exists $$siteAnalysisData{ $objPeptide->getRank }) {
          my $analysis = $$siteAnalysisData{ $objPeptide->getRank() };
          $string .= &formatElement($export_format, "", 0, "pep_var_mod_conf", sprintf("%.2f%%", 100 * $$analysis{'confidence'}) ) . $delimiter;
        } elsif ($export_format eq "CSV") {
          $string .= $delimiter;
        }
      }
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter . $delimiter . $delimiter;
      $string .= $delimiter if $siteAnalysisData;
    }
  }
  if ($thisScript->param($urlParams{'pep_num_match'})) {
    if ($objPeptide->getAnyMatch()) {
      $string .= &formatElement($export_format, "", 0, "pep_num_match", 
        $objPeptide->getNumIonsMatched()) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
  }
  if ($thisScript->param($urlParams{'pep_scan_title'})) {
  # creating query objects is much too slow
    my $title = $objResFile->getQuerySectionValueStr($queryNum, "Title");
    $title =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex($1))/eg;
    $string .= &formatElement($export_format, "", 1, "pep_scan_title", $title) . $delimiter;
  }
  
  return $string;

}

###############################################################################
# &fmtQuantDetails()
# returns quant fields for a single peptide match
# $_[0] peptide object
# $_[1] query number
# $_[2] peptide number
# $_[3] protein object
# $_[4] hit number
# globals:
# my($objSummary, %urlParams, $objResFile, $massDP, $thisScript, 
# $export_format, $sigThreshold);
###############################################################################

sub fmtQuantDetails {
  
  my ($objPeptide, 
    $queryNum,
    $i,
    $objProtein,
    $hitNum,
    ) = @_;

  my @ratio_names = @{ $objMS2Quant->getCombinedReportRatioNames() };
  my $key = msparser::ms_peptide_quant_key->new($objPeptide->getQuery(), $objPeptide->getRank());

  my @intensityValues = @{ $objMS2Quant->getComponentIntensities($key) };
  my @ratioValues = ();

  if (0 == @intensityValues) {
    @ratioValues = ('---') x @ratio_names;
  } else {
    for my $i (0 .. $#ratio_names) {
      my $ratio = $objMS2Quant->getPeptideRatio($key, $ratio_names[$i]);
      if ($ratio->isMissing()) {
        $ratioValues[$i] = '---';
      } elsif ($ratio->isInfinite()) {
        $ratioValues[$i] = '###';
      } else {
        $ratioValues[$i] = sprintf("%.3f", $ratio->getValue());
      }
    }
  }

  my $string = "";
  if ($thisScript->param($urlParams{'pep_quant'})) {
    for my $i (0 .. $#ratio_names) {
      my $name = $ratio_names[$i];
      if ($export_format eq "CSV") {
        $string .= "\"" . &noQuotes($name) . "\"" . $delimiter;
        if ($ratioValues[$i] eq "---" || $ratioValues[$i] eq "###") {
          $string .= "\"" . $ratioValues[$i] . "\"" . $delimiter;
        } else {
          $string .= $ratioValues[$i] . $delimiter;
        }
      } elsif ($export_format eq "XML") {
        $string .= "<quant_pep_ratio name=\""
          . &noXmlTag($name)
          . "\" ratio=\""
          . $ratioValues[$i]
          . "\" />" 
          . $delimiter;
      }
    }
    
    for (my $i = 0; $i < $objQuantMethod->getNumberOfComponents(); $i++) {
      my $name = $objQuantMethod->getComponentByNumber($i)->getName();
      my $intensity = $intensityValues[$i];
      $intensity = '---' if (not defined $intensityValues[$i] or $intensityValues[$i] eq '');
      if ($export_format eq "CSV") {
        $string .= "\"" . &noQuotes($name) . "\"" . $delimiter;
        if ($intensity eq '---') {
          $string .= '"---"' . $delimiter;
        } else {
          $string .= $intensity . $delimiter;
        }
      } elsif ($export_format eq "XML") {
        $string .= "<quant_pep_intensity name=\""
          . &noXmlTag($name)
          . "\" intensity=\""
          . $intensity
          . "\" />" 
          . $delimiter;
      }
    }
  }
  
  return $string;
  
}

###############################################################################
# &outputHeader()
# output header block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputHeader {
  my $j = 0;
  
  if ($export_format eq "XML") {
    print "<header>\n";
    $j = outputHeaderOfFile($export_format, $objResFile, $objParams, $j, 'full');
    $j = outputHeaderWarnings($j);
    print "</header>\n";
  } elsif ($export_format eq "CSV") {
    my $line = '--------------------------------------------------------';
    if (@fileIn == 1) {
      print "\n\"Header\"$delimiter\"$line\"\n\n";
      $j = outputHeaderOfFile($export_format, $objResFile, $objParams, $j, 'full');
      $j = outputHeaderWarnings($j);
    } else {
      my $l = $line;
      my $header_str = sprintf('results collection set of %d files', scalar(@fileIn));
      substr($l, 4, length($header_str)) = $header_str;
      print "\n\"Header\"$delimiter\"$l\"\n\n";
      $j = outputHeaderOfFile($export_format, $objResFile, $objParams, $j, 'combined');
      $j = outputHeaderWarnings($j);
      for my $i (1 .. $objResFile->getNumberOfResfiles()) {
        my $l = $line;
        my $fileIn = $fileIn[$i-1];
        substr($l, 4, length($fileIn)) = $fileIn;
        print "\n\"Header\"$delimiter\"$l\"\n\n";
        my $resfile = $objResFile->getResfile($i);
        $j = outputHeaderOfFile($export_format, $resfile, $resfile->params, $j, 'part');
      }
    }
  }
}

sub outputHeaderOfFile {
  my ($export_format, $objResFile, $objParams, $j, $header_type) = @_;

  if ($header_type ne 'combined') {
    print &formatElement($export_format, "Search title", 1, "COM", $objParams->getCOM()) . "\n";
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = gmtime ($objResFile->getDate);
    my $timestamp = sprintf("%04d-%02d-%02dT%02d:%02d:%02dZ", 
      $year+1900, $mon+1, $mday, $hour, $min, $sec);
    print &formatElement($export_format, "Timestamp", 1, "Date", $timestamp) . "\n";
    print &formatElement($export_format, "User", 1, "USERNAME", $objParams->getUSERNAME()) . "\n";
    print &formatElement($export_format, "Email", 1, "USEREMAIL", $objParams->getUSEREMAIL()) . "\n";
  }
  if ($header_type eq 'combined' or $header_type eq 'full') {
    if (my $fileset = $thisScript->param($urlParams{'fileset'})) {
      print &formatElement($export_format, "Report URI", 1, "URI", $URI . "cgi/master_results_2.pl?fileset=" . $fileset) . "\n";
    } else {
      print &formatElement($export_format, "Report URI", 1, "URI", $URI . "cgi/master_results.pl?file=" . $fileIn[0]) . "\n";
    }
  }
  if ($header_type ne 'combined') {
    if ($objParams->getDATAURL() eq '') {
      print &formatElement($export_format, "Peak list data path", 1, "FILENAME", $objParams->getFILENAME()) . "\n";
    } else {
      print &formatElement($export_format, "Peak list data URL", 1, "DATAURL", $objParams->getDATAURL()) . "\n";
    }
    print &formatElement($export_format, "Peak list format", 1, "FORMAT", $objParams->getFORMAT()) . "\n";
    print &formatElement($export_format, "Search type", 0, "SEARCH", $objParams->getSEARCH()) . "\n";
    print &formatElement($export_format, "Mascot version", 1, "MascotVer", $objResFile->getMascotVer()) . "\n";
    my($dbString, $fastaVerString);
    if ($objParams->getNumberOfDatabases == 1) {
      $dbString = $objParams->getDB(1);
      $fastaVerString = $objResFile->getFastaVer(1);
    } else {
      my(@tempList1, @tempList2);
      for (my $i=1; $i <= $objParams->getNumberOfDatabases; $i++) {
        push @tempList1, $i . "::" . $objParams->getDB($i);
        push @tempList2, $i . "::" . $objResFile->getFastaVer($i);
      }
      $dbString = join(" ", @tempList1);
      $fastaVerString = join(" ", @tempList2);
    }
    print &formatElement($export_format, "Database", 1, "DB", $dbString) . "\n";
    print &formatElement($export_format, "Fasta file", 1, "FastaVer", $fastaVerString) . "\n";
  # wait for parser function ### debug ###
    print &formatElement($export_format, "Total sequences", 0, "NumSeqs", $objResFile->getNumSeqs()) . "\n";
    print &formatElement($export_format, "Total residues", 0, "NumResidues", $objResFile->getNumResidues()) . "\n";
    print &formatElement($export_format, "Sequences after taxonomy filter", 0, "NumSeqsAfterTax", $objResFile->getNumSeqsAfterTax()) . "\n";
    if ($objResFile->isErrorTolerant()) {
      if (my $temp = $objParams->getACCESSION()) {
        my $i = $temp =~ tr/,//;
        print &formatElement($export_format, "Number of sequences searched in error tolerant mode", 
          0, "error_tolerant_num", $i + 1) . "\n";
      } else {
=pod Wrong number -- see bug 11120. Correct figure needs Parser support.
        print &formatElement($export_format, "Number of sequences searched in error tolerant mode", 
          0, "error_tolerant_num", $objResFile->getNumSeqsAfterTax()) . "\n";
=cut
      }
    }
  }
  if ($header_type eq 'combined' or $header_type eq 'full') {
    print &formatElement($export_format, "Number of queries", 0, "NumQueries", $objResFile->getNumQueries()) . "\n";
    if ($targetFDR) {
      my $FDR;
      if ($use_homology) {
        my $HF = $objSummary->getNumHitsAboveHomology(1 / $sigThreshold);
        my $HD = $objSummary->getNumDecoyHitsAboveHomology(1 / $sigThreshold);
        if ($HF > 0) {
          $FDR = sprintf("%.2f", 100 * $HD / $HF);
        }
      } else {
        my $IF = $objSummary->getNumHitsAboveIdentity(1 / $sigThreshold);
        my $ID = $objSummary->getNumDecoyHitsAboveIdentity(1 / $sigThreshold);
        if ($IF > 0) {
          $FDR = sprintf("%.2f", 100 * $ID / $IF);
        }
      }
      if (defined $FDR) {
        print &formatElement($export_format, "Target false discovery rate", 0, "TargetFDR", $FDR, type => ($use_homology ? 'homology' : 'identity') ) . "\n";
      }
    }
  }
# Display any warnings from result file header

  if ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn0')){
    while ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j)){
      print &formatElement($export_format, "Warning", 1, "warning", 
        $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j), number => $j) . "\n";
      $j++;
    }
  }
# Display a warning if this is an orphan error tolerant search
  if ($objResFile->isErrorTolerant() && !$objParams->getErrTolParentFilename()) {
    print &formatElement($export_format, "Warning", 1, "warning", 
      "Scores from standard search not available. Better matches may exist in other proteins.", number => $j) . "\n";
    $j++;
  }

  return $j;
}

sub outputHeaderWarnings {
  my ($j) = @_;

# Display any warnings from Parser
  for (my $k = 1; $k <= $objResFile->getNumberOfErrors(); $k++) {
    print &formatElement($export_format, "Warning", 1, "warning", 
      $objResFile->getErrorString($k), number => $j) . "\n";
    $j++;
  }
# Display any warnings from quantitation
  if (@quantWarnings) {
    foreach (@quantWarnings) {
      print &formatElement($export_format, "Warning", 1, "warning", 
        $_, number => $j) . "\n";
      $j++;
    }
  }
  if ($objMS2Quant) {
    my $objErr = $objMS2Quant->getErrorHandler();
    for (my $i = 1; $i <= $objErr->getNumberOfErrors(); $i++) {
      print formatElement($export_format, 'Warning', 1, 'warning', 
        sprintf('%s (%s)', $objErr->getErrorString($i), $objErr->getErrorNumber($i)),
        number => $j++
      );
    }
  }

  return $j;
}

###############################################################################
# &outputDecoy()
# output decoy database search statistics
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputDecoy {
  
  if ($export_format eq "XML") {
    print "<decoy>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Decoy\"$delimiter\"--------------------------------------------------------\"\n\n";
  }

  print &formatElement($export_format, "Number of matches above identity threshold in search of real database", 
    0, "NumHitsAboveIdentity", $objSummary->getNumHitsAboveIdentity(1 / $sigThreshold)) . "\n";
  print &formatElement($export_format, "Number of matches above identity threshold in search of decoy database", 
    0, "NumDecoyHitsAboveIdentity", $objSummary->getNumDecoyHitsAboveIdentity(1 / $sigThreshold)) . "\n";
  if ($objResFile->isPMF()){
    my $TF = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_SUMMARY, "h1");
    $TF = (split(/,/, $TF))[1];
    $TF = sprintf("%.0f", $TF);
    my $TD = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_DECOYSUMMARY, "h1");
    $TD = (split(/,/, $TD))[1];
    $TD = sprintf("%.0f", $TD);
    print &formatElement($export_format, "Highest score in search of real database", 
      0, "HighestScoreProtein", $TF) . "\n";
    print &formatElement($export_format, "Highest score in search of decoy database", 
      0, "HighestScoreDecoyProtein", $TD) . "\n";
  } elsif ($percolate != 1) {
    print &formatElement($export_format, "Number of matches above homology threshold in search of real database", 
      0, "NumHitsAboveHomology", $objSummary->getNumHitsAboveHomology(1 / $sigThreshold)) . "\n";
    print &formatElement($export_format, "Number of matches above homology threshold in search of decoy database", 
      0, "NumDecoyHitsAboveHomology", $objSummary->getNumDecoyHitsAboveHomology(1 / $sigThreshold)) . "\n";
  }
  
  if ($export_format eq "XML") {
    print "</decoy>\n";
  } 

}

###############################################################################
# &outputFixedMods()
# output fixed modifications block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputFixedMods {
  
    if ($export_format eq "XML") {
      print "<fixed_mods>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Fixed modifications\"$delimiter\"--------------------------------------------------------\"\n\n";
      print "\"Identifier\"$delimiter";
      print "\"Name\"$delimiter";
      print "\"Delta\"$delimiter";
      print "\"Neutral loss\"\n";
    } 

  my $string = "";
  my $i = 1;
  my $letter;

# $objParams->getFixedModsName() will only work with files from 2.1 and later
  if ($objParams->getFixedModsName(1)) {
    while ($objParams->getFixedModsName($i)) {
      $letter = $i;
      if ($i > 9) {
        $letter = chr($i + 55);
      }
      if ($export_format eq "CSV") {
        $string .= $letter . $delimiter;
      } elsif ($export_format eq "XML") {
        $string .= "<modification identifier=\"$letter\">$delimiter";
      }
      $string .= &formatElement($export_format, "", 1, "name", $objParams->getFixedModsName($i)) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "delta", $objParams->getFixedModsDelta($i)) . $delimiter;
      my $loss = $objParams->getFixedModsNeutralLoss($i);
      if ($loss) {
        $string .= &formatElement($export_format, "", 0, "neutral_loss", $loss) . $delimiter;
      }
      chop $string;
      $i++;
      if ($export_format eq "XML") {
        $string .= "$delimiter</modification>";
      }
      $string .= "\n";
    }
  } else {
    my @modList = split(/,/, $objParams->getMODS);
    foreach my $mod (@modList) {
      $letter = $i;
      if ($i > 9) {
        $letter = chr($i + 55);
      }
      if ($export_format eq "CSV") {
        $string .= $letter . $delimiter;
      } elsif ($export_format eq "XML") {
        $string .= "<modification identifier=\"$letter\">$delimiter";
      }
      my @modParams = &getResidueMods($mod, "N");
      unless (@modParams) {
        @modParams = &getTerminalMods($mod, "N");
      }
      $string .= &formatElement($export_format, "", 1, "name", $mod) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "delta", $modParams[1]) . $delimiter;
      chop $string;
      $i++;
      if ($export_format eq "XML") {
        $string .= "$delimiter</modification>";
      }
      $string .= "\n";
    }
  }
  
  chop $string;
  if ($export_format eq "XML") {
    $string =~ s/>$delimiter</>\n</g;
  }
  if ($string) {
    print $string . "\n";
  }

  if ($export_format eq "XML") {
    print "</fixed_mods>\n";
  } 

}

###############################################################################
# &outputVarMods()
# output variable modifications block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputVarMods {
  
    if ($export_format eq "XML") {
      print "<variable_mods>\n";
    } elsif ($export_format eq "CSV") {
      print "\n\"Variable modifications\"$delimiter\"--------------------------------------------------------\"\n\n";
      print "\"Identifier\"$delimiter";
      print "\"Name\"$delimiter";
      print "\"Delta\"$delimiter";
      print "\"Neutral loss(es)\"\n";
    } 

  my $string = "";
  my $i = 1;
  while ($objParams->getVarModsName($i)) {
    my $letter = $i;
    if ($i > 9) {
      $letter = chr($i + 55);
    }
    if ($export_format eq "CSV") {
      $string .= $letter . $delimiter;
    } elsif ($export_format eq "XML") {
      $string .= "<modification identifier=\"$letter\">$delimiter";
    }
    $string .= &formatElement($export_format, "", 1, "name", $objParams->getVarModsName($i)) . $delimiter;
    $string .= &formatElement($export_format, "", 0, "delta", $objParams->getVarModsDelta($i)) . $delimiter;
    my $losses = $objParams->getVarModsNeutralLosses($i);
    if ($losses
      && !(scalar(@{ $losses }) == 1 && ${ $losses }[0] == 0)) {
      my $j = 1;
      foreach(@{ $losses }) {
        my $letter = $j;
        if ($j > 9) {
          $letter = chr($j + 55);
        }
        $string .= &formatElement($export_format, "", 0, "neutral_loss", $_, identifier => $letter) . $delimiter;
        $j++;
      }
    }
    chop $string;
    $i++;
    if ($export_format eq "XML") {
      $string .= "$delimiter</modification>";
    }
    $string .= "\n";
  }
  
  chop $string;
  if ($export_format eq "XML") {
    $string =~ s/>$delimiter</>\n</g;
  }
  if ($string) {
    print $string . "\n";
  }

  if ($export_format eq "XML") {
    print "</variable_mods>\n";
  } 

}

###############################################################################
# &outputParams()
# output search parameters block
# globals:
# my($objParams, $objResFile, $export_format);
###############################################################################

sub outputParams {
  
  if ($export_format eq "XML") {
    print "<search_parameters>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Search Parameters\"$delimiter\"--------------------------------------------------------\"\n\n";
  }
  
  print &formatElement($export_format, "Taxonomy filter", 1, "TAXONOMY", $objParams->getTAXONOMY()) . "\n";
  print &formatElement($export_format, "Enzyme", 1, "CLE", $objParams->getCLE()) . "\n";
  print &formatElement($export_format, "Maximum Missed Cleavages", 0, "PFA", $objParams->getPFA()) . "\n";
  my $local_mods = $objParams->getMODS();
  if ($objQuantMethod) {
  # add in any fixed mods specified in the quant method
    my $qMods = &quant_subs::getQuantFixedModsAsString($objQuantMethod);
    if ($qMods && $local_mods) {
      $local_mods .= "," . $qMods;
    } elsif ($qMods) {
      $local_mods = $qMods;
    }
  }
  print &formatElement($export_format, "Fixed modifications", 1, "MODS", $local_mods) . "\n";
  if ($objParams->getQUANTITATION()) {
    print &formatElement($export_format, "Quantitation method", 1, "QUANTITATION", $objParams->getQUANTITATION()) . "\n";
  } elsif ($objParams->getICAT()) {
    print &formatElement($export_format, "ICAT experiment", 0, "ICAT", $objParams->getICAT()) . "\n";
  }
  my $local_it_mods = $objParams->getIT_MODS();
  if ($objParams->getICAT()){
    my $ICATLight = $objMascotOptions->getICATLight;
    $ICATLight = "ICAT_light" unless $ICATLight;
    my $ICATHeavy = $objMascotOptions->getICATHeavy;
    $ICATHeavy = "ICAT_heavy" unless $ICATHeavy;
    my @tempMods = split(/,/, $local_it_mods);
    for (my $i = $#tempMods; $i >= 0; $i--) {
      if (($tempMods[$i] eq $ICATLight) || ($tempMods[$i] eq $ICATHeavy)) {
        splice @tempMods, $i, 1;
      }
    }
    if (@tempMods) {
      $local_it_mods = join(',', @tempMods);
    } else {
      $local_it_mods = "";
    }
  }
  print &formatElement($export_format, "Variable modifications", 1, "IT_MODS", $local_it_mods) . "\n";
  print &formatElement($export_format, "Peptide Mass Tolerance", 0, "TOL", $objParams->getTOL()) . "\n";
  print &formatElement($export_format, "Peptide Mass Tolerance Units", 0, "TOLU", $objParams->getTOLU()) . "\n";
  if ($objResFile->isPMF()){
    print &formatElement($export_format, "Peptide Charge State", 0, "CHARGE", $objParams->getCHARGE()) . "\n";
  } else {
    print &formatElement($export_format, "Fragment Mass Tolerance", 0, "ITOL", $objParams->getITOL()) . "\n";
    print &formatElement($export_format, "Fragment Mass Tolerance Units", 0, "ITOLU", $objParams->getITOLU()) . "\n";
  }
  print &formatElement($export_format, "Mass values", 0, "MASS", $objParams->getMASS()) . "\n";
  if ($objParams->getSEG() > 0){
    print &formatElement($export_format, "Protein Mass", 0, "SEG", $objParams->getSEG()) . "\n";
  }
  if (!$objResFile->isPMF() && $objParams->getINSTRUMENT()) {
    print &formatElement($export_format, "Instrument type", 1, "INSTRUMENT", $objParams->getINSTRUMENT()) . "\n";
  }
  if ($objParams->getPEP_ISOTOPE_ERROR() > -1) {
    print &formatElement($export_format, "Isotope error mode", 0, "PEP_ISOTOPE_ERROR", $objParams->getPEP_ISOTOPE_ERROR()) . "\n";
  }
  if ($objParams->getDECOY() > -1) {
    print &formatElement($export_format, "Decoy database also searched", 0, "DECOY", $objParams->getDECOY()) . "\n";
  }

# Print out any user parameters with defined values (USER00 to USER12 and anything beginning with underscore) 
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_PARAMETERS, $i)) {
    if ($nextKey =~ /^(_.*)/i || $nextKey =~ /^(user\d\d)$/i) {
      my $parameter_name = $1;
      my $nextValue = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, $nextKey);
      $nextKey =~ s/\s*$//;
      if ($nextValue) {
        print &formatElement($export_format, $nextKey, 1, "user_parameter", $nextValue, name => $parameter_name) . "\n";
      }
    }
    $i++;
  }

  if ($export_format eq "XML") {
    print "</search_parameters>\n";
  } 

}

###############################################################################
# &outputFormat()
# output format block
# globals:
# my($objResFile, $export_format, %urlParams, $thisScript, $sigThreshold,
# $numHits, $ignoreIonsScoreBelow, $RequireBoldRed, $mudpitSwitch, $ShowSubSets);
###############################################################################

sub outputFormat {
  
  if ($export_format eq "XML") {
    print "<format_parameters>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Format parameters\"$delimiter\"--------------------------------------------------------\"\n\n";
  }
  
  print &formatElement($export_format, "Significance threshold", 0, "sigthreshold", $sigThreshold) . "\n";
  print &formatElement($export_format, "Max. number of hits", 0, "REPORT", $numHits) . "\n";
  if ($anyPepSumMatch) {
    if ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch){
      print &formatElement($export_format, "Use MudPIT protein scoring", 0, "mudpit", 1) . "\n";
    } else {
      print &formatElement($export_format, "Use MudPIT protein scoring", 0, "mudpit", 0) . "\n";
    }
    print &formatElement($export_format, "Ions score cut-off", 0, "ignoreionsscorebelow", $ignoreIonsScoreBelow) . "\n";
  }
  print &formatElement($export_format, "Include same-set proteins", 
    0, "show_same_sets", $ShowSameSets) . "\n";
  print &formatElement($export_format, "Include sub-set proteins", 
    0, "showsubsets", $ShowSubSets) . "\n";
  if ($anyPepSumMatch) {
    print &formatElement($export_format, "Include unassigned", 
      0, "show_unassigned", $thisScript->param($urlParams{'show_unassigned'}) || 0) . "\n";
    print &formatElement($export_format, "Require bold red", 0, "requireboldred", $RequireBoldRed) . "\n";
    if ($thisScript->param($urlParams{'unigene'})) {
      print &formatElement($export_format, "UniGene index", 1, "UNIGENE", $thisScript->param($urlParams{'unigene'}) + 0) . "\n";
    }
   print &formatElement($export_format, "Use homology threshold", 0, "use_homology", $use_homology) . "\n";
   print &formatElement($export_format, "Group protein families", 0, "group_family", $group_family) . "\n";
   unless ($percolate == -1) {
     print &formatElement($export_format, "Re-score using Percolator", 0, "percolate", $percolate) . "\n";
    }
  }
  if ($preferTaxonomy > 0) {
    my $tax = msparser::ms_taxonomyfile->new();
    my $taxonomy_entry;

    if ($preferTaxonomy < $tax->getNumberOfEntries()) {
        $taxonomy_entry = $tax->getEntryByNumber($preferTaxonomy);
    }

    if ($taxonomy_entry) {
      print formatElement($export_format, "Preferred taxonomy", 1, 'preferred_taxonomy', $taxonomy_entry->getTitle()), "\n";
    }
  }
  print formatElement($export_format, 'Show duplicate peptides', 0, 'show_pep_dupes', $thisScript->param($urlParams{'show_pep_dupes'}) ? 1 : 0) . "\n";

  if ($export_format eq "XML") {
    print "</format_parameters>\n";
  } 

}

###############################################################################
# &outputMasses()
# output masses block
# globals:
# my($objResFile, $export_format);
###############################################################################

sub outputMasses {
  
  if ($export_format eq "XML") {
    print "<masses>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Mass values\"$delimiter\"--------------------------------------------------------\"\n\n";
  }
  
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_MASSES, $i)) {
    if ($nextKey =~ /^delta/i
      || $nextKey =~ /^ignore/i
      || $nextKey =~ /^neutralloss/i
      || $nextKey =~ /^fixedmod/i
      || $nextKey =~ /^reqpep/i
      || $nextKey =~ /^pepneutral/i) {
      $i++;
      next;
    }
    print &formatElement($export_format, $nextKey, 0, "mass", 
      $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $nextKey), name => $nextKey) . "\n";
    $i++;
  }

  if ($export_format eq "XML") {
    print "</masses>\n";
  } 

}

###############################################################################
# &outputQuantitationHeader()
# output quantitation level details, such as normalisation bases
# globals:
# my($objMS2Quant, $export_format);
###############################################################################

sub outputQuantitationHeader {

  if (not $objMS2Quant) {
    return;
  }

  if ($export_format eq "XML") {
    print "<quantitation_header>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Quantitation header\"$delimiter\"--------------------------------------------------------\"\n\n";
  }
  
  if ($objMS2Quant->hasPeptideRatioNormalisation()) {
    my @names = @{ $objMS2Quant->getCombinedReportRatioNames() };
    for my $name (@names) {
      my $value = $objMS2Quant->getPeptideRatioNormalisationBase($name);
      print formatElement($export_format, 
          sprintf('Normalisation base,"%s"', noQuotes($name)), 
          0, 
          'normalisation_base', 
          $value, 

          type => 'ratio', 
          name => $name
      ), "\n";
    }
  }

  if ($objMS2Quant->hasIntensityNormalisation()) {
    my $qm = $objMS2Quant->getQuantitationMethod();
    for my $i (0 .. $qm->getNumberOfComponents()-1) {
      my $name = $qm->getComponentByNumber($i)->getName();
      my $value = $objMS2Quant->getIntensityNormalisationBase($name);
      print formatElement($export_format, 
          sprintf('Normalisation base (%s)', $name), 
          0, 
          'normalisation_base', 
          $value, 

          type => 'component',
          name => $name
      ), "\n";
    }
  }

  if ($export_format eq "XML") {
    print "</quantitation_header>\n";
  } 

}

###############################################################################
# &outputQueries()
# output queries block
# globals:
# my($objResFile, $objSummary, $objParams, $objUmodConfigFile, $msresFlags2,
# $export_format, %urlParams, $thisScript, $sigThreshold, $numHits,
# $ignoreIonsScoreBelow, $RequireBoldRed, $mudpitSwitch, $ShowSubSets);
###############################################################################

sub outputQueries {

  my $isPMF = $objResFile->isPMF();
  
  if ($export_format eq "XML") {
    print "<queries>\n";
  } elsif ($export_format eq "CSV") {
    print "\n\"Queries\"$delimiter\"--------------------------------------------------------\"\n\n";
    print "\"query_number\"";
    print "$delimiter\"moverz\"";
    print "$delimiter\"charge\"";
    print "$delimiter\"intensity\"";
    if ($thisScript->param($urlParams{'query_title'})) {
      print "$delimiter\"StringTitle\"";
      print "$delimiter\"Scan number range\"";
      print "$delimiter\"Retention time range\"";
    }
    if ($thisScript->param($urlParams{'query_qualifiers'})) {
      print "$delimiter\"qualifiers\"";
    }
    if ($thisScript->param($urlParams{'query_params'})) {
      print "$delimiter\"Peptide Mass Tolerance\"";
      print "$delimiter\"Peptide Mass Tolerance Units\"";
      print "$delimiter\"Variable modifications\"";
      print "$delimiter\"Instrument type\"";
    }
    if ($thisScript->param($urlParams{'query_peaks'})) {
      print "$delimiter\"TotalIonsIntensity\"";
      print "$delimiter\"NumVals\"";
      print "$delimiter\"StringIons1\"";
      print "$delimiter\"StringIons2\"";
      print "$delimiter\"StringIons3\"";
    }
    if ($thisScript->param($urlParams{'query_raw'})) {
      foreach my $field (@columns) {
        if ($field !~ /^pep/) {
          next;
        }
        if ($thisScript->param($urlParams{lc($field)})) {
          if ($field eq "pep_seq") {
            print $delimiter . "pep_res_before" . $delimiter . $field . $delimiter . "pep_res_after";
          } elsif ($field eq "pep_var_mod") {
            print $delimiter . $field 
                . $delimiter . "pep_var_mod_pos"
                . $delimiter . "pep_summed_mod_pos"
                . $delimiter . "pep_local_mod_pos"
                . $delimiter . "pep_var_mod_conf"
                ;
          } elsif ($field eq "pep_query") {
            next;
          } else {
            print $delimiter . $field;
          }
        }
      }
    }
    print "\n";
  }

  # Site analysis initialisation.
  my $modsByDelta = Util::SiteAnalysis::modifications_by_delta(
    $objUmodConfigFile, $objParams,
  );
  my %query_level_mod_cache = ();
  # $objUmodConfigFile could be the reduced file embedded in the results file.
  my $cs = new msparser::ms_connection_settings();
  $cs->setSessionID($sessionID);
  my $full_unimod_file = new msparser::ms_umod_configfile(
    "../config/unimod.xml", "../html/xmlns/schema/unimod_2/unimod_2.xsd",
    $cs
  );
  my $query_is_sig = $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH
    ? sub { Util::Peptides::above_homology($objSummary, $objSummary->getPeptide($_[0], 1)) }
    : sub { Util::Peptides::above_identity($objSummary, $objSummary->getPeptide($_[0], 1)) }
    ;
  my $max_p = $objSummary->getMaxRankValue();
  
  for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries; $queryNum++) {
    my $string = "";
    $queryColumnCount = 0;
    my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
    if ($export_format eq "CSV") {
      $string .= $queryNum . $delimiter;
    } elsif ($export_format eq "XML") {
      $string .= "<query number=\"$queryNum\">$delimiter";
    }
    $queryColumnCount++;
    
    $string .= &formatElement($export_format, "", 0, "query_moverz", $objResFile->getObservedMass($queryNum) + 0) . $delimiter;
    $string .= &formatElement($export_format, "", 1, "query_charge", $objQuery->getCharge()) . $delimiter;
    my $intensity = $objResFile->getObservedIntensity($queryNum) + 0;
    if ($intensity != 0) {
      $string .= &formatElement($export_format, "", 0, "query_intensity", $intensity) . $delimiter;
    } elsif ($export_format eq "CSV") {
      $string .= $delimiter;
    }
    $queryColumnCount += 3;

  # query_title
    if ($thisScript->param($urlParams{'query_title'})) {
      if ($objQuery->getStringTitle(1)) {
        $string .= &formatElement($export_format, "", 1, "StringTitle", 
          $objQuery->getStringTitle(1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
      if ($objResFile->getQuerySectionValueStr($queryNum, "SCANS")) {
        $string .= &formatElement($export_format, "", 1, "SCANS", $objResFile->getQuerySectionValueStr($queryNum, "SCANS")) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
      if ($objResFile->getQuerySectionValueStr($queryNum, "RTINSECONDS")) {
        $string .= &formatElement($export_format, "", 1, "RTINSECONDS", $objResFile->getQuerySectionValueStr($queryNum, "RTINSECONDS")) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
    }
    
  # query_qualifiers
    if ($thisScript->param($urlParams{'query_qualifiers'})) {
      my @qualifiers;
      if ($objQuery->getPepTolString()) {                                  
        if ($export_format eq "XML") {
          push @qualifiers, &formatElement($export_format, "", 1, "qual_tol", $objQuery->getPepTolString());
        } else {
          push @qualifiers, $objQuery->getPepTolString();
        }
      }
      for (my $j = 1; $j <= 20; $j++) {
        if ($objQuery->getSeq($j)) {
          if ($export_format eq "XML") {
            push @qualifiers, &formatElement($export_format, "", 1, "qual_seq", "seq(" . $objQuery->getSeq($j) . ")", number => $j);
          } else {
            push @qualifiers, "seq(" . $objQuery->getSeq($j) . ")";
          }
        } else {
          last;
        }
      }
      for (my $j = 1; $j <= 20; $j++) {
        if ($objQuery->getComp($j)) {
          if ($export_format eq "XML") {
            push @qualifiers, &formatElement($export_format, "", 1, "qual_comp", "comp(" . $objQuery->getComp($j) . ")", number => $j);
          } else {
            push @qualifiers, "comp(" . $objQuery->getComp($j) . ")";
          }
        } else {
          last;
        }
      }
      for (my $j = 1; $j <= 20; $j++) {
        if (my $local_tag_string = $objQuery->getTag($j)) {
          $local_tag_string =~ /([et]),(.*)/i;
          if ($1 eq "e") {
            $local_tag_string = "etag($2)";
          } else {
            $local_tag_string = "tag($2)";
          }
          if ($export_format eq "XML") {
            push @qualifiers, &formatElement($export_format, "", 1, "qual_tag", $local_tag_string, number => $j);
          } else {
            push @qualifiers, $local_tag_string;
          }
        } else {
          last;
        }
      }
      if ($export_format eq "XML") {
        if (@qualifiers) {
          $string .= join($delimiter, @qualifiers) . $delimiter;
        }
      } else {
        $string .= &formatElement($export_format, "", 1, "pep_qual", join(" ", @qualifiers)) . $delimiter;
      }
      $queryColumnCount++;
    }

  # query_params
    if ($thisScript->param($urlParams{'query_params'})) {
      if ($objQuery->getPepTol() > 0) {
        $string .= &formatElement($export_format, "", 0, "query_TOL", $objQuery->getPepTol()) . $delimiter;
        $string .= &formatElement($export_format, "", 0, "query_TOLU", $objQuery->getPepTolUnits()) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter . $delimiter;
      }
      $queryColumnCount++;
      $queryColumnCount++;
      if ($objQuery->getIT_MODS(1)) {
        $string .= &formatElement($export_format, "", 1, "query_IT_MODS", $objQuery->getIT_MODS(1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
      if ($objQuery->getINSTRUMENT(1)) {
        $string .= &formatElement($export_format, "", 1, "query_INSTRUMENT", $objQuery->getINSTRUMENT(1)) . $delimiter;
      } elsif ($export_format eq "CSV") {
        $string .= $delimiter;
      }
      $queryColumnCount++;
    }

  # query_peaks
    if ($thisScript->param($urlParams{'query_peaks'})) {
      $string .= &formatElement($export_format, "", 0, "TotalIonsIntensity", $objQuery->getTotalIonsIntensity() + 0) . $delimiter;
      $string .= &formatElement($export_format, "", 0, "NumVals", $objQuery->getNumVals() + 0) . $delimiter;
      $string .= &formatElement($export_format, "", 1, "StringIons1", $objQuery->getStringIons1()) . $delimiter;
      $string .= &formatElement($export_format, "", 1, "StringIons2", $objQuery->getStringIons2()) . $delimiter;
      $string .= &formatElement($export_format, "", 1, "StringIons3", $objQuery->getStringIons3()) . $delimiter;
      $queryColumnCount += 5;
    }
    
  # query_raw
    if ($thisScript->param($urlParams{'query_raw'})) {
      my $siteAnalysisData;
      if (&$query_is_sig($queryNum)) {
        $siteAnalysisData = analyse_sites_in_query(
          \%query_level_mod_cache, $objResFile, $objSummary, $objParams,
          $queryNum, $modsByDelta, $full_unimod_file, $max_p,
        );
      }
      my $needSpacer = 0;
      for my $rank (1 .. $max_p) {
        my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
        next unless $objPeptide;
        next unless $objPeptide->getAnyMatch();

        if ($export_format eq "CSV" && $needSpacer) {
          $string .= $queryNum . $delimiter x $queryColumnCount;
        } else {
          $needSpacer = 1;
        }
        if ($export_format eq "XML") {
          $string .= "<q_peptide query=\"$queryNum\"";
          if ($objPeptide->getAnyMatch() && !$isPMF) {
            $string .= " rank=\"" . $objPeptide->getPrettyRank() . "\"";
          }
          $string .= ">\n";
        }
        $string .= fmtPeptideDetails($objPeptide, $queryNum, $rank, "", 0, $isPMF, $siteAnalysisData || {}) . "\n";
        if ($export_format eq "XML") {
          chop $string;
          $string .= "</q_peptide>\n";
        } 
      }
    }

    chop $string;
    if ($export_format eq "XML") {
      $string .= "$delimiter</query>";
      $string =~ s/>$delimiter</>\n</g;
    }
    if ($string) {
      print $string . "\n";
    }

  }
  
  if ($export_format eq "XML") {
    print "</queries>\n";
  } 

}

###############################################################################
# &formatElement()
# returns formatted element
#
# $_[0] "XML" or "CSV" or ...
# $_[1] CSV label, empty if none required
# $_[2] 1 if text element could contain delimiter or illegal character
# $_[3] xml element name
# $_[4] element value
# $_[5..n] optional XML attribute hash
###############################################################################

sub formatElement {
  
  my($format, $label, $need_quotes, $name, $value, %xml_attribs) = @_;
  
  my $return = "";
  
  if ($format eq "XML") {
    $return .= "<$name";
    for my $key (sort keys %xml_attribs) {
      $return .= " $key=\"" . &noXmlTag($xml_attribs{$key}) . "\"";
    }
    if ($value eq "") {
      $return .= " />";
    } else {
      if ($need_quotes) {
        $return .= ">" . &noXmlTag($value) . "</$name>";
      } else {
        $return .= ">$value</$name>";
      }
    }
  } elsif ($format eq "CSV") {
    if ($label) {
      $return .= "$label$delimiter";
    }
    if ($need_quotes) {
      $return .= "\"" . &noQuotes($value) . "\"";
    } else {
      $return .= $value;
    }
  }
  
  return $return;

}

###############################################################################
# &noXmlTag()
# returns de-tagged string
# $_[0] string which may contain XML illegal characters
###############################################################################

sub noXmlTag {

  my $temp = shift;
  return $temp if not defined($temp);

  $temp =~ s/&/&amp;/g;
  $temp =~ s/</&lt;/g;
  $temp =~ s/>/&gt;/g;
  $temp =~ s/\"/&quot;/g;
  $temp =~ s/\'/&apos;/g;

  return $temp;

}

###############################################################################
# &noQuotes()
# returns string with quotes removed
# $_[0] string which may contain quotes
###############################################################################

sub noQuotes {

  my $temp = shift;

  $temp =~ s/[\"\']/~/g;

  return $temp;

}

###############################################################################
# &getResidueMods()
# returns aminoacid_modification elements
# $_[0] Mascot modification name
# $_[1] "N" or "Y" for variable mods
###############################################################################

sub getResidueMods {
  
  my ($modName, $variable) = @_;

  my @retVal;
  my $treatasvariable = $variable;
  my $terminalMod = "";
  my $objMod = $objModFile->getModificationByName($modName);
  if ($objMod) {
    if ($objMod->getModificationType() == $msparser::MOD_TYPE_RESIDUE
      || $objMod->getModificationType() == $msparser::MOD_TYPE_N_TERM_RESIDUE
      || $objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM_RESIDUE) {
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_N_TERM_RESIDUE) {
        $terminalMod = "n";
        $treatasvariable = "Y";
      }
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM_RESIDUE) {
        $terminalMod = "c";
        $treatasvariable = "Y";
      }
      for (my $i = 0; $i < $objMod->getNumberOfModifiedResidues(); $i++) {
        push @retVal, $objMod->getModifiedResidue($i);
        push @retVal, sprintf("%+." . $massDP . "f", $objMod->getDelta($mass_type));
        if ($terminalMod) {
          push @retVal, ($objMassesFile->getResidueMass($mass_type, $objMod->getModifiedResidue($i))
            + $objMod->getDelta($mass_type));
        } else {
          push @retVal, $objMod->getResidueMass($mass_type, $objMod->getModifiedResidue($i));
        }
        push @retVal, $treatasvariable;
        push @retVal, $terminalMod;
        push @retVal, $modName;
      # print "<!--\n";
      # print $objMod->getResidueMass($mass_type, $objMod->getModifiedResidue($i)) . "\n";
      # print $objMod->getDelta($mass_type) . "\n";
      # print $objMod->isResidueModified($objMod->getModifiedResidue($i)) . "\n";
      # print $objMod->getCTerminusMass($mass_type) . "\n";
      # print $objMassesFile->getResidueMass($mass_type, $objMod->getModifiedResidue($i)) . "\n";
      # print $objMassesFile->isResidueModified($objMod->getModifiedResidue($i)) . "\n";
      # print $objMassesFile->getCtermDelta($mass_type) . "\n";
      # print $objMassesFile->getCterminalMass($mass_type) . "\n";
      # print "-->\n";
      }
    }
  } else {
  # mod not found in file, extract values from result file
    unless ($modName =~ /\(c-term\)/i
      || $modName =~ /\(n-term\)/i
      || $modName =~ /\(protein\)/i) {
      (my $specificity) = $modName =~ /\((.*)\)/;
    # special cases
      if ($specificity =~ /([cn])-term (.+)/i) {
        $terminalMod = lc($1);
        $specificity = $2;
        $treatasvariable = "Y";
      }
      if ($specificity eq "") {
      # must be ICAT_light or ICAT_heavy
        $specificity = "C";
      } elsif ($specificity eq "camC") {
      # Pyro-cmC
        $specificity = "C";
      }
      my @spec_list = split(//, $specificity);
      foreach my $this_spec (@spec_list) {
        my ($mass_diff, $mass_total);
        if ($treatasvariable eq "Y") {
          while (my($key, $value) = each %vmString) {
            if ($value eq $modName) {
              $mass_diff = $vmMass{$key};
              $mass_total = $mass_diff + $objMassesFile->getResidueMass($mass_type, uc $this_spec);
            }
          }
        } else {
          $mass_total = $groupMasses{lc($this_spec)};
          $mass_diff = $mass_total - $objMassesFile->getResidueMass($mass_type, uc $this_spec);
        }
        push @retVal, $this_spec;
        push @retVal, sprintf("%+." . $massDP . "f", $mass_diff);
        push @retVal, $mass_total;
        push @retVal, $treatasvariable;
        push @retVal, $terminalMod;
        push @retVal, $modName;
      }
    }
  }
  return @retVal;

}

###############################################################################
# &getTerminalMods()
# returns terminal_modification elements
# $_[0] Mascot modification name
# $_[1] "N" or "Y" for variable mods
###############################################################################

sub getTerminalMods {
  
  my ($modName, $variable) = @_;

  my @retVal;
  my $treatasvariable = $variable;
  my $protein_terminus = "";
  my $objMod = $objModFile->getModificationByName($modName);
  if ($objMod) {
    if ($objMod->getModificationType() == $msparser::MOD_TYPE_N_TERM
      || $objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM
      || $objMod->getModificationType() == $msparser::MOD_TYPE_N_PROTEIN
      || $objMod->getModificationType() == $msparser::MOD_TYPE_C_PROTEIN) {
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_N_PROTEIN) {
        $treatasvariable = "Y";
        $protein_terminus = "n";
      }
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_C_PROTEIN) {
        $treatasvariable = "Y";
        $protein_terminus = "c";
      }
      my $term_end = "n";
      my $term_mass = $objMassesFile->getNterminalMass($msparser::MASS_TYPE_MONO);
      if ($objMod->getModificationType() == $msparser::MOD_TYPE_C_TERM 
        || $objMod->getModificationType() == $msparser::MOD_TYPE_C_PROTEIN) {
        $term_end = "c";
        $term_mass = $objMassesFile->getCterminalMass($mass_type);
      }
      push @retVal, $term_end;
      push @retVal, sprintf("%+." . $massDP . "f", $objMod->getDelta($mass_type));
      push @retVal, ($term_mass + $objMod->getDelta($mass_type));
      push @retVal, $treatasvariable;
      push @retVal, $protein_terminus;
      push @retVal, $modName;
    }
  } else {
  # mod not found in file, try to guess values
    if ($modName =~ /\(c-term\)/i
      || $modName =~ /\(n-term\)/i
      || $modName =~ /\(protein\)/i) {
      (my $specificity) = $modName =~ /\((.*)\)/;
      my ($this_spec, $term_end);
    # special cases
      if ($specificity =~ /protein/i) {
        $treatasvariable = "Y";
        $this_spec = "N_term";   # assumption
        $term_end = "n";         # assumption
        $protein_terminus = "n"; # assumption
      } elsif ($specificity =~ /c-term/i) {
        $this_spec = "C_term";
        $term_end = "c";
      } elsif ($specificity =~ /n-term/i) {
        $this_spec = "N_term";
        $term_end = "n";
      }
      my ($mass_diff, $mass_total);
      if ($treatasvariable eq "Y") {
        while (my($key, $value) = each %vmString) {
          if ($value eq $modName) {
            $mass_diff = $vmMass{$key};
            if ($this_spec eq 'C_term') {
              $mass_total = $mass_diff + $objMassesFile->getCterminalMass($mass_type);
            } else {
              $mass_total = $mass_diff + $objMassesFile->getNterminalMass($mass_type);
            }
          }
        }
      } else {
        $mass_total = $groupMasses{lc($this_spec)};
        if ($this_spec eq "C_term") {
          $mass_diff = $mass_total - $objMassesFile->getCterminalMass($mass_type);
        } else {
          $mass_diff = $mass_total - $objMassesFile->getNterminalMass($mass_type);
        }
      }
      push @retVal, $term_end;
      push @retVal, sprintf("%+." . $massDP . "f", $mass_diff);
      push @retVal, $mass_total;
      push @retVal, $treatasvariable;
      push @retVal, $protein_terminus;
      push @retVal, $modName;
    }
  }
  return @retVal;

}

###############################################################################
# &pepXML()
# output ISB pepXML elements
# my($objParams);
###############################################################################

sub pepXML {  

  my %query_level_mod_cache = ();

# msms_run_summary
  print "<msms_run_summary";
  print " base_name=\"\"";
  print " raw_data_type=\"\"";
  print " raw_data=\"\"";
  # optional msManufacturer
  # optional msModel
  # optional msIonization
  # optional msMassAnalyzer
  # optional msDetector
  print ">\n";

# sample_enzyme
  print "<sample_enzyme";
  print " name=\"" . &noXmlTag($objParams->getCLE()) . "\"";
  print " description=\"\"";
  my %enzymeDetails;
  eval { %enzymeDetails = getEnzymeDef($objParams->getCLE()) };
  if (my $err = $@) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
      print $err;
      common_subs::multipart_final($MIME_boundary);
    }
    die $err;
  }

  if ($enzymeDetails{'semispecific'}) {
    print " fidelity=\"semispecific\"";
  } elsif ($objParams->getCLE() eq "None") {
    print " fidelity=\"nonspecific\"";
  } else {
    print " fidelity=\"specific\"";
  }
  print " independent=\"" . $enzymeDetails{'independent'} . "\"";
  if ($objParams->getCLE() eq "None") {
    print " />\n";
  } else {
  # specificities
    print ">\n";
    for my $i (0 .. $enzymeDetails{'num_spec'}-1) {
      print "<specificity";
      if (defined $enzymeDetails{'spec_cleavage_end'}[$i]) {
        print " sense=\"" . $enzymeDetails{'spec_cleavage_end'}[$i] . "\"";
      } else {
        print " sense=\"\"";
      }
      print " min_spacing=\"1\"";
      if (defined $enzymeDetails{'spec_cleavage'}[$i]) {
        print " cut=\"" . $enzymeDetails{'spec_cleavage'}[$i] . "\"";
      } else {
        print " cut=\"\"";
      }
      if (defined $enzymeDetails{'spec_restrict'}[$i]) {
        print " no_cut=\"" . $enzymeDetails{'spec_restrict'}[$i] . "\"";
      } else {
        print " no_cut=\"\"";
      }
      print " />\n";
    }
    print "</sample_enzyme>\n";
  }

# search_summary
  print "<search_summary";
  if (my $fileset = $thisScript->param($urlParams{'fileset'})) {
    print " base_name=\"" . $URI . "cgi/master_results_2.pl?fileset=" . $fileset. "\"";
  } else {
    print " base_name=\"" . $URI . "cgi/master_results.pl?file=" . $fileIn[0] . "\"";
  }
  print " search_engine=\"MASCOT\"";
  print " precursor_mass_type=\"" . lc($objParams->getMASS()) . "\"";
  print " fragment_mass_type=\"" . lc($objParams->getMASS()) . "\"";
  print " out_data_type=\"\"";
  print " out_data=\"\"";
  print " search_id=\"1\">\n";
 
# search_database
# NB pepXML does not support multiple databases, so just provide details for first
# and list any others in search_summary/parameter elements
  print "<search_database";
  my $databasePath = &noXmlTag($objResFile->getFastaPath(1));
  print " local_path=\"$databasePath\"";
  # optional URL
  print " database_name=\"" . &noXmlTag($objParams->getDB(1)) . "\"";
  # optional orig_database_url
  # optional database_release_date
  print " database_release_identifier=\"" . &noXmlTag($objResFile->getFastaVer(1)) . "\"";
  # wait for parser function ### debug ###
  print " size_in_db_entries=\"" . $objResFile->getNumSeqs . "\"";
  print " size_of_residues=\"" . $objResFile->getNumResidues . "\"";
  if ($objSummary->isNA) {
    print " type=\"NA\"";
  } else {
    print " type=\"AA\"";
  }
  print " />\n";

# enzymatic_search_constraint
  print "<enzymatic_search_constraint";
  print " enzyme=\"" . &noXmlTag($objParams->getCLE()) . "\"";
  print " max_num_internal_cleavages=\"" . $objParams->getPFA() . "\"";
  my $min_number_termini = 2;
  if ($enzymeDetails{'semispecific'}) {
    $min_number_termini = 1;
  } elsif ($objParams->getCLE() eq "None") {
    $min_number_termini = 0;
  }
  print " min_number_termini=\"$min_number_termini\" />\n";

# sequence_search_constraint
# output at query level

# aminoacid_modifications
# fixed
  my(@res_mods, @term_mods, $iCount);
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;  
    while ($objParams->getFixedModsName($iCount)) {
      push @res_mods, &getResidueMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @res_mods, &getResidueMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;  
  while ($objParams->getVarModsName($iCount)) {
    push @res_mods, &getResidueMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  while (@res_mods) {
    print "<aminoacid_modification";
    print " aminoacid=\"" . shift(@res_mods) . "\"";
    print " massdiff=\"" . shift(@res_mods) . "\"";
    print " mass=\"" . shift(@res_mods) . "\"";
    print " variable=\"" . shift(@res_mods) . "\"";
    my $terminalMod = shift(@res_mods);
    if ($terminalMod) {
      print " peptide_terminus=\"$terminalMod\"";
    }
    # optional symbol
    print " binary=\"N\"";
    print " description=\"" . &noXmlTag(shift(@res_mods)) . "\" />\n";
  }
  
# terminal_modifications
# fixed
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;  
    while ($objParams->getFixedModsName($iCount)) {
      push @term_mods, &getTerminalMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @term_mods, &getTerminalMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;  
  while ($objParams->getVarModsName($iCount)) {
    push @term_mods, &getTerminalMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  while (@term_mods) { 
    print "<terminal_modification";
    print " terminus=\"" . shift(@term_mods) . "\"";
    print " massdiff=\"" . shift(@term_mods) . "\"";
    print " mass=\"" . shift(@term_mods) . "\"";
    print " variable=\"" . shift(@term_mods) . "\"";
    # optional symbol
    print " protein_terminus=\"" . shift(@term_mods) . "\"";
    print " description=\"" . &noXmlTag(shift(@term_mods)) . "\" />\n";
  }   

# parameter
  if($objParams->getNumberOfDatabases > 1) {
    for (my $i=2; $i <= $objParams->getNumberOfDatabases; $i++) {
      print "<parameter name=\"local_path_$i\" value=\"" . &noXmlTag($objResFile->getFastaPath($i)) . "\" />\n";
      print "<parameter name=\"database_name_$i\" value=\"" . &noXmlTag($objParams->getDB($i)) . "\" />\n";
      print "<parameter name=\"database_release_identifier_$i\" value=\"" . &noXmlTag($objResFile->getFastaVer($i)) . "\" />\n";
      # wait for parser function ### debug ###
      print "<parameter name=\"size_in_db_entries_$i\" value=\"" . $objResFile->getNumSeqs . "\" />\n";
      print "<parameter name=\"size_of_residues_$i\" value=\"" . $objResFile->getNumResidues . "\" />\n";
    }
  }
  my $index = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_PARAMETERS, $index)) {
    my $nextValue = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, $nextKey);
    $nextKey =~ s/\s*$//;
    if ($nextValue) {
      print "<parameter name=\"" . &noXmlTag($nextKey) . "\" value=\"" . &noXmlTag($nextValue) . "\" />\n";
    }
    $index++;
  }

# end of search_summary
  print "</search_summary>\n";
  
  my $print_prot_desc = $thisScript->param($urlParams{'prot_desc'});
  my $print_prot_mass = $thisScript->param($urlParams{'prot_mass'});
  my $print_prot_pI = $thisScript->param($urlParams{'prot_pi'});

  my $max_p = $objSummary->getMaxRankValue();

# loop through all matched queries
  my $string = "";
  my $needClosingTag;
  for my $queryNum (1 .. $objResFile->getNumQueries()) {
    $needClosingTag = 0;
    for my $rank (1 .. $max_p) {
      my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
      if (defined($objPeptide)) {
        my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
        if ($objPeptide->getAnyMatch()) {
          my $srcRank = $objSummary->getSrcRank($queryNum, $rank);
          my $srcSection = $objSummary->getSrcSection($queryNum, $rank);
          my $tempString = $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank");
          my($left, $right) = split(/;/, $tempString, 2);
          my @protein_list = $right =~ /"(.+?)"/g;
          $tempString = $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank\_db");
          my @db_list;
          if ($tempString) {
            @db_list = $tempString =~ /(\d\d)/g;
          } else {
            for (my $i = 0; $i <= $#protein_list; $i++) {
              $db_list[$i] = "01";
            }
          }
          $tempString = $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank\_terms");
          my($peptide_prev_aa, $peptide_next_aa) = $tempString =~ /^(.),(.)/;
          if ($rank == 1) {
          # spectrum_query
            $needClosingTag = 1;
            print "<spectrum_query";
            print " spectrum=\"" . &noXmlTag($objQuery->getStringTitle(1)) . "\"";
            print " start_scan=\"0\"";
            print " end_scan=\"0\"";
            print " precursor_neutral_mass=\"" . $objPeptide->getMrExperimental . "\"";
            print " assumed_charge=\"" . $objResFile->getObservedCharge($queryNum) . "\"";
            my $repeatSearchString = $objResFile->getRepeatSearchString($queryNum);
            $repeatSearchString =~ s/^[0-9.]+\s*//;
            $repeatSearchString =~ s/from\(.+?\)\s*//;
            $repeatSearchString =~ s/title\(.+?\)\s*//;
            $repeatSearchString =~ s/query\(.+?\)\s*//;
            $repeatSearchString =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex($1))/eg;
            $repeatSearchString =~ s/^\s*//;
            $repeatSearchString =~ s/\s*$//;
            if ($repeatSearchString) {
              print " search_specification=\"" . &noXmlTag($repeatSearchString) . "\"";
            }
            print " index=\"$queryNum\">\n";
          # search_result
            print "<search_result";
            print " search_id=\"1\">\n";
          }
          
        # search_hit
          print "<search_hit";
          print " hit_rank=\"$rank\"";
          print " peptide=\"" . $objPeptide->getPeptideStr() . "\"";
          if ($peptide_prev_aa) {
          # only available for 2.1 and later
            print " peptide_prev_aa=\"$peptide_prev_aa\"";
            print " peptide_next_aa=\"$peptide_next_aa\"";
          }
          if ($db_list[0] > 1) {
            print " protein=\"" . &noXmlTag(($db_list[0] + 0) . "::" . $protein_list[0]) . "\"";
          } else {
            print " protein=\"" . &noXmlTag($protein_list[0]) . "\"";
          }
          print " num_tot_proteins=\"" . scalar(@protein_list) . "\"";
          print " num_matched_ions=\"" . $objPeptide->getNumIonsMatched() . "\"";
          # optional tot_num_ions
          print " calc_neutral_pep_mass=\"" . $objPeptide->getMrCalc . "\"";
          print " massdiff=\"" 
            . sprintf("%+." . $massDP . "f", $objPeptide->getDelta) 
            . "\"";
          # optional num_tol_term
          print " num_missed_cleavages=\"" . $objPeptide->getMissedCleavages . "\"";
          print " is_rejected=\"0\"";
          my $prot_desc = "";
          if ($print_prot_desc) {
            $prot_desc = &noXmlTag(common_subs::mustGetProteinDescription($objSummary, $objParams, $shipper, $protein_list[0], $db_list[0]));
          }
          if ($prot_desc) {
            print " protein_descr=\"$prot_desc\"";
          }
          my $frame = 0;
          if (($print_prot_pI || $print_prot_mass) && $objSummary->isNA) {
            my $objProtein = $objSummary->getProtein($protein_list[0], $db_list[0]);
            if ($objProtein) {
              $frame = $objProtein->getFrame();
            }
          }
          if ($print_prot_pI) {
            my $pI = getProteinpI($objParams, $protein_list[0], $frame, $db_list[0]);
            if ($pI) {
              print " calc_pI=\"$pI\"";
            }
          }
          my $protein_mw = 0;
          if ($print_prot_mass) {
            $protein_mw = mustGetProteinMass($protein_list[0], $objSummary, $objParams, $frame, $db_list[0]);
          }
          if ($protein_mw) {
            print " protein_mw=\""
              . sprintf("%.0f", $protein_mw)
              . "\"";
          }
          print ">\n";
          
          if (scalar(@protein_list) > 1) {
          # alternative_protein
            for (my $prot_num = 1; $prot_num <= $#protein_list; $prot_num++) {
              print "<alternative_protein";
              if ($db_list[$prot_num] > 1) {
                print " protein=\"" . &noXmlTag(($db_list[$prot_num] + 0) . "::" . $protein_list[$prot_num]) . "\"";
              } else {
                print " protein=\"" . &noXmlTag($protein_list[$prot_num]) . "\"";
              }
              if ($print_prot_desc) {
              # only look for description in result file for alternative proteins
                if ($tempString = &noXmlTag($objSummary->getProteinDescription($protein_list[$prot_num], $db_list[$prot_num]))) {
                  print " protein_descr=\"$tempString\"";
                }
              }
              # optional num_tol_term
              if ($print_prot_mass) {
              # only look for mass in result file for alternative proteins
                if ($tempString = $objSummary->getProteinMass($protein_list[$prot_num], $db_list[$prot_num])) {
                  print " protein_mw=\""
                    . sprintf("%.0f", $tempString)
                    . "\"";
                }
              }
              print " />\n";
            }
          }
          
          my $varModsStr = $objPeptide->getVarModsStr();
          my $summedModsStr = $objPeptide->getSummedModsStr() eq '' ? '0' x length($varModsStr) : $objPeptide->getSummedModsStr();
          my $local_mods = lookup_query_level_mods(\%query_level_mod_cache, $objResFile, $objPeptide, $objUmodConfigFile2) || {};
          if ($varModsStr =~ /[1-9A-X]/ or $summedModsStr =~ /[1-9A-X]/ or %$local_mods) {
          # modification_info
          # grab any mod found in error tolerant search
            $tempString = 
                $objResFile->getSectionValueStr($srcSection, "q$queryNum\_p$srcRank\_et_mods");
            if ($tempString) {
              ($vmMass{'X'}, undef, $vmString{'X'}) = split(/,/, $tempString, 3);
            }
            print "<modification_info";
          # N term
            if (substr($varModsStr, 0, 1) =~ /[1-9A-X]/ or substr($summedModsStr, 0, 1) =~ /[1-9A-X]/ or $$local_mods{0}){
              my $mass = $objMassesFile->getNterminalMass($mass_type);
              for my $m (substr($varModsStr, 0, 1), substr($summedModsStr, 0, 1)) {
                $mass += $vmMass{$m} if $m =~ /[1-9A-X]/;
              }
              if ($$local_mods{0}) {
                $mass += $$local_mods{0}{'delta'};
              }
              print " mod_nterm_mass=\"" . $mass . "\"";
            }
          # C term
            if (substr($varModsStr, -1, 1) =~ /[1-9A-X]/ or substr($summedModsStr, -1, 1) or $$local_mods{length($varModsStr)-1}){
              my $mass = $objMassesFile->getCterminalMass($mass_type);
              for my $m (substr($varModsStr, -1, 1), substr($summedModsStr, -1, 1)) {
                $mass += $vmMass{$m} if $m =~ /[1-9A-X]/;
              }
              if ($$local_mods{length($varModsStr)-1}) {
                $mass += $$local_mods{length($varModsStr)-1}{'delta'};
              }
              print " mod_cterm_mass=\"" . $mass . "\"";
            }
            # optional modified_peptide
            print ">\n";
            
          # residues
            for (my $i = 1; $i < (length($varModsStr) - 1); $i++){
              if (substr($varModsStr, $i, 1) =~ /[1-9A-X]/ or substr($summedModsStr, $i, 1) =~ /[1-9A-X]/ or $$local_mods{$i}) {
                my $mass = $objMassesFile->getResidueMass($mass_type, uc substr($objPeptide->getPeptideStr(), $i - 1, 1));
                for my $m (substr($varModsStr, $i, 1), substr($summedModsStr, $i, 1)) {
                  $mass += $vmMass{$m} if $m =~ /[1-9A-X]/;
                }
                if ($$local_mods{$i}) {
                  $mass += $$local_mods{$i}{'delta'};
                }
                print "<mod_aminoacid_mass";
                print " position=\"$i\"";
                print " mass=\"" . $mass . "\" />\n";
              }
            }
            
            print "</modification_info>\n";
          }
        
        # search_score
          print "<search_score name=\"ionscore\" value=\"" 
            . sprintf("%.2f", $objPeptide->getIonsScore) . "\" />\n";
          print "<search_score name=\"identityscore\" value=\"" 
            . $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold) . "\" />\n";
          if ($objSummary->getHomologyThreshold($queryNum, 1 / $sigThreshold)) {
            print "<search_score name=\"homologyscore\" value=\"" 
              . $objSummary->getHomologyThreshold($queryNum, 1 / $sigThreshold) . "\" />\n";
          }
          print "<search_score name=\"expect\" value=\"" 
            . sprintf("%.2g", $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $queryNum)) . "\" />\n";
          # optional type
          print "</search_hit>\n";
          
        } else {
          if ($rank == 1) {
            print "<spectrum_query";
            print " spectrum=\"" . &noXmlTag($objQuery->getStringTitle(1)) . "\"";
            print " start_scan=\"0\"";
            print " end_scan=\"0\"";
            print " precursor_neutral_mass=\"" . $objResFile->getObservedMrValue($queryNum) . "\"";
            print " assumed_charge=\"" . $objResFile->getObservedCharge($queryNum) . "\"";
            my $repeatSearchString = $objResFile->getRepeatSearchString($queryNum);
            $repeatSearchString =~ s/^[0-9.]+\s*//;
            $repeatSearchString =~ s/from\(.+?\)\s*//;
            $repeatSearchString =~ s/title\(.+?\)\s*//;
            $repeatSearchString =~ s/query\(.+?\)\s*//;
            $repeatSearchString =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex($1))/eg;
            $repeatSearchString =~ s/^\s*//;
            $repeatSearchString =~ s/\s*$//;
            if ($repeatSearchString) {
              print " search_specification=\"" . &noXmlTag($repeatSearchString) . "\"";
            }
            print " index=\"$queryNum\" />\n";
          }
          last;
        }
      } else {
        last;
      }
    }
    if ($needClosingTag) {
      print "</search_result>\n";
      print "</spectrum_query>\n";
    }
  }

  print "</msms_run_summary>\n";
  print "</msms_pipeline_analysis>\n";
  
}

###############################################################################
# &getEnzymeDef()
# $_[0] enzyme title
# returns a hash:
#   - 'independent' (1 or 0)
#   - 'semispecific' (1 or 0)
#   - 'num_spec' (number of specificities)
#   - 'spec_cleavage' (array of cleavage strings for each specificity)
#   - 'spec_restrict' (array of restriction strings for each specificity)
#   - 'spec_cleavage_end' (array of C/N for each specificity)
###############################################################################

sub getEnzymeDef {
  my ($title) = @_;

# get enzyme cleavage rules
  open(my $fh, "../config/enzymes") or die "cannot open enzymes file: $! ($^E)";

  my $title_re = qr{^Title:\Q$title\E\s*$}i;

  my $found = 0;
  my $independent = 0;
  my $semispecific = 0;
  my (@cleavage, @restrict, @cleaveEnd);

  while (<$fh>) {
    next unless $_ =~ $title_re;

    # We've found the correct title line; now let's read the attributes.
    $found = 1;

    while (<$fh>) {
      last if /^\*/; # end of definition

      chomp;
      s/\s*$//;  # delete trailing white space

      if (/^Cleavage:(\w+)/i){
        $cleavage[0] = $1;
      } elsif (/^Restrict:(\w+)/i){
        $restrict[0] = $1;
      } elsif (/^Restrict:/i){
        # ignore
      } elsif (/^(\w)term$/i){
        $cleaveEnd[0] = $1;
      } elsif (/^Cleavage\[(\d+)\]:(\w+)/i){
        $cleavage[$1] = $2;
      } elsif (/^Restrict\[(\d+)\]:(\w+)/i){
        $restrict[$1] = $2;
      } elsif (/^Restrict\[(\d+)\]:/i){
        # ignore
      } elsif (/^(\w)term\[(\d+)\]/i){
        $cleaveEnd[$2] = $1;
      } elsif (/^Independent:(\d)/i){
        $independent = $1;
      } elsif (/^SemiSpecific:(\d)/i){
        $semispecific = $1;
      } else {
        die "Don't understand the enzyme file format (line $.): $_";
      }

      next;
    }

    last;
  }

  close($fh) or die "cannot close enzymes file: $! ($^E)";

  if (not $found) {
    die "could not find enzyme definition for $title";
  }

# Sanity checks
  unless (defined($cleavage[0]) && $cleavage[0]) {
    die "invalid enzyme definition for $_[0]";
  }
  if (@cleaveEnd != @cleavage || @restrict > @cleavage) {
    die "invalid enzyme definition for $_[0]";
  }
  foreach (@cleavage) {
    unless ($_) {
      die "invalid enzyme definition for $_[0]";
    }
  }
  foreach (@cleaveEnd) {
    unless ($_ eq "N" || $_ eq "C") {
      die "invalid enzyme definition for $_[0]";
    }
  }

  return (
    independent => $independent,
    semispecific => $semispecific,
    num_spec => scalar(@cleavage),
    spec_cleavage => \@cleavage,
    spec_restrict => \@restrict,
    spec_cleavage_end => \@cleaveEnd,
  );
}

###############################################################################
# &DTASelect()
# output Scripps DTASelect file
# my($objParams, $objResFile, $objSummary, $msresFlags);
###############################################################################

sub DTASelect {

  my $sameset_ordering = common_subs::samesetOrderingByPreferredTaxonomy() || {};

# pre-calculate total intensity values
  my @totalIntensity;
  for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries(); $queryNum++){
    my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
    $totalIntensity[$queryNum] = $objQuery->getTotalIonsIntensity();
  }

# terminal_modifications
# fixed
  my(@term_mods, $iCount);
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;  
    while ($objParams->getFixedModsName($iCount)) {
      push @term_mods, &getTerminalMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @term_mods, &getTerminalMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;  
  while ($objParams->getVarModsName($iCount)) {
    push @term_mods, &getTerminalMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }
  
  my $fixProtC = 0;
  my $fixProtN = 0;
  my $fixPepC = 0;
  my $fixPepN = 0;
  while (@term_mods) { 
    my $terminus = shift(@term_mods);
    my $massdiff = shift(@term_mods);
    my $mass = shift(@term_mods);
    my $variable = shift(@term_mods);
    my $protein_terminus = shift(@term_mods);
    my $description = shift(@term_mods);
    if ($variable eq "N") {
      if ($protein_terminus eq "c") {
        $fixProtC = $massdiff;
      } elsif ($protein_terminus eq "n") {
        $fixProtN = $massdiff;
      } elsif ($terminus eq "c") {
        $fixPepC = $massdiff;
      } elsif ($terminus eq "n") {
        $fixPepN = $massdiff 
      }
    }
  }   

# output header lines
  print  "DTASelect v1.9\n";
  print  "\n";
  print  "\n";
  print  "S\t" . $objResFile->getFastaVer(1) . "\t$fixProtC\t$fixProtN\t$fixPepC\t$fixPepN\tfalse\ttrue\n";
    
# aminoacid_modifications
# fixed
  my @res_mods;
  if ($objParams->getFixedModsName(1)) {
    $iCount = 1;  
    while ($objParams->getFixedModsName($iCount)) {
      push @res_mods, &getResidueMods($objParams->getFixedModsName($iCount), "N");
      $iCount++;
    }
  } else {
    my $mod_string = $objParams->getMODS();
    if ($mod_string) {
      my @mod_list = split(/,/, $mod_string);
      foreach my $modName (@mod_list) {
        push @res_mods, &getResidueMods($modName, "N");
      }
    }
  }
# variable
  $iCount = 1;  
  while ($objParams->getVarModsName($iCount)) {
    push @res_mods, &getResidueMods($objParams->getVarModsName($iCount), "Y");
    $iCount++;
  }

  # 'use warnings' complains if # is inside qw().
  my @symbols = (qw(* @ $ % & ~ ? / + = 1 2 3 4 5 6 7 8 9), '#');
  my %symLink;
  while (@res_mods) {
    my $aminoacid = shift(@res_mods);
    my $massdiff = shift(@res_mods);
    my $mass = shift(@res_mods);
    my $variable = shift(@res_mods);
    my $terminalMod = shift(@res_mods);
    my $description = shift(@res_mods);
    if ($variable eq "N") {
      print "SM\t$massdiff\t$aminoacid\n";
    } else {
      my $symbol = shift(@symbols);
      if (not defined $symbol) {
        common_subs::fatal("DTASelect cannot handle this many variable mods", __LINE__, __FILE__, $myName, $idForFatalError, $MIME_boundary);
      }
      ${ $symLink{$description} }{$aminoacid} = $symbol;
      print "DM\t$massdiff\t$symbol\t$aminoacid\t0.0\t0.0\n";
    }
  }

  print  "Source\tMascot\t" . $objResFile->getMascotVer() . "\n";
  print  "Format\tMascot\n";

# start of MAIN LOOP to enumerate the protein hits
  my $numL = 0;
  my $numD = 0;
  my $thisHit = 1;  
  my $objProtein = $objSummary->getHit($thisHit);
  while (defined($objProtein)) {
    my @sameset = common_subs::compileSameset($objSummary, $objProtein, 
      $msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS,
      $$sameset_ordering{$thisHit}
    );
    
    &outputL($sameset[0]);
    $numL++;
    &outputD($sameset[0], \@totalIntensity, \%symLink);
    $numD++;
    if ($msresFlags & $msparser::ms_mascotresults::MSRES_GROUP_PROTEINS) {
      for (@sameset[1 .. $#sameset]) {
        &outputL($_);
        $numL++;
        &outputD($_, \@totalIntensity, \%symLink);
        $numD++;
      }
    # $ShowSubSets hard coded to 1 for DTASelect
      $i = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($thisHit, $i++)) {
        if ($objProtein->getScore() > 0) {
          &outputL($objProtein);
          $numL++;
          &outputD($objProtein, \@totalIntensity, \%symLink);
          $numD++;
        }
      }
    }

    $thisHit++;
    $objProtein = $objSummary->getHit($thisHit);
    
  }
# end of MAIN LOOP to enumerate the protein hits
  
  print  "C\t$numL\t" . $objResFile->getNumQueries() . "\t$numD\n";

}

###############################################################################
# &outputL()
# output DTASelect L line
# $_[0] $objProtein
# my($objParams, $objSummary);
###############################################################################

sub outputL {
  
  my $objProtein = shift;

  my ($Length, $Description, $MW, $pI);
  my $preferred_acc = common_subs::preferredAccession($objProtein->getDB(), $objProtein->getAccession());
  if ($thisScript->param($urlParams{'prot_desc'})) {
    $Description = common_subs::mustGetProteinDescription($objSummary, $objParams, $shipper, $preferred_acc, $objProtein->getDB);
  }
  $Description = "no description" unless ($Description);
  if ($thisScript->param($urlParams{'prot_mass'})) {
    $MW = mustGetProteinMass($objProtein->getAccession(), $objSummary, $objParams, $objProtein->getFrame(), $objProtein->getDB);
  }
  $MW = 0 unless ($MW);
  if ($thisScript->param($urlParams{'prot_len'})) {
    $Length = getProteinLen($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  } elsif ($MW) {
    $Length = int($MW / 110);
  }
  $Length = 0 unless ($Length);
  if ($thisScript->param($urlParams{'prot_pi'})) {
    $pI = getProteinpI($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  }
  $pI = 0 unless ($pI);

  my $dbNumPrefix = "";
  if ($objParams->getNumberOfDatabases > 1 && $objProtein->getDB > 1) {
    $dbNumPrefix = $objProtein->getDB . "::";
  }
  
  print  join("\t", 
    "L", 
    $dbNumPrefix . $preferred_acc,
    $Length, 
    $MW, 
    $pI, 
    $Description, 
    "U\n");
  
}

###############################################################################
# &outputD()
# output DTASelect D line
# $_[0] $objProtein
# $_[1] reference to @totalIntensity
# $_[2] reference to %symLink
# my($objResFile, $objSummary, $fileIn);
###############################################################################

sub outputD {
  
  my ($objProtein, $totalIntensity_ref, $symLink_ref) = @_;
  
# Enumerate the peptides that matched this protein
  for (my $i = 1; $i <= $objProtein->getNumPeptides(); $i++) {
    my $queryNum = $objProtein->getPeptideQuery($i);
    my $objPeptide = $objSummary->getPeptide($queryNum, $objProtein->getPeptideP($i));
    my $ionSeriesString = $objPeptide->getSeriesUsedStr();
    my $numScoreSeries = $ionSeriesString =~ tr/2//;
    my $unique2locus = "true";
    if ($objPeptide->getNumProteins() > 0) {
      $unique2locus = "false";
    }
    
    my $local_pep = $objPeptide->getPeptideStr();

  # use the variable mods string to interpolate variable mod symbols into peptide string
  # Note: neither summed nor query-level modifications can be exported here
    my $varModsString = $objPeptide->getVarModsStr();
    unless ($varModsString eq "0" x length($varModsString)) {
      for (my $i = length($varModsString) - 2; $i > 0; $i--) {
        if (substr($varModsString, $i, 1) =~ /[1-9A-X]/) {
          my $symbol = ${ ${ $symLink_ref }{$vmString{substr($varModsString, $i, 1)}} }{substr($local_pep, $i - 1, 1)};
          substr($local_pep, $i - 1, 1) = substr($local_pep, $i - 1, 1) . $symbol;
        }
      }
    }

    my $aa_before = $objProtein->getPeptideResidueBefore($i);
    my $aa_after = $objProtein->getPeptideResidueAfter($i);
 #  unless ($aa_before eq "?" || $aa_after eq "?") {
      $aa_before =~ s/@/-/;
      $aa_after =~ s/@/-/;
      $local_pep = "$aa_before.$local_pep.$aa_after";
 #  }

    print  join("\t", 
      "D",
      "$fileIn[0].$queryNum." . $objPeptide->getCharge(),
      $fileIn[0] =~ /^(.*).dat/,
      sprintf("%.2f", $objPeptide->getIonsScore),
      sprintf("%.1f", $objSummary->getProbOfPepBeingRandomMatch($objPeptide->getIonsScore(), $queryNum)),
      ($objPeptide->getMrExperimental + $elementalMasses{'hydrogen'} - $elementalMasses{'electron'}),
      ($objPeptide->getMrCalc() + $elementalMasses{'hydrogen'} - $elementalMasses{'electron'}),
      sprintf("%.1f", ${ $totalIntensity_ref }[$queryNum]),
      $objPeptide->getPrettyRank(),
      $objSummary->getPeptideIdentityThreshold($queryNum, 1 / $sigThreshold),
      sprintf("%.3f", $objPeptide->getNumIonsMatched() / $numScoreSeries / (length($objPeptide->getPeptideStr()) - 1)),
      $local_pep,
      ($objProtein->getPeptideStart($i) - 1),
      "0",
      $unique2locus,
      "1",
      "U\n");
        
  }

}

###############################################################################
# &mascotDAT($thisScript, $isCGI, $fileIn, $objMascotOptions)
# download raw Mascot result file
###############################################################################

sub mascotDAT {
  my ($thisScript, $isCGI, $fileIn, $objMascotOptions, $MIME_boundary) = @_;

  # The Mascot data directory can be anywhere, so it's best to look at 
  # the actual directory the input filepath points to and see if it's
  # the same directory as the Mascot data directory. (See bug #11057.)

  my $dirjoin;

  if ($objMascotOptions->getInterFileBasePath() =~ /\\/) {
    # Assume Windows.
    $dirjoin = '\\';
    $fileIn =~ s#/#\\#g;
  } else {
    $dirjoin = '/';
    $fileIn =~ s#\\#/#g;
  }

  my ($last_bit) = $fileIn =~ m#([/\\]\d{8}[/\\]F[a-zA-Z\d]+\.dat)$#;

  if (not defined $last_bit) {
    return _daily_only($thisScript, $isCGI, $MIME_boundary);
  }

  my $parentdir = Cwd::abs_path(
    substr($fileIn, 0, length($fileIn) - length($last_bit))
  );

  if ($parentdir eq '') {
    return _daily_only($thisScript, $isCGI, $MIME_boundary);
  }

  my $datadir = do {
    my $base = $objMascotOptions->getInterFileBasePath(),
    # Despite the relativity implied by the name, InterFileRelPath can
    # actually be an absolute pathname.
    my $rel = $objMascotOptions->getInterFileRelPath();

    if ($rel =~ m#^[.]#) {
      Cwd::abs_path(join($dirjoin, $base, $rel))
    } else {
      Cwd::abs_path($rel)
    }
  };

  if ($datadir eq '') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
    } else {
      print $thisScript->header( 
        -charset => "utf-8", 
        -type => 'text/plain',
        -X_Content_Type_Options => 'nosniff',
        -Content_Disposition => 'inline;filename=error.txt',
      );
    }
    print "Unable to determine the location of the Mascot data directory\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    return 1;
  }

  # Bug #11354: if InterFileBasePath or fileIn are absolute pathnames,
  # and the paths are correct, then Cwd::abs_path() retains the case of
  # the input strings. On Linux, which has case-sensitive filesystems,
  # we can be sure that if the two strings are equal, the paths are the
  # same. However, on Windows, two paths with different cases can be the same
  # path; for example, one could start with C:/, while the other starts
  # with c:/. We must force the strings into the same case before comparison.

  if ($^O =~ /Win/) {
    $parentdir = uc($parentdir);
    $datadir = uc($datadir);
  }

  if ($parentdir ne $datadir) {
    return _daily_only($thisScript, $isCGI, $MIME_boundary);
  }

  my $filepath = common_subs::decompress($fileIn);

  if ($filepath eq '') {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
    } else {
      print $thisScript->header( 
        -charset => "utf-8", 
        -type => 'text/plain',
        -X_Content_Type_Options => 'nosniff',
        -Content_Disposition => 'inline;filename=error.txt',
      );
    }
    print "Results file $fileIn does not exist\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    return 1;
  }

  open(my $fh, '<', $filepath) or do {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary);
    } else {
      print $thisScript->header( 
        -charset => "utf-8", 
        -type => 'text/plain',
        -X_Content_Type_Options => 'nosniff',
        -Content_Disposition => 'inline;filename=error.txt',
      );
    }
    print "Unable to open results file for reading: $! ($^E)\n";
    if (defined $MIME_boundary) {
      common_subs::multipart_final($MIME_boundary);
    }
    return 1;
  };
  
  if ($isCGI or defined $MIME_boundary) {
    my ($filename) = $last_bit =~ m#[/\\](F[a-zA-Z\d]+.dat)$#;

    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/plain; charset=utf-8', $filename);
    } else {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => $filename);
    }
  }

  print while <$fh>;

  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }

  return 0;
}

sub _daily_only {
  my ($thisScript, $isCGI, $MIME_boundary) = @_;

  if (defined $MIME_boundary) {
    common_subs::multipart_end($MIME_boundary);
    common_subs::multipart_start($MIME_boundary);
  } elsif ($isCGI) {
    print $thisScript->header(
      -charset => "utf-8",
      -type => 'text/plain',
      -X_Content_Type_Options => 'nosniff',
      -Content_Disposition => 'inline;filename=error.txt',
    );
  }
  print "You can only download result files from daily folders under the Mascot data directory\n";
  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }
  return 1;
}

###############################################################################
# &exportMGF()
# download MGF Peak List
# my($objResFile, $fileIn)
###############################################################################

sub exportMGF {
  my ($thisScript, $MIME_boundary) = @_;
  
# parse out result file name root
  my ($fileIn) = $fileIn[0] =~ /[\/\\](F.*)\.dat$/;

  if ($isCGI or defined $MIME_boundary) {
    if (defined $MIME_boundary) {
      common_subs::multipart_end($MIME_boundary);
      common_subs::multipart_start($MIME_boundary, 'text/plain; charset=utf-8', "$fileIn.mgf");
    } else {
      print $thisScript->header( -type => "text/plain",
                                 -charset => "utf-8",
                                 -attachment => "$fileIn.mgf");
    }
  }

  my $something_was_printed = 0;
  my $i = 1;
  while (my $nextKey = $objResFile->enumerateSectionKeys($msparser::ms_mascotresfile::SEC_PARAMETERS, $i)) {
    if ($nextKey =~ /^(?:_.*)/i || $nextKey =~ /^(?:user\d\d)$/i) {
      my $nextValue = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_PARAMETERS, $nextKey);
      $nextKey =~ s/\s*$//;
      if (defined $nextValue and $nextValue ne '') {
        print $nextKey.'='.$nextValue."\n";
        $something_was_printed = 1;
      }
    }
    $i++;
  }

  print "\n" if $something_was_printed;

  # List of parameters that can be embedded as query level parameters:
  # CHARGE
  # COMP
  # ETAG
  # INSTRUMENT
  # IT_MODS
  # LOCUS
  # PEPMASS
  # RAWFILE
  # RAWSCANS
  # RTINSECONDS
  # SCANS
  # SEQ
  # TAG
  # TITLE
  # TOL
  # TOLU

  if ($objResFile->anyMSMS() || $objResFile->anySQ() || $objResFile->anyTag()) {
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries; $queryNum++) {
      my $objQuery = new msparser::ms_inputquery($objResFile, $queryNum);
      print "BEGIN IONS\n";
      if ($objQuery->getCharge() ne '') {
        print "CHARGE=" . $objQuery->getCharge() . "\n";
      }
      _print_mgf_parameter_array($objQuery, 'getComp', 'COMP', 1, 20);
      if ($objQuery->getINSTRUMENT(1) ne '') {
        print "INSTRUMENT=" . $objQuery->getINSTRUMENT(1) . "\n";
      }
      if ($objQuery->getIT_MODS(1)) {
        print "IT_MODS=" . $objQuery->getIT_MODS(1) . "\n";
      }
      if ($objQuery->getLocus() ne '') {
        print "LOCUS=" . $objQuery->getLocus . "\n";
      }
      my $intensity = $objResFile->getObservedIntensity($queryNum) + 0;
      print "PEPMASS=" . $objResFile->getObservedMass($queryNum);
      if ($intensity != 0) {
        print " $intensity\n";
      } else {
        print "\n";
      }
      if ($objQuery->getRawfile() ne '') {
        print "RAWFILE=" . $objQuery->getRawfile() . "\n";
      }
      _print_mgf_parameter_array($objQuery, 'getRawScans', 'RAWSCANS', -1);
      _print_mgf_parameter_array($objQuery, 'getRetentionTimes', 'RTINSECONDS', -1);
      _print_mgf_parameter_array($objQuery, 'getScanNumbers', 'SCANS', -1);
      _print_mgf_parameter_array($objQuery, 'getSeq', 'SEQ', 1, 20);
      for (my $j = 1; $j <= 20; $j++) {
        if ($objQuery->getTag($j)) {
          $objQuery->getTag($j) =~ /([et]),(.*)/i;
          if (lc($1) eq "e") {
            print "ETAG=$2\n";
          } else {
            print "TAG=$2\n";
          }
        } else {
          last;
        }
      }
      if ($objQuery->getStringTitle(1) ne '') {
        print "TITLE=" . $objQuery->getStringTitle(1) . "\n";
      }
      if ($objQuery->getPepTol()) {
        print "TOL=" . $objQuery->getPepTol() . "\n";
      }
      if ($objQuery->getPepTolUnits()) {
        print "TOLU=" . $objQuery->getPepTolUnits() . "\n";
      }

=pod Not needed; Parser has methods for all supported parameters
      my $i = 1;
      while (my $nextKey = $objResFile->enumerateQuerySectionKeys($queryNum, $i++)) {
        my $key = $nextKey;
        $key =~ s/\s+$//;
        $key =~ s/^\s+//;
        next unless $embeddable_key{uc $key};

        my $nextValue = $objResFile->getQuerySectionValueStr($queryNum, $nextKey);
        if (defined $nextValue and $nextValue ne '') {
          print uc($key).'='.$nextValue."\n";
        }
      }
=cut

      my $peakStr = $objQuery->getStringIons1();
    # no way to represent ions(b- or ions(y- in MGF format, so just output values as unassigned
      if ($objQuery->getStringIons2()) {
        $peakStr .= "," . $objQuery->getStringIons2();
      }
      if ($objQuery->getStringIons3()) {
        $peakStr .= "," . $objQuery->getStringIons3();
      }
      if ($peakStr) {
        $peakStr =~ s/:/ /g;
        $peakStr =~ s/,/\n/g;
        print "$peakStr\n";
      }
      print "END IONS\n\n";
    }
  } else {
  # PMF
    for (my $queryNum = 1; $queryNum <= $objResFile->getNumQueries; $queryNum++) {
      my $intensity = $objResFile->getObservedIntensity($queryNum) + 0;
      print $objResFile->getObservedMass($queryNum);
      if ($intensity != 0) {
        print " $intensity\n";
      } else {
        print "\n";
      }
    }
  }
  
  if (defined $MIME_boundary) {
    common_subs::multipart_final($MIME_boundary);
  }

}

sub _print_mgf_parameter_array {
  my ($objQuery, $fun, $param, $min, $max) = @_;

  my $i = $min;
  while ((my $str = $objQuery->$fun($i++)) ne '') {
    if (defined $max) {
      return if $i > $max;
    }
    if ($i == -1) {
      print "$param=$str\n";
    } else {
      print "$param\[$i\]=$str\n";
    }
  }
}

###############################################################################
# &display_args()
# display command line arguments for copy and paste
###############################################################################

sub display_args {

  &common_subs::printHeader1("Matrix Science - Mascot - Export search results");
  &common_subs::printHeader2("", "Export search results", 1, $session);

  print <<'end_of_static_HTML_text_block';
  <TABLE BORDER=0 CELLSPACING=1 CELLPADDING=3>
    <TR>
      <TD NOWRAP ALIGN=LEFT><h2>Command line arguments equivalent to current form settings</h2></TD>
    </TR>
     <TR> 
       <TD>&nbsp;</TD> 
     </TR> 
    <TR>
      <TD ALIGN=LEFT><tt>
end_of_static_HTML_text_block

  my $args = _cmdline_arguments($thisScript, \%urlParams);
      
  for (sort keys %$args) {
    if ($$args{$_} =~ /\s/) {
      print lc($_) . "=\"" . $$args{$_} . "\" ";
    } else {
      print lc($_) . "=" . $$args{$_} . " ";
    }
  }
      
  print <<'end_of_static_HTML_text_block';
      </tt></TD>
    </TR>
  </TABLE>
end_of_static_HTML_text_block

  &common_subs::printFooter;

  print "</BODY>\n";
  print "</HTML>\n";
    
}

sub _cmdline_arguments {
  my ($thisScript, $urlParams) = @_;

  my %args = ();

  foreach ($thisScript->param) {
    if (lc($_) eq "do_export") {
      $args{'do_export'} = 1;
    } elsif (!$thisScript->param($$urlParams{lc $_})) {
    # skip
    } else {
      $args{lc($_)} = $thisScript->param($$urlParams{lc $_});
    }
  }

  return \%args;
}

###############################################################################
# &mzIdentML()
# output PSI mzIdentML elements
# my($objParams);
###############################################################################

sub mzIdentML {
  my $isPMF = $objResFile->isPMF;
  my $isSQ = $objResFile->isSQ;

  my $vecFixed = _make_mod_vector($objResFile, $objModFile, 'FixedMod');
  my $vecVariable = _make_mod_vector($objResFile, $objModFile, 'delta');

# make hash for fast retrieval of fixed mods
  my %fixedMods = _make_fixed_mods_map($vecFixed);

  my (@componentNames, $numcomponents, %aaHelpers);

  if ($objQuantMethod) {
    $numcomponents = $objQuantMethod->getNumberOfComponents; 

    for my $c (0 .. $numcomponents-1) {
      my $component = $objQuantMethod->getComponentByNumber($c);
      $componentNames[$c] = $component->getName;
      my $objLocalMassesFile = new msparser::ms_masses($objMassesFile); 
      if ($component->getNumberOfIsotopes > 0) {
        $objLocalMassesFile->applyIsotopes($objUmodConfigFile, $component);
      }
      $aaHelpers{$component->getName} = new msparser::ms_aahelper($objResFile, "../config/enzymes");
      $aaHelpers{$component->getName}->setMasses($objLocalMassesFile);
      $aaHelpers{$component->getName}->setAvailableModifications($vecFixed, $vecVariable);
    # success?
      if (!$aaHelpers{$component->getName}->isValid()) {
        die "Unable to create aahelper object: " . common_subs::htmlEscape($aaHelpers{$component->getName}->getLastErrorString());
      }
    }
  } else {
    $numcomponents = 1;
    $componentNames[0] = "1";
    my $objLocalMassesFile = new msparser::ms_masses($objMassesFile); 
    $aaHelpers{"1"} = new msparser::ms_aahelper($objResFile, "../config/enzymes");
    $aaHelpers{"1"}->setMasses($objLocalMassesFile);
    $aaHelpers{"1"}->setAvailableModifications($vecFixed, $vecVariable);
  # success?
    if (!$aaHelpers{"1"}->isValid()) {
      die "Unable to create aahelper object: " . common_subs::htmlEscape($aaHelpers{"1"}->getLastErrorString());
    }
  }

  my @resfile_params = ();
  my @resfile_children = ();

  for my $i (1 .. $objResFile->getNumberOfResfiles()) {
    my $resfile = $objResFile->getResfile($i);
    $resfile_children[$i] = $resfile;
    $resfile_params[$i] = $resfile->params;
  }

  printCVList();

  printAnalysisSoftwareList($thisScript, \%urlParams, $objResFile);

  printProvider();

  printAuditCollection(\@resfile_params);
  
  printAnalysisSampleCollection($objQuantMethod);

  my $sameset_ordering = common_subs::samesetOrderingByPreferredTaxonomy() || {};
  my $queryAll = $thisScript->param($urlParams{'query_all'});

  my $protLen = $thisScript->param($urlParams{'prot_len'});
  my $protSeq = $thisScript->param($urlParams{'prot_seq'});
  my $protDesc = $thisScript->param($urlParams{'prot_desc'});
  my $protTaxStr = $thisScript->param($urlParams{'prot_tax_str'});
  my $protTaxId = $thisScript->param($urlParams{'prot_tax_id'});
  my $filter_duplicates = 1;
  if ($thisScript->param($urlParams{'show_pep_dupes'})) {
    $filter_duplicates = 0;
  }

  my %query_mod_and_dupe_cache = ();

  my ($exported_sequences, $exported_queryList, $peptide_evidence_list) = printSequenceCollection(
    \%query_mod_and_dupe_cache, $objResFile, $objSummary, $objUniGene,
    $ShowSameSets, $ShowSubSets, $queryAll, $protLen, $protSeq, $protDesc,
    $protTaxStr, $protTaxId, $objUmodConfigFile2, $objModFile,
    $sameset_ordering, \%vmMass, \%vmString, \%fixedMods, $filter_duplicates,
  );

  printAnalysisCollection($objResFile, $objParams, _tolerance_cv_param_map());

  printAnalysisProtocolCollection(
    $objResFile, $objParams, \@resfile_params, $objSummary, $objQuantMethod,
    $objUmodConfigFile, $numHits, $sigThreshold, $ignoreIonsScoreBelow,
    $use_homology, $group_family, $mudpitSwitch, $UniGeneFile, $ShowSameSets,
    $ShowSubSets, $RequireBoldRed, $NoErrorTolerant, $ShowDecoyReport,
    $preferTaxonomy, $percolate, $percolate_rt, $protTaxStr, $targetFDR,
    \%taxNames,
    _tolerance_cv_param_map()
  );

  my $queryRaw = $thisScript->param($urlParams{'query_raw'});

  printDataCollection(
    \%query_mod_and_dupe_cache, \@fileIn, $objResFile, $objParams,
    \@resfile_params, \@resfile_children, $objSummary, $UniGeneFile,
    $objUmodConfigFile2, $sessionID, $sigThreshold, $msresFlags2, $percolate,
    $exported_queryList, $exported_sequences, $peptide_evidence_list,
    \%aaHelpers, \%elementalMasses, _tolerance_cv_param_map(), $ShowSameSets,
    $ShowSubSets, $protTaxStr, $protTaxId, $queryAll, $reportType, $protLen,
    $queryRaw, $sameset_ordering, $numcomponents, $componentNames[0],
  );
  
  printBibliographicReference();
}

sub _make_mod_vector {
  my ($objResFile, $objModFile, $param_name) = @_;

  my $vec = new msparser::ms_modvector();
  my $i = 1;
  while (my $modText = $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_MASSES, $param_name . $i)) {
    my ($mass, $name) = split(/,/, $modText, 2);
    my $objMod = $objModFile->getModificationByName($name);
    if ($objMod) {
      $vec->appendModification($objMod);
    } else {
      die "Modification " . $name . " cannot be found";
    }
    $i++;
  }
  return $vec;
}

sub _make_fixed_mods_map {
  my ($vecFixed) = @_;
  
  my %fixedMods;
  for my $i (0 .. $vecFixed->getNumberOfModifications()-1) {
    my $objMod = $vecFixed->getModificationByNumber($i);
    if ($objMod->getModificationType == $msparser::MOD_TYPE_N_TERM) {
      $fixedMods{'n-term'} = { name => $objMod->getTitle, delta => $objMod->getDelta($mass_type) };
    } elsif ($objMod->getModificationType == $msparser::MOD_TYPE_C_TERM) {
      $fixedMods{'c-term'} = { name => $objMod->getTitle, delta => $objMod->getDelta($mass_type) };
    } else {
      for my $j (0 .. $objMod->getNumberOfModifiedResidues()-1) {
        $fixedMods{lc($objMod->getModifiedResidue($j))} = { name => $objMod->getTitle, delta => $objMod->getDelta($mass_type) };
      } 
    }
  }

  return %fixedMods;
}


sub printCVList {
  print <<EOF;
  <cvList>
    <cv id="PSI-MS" fullName="Proteomics Standards Initiative Mass Spectrometry Vocabularies"   uri="http://psidev.cvs.sourceforge.net/viewvc/*checkout*/psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo" version="3.64.0"></cv>
    <cv id="UNIMOD" fullName="UNIMOD"        uri="http://www.unimod.org/obo/unimod.obo"></cv>
    <cv id="UO"     fullName="UNIT-ONTOLOGY" uri="http://obo.cvs.sourceforge.net/*checkout*/obo/obo/ontology/phenotype/unit.obo"></cv>
  </cvList>
EOF
}

sub printAnalysisSoftwareList {
  my ($thisScript, $urlParams, $objResFile) = @_;

  # (When results are merged: we pick Mascot version from the first results
  # file.)
  my $mascotver = noXmlTag($objResFile->getMascotVer());

  my $parserver = noXmlTag($objResFile->getMSParserVersion());

  my $args = _cmdline_arguments($thisScript, $urlParams);

  my $pretty_args = noXmlTag(join(' ', map {
    $$args{$_} =~ /\s/
      ? sprintf("%s='%s'", $_, $$args{$_})
      : sprintf("%s=%s", $_, $$args{$_})
  } sort keys %$args));

  print <<EOF;
  <AnalysisSoftwareList>
    <AnalysisSoftware id="AS_mascot_server" name="Mascot Server" version="$mascotver" uri="http://www.matrixscience.com/search_form_select.html">
      <ContactRole contact_ref="ORG_MSL">
        <Role>
          <cvParam accession="MS:1001267" name="software vendor" cvRef="PSI-MS" />
        </Role>
      </ContactRole>
      <SoftwareName>
        <cvParam accession="MS:1001207" name="Mascot" cvRef="PSI-MS" />
      </SoftwareName>
      <Customizations>
        No customisations
      </Customizations>
    </AnalysisSoftware>
    <AnalysisSoftware id="AS_mascot_parser" name="Mascot Parser" version="$parserver" uri="http://www.matrixscience.com/msparser.html">
      <ContactRole contact_ref="ORG_MSL">
        <Role>
          <cvParam accession="MS:1001267" name="software vendor" cvRef="PSI-MS" />
        </Role>
      </ContactRole>
      <SoftwareName>
        <cvParam accession="MS:1001478" name="Mascot Parser" cvRef="PSI-MS" />
      </SoftwareName>
      <Customizations>
        No customisations
      </Customizations>
    </AnalysisSoftware>
  </AnalysisSoftwareList>
EOF
}
  
sub printProvider {
  print <<EOF;
  <Provider id="PROVIDER">
    <ContactRole contact_ref="PERSON_DOC_OWNER">
      <Role>
        <cvParam accession="MS:1001271" name="researcher" cvRef="PSI-MS" />
      </Role>
    </ContactRole>
  </Provider>
EOF
}
  
sub printAuditCollection {
  my ($resfile_params) = @_;

  my @users = ();
  my %users = ();

  # (When results are merged: the contact person is the user in the first
  # results file.)

  for my $i (1 .. $#$resfile_params) {
    my $params = $$resfile_params[$i];
    my $username = noXmlTag($params->getUSERNAME);
    my $email = noXmlTag($params->getUSEREMAIL);

    if (not $users{$username}) {
        push @users, $username;
    }

    $users{$username}{$email} = 1;
  }

  print <<EOF;
  <AuditCollection>
    <Person id="PERSON_MSL">
      <Affiliation organization_ref="ORG_MSL" />
    </Person>
EOF

  for my $i (0 .. $#users) {
    my $id = $i == 0 ? 'PERSON_DOC_OWNER' : sprintf('PERSON_DOC_OWNER_%d', $i);
    my $username = $users[$i];
    print <<EOF;
    <Person id="$id" name="$username">
EOF

    for my $email (sort keys %{ $users{$username} }) {
      next if $email eq '';
      print <<EOF;
      <cvParam accession="MS:1000589" name="contact email" value="$email" cvRef="PSI-MS" />
EOF
    }

    print <<EOF;
      <Affiliation organization_ref="ORG_DOC_OWNER" />
    </Person>
EOF
  }

  print <<EOF;
    <Organization id="ORG_MSL" name="Matrix Science Limited">
      <cvParam accession="MS:1000589" name="contact email" value="support\@matrixscience.com" cvRef="PSI-MS" />
      <cvParam accession="MS:1000587" name="contact address" value="64 Baker Street, London W1U 7GB, UK" cvRef="PSI-MS" />
      <cvParam accession="MS:1001755" name="contact phone number" value="+44 (0)20 7486 1050" cvRef="PSI-MS" />
      <cvParam accession="MS:1001756" name="contact fax number" value="+44 (0)20 7224 1344" cvRef="PSI-MS" />
    </Organization>
    <Organization id="ORG_DOC_OWNER" />
  </AuditCollection>
EOF
}

sub printAnalysisSampleCollection {
  my ($objQuantMethod) = @_;

  return if not $objQuantMethod;

  my $numcomponents = $objQuantMethod->getNumberOfComponents; 

  return if $numcomponents <= 1;

  print "  <AnalysisSampleCollection>\n";

  for my $c (0 .. $numcomponents-1) {
    my $component = $objQuantMethod->getComponentByNumber($c);
    print '    <Sample id="' . _mzid_Sample_id($component->getName) . "\">\n";
    print "      <userParam name=\"" . noXmlTag($component->getName) . "\" />\n";
    if ($component->getNumberOfIsotopes > 0) {
      for my $i (0 .. $component->getNumberOfIsotopes()-1) {
        my $isotope = $component->getIsotope($i);
        if ($isotope->haveOld && $isotope->haveNew) {
          print "      <userParam name=\"Isotope: " . $isotope->getOld . " -> " .  $isotope->getNew . "\" />\n";
        } else {
          print "      <userParam name=\"Isotope: None\" />\n";
        } 
      }
    }
    print "    </Sample>\n";
  }

  print '    <Sample id="' . _mzid_Sample_id('Combined') . "\">\n";
  for my $c (0 .. $numcomponents-1) {
    my $component = $objQuantMethod->getComponentByNumber($c);
    print '      <SubSample sample_ref="' . _mzid_Sample_id($component->getName) . "\" />\n";
  }
  print "    </Sample>\n";

  print "  </AnalysisSampleCollection>\n";
}

sub _mzid_bool { $_[0] ? "1" : "0" }

sub _mzid_DBSeq_id { sprintf("DBSeq_%d_%s", $_[0], noXmlTag($_[1])) }
sub _mzid_Enzyme_id { sprintf("ENZ_%d", $_[0]) }
sub _mzid_MassTable_id { sprintf("MT_%s", noXmlTag($_[0])) }
sub _mzid_Measure_id { sprintf("m_%s", noXmlTag($_[0])) }
sub _mzid_PeptideEvidence_id { sprintf("PE_%d_%d_%s_%d_%d_%d", $_[0], $_[1], noXmlTag($_[2]), $_[3], $_[4], $_[5]) }
sub _mzid_Peptide_id { sprintf("peptide_%d_%d", $_[0], $_[1]) }
sub _mzid_ProteinAmbiguityGroup_id { sprintf("PAG_hit_%d", $_[0]) }
sub _mzid_ProteinDetectionHypothesis_id { sprintf("PDH_%s_%d", noXmlTag($_[0]), $_[1]) }
sub _mzid_ProteinDetectionList_id { sprintf("PDL_%d", $_[0]) }
sub _mzid_ProteinDetectionProtocol_id { sprintf("PDP_%s_%d", noXmlTag($_[0]), $_[1]) }
sub _mzid_ProteinDetection_id { sprintf("PD_%d", $_[0]) }
sub _mzid_Sample_id { sprintf("Sample_%s", noXmlTag($_[0])) }
sub _mzid_SearchDatabase_id { sprintf("SDB_%s", noXmlTag($_[0])) }
sub _mzid_SourceFile_id { sprintf("SF_%d", $_[0]) }
sub _mzid_SpectraData_id { sprintf("SD_%d", $_[0]) }
sub _mzid_SpectrumIdentificationItem_id { sprintf("SII_%d_%d", $_[0], $_[1]) }
sub _mzid_SpectrumIdentificationList_id { sprintf("SIL_%d", $_[0]) }
sub _mzid_SpectrumIdentificationProtocol_id { sprintf("SIP") }
sub _mzid_SpectrumIdentificationResult_id { sprintf("SIR_%d", $_[0]) }
sub _mzid_SpectrumIdentification_id { sprintf("SI") }
sub _mzid_TranslationTable_id { sprintf("TT_%s", noXmlTag($_[0])) }

sub printSequenceCollectionPeptides {
  my ($query_mod_and_dupe_cache, $objResFile, $objParams, $objSummary,
      $objUniGene, $objUmodConfigFile2, $objModFile, $exported_sequences,
      $exported_queryList, $queryAll, $vmMass, $vmString, $fixedMods,
      $protTaxId, $protTaxStr, $filter_duplicates) = @_;

  my $isPMF = $objResFile->isPMF;
  my $max_p = $objSummary->getMaxRankValue();

  # We need to make two loops over the peptides, because PeptideEvidence
  # elements must follow all Peptide elements. Additionally, we need to
  # collect peptide evidence in the first loop and skip peptides for which
  # there is no evidence; otherwise we'll end up outputting Peptide elements
  # with no corresponding PeptideEvidence. See bug #10759.
  my %peptide_evidence_list = ();

  # We need to keep some further details about peptide evidence (residue
  # before, residue after, etc.) for the second loop.
  my %peptide_evidence_details = ();
  my %umod_mod_cache = ();
  my %dupe_key_cache = ();

  for my $query (1 .. $objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    for my $rank (1 .. $max_p) {
      my $objPeptide = $objSummary->getPeptide($query, $rank);
      next if $objPeptide->getPeptideStr(0) eq '';

      collectPeptideEvidence(
          $objSummary, $objUniGene, $query, $rank, $exported_sequences,
          \%peptide_evidence_list, \%peptide_evidence_details,
      );

      if ($filter_duplicates) {
        $dupe_key_cache{$query}{$rank} = {
          obj => $objPeptide,
          key => duplicate_group_main($query_mod_and_dupe_cache, $objResFile, $objParams, $objSummary, $objPeptide),
        };
      }
    }
  }

  if ($filter_duplicates) {
    # Bug #10777: If duplicates should be filtered out, we have some extra
    # work to do. The duplicate status of a peptide match depends on which
    # protein hit we are considering. Thus, for each protein hit, we need
    # to find its "duplicate groups" of peptide matches. We then remove
    # the evidence of a peptide match from protein hits where the peptide match
    # is not the highest scoring match in its duplicate group within the hit.


    # Step 1: reconstruct the list of matches for each protein hit, with
    # each peptide match grouped in its duplicate group.
    my %protein_data = ();

    for my $query (keys %peptide_evidence_details) {
      for my $rank (keys %{ $peptide_evidence_details{$query} }) {
        my $details = $peptide_evidence_details{$query}{$rank};
        my $key = $dupe_key_cache{$query}{$rank}{'key'};
        my $objPeptide = $dupe_key_cache{$query}{$rank}{'obj'};

        for my $d (@$details) {
          push @{ $protein_data{$$d{'db'}}{$$d{'acc'}}{$key} }, $objPeptide;
        }
      }
    }

    # Step 2: find the master peptide in each duplicate group.
    my %surviving_peptide = ();

    for my $db (keys %protein_data) {
      for my $acc (keys %{ $protein_data{$db} }) {
        for my $key (keys %{ $protein_data{$db}{$acc} }) {
          my $sorted = sort_duplicate_group($protein_data{$db}{$acc}{$key});
          my $master_peptide = $$sorted[0];
          $surviving_peptide{$master_peptide->getQuery()}{$master_peptide->getRank()}{$db}{$acc} = 1;
        }
      }
    }

    # Step 3: remove peptide match details from protein hits where the peptide
    # match is not the master peptide in the protein hit.
    for my $query (keys %peptide_evidence_details) {
      for my $rank (keys %{ $peptide_evidence_details{$query} }) {
        my $details = $peptide_evidence_details{$query}{$rank};
        my @filtered = ();

        for my $d (@$details) {
          if ($surviving_peptide{$query}{$rank}{$$d{'db'}}{$$d{'acc'}}) {
            push @filtered, $d;
          }
        }

        if (@filtered) {
          $peptide_evidence_details{$query}{$rank} = \@filtered;
          $peptide_evidence_list{$query}{$rank} = {};
          for my $d (@filtered) {
            $peptide_evidence_list{$query}{$rank}{$$d{'id'}} = 1;
          }
        } else {
          delete $peptide_evidence_details{$query}{$rank};
          if (not %{ $peptide_evidence_details{$query} }) {
            delete $peptide_evidence_details{$query};
          }
          delete $peptide_evidence_list{$query}{$rank};
          if (not %{ $peptide_evidence_list{$query} }) {
            delete $peptide_evidence_list{$query};
          }
        }
      }
    }
  }

  for my $query (1 .. $objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    for my $rank (1 .. $max_p) {
      next unless exists $peptide_evidence_list{$query}{$rank};

      my $objPeptide = $objSummary->getPeptide($query, $rank);

      printSequenceCollectionPeptide(
        $objResFile, $objParams, $objSummary, $objPeptide, $objUmodConfigFile2,
        $objModFile, $query, $rank, $vmMass, $vmString, $fixedMods, 
        \%umod_mod_cache, $query_mod_and_dupe_cache,
      );
    }
  }

  for my $query (1 .. $objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    for my $rank (1 .. $max_p) {
      my $objPeptide = $objSummary->getPeptide($query, $rank);
      next if $objPeptide->getPeptideStr(0) eq '';

      next if not $peptide_evidence_details{$query}{$rank};

      printSequenceCollectionPeptideEvidence(
        $objParams, $objSummary, $objPeptide, $query, $rank,
        $protTaxId, $protTaxStr, \%peptide_evidence_details
      );
    }
  }

  return \%peptide_evidence_list;
}

sub _find_unigene_ids {
  my ($objUniGene, $accession) = @_;

  my $i = 0;
  my @ids = ();

  while (1) {
    my $str = $objUniGene->getUnigeneForAccession($accession, ++$i);
    last if $str eq '';

    push @ids, $str;
  }

  return \@ids;
}

sub collectPeptideEvidence {
  my ($objSummary, $objUniGene, $query, $rank, $exported_sequences,
      $peptide_evidence_list, $peptide_evidence_details) = @_;

  my $start = new msparser::vectori;
  my $end   = new msparser::vectori;
  my $pre   = new msparser::VectorString;
  my $post  = new msparser::VectorString;
  my $frame = new msparser::vectori;
  my $multiplicity = new msparser::vectori;
  my $db    = new msparser::vectori;

  my $accessions = $objSummary->getAllProteinsWithThisPepMatch($query, $rank, $start, $end, $pre, $post, $frame, $multiplicity, $db);
  # print "Errors: " . $resfile->getLastErrorString() . "\n"; 

  my @unigene_ids = ();

  if ($objUniGene) {
    for my $i (0 .. $#$accessions) {
      $unigene_ids[$i] = _find_unigene_ids($objUniGene, $$accessions[$i]);
    }
  }

  for my $i (0 .. $#$accessions) {
    my $preStr  = $pre->get($i);
    my $postStr = $post->get($i);

    if ($preStr eq '@') {
      $preStr = '-';
    }
    if ($postStr eq '@') {
      $postStr = '-';
    }

    if (my @ids = @{ $unigene_ids[$i] || [] }) {
      for my $unigene_id (@ids) {
        next unless $$exported_sequences{ key_acc_str(0, $unigene_id) };

        my $id = _mzid_PeptideEvidence_id(
          $query, $rank, $unigene_id, $frame->get($i), 
          $start->get($i), $end->get($i)
        );

        # Several accessions can (and usually do) map to the same UniGene ID.
        # We only want one peptide evidence item in this case.
        next if $$peptide_evidence_list{$query}{$rank}{$id};
        $$peptide_evidence_list{$query}{$rank}{$id} = 1;
    
        push @{ $$peptide_evidence_details{$query}{$rank} }, {
          id => $id,
          acc => $unigene_id,
          db => 0,
          preferred_acc => $unigene_id,
          pre => $preStr,
          post => $postStr,
          start => $start->get($i),
          end => $end->get($i),
          frame => $frame->get($i),
        };
      }

      next;
    } 

    next unless $$exported_sequences{ key_acc_str($db->get($i), $$accessions[$i]) };

    my $preferred_acc = common_subs::preferredAccession($db->get($i), $$accessions[$i]);

    my $id = _mzid_PeptideEvidence_id(
      $query, $rank, $preferred_acc, $frame->get($i), 
      $start->get($i), $end->get($i)
    );

    $$peptide_evidence_list{$query}{$rank}{$id} = 1;

    push @{ $$peptide_evidence_details{$query}{$rank} }, {
      id => $id,
      acc => $$accessions[$i],
      db => $db->get($i),
      preferred_acc => $preferred_acc,
      pre => $preStr,
      post => $postStr,
      start => $start->get($i),
      end => $end->get($i),
      frame => $frame->get($i),
    };
  }
}

sub printSequenceCollectionPeptide {
  my ($objResFile, $objParams, $objSummary, $objPeptide, $objUmodConfigFile2,
      $objModFile, $query, $rank, $vmMass, $vmString, $fixedMods,
      $umod_mod_cache, $query_mod_and_dupe_cache) = @_;

  my %vmMass = %$vmMass;
  my %vmString = %$vmString;

  my $pepstr = $objPeptide->getPeptideStr(0);

  print '    <Peptide id="' . _mzid_Peptide_id($query, $rank) . "\">\n";
  print "      <PeptideSequence>" . $pepstr . "</PeptideSequence>\n";
  
  # Any modifications?
  my $varModsStr = $objPeptide->getVarModsStr();
  my $summedModsStr = $objPeptide->getSummedModsStr();
  $summedModsStr = '0' x length($varModsStr) if $summedModsStr eq '';
  my $local_mods = lookup_query_level_mods($query_mod_and_dupe_cache, $objResFile, $objPeptide, $objUmodConfigFile2);

  my $nlStr = $objPeptide->getPrimaryNlStr();
  $nlStr = '0' x length($varModsStr) if $nlStr eq '';
  my $localNlStr = $objPeptide->getLocalModsNlStr();
  $localNlStr = '0' x length($varModsStr) if $localNlStr eq '';

  my %varmod_index_map = do {
    my @indices = (0..9, 'A'..'W');
    map { $indices[$_] => $_ } 0 .. $#indices
  };

  my @pepRes = split(//, lc($objPeptide->getPeptideStr(1)));
  
  for my $p (0 .. length($varModsStr)-1) {
    my $has_var_mod = 0;

    if (my $varmod_id = substr($varModsStr, $p, 1)) {
      # $varmod_id is not '' and not '0' (both evaluate to false).
      my $srcRank = $rank;
      if ($varmod_id eq "X") {
        $srcRank = $objSummary->getSrcRank($query, $rank);
        my $tempString = 
            $objResFile->getSectionValueStr($objSummary->getSrcSection($query, $rank), "q$query\_p$srcRank\_et_mods");
        if ($tempString) {
          ($vmMass{'X'}, undef, $vmString{'X'}) = split(/,/, $tempString, 3);
        }
      } 
      my @cv_params = ();
      if ($varmod_id eq "X" && $vmMass{'X'} == 0 && $vmString{'X'} =~ /^NA_/) { 
        my %pretty_value = (
            NA_SUBSTITUTION => 'substitution',
            NA_INSERTION => 'insertion',
            NA_DELETION => 'deletion',
        );

        push @cv_params, "        <cvParam accession=\"MS:1002028\" name=\"nucleic acid base modification\" cvRef=\"PSI-MS\" value=\"" . $pretty_value{uc $vmString{'X'}} . "\" />\n";

        my $tempString = 
            $objResFile->getSectionValueStr($objSummary->getSrcSection($query, $rank), "q$query\_p$srcRank\_na_diff");
        if ($tempString) {
          my ($old, $new) = split(/,/, $tempString, 2);
          push @cv_params, "        <cvParam accession=\"MS:1002029\" name=\"original nucleic acid sequence\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($old) . "\" />\n";
          push @cv_params, "        <cvParam accession=\"MS:1002030\" name=\"modified nucleic acid sequence\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($new) . "\" />\n";
        }
      }
      my $nl_id = substr($nlStr, $p, 1);
      if (defined $nl_id) {
        $nl_id = $varmod_index_map{$nl_id};
      }
      my $nl_backup_value = undef;
      if (defined($nl_id) and $nl_id > 0) {
        # Old results files without Unimod section store neutral loss
        # information in the parameter section.
        my @arr = @{ $objParams->getVarModsNeutralLosses($varmod_index_map{$varmod_id}) };
        if ($nl_id-1 < @arr) {
          $nl_backup_value = $arr[$nl_id-1];
        }
      }
      printModificationElement(
        $umod_mod_cache, $objParams, $pepstr, $p, $vmMass{$varmod_id},
        $vmString{$varmod_id}, $objUmodConfigFile2, $nl_id, \@cv_params,
        $nl_backup_value,
      );
      $has_var_mod = 1;
    }

    if (my $varmod_id = substr($summedModsStr, $p, 1)) {
      printModificationElement(
        $umod_mod_cache, $objParams, $pepstr, $p, $vmMass{$varmod_id},
        $vmString{$varmod_id}, $objUmodConfigFile2, undef, [], 
      );
      $has_var_mod = 1;
    }

    if ($local_mods and $$local_mods{$p}) {
      my $nl_id = substr($localNlStr, $p, 1);
      if (defined $nl_id) {
        $nl_id = $varmod_index_map{$nl_id};
      }
      printModificationElement(
        $umod_mod_cache, $objParams, $pepstr, $p, $$local_mods{$p}{'delta'},
        $$local_mods{$p}{'name'}, $objUmodConfigFile2, $nl_id, [], 
      );
      $has_var_mod = 1;
    }

    if (not $has_var_mod) {
      my $mod_name;
    # check for fixed mod
      if ($p == 0) {
      # n-term
        $mod_name = $$fixedMods{'n-term'}{'name'};
      } elsif ($p == (length($varModsStr) - 1)) {
      # c-term
        $mod_name = $$fixedMods{'c-term'}{'name'};
      } else {
      # residue $pepRes[$p-1];
        $mod_name = $$fixedMods{$pepRes[$p-1]}{'name'};
      }
      if ($mod_name) {
        my $objMod = $objModFile->getModificationByName($mod_name);
        printModificationElement(
          $umod_mod_cache, $objParams, $pepstr, $p,
          $objMod->getDelta($objParams->getMassType()), $mod_name,
          $objUmodConfigFile2, 1, [], 
        );
      }
    }
  }
  # Any substitutions?
  my $ambiguity = $objPeptide->getAmbiguityString();
  if ($ambiguity) {
    # Format: 4,X,A
    my ($location, $before, $after) = split(/,/, $ambiguity);
    print "      <SubstitutionModification location=\"". $location . "\" originalResidue=\"" . $before . "\" replacementResidue=\"" . $after . "\" />\n";
  }

  print "    </Peptide>\n";
}

sub printModificationElement {
  my ($mod_cache, $objParams, $pepstr, $p, $vmMass, $vmString,
      $objUmodConfigFile2, $nl_id, $cv_params, $nl_backup_value) = @_;

  print "      <Modification location=\"" . $p . "\" ";
  if ($p > 0 and $p <= length($pepstr)) {
    print "residues=\"" . substr($pepstr, $p-1, 1) . "\" ";
  }

  if (@$cv_params) {
  # NA_SUBSTITUTION, NA_INSERTION, NA_DELETION
    print ">\n";
    print @$cv_params;
  } else {
    if ($objParams->getMassType == $msparser::MASS_TYPE_MONO) {
      print "monoisotopicMassDelta=\"" . $vmMass . "\"";
    } else {
      print "avgMassDelta=\"" . $vmMass . "\"";
    }
    print ">\n";
    my ($mod_name) = $vmString =~ /(.*)\s+\((.*)\)/;
    my ($cv_acc, $cv_name, $cv_ref, $nl_arr);
    if (my $mod2 = $objUmodConfigFile2->getModificationByName($mod_name)) {
      $cv_acc = 'UNIMOD:' . noXmlTag($mod2->getRecordID);
      $cv_name = noXmlTag($mod_name);
      $cv_ref = 'UNIMOD';
      if (exists $$mod_cache{$vmString}) {
        if ($$mod_cache{$vmString}) {
          $nl_arr = $$mod_cache{$vmString};
        }
      } else {
        my $modfile_mod = msparser::ms_modification->new;
        if ($modfile_mod->getFromUnimod($vmString, $objUmodConfigFile2)) {
          $nl_arr = $$mod_cache{$vmString} = $modfile_mod->getNeutralLoss($objParams->getMassType());
        }
      }
    } else {
      $cv_acc = "MS:1001460";
      $cv_name = 'unknown modification';
      $cv_ref = 'PSI-MS';
    }
    print "        <cvParam accession=\"$cv_acc\" name=\"$cv_name\" cvRef=\"$cv_ref\" />\n";
    
    my $nl_value = $nl_backup_value; # could be undef
    if (defined $nl_id) {
      if ($nl_arr and @$nl_arr and $nl_id-1 < @$nl_arr and $$nl_arr[$nl_id-1] > 0) {
        $nl_value = $$nl_arr[$nl_id-1];
      }
    }
    if (defined $nl_value) {
      print "        <cvParam accession=\"MS:1001524\" name=\"fragment neutral loss\" cvRef=\"PSI-MS\" value=\"" . $nl_value . "\" unitAccession=\"UO:0000221\" unitName=\"dalton\" unitCvRef=\"UO\" />\n";
    }
  }
  print "      </Modification>\n";
}

sub printSequenceCollectionPeptideEvidence {
  my ($objParams, $objSummary, $objPeptide, $query, $rank, 
      $protTaxId, $protTaxStr, $peptide_evidence_details) = @_;

  # The evidence needs to be printed in sorted order to keep the output
  # constant across different runs with the same parameters.
  for my $evidence (sort { $$a{'id'} cmp $$b{'id'} } @{ $$peptide_evidence_details{$query}{$rank} }) {
    my ($id, $start, $end, $preStr, $postStr) = @$evidence{qw(id start end pre post)};

    print '    <PeptideEvidence id="' . $id
      . '" start="' . $start
      . '" end="'   . $end
      . '" pre="'   . $preStr
      . '" post="'  . $postStr
      . '" peptide_ref="' . _mzid_Peptide_id($query, $rank)
      ;

    my ($db, $preferred_acc) = @$evidence{qw(db preferred_acc)};
    
    if ($objSummary->isNA) {
      my $ttNum = 1;  # if unspecified, Mascot uses standard genetic code
      my ($acc, $framenum) = @$evidence{qw(acc frame)};
      if ($protTaxStr || $protTaxId) {
        $ttNum = common_subs::getProteinTranslationTableID($objParams, $acc, $db);
      }
      print '" translationTable_ref="' . _mzid_TranslationTable_id($ttNum);
      if ($framenum >= 1 && $framenum <=3) {
        print '" frame="' . $framenum;
      } else {
        print '" frame="' . ($framenum -7);  # 6 -> -1, 5 -> -2, 4 -> -3
      }
    }
    print "\" isDecoy=\"" . "false"
      . "\" dBSequence_ref=\"" . _mzid_DBSeq_id($db, $preferred_acc) . "\" />\n";
  }
}

sub printSequenceCollection {
  my ($query_mod_and_dupe_cache, $objResFile, $objSummary, $objUniGene,
      $ShowSameSets, $ShowSubSets, $queryAll, $protLen, $protSeq, $protDesc,
      $protTaxStr, $protTaxId, $objUmodConfigFile2, $objModFile,
      $sameset_ordering, $vmMass, $vmString, $fixedMods,
      $filter_duplicates) = @_;

  print "  <SequenceCollection>\n";

  my ($exported_sequences, $exported_queryList) = printDBSequences(
    $objResFile, $objSummary, $objUniGene, $ShowSameSets, $ShowSubSets,
    $queryAll, $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
    $sameset_ordering, 
  );

  my $peptide_evidence_list = printSequenceCollectionPeptides(
    $query_mod_and_dupe_cache, $objResFile, $objParams, $objSummary,
    $objUniGene, $objUmodConfigFile2, $objModFile, $exported_sequences,
    $exported_queryList, $queryAll, $vmMass, $vmString, $fixedMods, $protTaxId,
    $protTaxStr, $filter_duplicates
  );

  print "  </SequenceCollection>\n";

  return ($exported_sequences, $exported_queryList, $peptide_evidence_list);
}

sub printDBSequences {
  my ($objResFile, $objSummary, $objUniGene, $ShowSameSets, $ShowSubSets,
      $queryAll, $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
      $sameset_ordering) = @_;

  my $isPMF = $objResFile->isPMF;
  my $objParams = $objResFile->params;

  my %exported_sequences = ();
  my @queryList = ();

  for my $h (1 .. $objSummary->getNumberOfHits()) {
    
    my $objProtein = $objSummary->getHit($h);
    my @sameset = common_subs::compileSameset(
      $objSummary, $objProtein, $ShowSameSets, $$sameset_ordering{$h}
    );

    # Note that markExportableQueries() must match exactly the calls done later
    # for printProteinDetectionHypothesis(). Otherwise some PeptideEvidence
    # elements can be missing from the exported file. See bug #11013.

    unless ($exported_sequences{ key_acc($sameset[0]) }) {
      printDBSequence(
        $objSummary, $objParams,
        $sameset[0]->getDB, $sameset[0]->getAccession,
        $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
      );
      $exported_sequences{ key_acc($sameset[0]) } = 1;
      markExportableQueries($queryAll, $isPMF, $sameset[0], \@queryList);
    }

    if ($group_family) {

    # Other proteins that match the same set of peptides
      if ($ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          unless ($exported_sequences{ key_acc($_) }) {
            printDBSequence(
              $objSummary, $objParams,
              $_->getDB, $_->getAccession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
            );
            $exported_sequences{ key_acc($_) } = 1;
            markExportableQueries($queryAll, $isPMF, $_, \@queryList);
          }
        }
      }
      
    # Other family members
      my $k = 1;
      while (my $objProtein = $objSummary->getNextFamilyProtein($h, $k++)) {
        my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, 
          $$sameset_ordering{$h}
        );

        unless ($exported_sequences{ key_acc($sameset[0]) }) {
          printDBSequence(
            $objSummary, $objParams,
            $sameset[0]->getDB, $sameset[0]->getAccession,
            $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
          );
          $exported_sequences{ key_acc($sameset[0]) } = 1;
          markExportableQueries($queryAll, $isPMF, $sameset[0], \@queryList);
        }

      # And their same-sets
        if ($ShowSameSets) {
          for (@sameset[1 .. $#sameset]) {
            unless ($exported_sequences{ key_acc($_) }) {
              printDBSequence(
                $objSummary, $objParams,
                $_->getDB, $_->getAccession,
                $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
              );
              $exported_sequences{ key_acc($_) } = 1;
              markExportableQueries($queryAll, $isPMF, $_, \@queryList);
            }
          }
        }
      }
      
      
    # sub-sets and intersections for the entire family
      if ($ShowSubSets > 0) {
        my $k = 1;
        while (my $objProtein = $objSummary->getNextSubsetProtein($h, $k++, 1)) {
          my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, 
            $$sameset_ordering{$h}
          );

          unless ($exported_sequences{ key_acc($sameset[0]) }) {
            printDBSequence(
              $objSummary, $objParams,
              $sameset[0]->getDB, $sameset[0]->getAccession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
            );
            $exported_sequences{ key_acc($sameset[0]) } = 1;
            markExportableQueries($queryAll, $isPMF, $sameset[0], \@queryList);
          }
  
        # And their same-sets
          if ($ShowSameSets) {
            for (@sameset[1 .. $#sameset]) {
              unless ($exported_sequences{ key_acc($_) }) {
                printDBSequence(
                  $objSummary, $objParams,
                  $_->getDB, $_->getAccession,
                  $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
                );
                $exported_sequences{ key_acc($_) } = 1;
                markExportableQueries($queryAll, $isPMF, $_, \@queryList);
              }
            }
          }
        }
        
      }
      
    } else {

      if ($ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          unless ($exported_sequences{ key_acc($_) }) {
            printDBSequence(
              $objSummary, $objParams,
              $_->getDB, $_->getAccession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
            );
            $exported_sequences{ key_acc($_) } = 1;
            markExportableQueries($queryAll, $isPMF, $_, \@queryList);
          }
        }
      }
      my $j = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($h, $j++)) {
        my $key_acc = key_acc($objProtein);
        unless ($exported_sequences{ $key_acc }) {
          printDBSequence(
            $objSummary, $objParams,
            $objProtein->getDB, $objProtein->getAccession,
            $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
          );
          $exported_sequences{ $key_acc } = 1;
          markExportableQueries($queryAll, $isPMF, $objProtein, \@queryList);
        }
      }
      
    }

  }

  # if query_all, then want to output all possible DBSequence values so that we
  # can output <PeptideEvidence> for everything.
  if ($queryAll) {
    my @ranks = (1 .. $objSummary->getMaxRankValue());
    my ($query, $rank);
    my $db    = new msparser::vectori;
    my $tmp1  = new msparser::vectori;
    my $tmp2  = new msparser::vectori;
    my $tmp3  = new msparser::VectorString;
    my $tmp4  = new msparser::VectorString;
    my $tmp5  = new msparser::vectori;
    my $tmp6  = new msparser::vectori;
    for my $query (1 .. $objResFile->getNumQueries()) {
      for my $rank (@ranks) {
        $db->clear;
        $_->clear for ($tmp1, $tmp2, $tmp3, $tmp4, $tmp5, $tmp6);
        my $accessions = $objSummary->getAllProteinsWithThisPepMatch($query, $rank, 
                                                                     $tmp1, # discard (start)
                                                                     $tmp2, # discard (end)
                                                                     $tmp3, # discard (pre)
                                                                     $tmp4, # discard (post)
                                                                     $tmp5,      # discard (frame)
                                                                     $tmp6, # discard (multiplicity)
                                                                     $db);

        for my $i (0 .. $#$accessions) {
          if ($objUniGene) {
            my $unigene_ids = _find_unigene_ids($objUniGene, $$accessions[$i]);

            if (@$unigene_ids) {
              for my $unigene_id (@$unigene_ids) {
                my $key_acc = key_acc_str(0, $unigene_id);

                unless ($exported_sequences{ $key_acc }) {
                  printDBSequence(
                    $objSummary, $objParams,
                    0, $unigene_id,
                    $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
                  );
                  $exported_sequences{ $key_acc } = 1;
                }
              }

              next;
            }
          }

          my $accession = $$accessions[$i];
          my $key_acc = key_acc_str($db->get($i), $accession);

          unless ($exported_sequences{ $key_acc }) {
            printDBSequence(
              $objSummary, $objParams,
              $db->get($i), $accession,
              $protLen, $protSeq, $protDesc, $protTaxStr, $protTaxId,
            );
            $exported_sequences{ $key_acc } = 1;
          }
        }
      }
    }
  }

  return (\%exported_sequences, \@queryList);
}

sub printDBSequence {
  my ($objSummary, $objParams, $dbIdx, $accession, $protLen, $protSeq,
      $protDesc, $protTaxStr, $protTaxId) = @_;

  my $preferred_acc = common_subs::preferredAccession($dbIdx, $accession);

  print '    <DBSequence id="' . _mzid_DBSeq_id($dbIdx, $preferred_acc)."\" ";

  if ($protLen) {
    my $length = getProteinLen($objParams, $accession, 0, $dbIdx); # best to use NA length? $objProtein->getFrame()
    if ($length) {
      print "length=\"" . $length . "\" ";
    }
  }

  if ($dbIdx == 0) {
    # UniGene
    print 'searchDatabase_ref="' . _mzid_SearchDatabase_id('UniGene') . '" accession="' . $preferred_acc . '" ';
  } else {
    print 'searchDatabase_ref="' . _mzid_SearchDatabase_id($objParams->getDB($dbIdx)) . '" accession="' . $preferred_acc . '" ';
  }

  print ">\n";

  if ($protSeq) {
    # Have to use NA because PI doesn't support anything other than [A-Z] and
    # frame may be mixed.
    my $seq = mustGetProteinSeq($objParams, $accession, 0, $dbIdx); 
    if ($seq) {
        print "      <Seq>" . $seq . "</Seq>\n";
    }
  }

  if ($protDesc) {
    my $description = common_subs::mustGetProteinDescription($objSummary, $objParams, $shipper, $preferred_acc, $dbIdx);
    if ($description) {
        print "      <cvParam accession=\"MS:1001088\" name=\"protein description\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($description) . "\" />\n";
    }
  }
  
  if ($protTaxStr or $protTaxId) {
    my ($taxID, $species) = common_subs::getProteinTaxonomy($objSummary, $objParams, $preferred_acc, $dbIdx);

    if ($protTaxStr and $species) {
      print "      <cvParam accession=\"MS:1001469\" name=\"taxonomy: scientific name\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($species) . "\" />\n";
    }

    if ($protTaxId and $taxID) {
      print "      <cvParam accession=\"MS:1001467\" name=\"taxonomy: NCBI TaxID\" cvRef=\"PSI-MS\" value=\"" . $taxID . "\" />\n";
    }
  }

  print "    </DBSequence>\n";
}

sub printAnalysisCollection {
  my ($objResFile, $objParams, $tolCVparam) = @_;

  # (When results are merged: it's desirable to have a single master list of
  # spectrum identifications, which is why we do not export one per input
  # file. For this reason activityDate is the date from the first results
  # file.)

  print "  <AnalysisCollection>\n";
  print '    <SpectrumIdentification id="' . _mzid_SpectrumIdentification_id() . "\" spectrumIdentificationProtocol_ref=\"" . _mzid_SpectrumIdentificationProtocol_id() . "\"  spectrumIdentificationList_ref=\"" . _mzid_SpectrumIdentificationList_id(1) . "\" ";
  printf "activityDate=\"%s\">\n", _fmtdate_iso8601($objResFile->getDate());
  for my $i (1 .. $objResFile->getNumberOfResfiles()) {
    print '      <InputSpectra spectraData_ref="' . _mzid_SpectraData_id($i) . "\" />\n";
  }
  for my $i (1 .. $objParams->getNumberOfDatabases()) {
    print '      <SearchDatabaseRef searchDatabase_ref="' . _mzid_SearchDatabase_id($objParams->getDB($i)) . "\" />\n";
  }
  print "    </SpectrumIdentification>\n";
  
  # Protein determination is performed 'now'
  print '    <ProteinDetection id="' . _mzid_ProteinDetection_id(1) . "\" proteinDetectionProtocol_ref=\"" . _mzid_ProteinDetectionProtocol_id('MascotParser', '1') . "\" proteinDetectionList_ref=\"" . _mzid_ProteinDetectionList_id(1) . "\" ";
  printf "activityDate=\"%s\">\n", _fmtdate_iso8601(time);
  print '      <InputSpectrumIdentifications spectrumIdentificationList_ref="' . _mzid_SpectrumIdentificationList_id(1) . "\" />\n";
  print "    </ProteinDetection>\n";
  print "  </AnalysisCollection>\n";
}

sub printAnalysisProtocolCollection {
  my ($objResFile, $objParams, $resfile_params, $objSummary, $objQuantMethod,
      $objUmodConfigFile, $numHits, $sigThreshold, $ignoreIonsScoreBelow,
      $use_homology, $group_family, $mudpitSwitch, $UniGeneFile, $ShowSameSets,
      $ShowSubSets, $RequireBoldRed, $NoErrorTolerant, $ShowDecoyReport,
      $preferTaxonomy, $percolate, $percolate_rt, $protTaxStr, $targetFDR,
      $taxNames, $tolCVparam) = @_;

  # (When results are merged: nearly all parameters must be the same between
  # the input files, so we simply use $objParams. The exception is the search
  # title (COM).)

  print "  <AnalysisProtocolCollection>\n";

  print '    <SpectrumIdentificationProtocol id="' . _mzid_SpectrumIdentificationProtocol_id() . "\" analysisSoftware_ref=\"AS_mascot_server\">\n";

  print "      <SearchType>\n";
  if ($objResFile->anyMSMS) {
    print "        <cvParam accession=\"MS:1001083\" name=\"ms-ms search\" cvRef=\"PSI-MS\" value=\"\" />\n";
  } elsif ($objResFile->anySQ || $objResFile->anyTag) {
    print "        <cvParam accession=\"MS:1001082\" name=\"tag search\" cvRef=\"PSI-MS\" value=\"\" />\n";
  } else {
    print "        <cvParam accession=\"MS:1001081\" name=\"pmf search\" cvRef=\"PSI-MS\" value=\"\" />\n";
  }
  print "      </SearchType>\n";

  print "      <AdditionalSearchParams>\n";
  my %printed_COM = ();
  for my $i (1 .. $#$resfile_params) {
    my $params = $$resfile_params[$i];
    if (my $com = $params->getCOM) {
      next if $printed_COM{$com};
      $printed_COM{$com} = 1;
      print "        <userParam name=\"Mascot User Comment\" value=\"" . noXmlTag($com) . "\" />\n";
    }
  }
  if ($objParams->getINSTRUMENT) {
    print "        <userParam name=\"Mascot Instrument Name\" value=\"" . noXmlTag($objParams->getINSTRUMENT)  . "\" />\n";
  }
  
  if ($objParams->getACCESSION) {
    print "        <cvParam accession=\"MS:1001021\" name=\"DB filter on accession numbers\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($objParams->getACCESSION) . "\" />\n";
  }
  if (lc($objParams->getMASS) eq "monoisotopic") {
    print "        <cvParam accession=\"MS:1001211\" name=\"parent mass type mono\"    cvRef=\"PSI-MS\" />\n";
  } else {
    print "        <cvParam accession=\"MS:1001212\" name=\"parent mass type average\" cvRef=\"PSI-MS\" />\n";
  }
  if ($objResFile->anyMSMS || $objResFile->anySQ || $objResFile->anyTag) {
    my %cvParam = (
      # 1 => { name => "singly charged" },
      # 2 => { name => "doubly charged if precursor 2+ or higher (not internal or immonium)" },
      # 3 => { name => "doubly charged if precursor 3+ or higher (not internal or immonium)" },
      4 =>  { id => "MS:1001259", name => "param: immonium ion" },
      5 =>  { id => "MS:1001108", name => "param: a ion" },
      6 =>  { id => "MS:1001146", name => "param: a ion-NH3" },
      7 =>  { id => "MS:1001148", name => "param: a ion-H2O" },
      8 =>  { id => "MS:1001118", name => "param: b ion" },
      9 =>  { id => "MS:1001149", name => "param: b ion-NH3" },
      10 => { id => "MS:1001150", name => "param: b ion-H2O" },
      11 => { id => "MS:1001119", name => "param: c ion" },
      12 => { id => "MS:1001261", name => "param: x ion" },
      13 => { id => "MS:1001262", name => "param: y ion" },
      14 => { id => "MS:1001151", name => "param: y ion-NH3" },
      15 => { id => "MS:1001152", name => "param: y ion-H2O" },
      16 => { id => "MS:1001263", name => "param: z ion" },
      17 => { id => "MS:1001406", name => "param: internal yb ion" },
      18 => { id => "MS:1001407", name => "param: internal ya ion" },
      # 19 => { name => "", id => "" },
      # 20 => { name => "", id => "" },
      21 => { id => "MS:1001408", name => "param: z+1 ion" },
      22 => { id => "MS:1001258", name => "param: d ion" },
      23 => { id => "MS:1001257", name => "param: v ion" },
      24 => { id => "MS:1001260", name => "param: w ion" },
      25 => { id => "MS:1001409", name => "param: z+2 ion" },
    );
    my @rules_list = split(/,/, $objParams->getRULES);

    foreach my $ruleNum (@rules_list) {
      my $cv = $cvParam{$ruleNum};
      next unless $cv;

      print "        <cvParam accession=\"" . $$cv{'id'}. "\" name=\"" . noXmlTag($$cv{'name'}) . "\" cvRef=\"PSI-MS\" />\n";
    }
  }
  print "      </AdditionalSearchParams>\n";
  
  my $fixed_mods = $objParams->getMODS();

  if ($objQuantMethod) {
    # add in any fixed mods specified in the quant method
    my $qMods = quant_subs::getQuantFixedModsAsString($objQuantMethod);
    if ($qMods && $fixed_mods) {
      $fixed_mods .= "," . $qMods;
    } elsif ($qMods) {
      $fixed_mods = $qMods;
    }
  }

  my $variable_mods = $objParams->getIT_MODS();

  if ($objQuantMethod) {
    # add in any variable or exclusive mods specified in the quant method
    my $qMods = quant_subs::getQuantVariableModsAsString($objQuantMethod);
    if ($qMods && $variable_mods) {
      $variable_mods .= "," . $qMods;
    } elsif ($qMods) {
      $variable_mods = $qMods;
    }
  }

  if ($fixed_mods || $variable_mods) {
    print "      <ModificationParams>\n";
    if ($fixed_mods) {
      my @mod_list = split(/,/, $fixed_mods);
      foreach (@mod_list) {
        printSearchModification($_, "true");
      }
    }
    if ($variable_mods) {
      my @mod_list = split(/,/, $variable_mods);
      foreach (@mod_list) {
        printSearchModification($_, "false");
      }
    }
    print "      </ModificationParams>\n";
  }
  
# create an enzymes file object
  my $objEnzymesFile = new msparser::ms_enzymefile;

  if ($objResFile->getEnzyme($objEnzymesFile, "../config/enzymes")) {
    my $enzyme = $objEnzymesFile->getEnzymeByNumber(0);
    if ($enzyme->isValid) {
      print "      <Enzymes";
      if ($enzyme->getNumberOfCutters > 1) {
        print " independent=\"";
        if ($enzyme->isIndependent) {
          print "1\"";
        } else {
          print "0\"";
        }
      }
      print ">\n";

      my %cvParam = (
        '::none' => { id => "MS:1001955", name => "no cleavage" },
        'kr:p:trypsin' => { id => "MS:1001251", name => "Trypsin" },
        'r:p:arg-c' => { id => "MS:1001303", name => "Arg-C" },
        'bd::asp-n' => { id => "MS:1001304", name => "Asp-N" },
        'de::asp-n_ambic' => { id => "MS:1001305", name => "Asp-N_ambic" },
        'flwy:p:chymotrypsin' => { id => "MS:1001306", name => "Chymotrypsin" },
        'm::cnbr' => { id => "MS:1001307", name => "CNBr" },
        'd::formic_acid' => { id => "MS:1001308", name => "Formic_acid" },
        'k:p:lys-c' => { id => "MS:1001309", name => "Lys-C" },
        'k::lys-c/p' => { id => "MS:1001310", name => "Lys-C/P" },
        'fl::pepsina' => { id => "MS:1001311", name => "PepsinA" },
        'fklrwy:p:trypchymo' => { id => "MS:1001312", name => "TrypChymo" },
        'kr::trypsin/p' => { id => "MS:1001313", name => "Trypsin/P" },
        'bdez:p:v8-de' => { id => "MS:1001314", name => "V8-DE" },
        'ez:p:v8-e' => { id => "MS:1001315", name => "V8-E" },
      );

      for my $i (0 .. $enzyme->getNumberOfCutters()-1) {
        my $cTermGain = $objParams->getOxygenMass + $objParams->getHydrogenMass;
        my $nTermGain = $objParams->getHydrogenMass;          
        print '        <Enzyme id="' . _mzid_Enzyme_id($i)
            . "\" cTermGain=\"" . "OH" . "\" nTermGain=\"" . "H"
            . "\" missedCleavages=\"" . $objParams->getPFA . "\" semiSpecific=\"";
        if ($enzyme->isSemiSpecific) {
          print "1";
        } else {
          print "0";
        }
        print "\">\n";
        print "          <SiteRegexp>";
        print "<![CDATA[";
        if (length($enzyme->getCleave($i)) == 1) {
          print "(?<=" . $enzyme->getCleave($i) . ")";
        } elsif (length($enzyme->getCleave($i)) > 1) {
          print "(?<=[" . $enzyme->getCleave($i) . "])";
        }
        if (length($enzyme->getRestrict($i)) == 1) {
          print "(?!" . $enzyme->getRestrict($i) . ")";
        } elsif (length($enzyme->getRestrict($i)) > 1) {
          print "(?![" . $enzyme->getRestrict($i) . "])";
        }
        print "]]>";
        
      # sort multi-residue strings to simplify comparison
        my($cleaver, $restrictor, @tmpRes);
        if (length($enzyme->getCleave($i)) > 1) {
          @tmpRes = split(//, lc($enzyme->getCleave($i)));
          @tmpRes = sort { $a cmp $b } @tmpRes;
          $cleaver = join("", @tmpRes)
        } else {
          $cleaver = lc($enzyme->getCleave($i));
        }
        if (length($enzyme->getRestrict($i)) > 1) {
          @tmpRes = split(//, lc($enzyme->getRestrict($i)));
          @tmpRes = sort { $a cmp $b } @tmpRes;
          $restrictor = join("", @tmpRes)
        } else {
          $restrictor = lc($enzyme->getRestrict($i));
        }
        
        print "</SiteRegexp>\n";
        print "          <EnzymeName>\n";

        my $cv = $cvParam{
          join(':', $cleaver, $restrictor ? $restrictor : '', lc($enzyme->getTitle))
        };

        if ($cv) {
          my $name = noXmlTag($$cv{'name'});
          print "            <cvParam accession=\"$$cv{'id'}\" name=\"$name\" cvRef=\"PSI-MS\" />\n";
        } else {
          # Alternative is to use M:1001956 "unspecified cleavage".
          print "            <userParam name=\"Enzyme\" value=\"" . noXmlTag($enzyme->getTitle) . "\" />\n";
        }
        print "          </EnzymeName>\n";
        print "        </Enzyme>\n";
      }
      print "      </Enzymes>\n";
    }
  }

  my $multiMassTables = 1;

  if ($objQuantMethod and $objQuantMethod->getNumberOfComponents > 1) {
    for my $c (0 .. $objQuantMethod->getNumberOfComponents()-1) {
      my $component = $objQuantMethod->getComponentByNumber($c);
      if ($component->getNumberOfIsotopes > 0) {
        $multiMassTables = $objQuantMethod->getNumberOfComponents;
        last;
      }
    }
  }
  
  for my $c (0 .. $multiMassTables-1) {
    my $objLocalMassesFile = new msparser::ms_masses($objMassesFile); 
    if ($multiMassTables > 1) {
      $objLocalMassesFile->applyIsotopes($objUmodConfigFile, $objQuantMethod->getComponentByNumber($c));
      print '      <MassTable id="' . _mzid_MassTable_id($objQuantMethod->getComponentByNumber($c)->getName) . "\"";
    } else {
      print '      <MassTable id="' . _mzid_MassTable_id(1) . "\"";
    }
    #my $err = new msparser::ms_errs;
  #  my $dummy = $objLocalMassesFile->applyFixedMods($vecFixed, $err);  # decided MassTable should be completely unmodified
    #unless ($err->isValid) {
    #  my $errorList = "Mascot Parser error(s):<BR>\n";
    #  for (my $i = 1; $i <= $err->getNumberOfErrors(); $i++) {
    #    $errorList .= $err->getErrorNumber($i) 
    #      . ": " . $err->getErrorString($i) . "<BR>\n";
    #  }
    #  die($errorList, " ", $idForFatalError);
    #}
    print " msLevel=\"1 2\">\n";
    for my $residue ('A' .. 'Z') {
      my $mass = $objLocalMassesFile->getResidueMass($mass_type, $residue);
      if ($residue ne 'B' and $residue ne 'X' and $residue ne 'Z' and $mass != 0) {
        print "        <Residue code=\"" . $residue . "\" mass=\"" . $mass . "\" />\n";
      }
    }
    print <<EOF;
        <AmbiguousResidue code="B">
          <cvParam accession="MS:1001360" name="alternate single letter codes" cvRef="PSI-MS" value="D N" />
        </AmbiguousResidue>
        <AmbiguousResidue code="Z">
          <cvParam accession="MS:1001360" name="alternate single letter codes" cvRef="PSI-MS" value="E Q" />
        </AmbiguousResidue>
        <AmbiguousResidue code="X">
          <cvParam accession="MS:1001360" name="alternate single letter codes" cvRef="PSI-MS" value="A C D E F G H I K L M N O P Q R S T U V W Y" />
        </AmbiguousResidue>
      </MassTable>
EOF
  }
  
  if (!$objResFile->isPMF) {
    print "      <FragmentTolerance>\n";

    if (my $cv = $$tolCVparam{$objParams->getITOLU}) {
      my $val = $$cv{'fmt'}->($objParams->getITOL);
      print "        <cvParam accession=\"MS:1001412\" name=\"search tolerance plus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n";
      print "        <cvParam accession=\"MS:1001413\" name=\"search tolerance minus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n";
    }

    print "      </FragmentTolerance>\n";
  }

  print "      <ParentTolerance>\n";

  if (my $cv = $$tolCVparam{$objParams->getTOLU}) {
    my $val = $$cv{'fmt'}->($objParams->getTOL);
    print "        <cvParam accession=\"MS:1001412\" name=\"search tolerance plus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n";
    print "        <cvParam accession=\"MS:1001413\" name=\"search tolerance minus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n";
  }

  print "      </ParentTolerance>\n";
  
  print "      <Threshold>\n";
  if ($objResFile->isPMF) {
    print "        <cvParam accession=\"MS:1001494\" name=\"no threshold\" cvRef=\"PSI-MS\" />\n";
  } else {
    print "        <cvParam accession=\"MS:1001316\" name=\"Mascot:SigThreshold\" cvRef=\"PSI-MS\" value=\"" . $sigThreshold . "\" />\n";
    if ($use_homology) {
      print "        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\" cvRef=\"PSI-MS\" value=\"homology\" />\n";
    } else {
      print "        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\" cvRef=\"PSI-MS\" value=\"identity\" />\n";
    } 
  }
  print "      </Threshold>\n";

  print <<EOF;
      <DatabaseFilters>
        <Filter>
          <FilterType>
            <cvParam accession="MS:1001020" name="DB filter taxonomy" cvRef="PSI-MS" />
          </FilterType>
EOF

  my $taxonomy = new msparser::ms_taxonomyfile;
  if ($objResFile->getTaxonomy($taxonomy)) {
    if ($taxonomy->getNumberOfEntries == 0 ) {
      print <<EOF;
          <Include>
            <cvParam accession="MS:1001467" name="taxonomy: NCBI TaxID" cvRef="PSI-MS" value="1" />
          </Include>
EOF
    } else {
      my $choice = $taxonomy->getEntryByNumber(0);
      if ($choice->getNumberOfIncludeTaxonomies > 0) {
        print "          <Include>\n";
        for my $i (0 .. $choice->getNumberOfIncludeTaxonomies()-1) {
          print "            <cvParam accession=\"MS:1001467\" name=\"taxonomy: NCBI TaxID\" cvRef=\"PSI-MS\" value=\"" . $choice->getIncludeTaxonomy($i) . "\" />\n";
          if ($protTaxStr) {
            my $tax_title = getTaxonomyNameFromId($choice->getIncludeTaxonomy($i), $taxNames);
            if ($tax_title) {
              print "            <cvParam accession=\"MS:1001469\" name=\"taxonomy: scientific name\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($tax_title) . "\" />\n";
            }
          }
        }
        print "          </Include>\n";
      }
    
      if ($choice->getNumberOfExcludeTaxonomies > 0) {
        print "          <Exclude>\n";
        for my $i (0 .. $choice->getNumberOfExcludeTaxonomies()-1) {
          print "            <cvParam accession=\"MS:1001467\" name=\"taxonomy: NCBI TaxID\" cvRef=\"PSI-MS\" value=\"". $choice->getExcludeTaxonomy($i) . "\" />\n";
          if ($protTaxStr) {
            my $tax_title = getTaxonomyNameFromId($choice->getExcludeTaxonomy($i), $taxNames);
            if ($tax_title) {
              print "            <cvParam accession=\"MS:1001469\" name=\"taxonomy: scientific name\" cvRef=\"PSI-MS\" value=\"" . noXmlTag($tax_title) . "\" />\n";
            }
          }
        }
        print "          </Exclude>\n";
      }
    }
  }
  print <<EOF;
        </Filter>
      </DatabaseFilters>
EOF

  if ($objSummary->isNA) {
    open(my $fh, '<', "../taxonomy/gencode.dmp") 
      or die sprintf("Cannot open gencode.dmp (%s:%d): %s (%s)", __FILE__, __LINE__, $!, $^E);

    print "      <DatabaseTranslation frames=\"1 2 3 -1 -2 -3\">\n";
    while (defined(my $gencode = <$fh>)) {
      my ($id, $dummy, $name, $table, $startcodons) = split(/\s*\|\s*/, $gencode);
      if ($table && $startcodons) {
        print '        <TranslationTable id="' . _mzid_TranslationTable_id($id) . "\" name=\"" . $name . "\">\n";
        print "          <cvParam accession=\"MS:1001025\" name=\"translation table\" cvRef=\"PSI-MS\" value=\"" . $table . "\" />\n";
        print "          <cvParam accession=\"MS:1001410\" name=\"translation start codons\" cvRef=\"PSI-MS\" value=\"" . $startcodons . "\" />\n";
        print "          <cvParam accession=\"MS:1001423\" name=\"translation table description\" cvRef=\"PSI-MS\" value=\"http://www.ncbi.nlm.nih.gov/Taxonomy/taxonomyhome.html/index.cgi?chapter=cgencodes#SG" . $id . "\" />\n";
        print "        </TranslationTable>\n";
      }
    }
    print "      </DatabaseTranslation>\n";

    close($fh) 
      or die sprintf("Cannot close gencode.dmp (%s:%d): %s (%s)", __FILE__, __LINE__, $!, $^E);
  }

  print "    </SpectrumIdentificationProtocol>\n";

  my $scoringType;
  if ($objResFile->getNumSeqsAfterTax() > 0 and $objResFile->getNumQueries() / $objResFile->getNumSeqsAfterTax() > $mudpitSwitch){
    $scoringType = "MudPIT";
  } else {
    $scoringType = "Standard";
  }
  my $maxProteinHits = $numHits;
  if ($maxProteinHits == 0) {
    $maxProteinHits = "Auto";
  }
  my $useUnigene = "true";
  if ($UniGeneFile eq "") {
    $useUnigene = "false";
  }
  print '    <ProteinDetectionProtocol id="' . _mzid_ProteinDetectionProtocol_id('MascotParser', 1) . "\"  analysisSoftware_ref=\"AS_mascot_parser\">\n";
  print "      <AnalysisParams>\n";
  print "        <cvParam accession=\"MS:1001316\" name=\"Mascot:SigThreshold\"                               cvRef=\"PSI-MS\" value=\"" . $sigThreshold         . "\" />\n";
  print "        <cvParam accession=\"MS:1001317\" name=\"Mascot:MaxProteinHits\"                             cvRef=\"PSI-MS\" value=\"" . $maxProteinHits       . "\" />\n";
  unless ($objResFile->isPMF) {  
    print "        <cvParam accession=\"MS:1001318\" name=\"Mascot:ProteinScoringMethod\"                       cvRef=\"PSI-MS\" value=\"" . $scoringType          . "\" />\n";
    print "        <cvParam accession=\"MS:1001319\" name=\"Mascot:MinMSMSThreshold\"                           cvRef=\"PSI-MS\" value=\"" . $ignoreIonsScoreBelow . "\" />\n";
    if ($use_homology) {
      print "        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\"                           cvRef=\"PSI-MS\" value=\"homology\" />\n";
    } else {
      print "        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\"                           cvRef=\"PSI-MS\" value=\"identity\" />\n";
    } 
  }
  if ($group_family) {
    print "        <cvParam accession=\"MS:1001759\" name=\"Mascot:ProteinGrouping\"                            cvRef=\"PSI-MS\" value=\"family clustering\" />\n";
  } else {
    print "        <cvParam accession=\"MS:1001759\" name=\"Mascot:ProteinGrouping\"                            cvRef=\"PSI-MS\" value=\"Occam's razor\" />\n";
  }
  print "        <cvParam accession=\"MS:1001320\" name=\"Mascot:ShowHomologousProteinsWithSamePeptides\"     cvRef=\"PSI-MS\" value=\"" . _mzid_bool($ShowSameSets) . "\" />\n";
  print "        <cvParam accession=\"MS:1001321\" name=\"Mascot:ShowHomologousProteinsWithSubsetOfPeptides\" cvRef=\"PSI-MS\" value=\"" . _mzid_bool($ShowSubSets) . "\" />\n";
  unless ($objResFile->isPMF) {  
    print "        <cvParam accession=\"MS:1001322\" name=\"Mascot:RequireBoldRed\"                             cvRef=\"PSI-MS\" value=\"" . _mzid_bool($RequireBoldRed) . "\" />\n";
    print "        <cvParam accession=\"MS:1001323\" name=\"Mascot:UseUnigeneClustering\"                       cvRef=\"PSI-MS\" value=\"" . _mzid_bool($useUnigene) . "\" />\n";
    print "        <cvParam accession=\"MS:1001324\" name=\"Mascot:IncludeErrorTolerantMatches\"                cvRef=\"PSI-MS\" value=\"" . _mzid_bool(!$NoErrorTolerant) . "\" />\n";
  }
  print "        <cvParam accession=\"MS:1001325\" name=\"Mascot:ShowDecoyMatches\"                           cvRef=\"PSI-MS\" value=\"" . _mzid_bool($ShowDecoyReport) . "\" />\n";
  if ($preferTaxonomy > 0) {
    my $taxonomyfile = MascotDat->taxonomy_file();
    my $taxonomy_entry;

    if ($preferTaxonomy < $taxonomyfile->getNumberOfEntries()) {
      $taxonomy_entry = $taxonomyfile->getEntryByNumber($preferTaxonomy);
    }

    my @tax_ids = ();

    if ($taxonomy_entry) {
      for my $i (0 .. $taxonomy_entry->getNumberOfIncludeTaxonomies()-1) {
        push @tax_ids, $taxonomy_entry->getIncludeTaxonomy($i);
      }
    }

    if (@tax_ids) {
      print "        <cvParam accession=\"MS:1001793\" name=\"Mascot:PreferredTaxonomy\"                           cvRef=\"PSI-MS\" value=\"" . join(' ', @tax_ids). "\" />\n";
    }
  }

  if ($percolate == 1) {
    my @features = split /\s*,\s*/, $objMascotOptions->getPercolatorFeatures();
    if ($percolate_rt) {
      push @features, 'retentionTime' unless 0 < grep { $_ eq 'retentionTime' } @features;
    }
    print "        <cvParam accession=\"MS:1001760\" name=\"Percolator:features\" cvRef=\"PSI-MS\" value=\"".join(' ', map { noXmlTag($_) } @features)."\" />\n";
  }

  if ($targetFDR) {
    my $FDR;
    if ($use_homology) {
      my $HF = $objSummary->getNumHitsAboveHomology(1 / $sigThreshold);
      my $HD = $objSummary->getNumDecoyHitsAboveHomology(1 / $sigThreshold);
      if ($HF > 0) {
        $FDR = sprintf("%.4f", $HD / $HF);
      }
    } else {
      my $IF = $objSummary->getNumHitsAboveIdentity(1 / $sigThreshold);
      my $ID = $objSummary->getNumDecoyHitsAboveIdentity(1 / $sigThreshold);
      if ($IF > 0) {
        $FDR = sprintf("%.4f", $ID / $IF);
      }
    }
    # NB: Exported as an absolute value between 0 and 1, not as a percentage.
    # XXX: it's actually target FDR, and only reported when target is selected
    if (defined $FDR) {
      print "        <cvParam accession=\"MS:1001364\" name=\"distinct peptide-level global FDR\" cvRef=\"PSI-MS\" value=\"$FDR\" />\n";
    }
  }

  print "      </AnalysisParams>\n";
  
  print "      <Threshold>\n";

  if ($objResFile->isPMF) {
    print "        <cvParam accession=\"MS:1001316\" name=\"Mascot:SigThreshold\" cvRef=\"PSI-MS\" value=\"" . $sigThreshold . "\" />\n";
    if ($use_homology) {
      print "        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\" cvRef=\"PSI-MS\" value=\"homology\" />\n";
    } else {
      print "        <cvParam accession=\"MS:1001758\" name=\"Mascot:SigThresholdType\" cvRef=\"PSI-MS\" value=\"identity\" />\n";
    } 
  } else {
    print "        <cvParam accession=\"MS:1001494\" name=\"no threshold\" cvRef=\"PSI-MS\" />\n";
  }

  print <<EOF;
      </Threshold>
    </ProteinDetectionProtocol>
  </AnalysisProtocolCollection>
EOF
}

sub printSearchModification {
  my ($mod_file_name, $fixed) = @_;
  
  my ($mod_name, $specificity) = $mod_file_name =~ /(.*)\s+\((.*)\)/;

  my ($specificityAccession, $specificityName, $residues);

  if ($specificity =~ /^N-term/i) {
    $specificityAccession = "MS:1001189";
    $specificityName      = "modification specificity N-term";
    ($residues) = $specificity =~ /^N-term\s+(.*)/i;
  } elsif ($specificity =~ /^C-term/i) {
    $specificityAccession = "MS:1001190";
    $specificityName      = "modification specificity C-term";
    ($residues) = $specificity =~ /^C-term\s+(.*)/i;
  } elsif( $specificity =~ /^Protein N-term/i) {
    $specificityAccession = "MS:1001189";
    $specificityName      = "modification specificity N-term";
    ($residues) = $specificity =~ /^Protein N-term\s+(.*)/i;
  } elsif ($specificity =~ /^Protein C-term/i) {
    $specificityAccession = "MS:1001190";
    $specificityName      = "modification specificity C-term";
    ($residues) = $specificity =~ /^Protein C-term\s+(.*)/i;
  } else {
    $specificityAccession = "";
    $specificityName      = "";
    $residues             = $specificity;
  }

  $residues = '' unless defined($residues);

  my $acc;
  my $mod2 = $objUmodConfigFile2->getModificationByName($mod_name);

  if ($mod2) {
    $acc = "UNIMOD:" . $mod2->getRecordID;
  } else {
    $acc = "UNIMOD:unknown";
  }

  my $delta = $objModFile->getModificationByName($mod_file_name)->getDelta($mass_type);

  if (length($residues) > 1) {
    $residues = join(" ", split(//, $residues));
  }

  print "        <SearchModification fixedMod=\"" . $fixed . "\" massDelta=\"" . $delta . "\" residues=\"" . $residues . "\">\n";
  if ($specificityAccession) {
    print "          <SpecificityRules>\n";
    print "            <cvParam accession=\"" . $specificityAccession . "\" cvRef=\"PSI-MS\" name=\"" . $specificityName . "\" />\n";
    print "          </SpecificityRules>\n";
  }
  print "          <cvParam accession=\"" . $acc . "\" name=\"" . noXmlTag($mod_name) . "\" cvRef=\"UNIMOD\" />\n";
  print "        </SearchModification>\n";
}

sub printDataCollection {
  my ($query_mod_and_dupe_cache, $fileIn, $objResFile, $objParams,
      $resfile_params, $resfile_children, $objSummary, $UniGeneFile,
      $objUmodConfigFile2, $sessionID, $sigThreshold, $msresFlags2, $percolate,
      $exported_queryList, $exported_sequences, $peptide_evidence_list,
      $aaHelpers, $elementalMasses, $tolCVparam, $ShowSameSets, $ShowSubSets,
      $protTaxStr, $protTaxId, $queryAll, $reportType, $protLen, $queryRaw,
      $sameset_ordering, $numcomponents, $firstComponentName) = @_;

  print "  <DataCollection>\n";

  my ($indexIsQueryNumber) = printDataCollectionInputs(
    $fileIn, $objResFile, $objParams, $resfile_params, $resfile_children,
    $objSummary, $UniGeneFile
  );

  print "    <AnalysisData>\n";

  printSpectrumIdentificationList(
    $query_mod_and_dupe_cache, $objResFile, $objParams, $resfile_params,
    $objSummary, $msresFlags2, $percolate, $objUmodConfigFile2, $sessionID,
    $exported_queryList, $exported_sequences, $peptide_evidence_list,
    $protTaxStr, $protTaxId, $queryAll, $aaHelpers, $elementalMasses,
    $tolCVparam, $queryRaw, $indexIsQueryNumber, $numcomponents,
    $firstComponentName,
  );

  printProteinDetectionList(
    $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
    $ShowSameSets, $ShowSubSets, $reportType, $protLen, $sameset_ordering
  );
  
  print "    </AnalysisData>\n";
  print "  </DataCollection>\n";
}

sub printDataCollectionInputs {
  my ($fileIn, $objResFile, $objParams, $resfile_params, $resfile_children,
      $objSummary, $UniGeneFile) = @_;

  print "    <Inputs>\n";

  for my $i (1 .. @$fileIn) {
    print "      <SourceFile location=\"file:///" . noXmlTag($fileIn[$i-1]) . "\" id=\"" . _mzid_SourceFile_id($i) . "\" >\n";
    print "        <FileFormat>\n";
    print "          <cvParam accession=\"MS:1001199\" name=\"Mascot DAT format\" cvRef=\"PSI-MS\" />\n";
    print "        </FileFormat>\n";
    print "      </SourceFile>\n";
  }

  if ($UniGeneFile) {
    print "      <SearchDatabase location=\"file:///" . noXmlTag($UniGeneFile)
                 . "\" id=\"" . _mzid_SearchDatabase_id('UniGene') 
                 . "\" name=\"" . 'UniGene'
                 . "\">\n";
    print "        <ExternalFormatDocumentation>http://www.ncbi.nlm.nih.gov/unigene/</ExternalFormatDocumentation>\n";
    print "        <DatabaseName>\n";
    print "          <userParam name=\"" . noXmlTag($UniGeneFile) . "\" />\n";
    print "        </DatabaseName>\n";
    print "      </SearchDatabase>\n";
  }

  # (When results are merged: merged results have the same databases, so we
  # don't need to merge database information manually.)

  for my $i (1 .. $objParams->getNumberOfDatabases()) {
    # releaseDate is not easy to figure out. We could, in principle, check
    # what the last modified time of the local FASTA file is (if it's
    # available), but that's not reliable. It's best to omit.

    print "      <SearchDatabase location=\"file:///" . noXmlTag($objResFile->getFastaPath($i))
                 . "\" id=\"" . _mzid_SearchDatabase_id($objParams->getDB($i)) 
                 . "\" name=\"" . noXmlTag($objParams->getDB($i)) 
                 . "\" numDatabaseSequences=\"" . $objResFile->getNumSeqs    # wait for parser function ### debug ###
                 . "\" numResidues=\""          . $objResFile->getNumResidues
                 . "\" version=\""              . $objResFile->getFastaVer($i)
                 . "\">\n";
    print "        <FileFormat>\n";
    print "          <cvParam accession=\"MS:1001348\" name=\"FASTA format\" cvRef=\"PSI-MS\" />\n";
    print "        </FileFormat>\n";
    print "        <DatabaseName>\n";
    print "          <userParam name=\"" . $objResFile->getFastaVer($i) . "\" />\n";
    print "        </DatabaseName>\n";
    
    if ($objSummary->isNA) {
      print "        <cvParam accession=\"MS:1001079\" name=\"database type nucleotide\" cvRef=\"PSI-MS\" />\n";
    } else {
      print "        <cvParam accession=\"MS:1001073\" name=\"database type amino acid\" cvRef=\"PSI-MS\" />\n";
    }
    
    print "      </SearchDatabase>\n";
  }

  my @indexIsQueryNumber = ();

  for my $i (1 .. $#$resfile_children) {
    my $resfile = $$resfile_children[$i];
    my $params = $$resfile_params[$i];

    my $location;

    if ($params->getDATAURL() eq '') {
        my $filename = $params->getFILENAME();
        $filename =~ s/\\/\//g; 
        $location = "file:///" . noXmlTag($filename);
    } else {
        $location = noXmlTag($params->getDATAURL());
    }
  
    my $indexIsQueryNumber = 1;
    my $objQueryTmp = new msparser::ms_inputquery($resfile, 1);
    if ($objQueryTmp->getIndex > -1) {
      $indexIsQueryNumber = 0;
    }

    print "      <SpectraData location=\"$location\" id=\"" . _mzid_SpectraData_id($i) . "\">\n";
    print "        <FileFormat>\n";
    if ($params->getFORMAT eq "Mascot generic") {
      print "          <cvParam accession=\"MS:1001062\" name=\"Mascot MGF format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "Sequest (.DTA)") {
      print "          <cvParam accession=\"MS:1000613\" name=\"DTA format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "Finnigan (.ASC)") {
      print "          <cvParam accession=\"PI:99999\" name=\"TODO: Finnigan List file\" cvRef=\"MASCOT:PSI:PI\" />\n";   # not worth worrying about
    } elsif ($params->getFORMAT eq "Micromass (.PKL)") {
      print "          <cvParam accession=\"MS:1000565\" name=\"Micromass PKL format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "PerSeptive (.PKS)") {
      print "          <cvParam accession=\"MS:1001245\" name=\"PerSeptive PKS format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "Sciex API III") {
      print "          <cvParam accession=\"MS:1001246\" name=\"Sciex API III format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "Bruker (.XML)") {
      print "          <cvParam accession=\"MS:1001247\" name=\"Bruker XML format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "mzData (.XML)") {
      print "          <cvParam accession=\"MS:1000564\" name=\"PSI mzData format\" cvRef=\"PSI-MS\" />\n";
      $indexIsQueryNumber = 0;
    } elsif ($params->getFORMAT eq "mzML (.mzML)") {
      print "          <cvParam accession=\"MS:1000584\" name=\"mzML format\" cvRef=\"PSI-MS\" />\n";
      $indexIsQueryNumber = 0;
    } elsif ($resfile->isPMF || $resfile->isSQ) {
      print "          <cvParam accession=\"MS:1001369\" name=\"text format\" cvRef=\"PSI-MS\" />\n";
    }
    print "        </FileFormat>\n";
    print "        <SpectrumIDFormat>\n";
    if ($resfile->isPMF || $resfile->isSQ) {
      print "          <cvParam accession=\"MS:1000775\" name=\"single peak list nativeID format\" cvRef=\"PSI-MS\" />\n";
    } elsif ($indexIsQueryNumber) {
      print "          <cvParam accession=\"MS:1001528\" name=\"Mascot query number\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "mzML (.mzML)") {
      print "          <cvParam accession=\"MS:1001530\" name=\"mzML unique identifier\" cvRef=\"PSI-MS\" />\n";
    } elsif ($params->getFORMAT eq "mzData (.XML)") {
      print "          <cvParam accession=\"MS:1000777\" name=\"spectrum identifier nativeID format\" cvRef=\"PSI-MS\" />\n";
    } else {
      print "          <cvParam accession=\"MS:1000774\" name=\"multiple peak list nativeID format\" cvRef=\"PSI-MS\" />\n";
    }
    print "        </SpectrumIDFormat>\n";  
    print "      </SpectraData>\n";

    $indexIsQueryNumber[$i] = $indexIsQueryNumber;
  }

  print "    </Inputs>\n";

  return (\@indexIsQueryNumber);
}

sub _determine_scan_id {
  my ($indexIsQueryNumber, $objParams, $objQuery, $query) = @_;

  my $scanID;

  if ($indexIsQueryNumber) {
    $scanID = 'query=' . $query;
  } elsif ($objParams->getFORMAT eq 'mzML (.mzML)') {
    $scanID = $objQuery->getStringTitle(1);
    $scanID =~ s/^nativeID=//;
    $scanID = 'mzMLid=' . $scanID;
  } elsif ($objParams->getFORMAT eq 'mzData (.XML)') {
    $scanID = 'spectrum=' . $objQuery->getScanNumbers;
  } else {
    $scanID = 'index=' . $objQuery->getIndex;
  }

  return $scanID;
}

sub printSpectrumIdentificationList {
  my ($query_mod_and_dupe_cache, $objResFile, $objParams, $resfile_params,
      $objSummary, $msresFlags2, $percolate, $objUmodConfigFile2, $sessionID,
      $exported_queryList, $exported_sequences, $peptide_evidence_list,
      $protTaxStr, $protTaxId, $queryAll, $aaHelpers, $elementalMasses,
      $tolCVparam, $queryRaw, $indexIsQueryNumber, $numcomponents,
      $firstComponentName) = @_;

  my $isPMF = $objResFile->isPMF;

  my ($modsByDelta, $full_unimod_file, $query_is_sig);
  
  if (not $isPMF) {
    # Site analysis initialisation.
    $modsByDelta = Util::SiteAnalysis::modifications_by_delta(
      $objUmodConfigFile2, $objParams,
    );
    # $objUmodConfigFile2 could be the reduced file embedded in the results
    # file.
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($sessionID);
    $full_unimod_file = new msparser::ms_umod_configfile(
      "../config/unimod.xml", "../html/xmlns/schema/unimod_2/unimod_2.xsd",
      $cs
    );
  
    $query_is_sig = $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH
      ? sub { Util::Peptides::above_homology($objSummary, $objSummary->getPeptide($_[0], 1)) }
      : sub { Util::Peptides::above_identity($objSummary, $objSummary->getPeptide($_[0], 1)) }
      ;
  }
  
  print '      <SpectrumIdentificationList id="' . _mzid_SpectrumIdentificationList_id(1) . '"'
               . " numSequencesSearched=\"" . $objResFile->getNumSeqsAfterTax
               . "\">\n"; # . $componentNames[$c] . "\">\n";

  # Now output all peptide scores matches
  # For PMF, there's just one spectrum, so just one SpectrumIdentificationResult
  if ($isPMF) {
    my $location;

    if ($objParams->getDATAURL() eq '') {
        my $filename = $objParams->getFILENAME();
        $filename =~ s/\\/\//g; 
        $location = 'file:///' . noXmlTag($filename);
    } else {
        $location = noXmlTag($objParams->getDATAURL());
    }

    print '        <SpectrumIdentificationResult id="' . _mzid_SpectrumIdentificationResult_id(1) . "\" spectrumID=\"file=$location\" spectraData_ref=\"" . _mzid_SpectraData_id(1) . "\">\n";
  } else {
    print "        <FragmentationTable>\n";
    print '          <Measure id="' . _mzid_Measure_id('mz') . "\">\n";
    print "            <cvParam cvRef=\"PSI-MS\" accession=\"MS:1001225\" name=\"product ion m/z\" unitCvRef=\"PSI-MS\" unitAccession=\"MS:1000040\" unitName=\"m/z\" />\n";
    print "          </Measure>\n";
    print '          <Measure id="' . _mzid_Measure_id('intensity') . "\">\n";
    print "            <cvParam cvRef=\"PSI-MS\" accession=\"MS:1001226\" name=\"product ion intensity\" unitCvRef=\"PSI-MS\" unitAccession=\"MS:1000131\" unitName=\"number of counts\" />\n";
    print "          </Measure>\n";
    print '          <Measure id="' . _mzid_Measure_id('error') . "\">\n";
    print "            <cvParam cvRef=\"PSI-MS\" accession=\"MS:1001227\" name=\"product ion m/z error\" unitAccession=\"MS:1000040\" unitName=\"m/z\" unitCvRef=\"PSI-MS\" />\n";
    print "          </Measure>\n";
    print "        </FragmentationTable>\n";
  }

  for my $query (1 .. $objResFile->getNumQueries()) {
    next unless ($queryAll || $$exported_queryList[$query] || $isPMF);

    my $objQuery   = new msparser::ms_inputquery($objResFile, $query);
    my (undef, $src_q, $src_file_id) = $objResFile->getSrcQueryAndFileIdForMultiFile($query);
    
    # Try scanID first - for an mzML file, this will be the unique id of the spectrum.
    my $scanID = _determine_scan_id($$indexIsQueryNumber[$src_file_id], $$resfile_params[$src_file_id], $objQuery, $src_q);
    
    my $haveOutputSIRelement = 0;

    my ($identityThreshold,  $homologyThreshold, $thisThreshold);

    if (!$isPMF) {
      $identityThreshold = $objSummary->getPeptideIdentityThreshold($query, 1 / $sigThreshold);
      $homologyThreshold = $objSummary->getHomologyThreshold($query, 1 / $sigThreshold);
      if ($use_homology and $homologyThreshold > 0) {
        $thisThreshold = $homologyThreshold;
      } else {
        $thisThreshold = $identityThreshold;
      }
    }

    my $any_ET_in_query = 0;
    my $siteAnalysisData;

    if (not $isPMF) {
      if (&$query_is_sig($query)) {
        # (When results are merged: $objParams is the right thing to pass
        # to site analysis, because it must contain the same tolerances
        # and mods in all input files.)
        $siteAnalysisData = analyse_sites_in_query(
          $query_mod_and_dupe_cache, $objResFile, $objSummary, $objParams,
          $query, $modsByDelta, $full_unimod_file,
          $objSummary->getMaxRankValue(),
        );
      }

      for my $rank (1 .. $objSummary->getMaxRankValue()) {
        my $objPeptide = $objSummary->getPeptide($query, $rank);
        next if $objPeptide->getPeptideStr(0) eq '';

        if ($objPeptide->getIsFromErrorTolerant()) {
          $any_ET_in_query = 1;
          last;
        }
      }
    }

    my @frag_rules = ();

    if (not $isPMF and $queryRaw) {
      @frag_rules = split(/,/, $objResFile->params->getRULES);
      @frag_rules = sort { $a <=> $b } @frag_rules; # ensure sorted ascending
    }
    
    for my $rank (1 .. $objSummary->getMaxRankValue()) {
      my $objPeptide = $objSummary->getPeptide($query, $rank);
      my $pepstr     = $objPeptide->getPeptideStr;
      next if $pepstr eq '';

      my @peptide_evidence = ();

      # The evidence needs to be printed in sorted order to keep the output
      # constant across different runs with the same parameters.
      for my $id (sort keys %{ $$peptide_evidence_list{$query}{$rank} || {} }) {
        push @peptide_evidence, "            <PeptideEvidenceRef peptideEvidence_ref=\"$id\" />\n";
      }

      if (not @peptide_evidence) {
        # It used to be legal in mzIdentML 1.0 to not have peptide evidence
        # for a SpectrumIdentificationItem. In mzIdentML 1.1, we must omit
        # the SpectrumIdentificationItem entirely. See bug #10759.
        next;
      }

      if (!$isPMF && !$haveOutputSIRelement) {
        print '        <SpectrumIdentificationResult id="' . _mzid_SpectrumIdentificationResult_id($query)
            . "\" spectrumID=\"" . $scanID . "\" spectraData_ref=\"" . _mzid_SpectraData_id($src_file_id) . "\">\n";
        $haveOutputSIRelement = 1;
      }
      
      my ($pepCharge, $charge);

      if ($objPeptide->getCharge > 0) {
        $charge = $$elementalMasses{'hydrogen'} - $$elementalMasses{'electron'};
        $pepCharge = $objPeptide->getCharge();
      } elsif ($objPeptide->getCharge == 0) {
        $charge = 0;
        $pepCharge = 1;
      } else {
        $charge = - $$elementalMasses{'hydrogen'} + $$elementalMasses{'electron'};
        $pepCharge = $objPeptide->getCharge();
      }

      my $calcMz = ($objPeptide->getMrCalc / abs($pepCharge)) + $charge;
    
      print '          <SpectrumIdentificationItem id="' . _mzid_SpectrumIdentificationItem_id($query, $rank)
          . '"  calculatedMassToCharge="' . $calcMz 
          . '" chargeState="' . $objPeptide->getCharge 
          . '" experimentalMassToCharge="' . $objResFile->getObservedMass($query) 
          . '" peptide_ref="' . _mzid_Peptide_id($query, $rank)
          ;

      my $thisThreshold;
      my @threshold_and_expect_cv = ();

      if (!$isPMF) {
        if ($objResFile->isErrorTolerant) {
          if ($objPeptide->getIsFromErrorTolerant) {
            $thisThreshold = undef;
          } else {
            my $homologyThreshold = $objSummary->getHomologyThreshold($query, 1 / $sigThreshold, $rank);
            if ($use_homology and $homologyThreshold > 0) {
              $thisThreshold = $homologyThreshold;
            } else {
              $thisThreshold = $identityThreshold;
            }

            # Only non-error tolerant matches can have threshold and expect
            # values.

            @threshold_and_expect_cv = (
              { id => "MS:1001172", name => "Mascot:expectation value", 
                value => $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $query) },
            );

            if ($any_ET_in_query) {
              push @threshold_and_expect_cv, (
                { id => "MS:1001371", name => "Mascot:identity threshold",
                  value => $identityThreshold },
                $homologyThreshold > 0
                  ? { id => "MS:1001370", name => "Mascot:homology threshold",
                      value => $homologyThreshold }
                  : ()
                  ,
              );
            }
          }
        } else {
          if ($use_homology and $homologyThreshold > 0) {
            $thisThreshold = $homologyThreshold;
          } else {
            $thisThreshold = $identityThreshold;
          }

          @threshold_and_expect_cv = (
            { id => "MS:1001172", name => "Mascot:expectation value", 
              value => $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $query) },
            # Use query-level threshold terms.
          );
        }
      }

      if ($isPMF) {
        print "\" rank=\"0";
        print "\" passThreshold=\"true";     # according to spec, if no threshold then set true(!)
      } else {
        print "\" rank=\"" . $rank;

        if (defined $thisThreshold) {
          if ($objPeptide->getIonsScore > $thisThreshold) {
            print "\" passThreshold=\"true";
          } else {
            print "\" passThreshold=\"false";
          }
        } else {
          print "\" passThreshold=\"true";
        }
      }

      if ($objPeptide->getComponentStr) {
        print '" massTable_ref="' . _mzid_MassTable_id($objPeptide->getComponentStr);
        print '" sample_ref="' . _mzid_Sample_id($objPeptide->getComponentStr);
      }

      print "\">\n";

      print @peptide_evidence;

      if ($isPMF) {
        # Nothing required or desired
      } else {
        if ($queryRaw) {
          if ($numcomponents > 1 && $objPeptide->getComponentStr) {
            printFragmentation($query, $rank, $objPeptide, $$aaHelpers{$objPeptide->getComponentStr}, \@frag_rules);
          } elsif ($numcomponents > 1) {
            printFragmentation($query, $rank, $objPeptide, $$aaHelpers{$firstComponentName}, \@frag_rules);
          } else {
            printFragmentation($query, $rank, $objPeptide, $$aaHelpers{"1"}, \@frag_rules);
          }
        }

        print "            <cvParam accession=\"MS:1001171\" name=\"Mascot:score\" cvRef=\"PSI-MS\" value=\"" 
          . sprintf("%.2f", $objPeptide->getIonsScore) . "\" />\n";

        if ($percolate == 1) {
          my ($pep, $q) = $objPeptide->getPercolatorScores();
          print "            <cvParam accession=\"MS:1001491\" name=\"percolator:Q value\" cvRef=\"PSI-MS\" value=\"$q\" />\n";
          print "            <cvParam accession=\"MS:1001493\" name=\"percolator:PEP\" cvRef=\"PSI-MS\" value=\"$pep\" />\n";
        }

        for (@threshold_and_expect_cv) {
          print "            <cvParam accession=\"$$_{'id'}\" name=\"$$_{'name'}\" cvRef=\"PSI-MS\" value=\"$$_{'value'}\" />\n";
        }

        if ($objSummary->isPeptideUnique($query, $rank)) {
          print "            <cvParam accession=\"MS:1001363\" name=\"peptide unique to one protein\" cvRef=\"PSI-MS\" />\n";
        } else {
          print "            <cvParam accession=\"MS:1001175\" name=\"peptide shared in multiple proteins\" cvRef=\"PSI-MS\" />\n";
        }

        if ($siteAnalysisData) {
          if (exists $$siteAnalysisData{$rank}) {
            print "            <cvParam accession=\"MS:1002012\" name=\"Mascot:PTM site assignment confidence\" cvRef=\"PSI-MS\" value=\"" . sprintf("%.2f", 100 * $$siteAnalysisData{$rank}{'confidence'}) . "\" unitAccession=\"UO:0000187\" unitName=\"percent\" unitCvRef=\"UO\" />\n";
          }
        }
      }

      print "          </SpectrumIdentificationItem>\n";
    }

    if (!$isPMF && $haveOutputSIRelement) {
      my $qmatch = $objSummary->getQmatch($query);

      if ($any_ET_in_query) {
        # Individual matches have their own CV parameters for thresholds.
      } else {
        print "          <cvParam accession=\"MS:1001371\" name=\"Mascot:identity threshold\"  cvRef=\"PSI-MS\" value=\"" . $identityThreshold . "\" />\n";
        if ($homologyThreshold > 0) {
          print "          <cvParam accession=\"MS:1001370\" name=\"Mascot:homology threshold\"  cvRef=\"PSI-MS\" value=\"" . $homologyThreshold . "\" />\n";
        }
      }

      print "          <cvParam accession=\"MS:1001030\" name=\"number of peptide seqs compared to each spectrum\"  cvRef=\"PSI-MS\" value=\"" . $qmatch . "\" />\n";

      if ($objQuery->getScanNumbers) {
        print "          <cvParam accession=\"MS:1000797\" name=\"peak list scans\"  cvRef=\"PSI-MS\" value=\"" . $objQuery->getScanNumbers . "\" />\n";
      }

      if ($objQuery->getRetentionTimes) {
        print "          <cvParam accession=\"MS:1001114\" name=\"retention time(s)\"  cvRef=\"PSI-MS\" value=\"" . $objQuery->getRetentionTimes . "\" unitAccession=\"UO:0000010\" unitName=\"second\" unitCvRef=\"UO\" />\n";
      }

      if ($objQuery->getStringTitle(1)) {
        print "          <cvParam accession=\"MS:1000796\" name=\"spectrum title\"  cvRef=\"PSI-MS\" value=\"" . noXmlTag($objQuery->getStringTitle(1)) . "\" />\n";
      }

      if ($objQuery->getPepTol && $objQuery->getPepTolUnits) {
        my $cv = $$tolCVparam{$objQuery->getPepTolUnits};
        if ($cv) {
          my $val = $$cv{'fmt'}->($objQuery->getPepTol);
          print "        <cvParam accession=\"MS:1001412\" name=\"search tolerance plus value\" value=\"" . $val . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n";
          print "        <cvParam accession=\"MS:1001413\" name=\"search tolerance minus value\" value=\"" . $objQuery->getPepTol . "\" cvRef=\"PSI-MS\" unitAccession=\"$$cv{'unitAccession'}\" unitName=\"$$cv{'unitName'}\" unitCvRef=\"UO\" />\n";
        }
      }

      for my $j (1 .. 20) {
        if ($objQuery->getSeq($j)) {
          print "            <userParam name=\"SEQ_$j\" value=\"" . $objQuery->getSeq($j) . "\" />\n";
        } else {
          last;
        }
      }

      for my $j (1 .. 20) {
        if ($objQuery->getComp($j)) {
          print "            <userParam name=\"COMP_$j\" value=\"" . $objQuery->getComp($j) . "\" />\n";
        } else {
          last;
        }
      }

      for my $j (1 .. 20) {
        if ($objQuery->getTag($j)) {
          $objQuery->getTag($j) =~ /([et]),(.*)/i;
          if (lc($1) eq "e") {
            print "            <userParam name=\"ETAG_$j\" value=\"" . $2 . "\" />\n";
          } else {
            print "            <userParam name=\"TAG_$j\" value=\"" . $2 . "\" />\n";
          }
        } else {
          last;
        }
      }

      if ($objQuery->getIT_MODS(1)) {
      # Need reference to query level mods, otherwise cannot fully understand results.
      # Structured information for mods in matched peptides will be found in 
      # /mzIdentML/SequenceCollection/Peptide/Modification
        print "            <userParam name=\"IT_MODS\" value=\"" . noXmlTag($objQuery->getIT_MODS(1)) . "\" />\n";
      }
      if ($objQuery->getINSTRUMENT(1)) {
        print "            <userParam name=\"INSTRUMENT\" value=\"" . noXmlTag($objQuery->getINSTRUMENT(1)) . "\" />\n";
      }
      if ($objQuery->getRULES()) {
        print "            <userParam name=\"RULES\" value=\"" . $objQuery->getRULES() . "\" />\n";
      }
      print "        </SpectrumIdentificationResult>\n";
    }
  }
  if ($isPMF) {
    print "        </SpectrumIdentificationResult>\n";
  }
  print "      </SpectrumIdentificationList>\n";
}

BEGIN {
  my %fragCVparam = (
    $msparser::ms_fragmentationrules::FRAG_IMMONIUM => {
      accession => "MS:1001239", name => "frag: immonium ion"
    },
    $msparser::ms_fragmentationrules::FRAG_A_SERIES => {
      accession => "MS:1001229", name => "frag: a ion"
    },
    $msparser::ms_fragmentationrules::FRAG_A_MINUS_NH3 => {
      accession => "MS:1001235", name => "frag: a ion - NH3"
    },
    $msparser::ms_fragmentationrules::FRAG_A_MINUS_H2O => {
      accession => "MS:1001234", name => "frag: a ion - H2O"
    },
    $msparser::ms_fragmentationrules::FRAG_B_SERIES => {
      accession => "MS:1001224", name => "frag: b ion"
    },
    $msparser::ms_fragmentationrules::FRAG_B_MINUS_NH3 => {
      accession => "MS:1001232", name => "frag: b ion - NH3"
    },
    $msparser::ms_fragmentationrules::FRAG_B_MINUS_H2O => {
      accession => "MS:1001222", name => "frag: b ion - H2O"
    },
    $msparser::ms_fragmentationrules::FRAG_C_SERIES => {
      accession => "MS:1001231", name => "frag: c ion"
    },
    $msparser::ms_fragmentationrules::FRAG_X_SERIES => {
      accession => "MS:1001228", name => "frag: x ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Y_SERIES => {
      accession => "MS:1001220", name => "frag: y ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Y_MINUS_NH3 => {
      accession => "MS:1001233", name => "frag: y ion - NH3"
    },
    $msparser::ms_fragmentationrules::FRAG_Y_MINUS_H2O => {
      accession => "MS:1001223", name => "frag: y ion - H2O"
    },
    $msparser::ms_fragmentationrules::FRAG_Z_SERIES => {
      accession => "MS:1001230", name => "frag: z ion"
    },
    $msparser::ms_fragmentationrules::FRAG_INTERNAL_YB => {
      accession => "MS:1001365", name => "frag: internal yb ion"
    },
    $msparser::ms_fragmentationrules::FRAG_INTERNAL_YA => {
      accession => "MS:1001366", name => "frag: internal ya ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Z_PLUS_1 => {
      accession => "MS:1001367", name => "frag: z+1 ion"
    },
    $msparser::ms_fragmentationrules::FRAG_D_SERIES => {
      accession => "MS:1001236", name => "frag: d ion"
    },
    $msparser::ms_fragmentationrules::FRAG_V_SERIES  => {
      accession => "MS:1001237", name => "frag: v ion"
    },
    $msparser::ms_fragmentationrules::FRAG_W_SERIES => {
      accession => "MS:1001238", name => "frag: w ion"
    },
    $msparser::ms_fragmentationrules::FRAG_Z_PLUS_2  => {
      accession => "MS:1001368", name => "frag: z+2 ion"
    },
  );

  sub printFragmentation {
    my ($query, $rank, $peptide, $objAahelper, $rules) = @_;
  
    my $any_printed = 0;
  
    my ($singleCharge, $doubleCharge) = (0, 0);
  
    foreach my $rule (@$rules) {
      if ($rule == 1) {
        $singleCharge = 1;
        next;
      }
  
      if ($rule == 2 || $rule == 3) {
        if (($rule == 2 && $peptide->getCharge > 1)
          || ($rule == 3 && $peptide->getCharge > 2)) {
          $doubleCharge = 1;
        }
        next;
      } 
  
      for my $charge (1 .. 2) {
        if ($charge == 1 && $singleCharge || 
            $charge == 2 && $doubleCharge 
                         && $rule != $msparser::ms_fragmentationrules::FRAG_IMMONIUM
                         && $rule != $msparser::ms_fragmentationrules::FRAG_INTERNAL_YB
                         && $rule != $msparser::ms_fragmentationrules::FRAG_INTERNAL_YA) {
          my $cv = $fragCVparam{$rule};
          next if not $cv;
  
          my $fragments  = new msparser::ms_fragmentvector;
          my $err        = new msparser::ms_errs;
          if (!$objAahelper->calcFragmentsEx($peptide,
                                      $rule, 
                                      $charge,
                                      0,  # min fragment mass to return
                                      $peptide->getMrCalc, # max fragment mass to return 
                                      $objResFile->params->getMassType, # monoisotopic / average
                                      $fragments,
                                      $err)) {
            die "Error creating fragments";
          } 
  
          $fragments->addExperimentalData($objResFile, $query);
  
          my (@idx, @exptMasses, @exptIntensities, @exptErrors);
  
          for my $i (0 .. $fragments->getNumberOfFragments()-1) {
            my $fragment = $fragments->getFragmentByNumber($i);
  
            if ($fragment->getMatchedIonMass > 0) {
              if ($fragment->isInternal) {
                push @idx, $fragment->getStart, $fragment->getEnd;
              } else {
                push @idx, $fragment->getColumn;
              }
              push @exptMasses, $fragment->getMatchedIonMass();
              push @exptIntensities, $fragment->getMatchedIonIntensity();
              push @exptErrors, sprintf("%.4f", ($fragment->getMatchedIonMass - $fragment->getMass));
            }
          }
  
          if (@idx) {
            if (!$any_printed) {
              print "            <Fragmentation>\n";
              $any_printed = 1;
            }
            print "              <IonType index=\"" . join(' ', @idx) . "\" charge=\"" . $charge . "\">\n";
            print "                <FragmentArray values=\"" . join(' ', @exptMasses)      . '" measure_ref="' . _mzid_Measure_id('mz') . "\" />\n";
            print "                <FragmentArray values=\"" . join(' ', @exptIntensities) . '" measure_ref="' . _mzid_Measure_id('intensity') . "\" />\n";
            print "                <FragmentArray values=\"" . join(' ', @exptErrors)      . '" measure_ref="' . _mzid_Measure_id('error') . "\" />\n";
            print "                <cvParam cvRef=\"PSI-MS\" accession=\"" . $$cv{'accession'} . "\" name=\"" . $$cv{'name'} . "\" />\n";
            print "              </IonType>\n";
          }
        }
      }
    }
  
    if ($any_printed) {
      print "            </Fragmentation>\n";
    }
  
    return 1;
  }
}

sub printProteinDetectionList {
  my ($peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
      $ShowSameSets, $ShowSubSets, $reportType, $protLen, 
      $sameset_ordering) = @_;

  my %cv_anchor_protein     = ( acc => 'MS:1001591', name => 'anchor protein' );
  my %cv_family_member      = ( acc => 'MS:1001592', name => 'family member protein' );

  # In Mascot <= 2.2, grouping was based on sequence.
  my %cv_sequence_sameset   = ( acc => 'MS:1001594', name => 'sequence same-set protein' );
  my %cv_sequence_subset    = ( acc => 'MS:1001596', name => 'sequence sub-set protein' );
  my %cv_sequence_intersect = ( acc => 'MS:1001598', name => 'sequence subsumable protein' );

  # In Mascot >= 2.3, family grouping is based on spectrum.
  my %cv_spectrum_sameset   = ( acc => 'MS:1001595', name => 'spectrum same-set protein' );
  my %cv_spectrum_subset    = ( acc => 'MS:1001597', name => 'spectrum sub-set protein' );
  my %cv_spectrum_intersect = ( acc => 'MS:1001599', name => 'spectrum subsumable protein' );
  
  print '      <ProteinDetectionList id="' . _mzid_ProteinDetectionList_id(1) . "\">\n";

  my $isPMF = $objResFile->isPMF;

  for my $i (1 .. $objSummary->getNumberOfHits()) {
    my $objProtein = $objSummary->getHit($i);
    next if $isPMF and $objProtein->isPMFMixture();

    my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets, 
      $$sameset_ordering{$i}
    );

    print '        <ProteinAmbiguityGroup id="' . _mzid_ProteinAmbiguityGroup_id($i) . "\" >\n";

    printProteinDetectionHypothesis(
      $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
      $sameset[0], $reportType, $protLen, { %cv_anchor_protein }
    );

    my $cv_anchor = _mzid_DBSeq_id($sameset[0]->getDB, $sameset[0]->getAccession);

    if ($group_family) {

    # Other proteins that match the same set of peptides
      if ($ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          printProteinDetectionHypothesis(
            $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold, $_,
            $reportType, $protLen, 
            { %cv_spectrum_sameset, val => $cv_anchor }
          );
        }
      }
      
    # Other family members
      my $k = 1;
      while (my $objProtein = $objSummary->getNextFamilyProtein($i, $k++)) {
        my @sameset = common_subs::compileSameset($objSummary, $objProtein, $ShowSameSets,
          $$sameset_ordering{$i},
        );
        printProteinDetectionHypothesis(
            $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
            $sameset[0], $reportType, $protLen, 
            { %cv_family_member, val => $cv_anchor }
        );

      # And their same-sets
        if ($ShowSameSets) {
          my $cv_fm = _mzid_DBSeq_id($sameset[0]->getDB, $sameset[0]->getAccession);
          for (@sameset[1 .. $#sameset]) {
            printProteinDetectionHypothesis(
              $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
              $_, $reportType, $protLen, 
              { %cv_spectrum_sameset, val => $cv_fm }
            );
          }
        }
      }
      
    # sub-sets and intersections for the entire family
      if ($ShowSubSets > 0) {
        my $k = 1;
        while (my $objProtein = $objSummary->getNextSubsetProtein($i, $k++, 1)) {
          # We need the full sameset to filter out proteins from
          # getSimilarProteins().
          my @sameset = common_subs::compileSameset($objSummary, $objProtein, 1, 
            $$sameset_ordering{$i},
          );
          my %is_sameset = map { key_acc($_) => 1 } @sameset;

          my @supersets = _get_supersets($sameset[0], \%is_sameset);
          my %cv = @supersets > 1 ? %cv_spectrum_intersect : %cv_spectrum_subset;

          printProteinDetectionHypothesis(
            $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
            $sameset[0], $reportType, $protLen, 
            map { 
              { %cv, val => _mzid_DBSeq_id($_->{'db'}, $_->{'acc'}) }
            } @supersets
          );
  
        # And their same-sets
          if ($ShowSameSets) {
            my $cv_subset = _mzid_DBSeq_id($sameset[0]->getDB, $sameset[0]->getAccession);
            for (@sameset[1 .. $#sameset]) {
              printProteinDetectionHypothesis(
               $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
               $_, $reportType, $protLen, 
               { %cv_spectrum_sameset, val => $cv_subset }) 
            }
          }
        }
        
      }
      
    } else {
      if ($ShowSameSets) {
        for (@sameset[1 .. $#sameset]) {
          printProteinDetectionHypothesis(
            $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
            $_, $reportType, $protLen, 
            { %cv_sequence_sameset, val => $cv_anchor }) 
        }
      }
    
      my $j = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($i, $j++)) {
        # We need the full sameset to filter out proteins from
        # getSimilarProteins().
        my @sameset = common_subs::compileSameset($objSummary, $objProtein, 1);
        my %is_sameset = map { key_acc($_) => 1 } @sameset;

        my @supersets = _get_supersets($objProtein, \%is_sameset);
        my %cv = @supersets > 1 ? %cv_sequence_intersect : %cv_sequence_subset;

        printProteinDetectionHypothesis(
          $peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
          $objProtein, $reportType, $protLen, 
          map {
            { %cv, val => _mzid_DBSeq_id($_->{'db'}, $_->{'acc'}) }
          } @supersets
        );
      }
      
    }
    
    print "        </ProteinAmbiguityGroup>\n";
  }

  print "      </ProteinDetectionList>\n";
}

sub _get_supersets {
  my ($protein, $is_sameset) = @_;

  my $acc = msparser::VectorString->new;
  my $db = msparser::vectori->new;
  my $n = $protein->getSimilarProteins($acc, $db);
  my @supersets = ();

  for my $i (0 .. $n-1) {
    next if $$is_sameset{key_acc_str($db->get($i), $acc->get($i))};
    push @supersets, { db => $db->get($i), acc => $acc->get($i) };
  }

  return @supersets;
}

sub printProteinDetectionHypothesis {
  my ($peptide_evidence_list, $objResFile, $objSummary, $sigThreshold,
      $objProtein, $reportType, $protLen, @cv_relationships) = @_;

  my $passThreshold = "true";

  if ($objResFile->isPMF && $objProtein->getScore < $objSummary->getProteinThreshold(1 / $sigThreshold)) { 
    $passThreshold = "false"
  }

  my $preferred_acc = common_subs::preferredAccession($objProtein->getDB(), $objProtein->getAccession());
  
  print '          <ProteinDetectionHypothesis id="' . _mzid_ProteinDetectionHypothesis_id($preferred_acc, $objProtein->getFrame()) . '" dBSequence_ref="' . _mzid_DBSeq_id($objProtein->getDB, $preferred_acc) . "\"  passThreshold=\"" . $passThreshold . "\">\n";

  for my $pepnum (1 .. $objProtein->getNumPeptides()) {
    my $query     = $objProtein->getPeptideQuery($pepnum);
    my $rank      = $objProtein->getPeptideP($pepnum);
    my $startpos  = $objProtein->getPeptideStart($pepnum);
    my $endpos    = $objProtein->getPeptideEnd($pepnum);
    my $resbefore = $objProtein->getPeptideResidueBefore($pepnum);
    my $resafter  = $objProtein->getPeptideResidueAfter($pepnum);     

    my $id = _mzid_PeptideEvidence_id($query, $rank, $preferred_acc, $objProtein->getPeptideFrame($pepnum), $startpos, $endpos);
    next unless $$peptide_evidence_list{$query}{$rank}{$id};

    print "            <PeptideHypothesis peptideEvidence_ref=\"" . $id . "\">\n";
    print "              <SpectrumIdentificationItemRef spectrumIdentificationItem_ref=\"" . _mzid_SpectrumIdentificationItem_id($query, $rank) . "\" />\n";
    print "            </PeptideHypothesis>\n";
  }

  print "            <cvParam accession=\"MS:1001171\" name=\"Mascot:score\" cvRef=\"PSI-MS\" value=\"" . $objProtein->getScore() . "\" />\n";
  
  if ($reportType eq 'concise') {
    my $expect = $objSummary->getProteinExpectationValue($objProtein->getScore());
    print "            <cvParam accession=\"MS:1001172\" name=\"Mascot:expectation value\" cvRef=\"PSI-MS\" value=\"" . 
      sprintf("%.3g", $expect) . "\" />\n";
  }

  for (@cv_relationships) {
    my ($name, $acc, $val) = map { noXmlTag($_) } @$_{qw(name acc val)};
    next unless defined($name) and defined($acc);
    $val = '' unless defined $val;
    print "            <cvParam accession=\"$acc\" name=\"$name\" cvRef=\"PSI-MS\" value=\"$val\" />\n";
  }

  if ($protLen) {
    my $coverage = 0;  
    my $length = getProteinLen($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
    if ($length) {
    # accurate
      $coverage = $objProtein->getCoverage() * 100 / $length;
    } else {
      my $protMass = mustGetProteinMass($objProtein->getAccession(), $objSummary, $objParams, $objProtein->getFrame(), $objProtein->getDB);
      if ($protMass) {
      # approximate
        $coverage = $objProtein->getCoverage() * 100 * 110 / $protMass;
      }
    }
    $coverage = int $coverage;
    
    print "            <cvParam accession=\"MS:1001093\" name=\"sequence coverage\" cvRef=\"PSI-MS\" value=\"" . $coverage . "\" />\n";
  }
  
  if ($reportType eq 'concise') {
    my $numUnmatched = $objResFile->getNumQueries - $objProtein->getNumPeptides;
    print "            <cvParam accession=\"MS:1001362\" name=\"number of unmatched peaks\" cvRef=\"PSI-MS\" value=\"" . $numUnmatched . "\" />\n";
  } else {
    print "            <cvParam accession=\"MS:1001097\" name=\"distinct peptide sequences\" cvRef=\"PSI-MS\" value=\"" . $objProtein->getNumDistinctPeptides(0, $msparser::ms_protein::DPF_SEQUENCE) . "\" />\n";
  }

  print "          </ProteinDetectionHypothesis>\n";
}

sub printBibliographicReference {
  print <<EOF;
  <BibliographicReference 
    id="10.1002/(SICI)1522-2683(19991201)20:18&lt;3551::AID-ELPS3551&gt;3.0.CO;2-2"
    authors="David N. Perkins, Darryl J. C. Pappin, David M. Creasy, John S. Cottrell"
    editor="" 
    title="Probability-based protein identification by searching sequence databases using mass spectrometry data"
    name="Probability-based protein identification by searching sequence databases using mass spectrometry data"
    publication="Electrophoresis" 
    volume="20" 
    issue="18" 
    pages="3551-3567" 
    publisher="Wiley VCH"
    year="1999"
    />
EOF
}

sub _tolerance_cv_param_map {
  return {
    Da => { 
      unitAccession => "UO:0000221", unitName => "dalton",
      fmt => sub { $_[0] },
    },
    mmu => {
      unitAccession => "UO:0000221", unitName => "dalton",
      fmt => sub { $_[0]/1000 },
    },
    '%' => {
      unitAccession => "UO:0000187", unitName => "percent",
      fmt => sub { $_[0] },
    },
    ppm => {
      unitAccession => "UO:0000169", unitName => "parts per million",
      fmt => sub { $_[0] },
    },
  }
}



###############################################################################
# mzTab()
# output PSI mzTab metadata (MTD), protein table (PRH, PRT) and peptide table
# (PEH, PEP)
###############################################################################

sub mzTab {
  my $vecFixed = _make_mod_vector($objResFile, $objModFile, 'FixedMod');

# make hash for fast retrieval of fixed mods
  my %fixedMods = _make_fixed_mods_map($vecFixed);

  # We export only two kinds of mzTab 1.0 files:
  #   - Complete Identification
  #   - Complete Quantification
  # Quantification export is selected when the results file has a quantitation
  # method and the ms_ms2quantitation object has been initialised without
  # error. Otherwise the file is exported in Identification mode.

  # We can't print out the header block until we know exactly which
  # modifications the peptide and protein tables contain. This because the
  # 'mod' header in MTD must list all modifications mentioned in either data
  # table. The peptide table might not contain all modifications given as
  # search parameters, and it might contain error tolerant mods. Thus, we
  # need to iterate over proteins and peptides first, pool the data table
  # rows, and then print out everything at the very end.
  #
  # Also, we need to iterate over peptide data before constructing the protein
  # rows, because we need to map all variable modifications from the peptides
  # onto the protein sequence.

  my $sameset_ordering = common_subs::samesetOrderingByPreferredTaxonomy() || {};

  my $showPRH = $thisScript->param($urlParams{'protein_master'});
  my $protDesc = $thisScript->param($urlParams{'prot_desc'});
  my $protTaxStr = $thisScript->param($urlParams{'prot_tax_str'});
  my $protCover = $thisScript->param($urlParams{'prot_cover'});
  my $protEmPAI = $thisScript->param($urlParams{'prot_empai'});
  my $protpI = $thisScript->param($urlParams{'prot_pi'});
  my $protQuant = $thisScript->param($urlParams{'prot_quant'});

  my $showPEH = $thisScript->param($urlParams{'peptide_master'});
  my $pepDelta = $thisScript->param($urlParams{'pep_delta'});
  my $pepScore = $thisScript->param($urlParams{'pep_score'});
  my $pepHomThr = $thisScript->param($urlParams{'pep_homol'});
  my $pepIdThr = $thisScript->param($urlParams{'pep_ident'});
  my $pepFrame = $thisScript->param($urlParams{'pep_frame'});
  my $pepNumMatch = $thisScript->param($urlParams{'pep_num_match'});
  my $pepScanTitle = $thisScript->param($urlParams{'pep_scan_title'});
  my $pepQuant = $thisScript->param($urlParams{'pep_quant'});

  my $filter_duplicates = 1;
  if ($thisScript->param($urlParams{'show_pep_dupes'})) {
    $filter_duplicates = 0;
  }

  my $is_decoy = $objParams->getDECOY > 0 && $ShowDecoyReport;
  my $is_error_tolerant = $objResFile->isErrorTolerant();

  if (not $protQuant and not $pepQuant) {
    # Export Identification only.
    $objQuantMethod = undef;
    $objMS2Quant = undef;
  } else {
    # If either quantitation data has been selected, we must export both.
    # This is required by the specification. The user can choose not to
    # export the peptide tables if he does not want to see peptide ratios.
    $pepQuant = $protQuant = 1;
  }

  my ($ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs) = mzTabExperimentStructure(
      \@fileIn, $objResFile, $objParams, $objSummary, $objQuantMethod, 
      $objMS2Quant, $msresFlags, $sigThreshold, $numHits, $UniGeneFile,
      $ignoreIonsScoreBelow, $minPepLen, $msresFlags2, $ShowSubSets,
  );

  my %query_mod_and_dupe_cache = ();

  # We need to collect protein data even if $showPEH == 0, because the
  # peptide rows still report protein accessions.
  my ($protein_rows, $peptide_map) = collectMzTabProteins(
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $objMS2Quant, $sigThreshold, $ShowSameSets, $ShowSubSets, $reportType,
    $protDesc, $protTaxStr, $protCover, $protEmPAI, $protpI, $protQuant,
    \%query_mod_and_dupe_cache, $filter_duplicates, $sameset_ordering
  );

  my $indexIsQueryNumber = 1;
  my $objQueryTmp = new msparser::ms_inputquery($objResFile, 1);
  if ($objQueryTmp->getIndex > -1) {
    $indexIsQueryNumber = 0;
  } elsif ($objParams->getFORMAT eq "mzData (.XML)") {
    $indexIsQueryNumber = 0;
  } elsif ($objParams->getFORMAT eq "mzML (.mzML)") {
    $indexIsQueryNumber = 0;
  }

  # Note that we must collect peptide information even if $showPRH == 0
  # and $showPEH == 0, because we need the modification tally.
  my ($peptide_rows, $protein_mods, $mod_tally, $any_retention_time) = collectMzTabPeptides(
      $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
      $msresFlags2, $objMS2Quant, $sigThreshold, $objUmodConfigFile2, $sessionID,
      $indexIsQueryNumber, $is_decoy, $is_error_tolerant, $pepDelta, $pepScore,
      $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch, $pepScanTitle, $pepQuant,
      $massDP, \%vmMass, \%vmString, $percolate, \%fixedMods, 
      \%query_mod_and_dupe_cache, $peptide_map
  );

  my @database_map = ();

  if ($UniGeneFile) {
    # We know the UniGene file exists, because it has been validated before.
    my $mtime = (stat($UniGeneFile))[9];
    $database_map[0] = {
      full_name => $UniGeneFile,
      abbr_name => $UniGeneFile,
      version => _fmtdate_iso8601($mtime),
    };
  }

  for my $i (1 .. $objParams->getNumberOfDatabases()) {
    # The specification recommends using MIRIAM registered names, but that's
    # up to the user who created and named the database definition.
    my $name = $objParams->getDB($i);

    $database_map[$i] = {
      full_name => sprintf('%s (%d entries)', $name, $objResFile->getNumSeqs($i)),
      abbr_name => $name,
      version => $objResFile->getFastaVer($i),
    };
  }

  printMzTabMTD(
    $thisScript, $URI, \%urlParams, $ms_run_resfiles, $comp_idx_to_assay,
    $study_variable_defs, $objSummary, $objQuantMethod, $objMS2Quant,
    $objUmodConfigFile2, $indexIsQueryNumber, $reportType, $protTaxStr,
    \%taxNames, $pepDelta, \@quantWarnings, $mod_tally,
    $any_retention_time, $showPRH, $showPEH, $protQuant, $pepQuant,
    $percolate
  );

  my $num_ms_runs = $#$ms_run_resfiles;
  my $num_study_variables = scalar(@$study_variable_defs);

  if ($showPRH) {
    print "\n";
  
    printMzTabPRHandPRT(
      $protein_rows, $objResFile->isPMF, $objMS2Quant, $protQuant,
      $num_ms_runs, $num_study_variables, $protEmPAI, $protpI, $reportType,
      $protein_mods, \@database_map,
    );
  }

  if ($showPEH) {
    if ($objQuantMethod) {
      print "\n";
  
      # Only export a Peptides section in Quantification files.
      printMzTabPEHandPEP(
        $peptide_rows, $objMS2Quant, $num_ms_runs, $num_study_variables,
        $is_decoy, $is_error_tolerant, $percolate, \@database_map, $pepDelta,
        $pepScore, $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch,
        $pepScanTitle, $pepQuant
      );
    }

    print "\n";

    # Always export the PSM section if peptide output is desired, even in
    # Quantification files. We treat PSMs as if they were peptides (for
    # simplicity), so these two sections are nearly identical. The main
    # difference is that the PSM table contains data about start and end
    # residues and calculated mass, while the Peptides table contains
    # quantitation ratios.
    printMzTabPSHandPSM(
      $peptide_rows, $is_decoy, $is_error_tolerant, $percolate, \@database_map,
      $pepDelta, $pepScore, $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch,
      $pepScanTitle
    );
  }
}

sub mzTabExperimentStructure {
  my (
      $fileIn, $objResFile, $objParams, $objSummary, $objQuantMethod, 
      $objMS2Quant, $msresFlags, $sigThreshold, $numHits, $UniGeneFile,
      $ignoreIonsScoreBelow, $minPepLen, $msresFlags2, $ShowSubSets,
  ) = @_;

  # An index from resfile (ms_run) index to resfile and params is needed
  # for reporting ms_run-specific protein scores (which is a required column).
  my @ms_run_resfiles = (
      # combined resfile at index 0
      { resfile => $objResFile, summary => $objSummary, params => $objParams } 
  );

  for my $i (1 .. @$fileIn) {
    my $resfile = $objResFile;
    my $params = $objParams;
    my $summary = $objSummary;

    if (@$fileIn > 1) {
      # If there is more than one ms_run, we must open a summary for each one.
      # It's possible that some of the individual cache files do not exist, but
      # as this is a rather rare use case, we'll ignore the problem and simply
      # suppress keepalive to prevent output corruption.

      my $resfile = $objResFile->getResfile($i);
      $params = $resfile->params;
      $objResFile->resetKeepAlive(0, '', 1);

      if ($reportType eq 'concise') {
        $summary = new msparser::ms_proteinsummary(
          $resfile, 
          $msresFlags, 
          $sigThreshold,
          $numHits
        );
      } else {
        $summary = new msparser::ms_peptidesummary(
          $resfile, 
          $msresFlags,
          $sigThreshold, 
          $numHits, 
          $UniGeneFile,
          $ignoreIonsScoreBelow,
          $minPepLen,
          "",
          $msresFlags2
        );
        $summary->setSubsetsThreshold($ShowSubSets);
      }
    }

    $ms_run_resfiles[$i] = { 
        filepath => $$fileIn[$i-1],
        resfile => $resfile, 
        params => $params,
        summary => $summary,
    };
  }

  if (not $objQuantMethod or not $objMS2Quant) {
    # There is no need to report assay information in Identification-only
    # files.
    return (\@ms_run_resfiles, [], []);
  }

  # The relationship between samples, runs and assays is somewhat subtle.
  # If there is more than one results file (combined export), we assume each
  # file is an ms_run and a technical replicate. This is the most common
  # case for merging files and if it is not the case, we would need more
  # input from the user to determine the relationships. Further, we know
  # the quantitation method component definitions must be the same in all
  # of the files. Thus, if the quantitation method contains C components, we
  # assume there are C biological samples. If there are K files, we assume
  # there are K technical replicates of each sample, making K*C assays.
  #
  # In both protocols, with K input files, C channels and R ratios, we have
  # (for 1 <= k <= K and 1 <= c <= C and 1 <= r <= R)
  #
  #    assay[c+C*(k-1)]-sample_ref   = sample[c]
  #    assay[c+C*(k-1)]-ms_run_ref   = ms_run[k]
  #    study_variable[r]-ms_run_refs = ms_run[1], ..., ms_run[K]
  #    study_variable[r]-assay_refs  = assay[c+C(*k-1)] for some c's and all k
  #
  # For example, iTRAQ 4-plex with ratios 115/114, 116/114 and 117/114
  # and one file has (K = 1, C = 4, R = 3)
  #
  #   assay[c]-sample_ref           = sample[c] for c = 1,2,3,4
  #   assay[c]-ms_run_ref           = ms_run[1] for c = 1,2,3,4
  #   study_variable[r]-ms_run_refs = ms_run[1] for r = 1,2,3
  #   study_variable[1]-assay_refs  = assay[1], assay[2]
  #   study_variable[2]-assay_refs  = assay[1], assay[3]
  #   study_variable[3]-assay_refs  = assay[1], assay[4]
  #
  # In 18O Multiplex with two files, two heavy and one light component, and
  # quantitation ratio heavy/light = (18O1 + 18O2)/18O0, the definitions
  # are (K = 2, C = 3, R = 1)
  #
  #   assay[c]-sample_ref           = sample[1] for c = 1,4
  #   assay[c]-sample_ref           = sample[2] for c = 2,5
  #   assay[c]-sample_ref           = sample[3] for c = 3,6
  #   assay[c]-ms_run_ref           = ms_run[1] for c = 1,2,3
  #   assay[c]-ms_run_ref           = ms_run[2] for c = 4,5,6
  #   study_variable[1]-ms_run_refs = ms_run[1], ms_run[2]
  #   study_variable[1]-assay_refs  = assay[1], assay[2], assay[3], assay[4], assay[5], assay[6]
  #
  # Here assay[1] and assay[4] are 18O0, assay[2] and assay[5] are 18O1
  # and assay[3] and assay[6] are 18O2.
  #
  # Note 1: mzTab 1.0 does not provide any mechanism for defining the ratio
  # formula. We export the ratio name in the free-text field
  # study_variable[]-description and the peptide ratio formula in a custom
  # field. For example, iTRAQ 4-plex above has
  #
  #   MTD study_variable[1]-description  115/114
  #   MTD custom [,,study_variable[1]-peptide_ratio_formula, assay[2]/assay[1]]
  #
  # where the ratio name defined in the quantitation method is in quotes.
  # The two-file 18O Multiplex above has
  #
  #   MTD study_variable[1]-description  heavy/light
  #   MTD custom [,,study-variable[1]-peptide_ratio_formula, (assay[2] + assay[5] + assay[3] + assay[6])/(assay[1] + assay[4])]
  # 
  # because both assay[2] and assay[5] are 18O1 etc.
  #
  # For practical reasons, when we export a Peptides section, we treat peptides
  # as if they were PSMs. This means a peptide only ever has component
  # intensities from one ms_run, so in the above formula,
  #   either assay[4] = assay[5] = assay[6] = null 
  #   or     assay[1] = assay[2] = assay[3] = null.
  # Both simplify the formula to its usual form.
  #
  # Note 2: protein ratios are complicated functions of peptide ratios,
  # e.g. a geometric mean. The mzTab 1.0 specification does not provide
  # any way to describe how protein ratios are calculated. Instead, we
  # export a custom metadata field, e.g.:
  #
  #   MTD custom [,,study-variable[1]-protein_ratio_type, median]

  my @comp_names = ();

  for my $i (0 .. $objQuantMethod->getNumberOfComponents()-1) {
    my $comp = $objQuantMethod->getComponentByNumber($i);
    my $comp_name = '';
    if ($comp->haveName()) {
      $comp_name = $comp->getName();
    }
    push @comp_names, $comp_name;
  }

  my @comp_idx_to_assay = ();
  my %comp_name_to_assay = ();
  my $num_assays = 0;

  for my $i (1 .. @$fileIn) {
    for my $j (1 .. @comp_names) {
      $comp_name_to_assay{$comp_names[$j-1]}{$i} 
        = $comp_idx_to_assay[$i][$j] 
        = ++$num_assays;
    }
  }

  my @ratio_names = @{ $objMS2Quant->getCombinedReportRatioNames() };
  my @study_variable_defs = ();

  for my $i (1 .. @ratio_names) {
    my $ratio_def = $objMS2Quant->getCombinedReportRatio($ratio_names[$i-1]);
    my $ratio_name = '';
    if ($ratio_def->haveName()) {
      $ratio_name = $ratio_def->getName();
    }

    my %assay_indices = ();
    my @numerators = ();
    my @denominators = ();
    for my $j (0 .. $ratio_def->getNumberOfNumerators()-1) {
      my $numerator = $ratio_def->getNumeratorByNumber($j);
      next if not $numerator->haveName();
      for my $k (1 .. @$fileIn) {
        my $assay_idx = $comp_name_to_assay{$numerator->getName()}{$k};
        next if not defined $assay_idx;
        push @numerators, sprintf('assay[%d]', $assay_idx);
        $assay_indices{$assay_idx} = 1;
      }
    }
    for my $j (0 .. $ratio_def->getNumberOfDenominators()-1) {
      my $denominator = $ratio_def->getDenominatorByNumber($j);
      next if not $denominator->haveName();
      for my $k (1 .. @$fileIn) {
        my $assay_idx = $comp_name_to_assay{$denominator->getName()}{$k};
        next if not defined $assay_idx;
        push @denominators, sprintf('assay[%d]', $assay_idx);
        $assay_indices{$assay_idx} = 1;
      }
    }
    my $numerator = '1';
    if (@numerators == 0) {
      # Keep '1'
    } elsif (@numerators == 1) {
      $numerator = $numerators[0];
    } else {
      $numerator = sprintf('(%s)', join(' + ', @numerators));
    }
    my $denominator = '';
    if (@denominators == 0) {
      # Keep ''
    } elsif (@denominators == 1) {
      $denominator = sprintf('/%s', $denominators[0]);
    } else {
      $denominator = sprintf('/(%s)', join(' + ', @denominators));
    }

    push @study_variable_defs, {
        idx => $i, 
        ratio_name => $ratio_name,
        description => $ratio_name, 
        formula => $numerator.$denominator,
        assay_refs => join(', ', map { sprintf('assay[%d]', $_) } sort { $a <=> $b } keys %assay_indices),
    };
  }

  return (
      \@ms_run_resfiles,
      \@comp_idx_to_assay,
      \@study_variable_defs,
  );
}

sub str_or_null {
    my $str = defined($_[0]) ? $_[0] : ''; 
    $str =~ s/\s/ /sg; 
    return $str =~ m#^\s*$# ? 'null' : $str;
}
sub printTabbed { print join("\t", map { str_or_null($_) } @_), "\n" }
sub printMTD { printTabbed('MTD', $_[0], join('', @_[1..$#_])) }

sub _fmtdate_iso8601 {
  my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($_[0]);
  return sprintf('%4d-%02d-%02dT%02d:%02d:%02d',  $year+1900, $mon+1, $mday, $hour, $min, $sec);
}

# We print metadata headers in logical clumps, as apart from a few exceptions,
# the mzTab 1.0 specification does not require a fixed order. Some headers
# are not output, because they are not relevant here, e.g. sample processing
# details, publication, tissue, cell_type and disease.
sub printMzTabMTD {
  my (
    $thisScript, $URI, $urlParams, $ms_run_resfiles, $comp_idx_to_assay,
    $study_variable_defs, $objSummary, $objQuantMethod, $objMS2Quant,
    $objUmodConfigFile2, $indexIsQueryNumber, $reportType, $protTaxStr,
    $taxNames, $pepDelta, $quantWarnings, $mod_tally, $any_retention_time,
    $showPRH, $showPEH, $protQuant, $pepQuant, $percolate
  ) = @_;

  my $MISSING = 'null';

  my @custom = ();

  printMTD('mzTab-version', '1.0');
  printMTD('mzTab-mode', 'Complete');

  if ($objQuantMethod) {
    printMTD('mzTab-type', 'Quantification');
  } else {
    printMTD('mzTab-type', 'Identification');
  }

  for my $i (1 .. $#$ms_run_resfiles) {
    my $run_date = _fmtdate_iso8601($$ms_run_resfiles[$i]{'resfile'}->getDate());
    push @custom, sprintf('[,,Search date[%d], %s]', $i, $run_date);
  }

  do {
    my $export_date = _fmtdate_iso8601(time);
    push @custom, sprintf('[,,Export date, %s]', $export_date);
  };

  my $NEWT_vocabulary_used = 0;

  # Each input file $i becomes ms_run[$i] and sample[$i]. We can't know,
  # without more information from the user, whether the input files refer
  # to the same sample (technical replicates) or a different sample.
  # The mzTab 1.0 specification deliberately leaves the definition of
  # sample open to interpretation, so it is not actually incorrect to treat
  # technical replicates as different samples.
  do {
    my %file_formats = (
        'Mascot generic' => { param => 'MS:1001062', name => 'Mascot MGF format' },
        'Sequest (.DTA)' => { param => 'MS:1001063', name => 'DTA format' },
        #'Finnigan (.ASC)' => { param => '99999', name => 'TODO: Finnigan List file' }, # see mzIdentML output
        'Micromass (.PKL)' => { param => 'MS:1000565', name => 'Micromass PKL format' },
        'PerSeptive (.PKS)' => { param => 'MS:1001245', name => 'PerSeptive PKS format' },
        'Sciex API III' => { param => 'MS:1001246', name => 'Sciex API III format' },
        'Bruker (.XML)' => { param => 'MS:1001247', name => 'Bruker XML format' },
        'mzData (.XML)' => { param => 'MS:1000564', name => 'PSI mzData format' },
        'mzML (.mzML)' => { param => 'MS:1000584', name => 'mzML format' },
    );

    my @com = ();
    my @contact_name = ();
    my @contact_email = ();
    my @instrument = ();
    my @species_header = ();
    my @species_excluded = ();
    my @filename = ();
    my @file_format = ();
    my @id_format = ();

    for my $i (1 .. $#$ms_run_resfiles) {
      my $resfile = $$ms_run_resfiles[$i]{'resfile'};
      my $params = $$ms_run_resfiles[$i]{'params'};

      if ($params->getCOM ne '') {
        push @com, $params->getCOM
      } else {
        push @com, 'untitled search';
      }

      my $has_contact = 0;
    
      if ($params->getUSERNAME) {
        $contact_name[$i] = $params->getUSERNAME;
      }
  
      # (contact-affiliation[] omitted)
      # (publication[] omitted)
    
      if ($params->getUSEREMAIL) {
        $contact_email[$i] = $params->getUSEREMAIL;
      }

      if ($params->getINSTRUMENT) {
        $instrument[$i] = $params->getINSTRUMENT;
      }

      my $taxonomy = new msparser::ms_taxonomyfile;
      my @header = ();
      my @excluded = ();

      if ($resfile->getTaxonomy($taxonomy)) {
        if ($taxonomy->getNumberOfEntries == 0 ) {
          push @header, { id => 1, name => 'all' };
        } else {
          my $choice = $taxonomy->getEntryByNumber(0);
          if ($choice->getNumberOfIncludeTaxonomies > 0) {
            for my $i (0 .. $choice->getNumberOfIncludeTaxonomies()-1) {
              my $id = $choice->getIncludeTaxonomy($i);
              my $tax_title = getTaxonomyNameFromId($id, $taxNames);
              push @header, { id => $id, name => $tax_title };
            }
          }
        
          if ($choice->getNumberOfExcludeTaxonomies > 0) {
            # No way to specify excluded taxa in the standard species[k] header,
            # so we need to print a custom header.
            for my $i (0 .. $choice->getNumberOfExcludeTaxonomies()-1) {
              my $id = $choice->getExcludeTaxonomy($i);
              my $tax_title = getTaxonomyNameFromId($id, $taxNames);
              push @excluded, { id => $id, name => $tax_title };
            }
          }
        }
      }
  
      $species_header[$i] = \@header;
      $species_excluded[$i] = \@excluded;

      if (my $filename = $params->getFILENAME()) {
        $filename =~ s/\\/\//g; 
        $filename[$i] = 'file:///' . $filename;
      } elsif (my $url = $params->getDATAURL()) {
        $filename[$i] = $url;
      } else {
        # ms_run[]-location is mandatory, so if it's not present, we must
        # report it as missing.
        $filename[$i] = $MISSING;
      }

      if (my $cv = $file_formats{$params->getFORMAT}) {
        $file_format[$i] = sprintf('[MS, %s, %s, ]', $$cv{'param'}, $$cv{'name'});
      } elsif ($resfile->isPMF or $resfile->isSQ) {
        $file_format[$i] = sprintf('[MS, %s, %s, ]', 'MS:1001369', 'text format');
      } else {
        # A bit nonsensical, but we really don't know what the format is.
        $file_format[$i] = sprintf('[MS, %s, %s, ]', 'MS:1000560', 'mass spectrometer file format');
      }
    
      if ($resfile->isPMF or $resfile->isSQ) {
        $id_format[$i] = '[MS, MS:1000775, single peak list nativeID format, ]';
      } elsif ($indexIsQueryNumber) {
        $id_format[$i] = '[MS, MS:1001528, Mascot query number, ]';
      } elsif ($params->getFORMAT eq "mzML (.mzML)") {
        $id_format[$i] = '[MS, MS:1001530, mzML unique identifier, ]';
      } elsif ($params->getFORMAT eq "mzData (.XML)") {
        $id_format[$i] = '[MS, MS:1000777, spectrum identifier nativeID format, ]';
      } else {
        $id_format[$i] = '[MS, MS:1000774, multiple peak list nativeID format, ]';
      }
  
      # (ms_run[]-fragmentation_method omitted)
    }

    # The MTD section offers both description and title. Description is
    # mandatory, while title is not, so we'll put the search title in the
    # description field.
    if (@com == 1) {
      printMTD('description', $com[0]);
    } else {
      printMTD('description', 
          sprintf('Combined export of %d files: %s', scalar(@com), joinand(@com))
      );
    }

    for my $i (1 .. $#$ms_run_resfiles) {
      if (defined $contact_name[$i]) {
        printMTD('contact['.$i.']-name', $contact_name[$i]);
      }
      if (defined $contact_email[$i]) {
        printMTD('contact['.$i.']-email', $contact_email[$i]);
      }
    }

    for my $i (1 .. $#$ms_run_resfiles) {
      if (defined $instrument[$i]) {
        printMTD('instrument['.$i.']-analyzer', sprintf('[,,%s,]', $instrument[$i]));
      }
    }

    for my $i (1 .. $#$ms_run_resfiles) {
      my $run_str = "ms_run[$i]";
      printMTD($run_str.'-location', $filename[$i]);
      printMTD($run_str.'-format', $file_format[$i]);
      printMTD($run_str.'-id_format', $id_format[$i]);
    }

    # sample[]-species[]

    for my $i (1 .. $#$ms_run_resfiles) {
      my $excluded = $species_excluded[$i];

      for my $j (1 .. $#$excluded) {
        my $s = $$excluded[$j];
        push @custom, sprintf(
          '[,, sample[%d]-species_excluded[%d], [NEWT, %d, %s, ]]', 
          $i, $j, $$s{'id'}, defined($$s{'name'}) ? $$s{'name'} : ''
        );
        ++$NEWT_vocabulary_used;
      }

      my $header = $species_header[$i];
    
      for my $j (1 .. $#$header) {
        my $s = $$header[$j];
        # NEWT = NCBI taxonomy IDs.
        printMTD(sprintf('sample[%d]-species[%d]', $i, $j), 
            sprintf('[NEWT, %d, %s, ]', 
                $$s{'id'}, defined($$s{'name'}) ? $$s{'name'} : ''
            )
        );
        ++$NEWT_vocabulary_used;
      }
    }
  };

  # (mzTab-ID omitted)
  # (title omitted)
  # (sample_processing[] omitted)
  # (sample[]-tissue[] omitted)
  # (sample[]-cell_type[] omitted)
  # (sample[]-disease[] omitted)
  # (sample[]-description omitted)
  # (sample[]-custom[] omitted)
  
  #printMTD('contact[2]-name', 'Matrix Science');
  #printMTD('contact[2]-affiliation', '64 Baker Street, London W1U 7GB, UK');
  #printMTD('contact[2]-email', 'support@matrixscience.com');

  # (uri[] omitted; it's the URI to the source data, not to the Mascot report.)

  do {
    my $i = 1;

    # XXX should be a list of versions for combined files
    printMTD('software['.$i.']', '[MS, MS:1001207, Mascot, ', $$ms_run_resfiles[0]{'resfile'}->getMascotVer(), ']');

    # However, all the important parameters between the combined results files
    # must be the same, so it's fine to print out only one set of parameters.
    printMTDsearchParameters('software['.$i.']-setting', $$ms_run_resfiles[0]{'params'});

    ++$i;

    if ($percolate == 1) {
      printMTD('software['.$i.']', '[MS, MS:1001490, Percolator, ', ']');
      # The Percolator command line used for this results file is not saved
      # anywhere. We could report it as it appears in mascot.dat, but that
      # could have changed since.
      #printMTDpercolatorParameters('software['.$i.']-setting', );
      ++$i;
    }

    printMTD('software['.$i.']', '[MS, MS:1001478, Mascot Parser, ', $$ms_run_resfiles[0]{'resfile'}->getMSParserVersion(), ' (export_dat_2.pl)]');
    printMTDformatParameters('software['.$i.']-setting', $thisScript, $urlParams);
  };

  for (['fixed', 'fixed_mod', '[MS, MS:1002453, No fixed modifications searched,]'],
       ['variable', 'variable_mod', '[MS, MS:1002454, No variable modifications searched,]']) {
    my ($tally_type, $param_name, $cv_empty) = @$_;
    my $i = 1;
  
    for my $id (sort keys %{ $$mod_tally{$tally_type} || {} }) {
      my $mod = $$mod_tally{$tally_type}{$id};
      my $mtd = sprintf('%s[%d]', $param_name, $i);
      printMTD($mtd, $id);
      if (defined $$mod{'position'}) {
        printMTD($mtd.'-position', $$mod{'position'});
      }
      if (defined $$mod{'site'}) {
        printMTD($mtd.'-site', $$mod{'site'});
      }
      ++$i;
    }

    if ($i == 1) {
      # No mods, so report it.
      printMTD(sprintf("%s[1]", $param_name), $cv_empty);
    }
  }

  if ($objQuantMethod) {
    my $protocol = lc($objQuantMethod->getProtocol()->getType());

    # The mzTab 1.0 specification recommends outputting quantitation methods 
    # as CV parameters (e.g. [MS, MS:1001837, iTRAQ quantitation analysis]),
    # but we can't infer it from the Mascot quantitation method -- the latter
    # is far more generic. The best we can do is output a user parameter.
    printMTD('quantification_method', 
        sprintf('[,,"%s", "%s"]',
            $objQuantMethod->getName(), $objQuantMethod->getDescription()
        )
    );

    if ($showPRH and $protQuant) {
      printMTD('protein-quantification_unit', '[MS, MS:1001134, protein ratio, ]');
    }

    if ($showPEH and $pepQuant) {
      printMTD('peptide-quantification_unit', '[MS, MS:1001132, peptide ratio, ]');
    }

    my ($method_unimods, $method_chemmods) = _collect_mod_types($objQuantMethod, $objUmodConfigFile2);
    my @comp_mod_cache = ();

    my @assay_structure = ();
    my @assay_defs = ();

    for my $k (1 .. $#$ms_run_resfiles) {
      for my $i (1 .. $objQuantMethod->getNumberOfComponents()) {
        my $component = $objQuantMethod->getComponentByNumber($i-1);
        my $comp_name = '';
        if ($component->haveName()) {
          $comp_name = $component->getName();
        }

        my $assay_id = sprintf('assay[%d]', $$comp_idx_to_assay[$k][$i]);
  
        # As above, it's impossible to infer the right CV terms for
        # quantification_reagent. We have to use a user parameter.
        push @assay_defs, [$assay_id.'-quantification_reagent', '[,,'.$component->getName().',]'];
  
        my @mods = ();
        
        if ($comp_mod_cache[$i]) {
          @mods = @{ $comp_mod_cache[$i] };
        } else {
          @mods = _determine_component_quant_mods(
            $protocol, $method_unimods, $method_chemmods, $component,
            $objUmodConfigFile2
          );
          # Cache just so that we don't need to re-determine as many times
          # as there are resfiles.
          $comp_mod_cache[$i] = \@mods;
        }

        for my $j (1 .. @mods) {
          my $mod = $mods[$j-1];
          push @assay_defs, [$assay_id.'-quantification_mod['.$j.']', $$mod{'cv'}];
          if (defined $$mod{'position'}) {
            push @assay_defs, [$assay_id.'-quantification_mod['.$j.']-position', $$mod{'position'}];
          }
          if (defined $$mod{'site'}) {
            push @assay_defs, [$assay_id.'-quantification_mod['.$j.']-site', $$mod{'site'}];
          }
        }

        # At least one field should exist for each sample referred to here,
        # whether it is sample[i]-description or sample[i]-species[].
        my $sample_desc;
        if ($#$ms_run_resfiles > 1) {
          $sample_desc = sprintf('Component %d in file %d', $i, $k);
        } else {
          $sample_desc = sprintf('Component %d', $i);
        }
        push @assay_structure, [sprintf('sample[%d]-description', $i), ''];

        push @assay_structure, [$assay_id.'-sample_ref', sprintf('sample[%d]', $i)];
        push @assay_structure, [$assay_id.'-ms_run_ref', sprintf('ms_run[%d]', $k)];
      }
    }

    printMTD(@$_) for @assay_structure;
    printMTD(@$_) for @assay_defs;

    # Study variables always use information from all resfiles.
    my $sv_ms_run_refs = join(', ', map { sprintf('ms_run[%d]', $_) } 1 .. $#$ms_run_resfiles);
    my $protein_ratio_type = $objQuantMethod->getProteinRatioType();
  
    for my $sv (@$study_variable_defs) {
      my $sv_id = sprintf('study_variable[%d]', $$sv{'idx'});
      printMTD($sv_id.'-description', $$sv{'description'});
      printMTD($sv_id.'-assay_refs', $$sv{'assay_refs'});
      push @custom, sprintf('[,,study-variable[%d]-peptide_ratio_formula, %s]', $$sv{'idx'}, $$sv{'formula'});
      push @custom, sprintf('[,,study-variable[%d]-protein_ratio_type, %s]', $$sv{'idx'}, $protein_ratio_type);
    }
  }

  if ($objParams->getDECOY > 0 
          and not $ShowDecoyReport
          and $reportType ne 'protein'
          and $reportType ne 'concise') {
    my $hits_above_id = $objSummary->getNumHitsAboveIdentity(1 / $sigThreshold);
    my $hits_above_hom = $objSummary->getNumHitsAboveHomology(1 / $sigThreshold);
    my $decoy_above_id = $objSummary->getNumDecoyHitsAboveIdentity(1 / $sigThreshold);
    my $decoy_above_hom = $objSummary->getNumDecoyHitsAboveHomology(1 / $sigThreshold);

    my $FDR = 0;
    if ($use_homology) {
      if ($hits_above_hom) {
        $FDR = $decoy_above_hom / $hits_above_hom;
      }
    } else {
      if ($hits_above_id) {
        $FDR = $decoy_above_id / $hits_above_id;
      }
    }

    if ($FDR > 0) {
      printMTD('false_discovery_rate', sprintf('[MS, MS:1001364, distinct peptide-level global FDR, %.4f]', $FDR));
    }
  }

  if ($showPRH) {
    printMTD('protein_search_engine_score[1]', '[MS, MS:1001171, Mascot:score,]');
  }

  if ($showPEH) {
    if ($objQuantMethod) {
      printMTD('peptide_search_engine_score[1]', '[MS, MS:1001172, Mascot:expectation value,]');
    }
    printMTD('psm_search_engine_score[1]', '[MS, MS:1001172, Mascot:expectation value,]');
  }

  my $UO_vocabulary_used = 0;

  if ($showPEH) {
    if ($any_retention_time) {
      ++$UO_vocabulary_used;
      printMTD('colunit-peptide', sprintf('retention_time=[UO, %s, %s,]', 'UO:0000010', 'second'));
    }
  
    if ($pepDelta) {
      ++$UO_vocabulary_used;
      printMTD('colunit-peptide', sprintf('opt_%s_%s=[UO, %s, %s,]', 'MS:1001976', 'delta_M', 'UO:0000221', 'dalton'));
    }
  }

  do {
    my $i = 1;

    printMTD('cv['.$i.']-label', 'MS');
    printMTD('cv['.$i.']-full_name', 'Proteomics Standards Initiative Mass Spectrometry Vocabularies');
    printMTD('cv['.$i.']-version', '3.64.0');
    printMTD('cv['.$i.']-url', 'http://psidev.cvs.sourceforge.net/viewvc/*checkout*/psidev/psi/psi-ms/mzML/controlledVocabulary/psi-ms.obo');

    ++$i;
  
    printMTD('cv['.$i.']-label', 'UNIMOD');
    printMTD('cv['.$i.']-full_name', 'UNIMOD');
    printMTD('cv['.$i.']-url', 'http://www.unimod.org/obo/unimod.obo');

    ++$i;
  
    if ($UO_vocabulary_used) {
      printMTD('cv['.$i.']-label', 'UO');
      printMTD('cv['.$i.']-full_name', 'UNIT-ONTOLOGY');
      printMTD('cv['.$i.']-url', 'http://obo.cvs.sourceforge.net/*checkout*/obo/obo/ontology/phenotype/unit.obo');

      ++$i;
    }
  
    if ($NEWT_vocabulary_used) {
      printMTD('cv['.$i.']-label', 'NEWT');
      printMTD('cv['.$i.']-full_name', 'NEWT Uniprot Taxonomy Database');
      printMTD('cv['.$i.']-url', 'http://www.ebi.ac.uk/ontology-lookup/browse.do?ontName=NEWT');

      ++$i;
    }
    # NB: There is no OBO definition for CHEMMOD.
  };

  # custom headers
  do {
  # Display any warnings from result file header
    my $warning_idx = 1;
  
    if ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn0')){
      my $j = 0;
      while ($objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j)){
        push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++, 
          $objResFile->getSectionValueStr($msparser::ms_mascotresfile::SEC_HEADER, 'warn' . $j++)
        );
      }
    }
  
  # Display a warning if this is an orphan error tolerant search
    if ($objResFile->isErrorTolerant() and not $objParams->getErrTolParentFilename()) {
      push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++, 
        "Scores from standard search not available. Better matches may exist in other proteins."
      );
    }
  
  # Display any warnings from Parser
    for my $k (1 .. $objResFile->getNumberOfErrors()) {
      push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++, 
        $objResFile->getErrorString($k)
      );
    }
  
  # Display any warnings from quantitation
    if (@$quantWarnings) {
      foreach (@$quantWarnings) {
        push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++, 
          $_
        );
      }
    }
    if ($objMS2Quant) {
      my $objErr = $objMS2Quant->getErrorHandler();
      for my $i (1 .. $objErr->getNumberOfErrors()) {
        push @custom, sprintf('[,,Warning[%d], "%s"]', $warning_idx++, 
          sprintf('%s (%s)', $objErr->getErrorString($i), $objErr->getErrorNumber($i)),
        );
      }
    }
  
    if (my $fileset = $thisScript->param($$urlParams{'fileset'})) {
      push @custom, '[,,Report URI, ' . $URI . 'cgi/master_results_2.pl?fileset=' . $fileset . ']';
    } else {
      push @custom, '[,,Report URI, ' . $URI . 'cgi/master_results_2.pl?file=' . $fileIn[0] . ']';
    }

    my $i = 1;

    for (@custom) {
      printMTD('custom['.($i++).']', $_);
    }
  };

  return ($indexIsQueryNumber);
}

sub _determine_component_quant_mods {
  my ($protocol, $method_unimods, $method_chemmods, $component,
      $objUmodConfigFile2) = @_;

  # Many possible modification types:
  # - iTRAQ has modfile names iTRAQ4plex or iTRAQ8plex.
  # - TMT has modfile names TMT6plex or TMT2plex.
  # - DiLeu has modfile name DiLeu4plex.
  # - 18O has modfile names Label:18O(1) or Label:18O(2).
  # - SILAC K+6 R+6 has local definition 13C6_NL_K and 13C6_NL_R.
  # - IPTL has local definitions IMID_2H(0), Succ_2H(4)_NL or IMID_2H(4)_NL
  #   or modfile name Succinyl.
  #
  # Since the Reporter protocol does not specify reagents, we have to make
  # the following guesses:
  # - If it's a modfile (Unimod) mod, use  
  #     [UNIMOD, UNIMOD:$id, $title,]
  # - If it's a local definition (unlikely), use
  #     [CHEMMOD, CHEMMOD:$formula, $title,]
  # 
  # Since it's theoretically possible (but probably practically meaningless)
  # to mix and match, we export the full union of all found reagents.
  #
  # The Multiplex protocol defines mods (= reagents), which can be unlabelled,
  # labelled with modfile name, or labelled with a local definition.
  # - The standard 18O has no mods in component 18O0, Label:18O(1) in 18O1 and 
  #   Label:18O(2) in 18O2, and no other mods in any component.
  # - The standard SILAC K+6 R+6 method has no mods in light component
  #   and two local definitions on heavy, and no other mods.
  # - The standard IPTL method has two local definitions on n-label and
  #   one modfile and one local definition on c-label, and no other mods.
  #
  # Thus, we assume that all mods in Multiplex components are part of the
  # reagent.
  #
  # There is no standard chemical formula syntax for local definitions in
  # mzTab 1.0. We will use our own: IMID_2H(4)_NL, for example, becomes
  #   CHEMMOD:+(2H)4 -H4
  # where parentheses are used for disambiguating isotopes.

  my @mods = ();

  if ($protocol eq 'reporter') {
    @mods = _make_quant_mod_defs($method_unimods, $method_chemmods);
  } else { # multiplex
    my ($unimod_mods, $chemmods) = _collect_mod_types($component, $objUmodConfigFile2);
    @mods = _make_quant_mod_defs($unimod_mods, $chemmods);
  }

  return @mods;
}

sub _make_quant_mod_defs {
  my ($unimod_mods, $chemmods) = @_;

  my @mods = ();
  
  for (sort { $a <=> $b } keys %$unimod_mods) {
    my $mod = $$unimod_mods{$_};
    push @mods, {
      %$mod,
      cv => sprintf('[UNIMOD, UNIMOD:%d, %s,]', $_, $$mod{'title'}),
    };
  }
  
  for (sort keys %$chemmods) {
    my $mod = $$chemmods{$_};
    push @mods, {
      %$mod,
      cv => sprintf('[CHEMMOD, CHEMMOD:%s, %s,]', $_, $$mod{'title'}),
    };
  }

  return @mods;
}


sub _collect_mod_types {
  my ($obj, $ms_umod_configfile) = @_;

  my %unimod_mods = ();
  my %chemmods = ();

  for my $i (0 .. $obj->getNumberOfModificationGroups()-1) {
    my $modgroup = $obj->getModificationGroupByNumber($i);

    for my $j (0 .. $modgroup->getNumberOfModFiles()-1) {
      my $modfile_name = $modgroup->getModFile($j);
      my ($mod_name) = $modfile_name =~ /(.*)\s+\((?:.*)\)/;
      my $mod = $ms_umod_configfile->getModificationByName($mod_name);
      next if not $mod;
      my $id = $mod->getRecordID();
      next if $unimod_mods{$id};
      my ($position, $site) = _unimod_position_and_site($mod, $modfile_name);
      $unimod_mods{$id} = { 
        title => $mod->getTitle(),
        position => $position,
        site => $site,
      };
    }

    for my $j (0 .. $modgroup->getNumberOfLocalDefinitions()-1) {
      my $def = $modgroup->getLocalDefinition($j);
      my $comp = $def->getDelta();
      my (@plus, @minus);
      for my $k (0 .. $comp->getNumberOfElements()-1) {
        my $elt = $comp->getElement($k);
        next if not $elt->haveSymbol();
        my $symbol = $elt->getSymbol();
        my $num = 1;
        if ($elt->haveNumber()) {
          $num = 0 + $elt->getNumber();
        }
        my $arr = $num > 0 ? \@plus : \@minus;
        if ($symbol =~ /\d/) {
          push @$arr, sprintf('(%s)%d', $symbol, abs($num));
        } else {
          push @$arr, sprintf('%s%d', $symbol, abs($num));
        }
      }
      next if (not @plus and not @minus);
      if ($def->getNumberOfSpecificities() == 0) {
        $chemmods{join(' ', join('', @plus), join('', @minus))} = {
          title => $def->getTitle(),
        };
      } else {
        # We can only export the first specificity...
        my $spec = $def->getSpecificity(0);
        $chemmods{join(' ', join('', @plus), join('', @minus))} = {
          title => $def->getTitle(),
          position => $spec->havePosition() ? $spec->getPosition() : undef,
          site => $spec->haveSite() ? $spec->getSite() : undef,
        };
      }
    }
  }

  return (\%unimod_mods, \%chemmods);
}

sub printMzTabPRHandPRT {
  my ($protein_rows, $isPMF, $objMS2Quant, $protQuant, $num_ms_runs,
      $num_study_variables, $protEmPAI, $protpI, $reportType, $protein_mods,
      $database_map) = @_;

  # Order of columns matters.
  my @column_data_names_pre = qw(
    accession
    description
    taxid
    species
    database
    database_version
    search_engine
    best_search_engine_score[1]
    ambiguity_members
    modifications
    protein_coverage
  );

  if ($protQuant and $objMS2Quant) {
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('protein_abundance_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('protein_abundance_stdev_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('protein_abundance_std_error_study_variable[%d]', $i);
    }
  }

  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('search_engine_score[1]_ms_run[%d]', $i);
  }
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('num_psms_ms_run[%d]', $i);
  }
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('num_peptides_distinct_ms_run[%d]', $i);
  }
  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('num_peptides_unique_ms_run[%d]', $i);
  }

  if ($protQuant and $objMS2Quant) {
    my $n = $objMS2Quant->getQuantitationMethod()->getNumberOfComponents();
    for my $i (1 .. $n*$num_ms_runs) {
      push @column_data_names_pre, sprintf('protein_abundance_assay[%d]', $i);
    }
  }

  my @MS_available = ();

  push @MS_available, ['MS:1001301', 'protein rank'];

  if (not $isPMF) {
      push @MS_available, ['MS:1001592', 'family member protein'];
  }

  if ($protEmPAI) {
    push @MS_available, ['MS:1001905', 'emPAI value'];
  }

  if ($protpI) {
    push @MS_available, ['MS:1000862', 'isoelectric point'];
  }

  if ($reportType eq 'concise') {
    push @MS_available, ['MS:1001172', 'Mascot:expectation value'];
  }

  my @column_names = @column_data_names_pre;
  my @opt_data_names = ();

  for my $i (1 .. $num_study_variables) {
    for (qw(_sample_size _is_normal _is_significant)) {
      push @column_names, sprintf('opt_study_variable[%d]%s', $i, $_);
      push @opt_data_names, sprintf('study_variable[%d]%s', $i, $_);
    }
  }

  for (@MS_available) {
    my ($ms, $name) = @$_;
    $name =~ s/\s/_/g;
    push @column_names, sprintf('opt_global_cv_%s_%s', $ms, $name);
    push @opt_data_names, $ms;
  }

  if ($objParams->getNumberOfDatabases() != 1) {
    push @column_names, 'opt_global_ambiguity_member_databases';
    push @opt_data_names, 'ambiguity_member_databases';
  }

  # Optional columns all come before reliability.
  my @column_data_names_post = ();

  push @column_data_names_post, qw(
    reliability
  );
  push @column_names, qw(
    reliability
  );

  printTabbed('PRH', @column_names);

  for my $row (@$protein_rows) {
    my @data = ();

    my $db = $$row{'database'};
    my $db_entry = $$database_map[$db];
    $$row{'database'} = $$db_entry{'full_name'};
    $$row{'database_version'} = $$db_entry{'version'};

    $$row{'search_engine'} = '[MS, MS:1001207, Mascot, ]';

    my $mods = $$protein_mods{$db}{$$row{'accession'}};

    if ($mods) {
      my %m = ();

      for my $idx (keys %$mods) {
        for (@{ $$mods{$idx} }) {
          my $id = sprintf('%s%s', $$_{'mod_id'},
            defined($$_{'neutral_loss'}) ? '|'.$$_{'neutral_loss'} : ''
          );
          $m{$idx}{$id}++;
        }
      }

      my @m = ();

      # Ambiguity in modification position is not allowed, but ambiguity in
      # modification at a specified position is. Since there is no ambiguity
      # in position at peptide level, there is no ambiguity at protein level.

      for my $idx (sort { $a <=> $b } keys %m) {
        push @m, map { sprintf('%d-%s', $idx, $_) } sort keys %{ $m{$idx} };
      }

      if (@m) {
        $$row{'modifications'} = join(',', @m);
      } else {
        $$row{'modifications'} = '0';
      }
    }

    my (@amb_acc, @amb_db);

    for (sort { $$a{'accession'} cmp $$b{'accession'} } @{ $$row{'ambiguity_members'} }) {
      push @amb_acc, $$_{'accession'};
      my $db_entry = $$database_map[$$_{'db'}];
      push @amb_db, $$db_entry{'abbr_name'};
    }

    $$row{'ambiguity_members'} = join(',', @amb_acc);
    $$row{'opt'}{'ambiguity_member_databases'} = join(',', @amb_db);

    for (@column_data_names_pre) {
      push @data, $$row{$_};
    }

    for (@opt_data_names) {
      push @data, $$row{'opt'}{$_};
    }

    for (@column_data_names_post) {
      push @data, $$row{$_};
    }

    printTabbed('PRT', @data);
  }
}

sub printMzTabPEHandPEP {
  my ($peptide_rows, $objMS2Quant, $num_ms_runs, $num_study_variables,
      $is_decoy, $is_error_tolerant, $percolate, $database_map, $pepDelta,
      $pepScore, $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch, $pepScanTitle,
      $pepQuant) = @_;

  # Order of columns matters.
  my @column_data_names_pre = qw(
    sequence
    accession
    unique
    database
    database_version
    search_engine
    best_search_engine_score[1]
    modifications
    retention_time
    retention_time_window
    charge
    mass_to_charge
  );

  if ($pepQuant and $objMS2Quant) {
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('peptide_abundance_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('peptide_abundance_stdev_study_variable[%d]', $i);
    }
    for my $i (1 .. $num_study_variables) {
      push @column_data_names_pre, sprintf('peptide_abundance_std_error_study_variable[%d]', $i);
    }
  }

  for my $i (1 .. $num_ms_runs) {
    push @column_data_names_pre, sprintf('search_engine_score[1]_ms_run[%d]', $i);
  }

  if ($pepQuant and $objMS2Quant) {
    my $n = $objMS2Quant->getQuantitationMethod()->getNumberOfComponents();
    for my $i (1 .. $n*$num_ms_runs) {
      push @column_data_names_pre, sprintf('peptide_abundance_assay[%d]', $i);
    }
  }

  push @column_data_names_pre, qw(
    spectra_ref
  );

  my @column_names = @column_data_names_pre;
  my @opt_data_names = ();

  my @opt_available = ();

  push @opt_available, ['MS:1001528', 'Mascot query number'];

  if ($percolate == 1) {
    push @opt_available, (
      ['MS:1001491', 'percolator:Q value'],
      ['MS:1001493', 'percolator:PEP'],
    );
  }

  if ($is_decoy) {
    push @opt_available, ['MS:1002217', 'decoy peptide'];
  } elsif ($is_error_tolerant) {
    push @opt_available, ['MS:1002341', 'second-pass peptide identification'];
  }

  if ($pepDelta) {
    push @opt_available, ['MS:1001976', 'delta M'];
  }

  if ($pepScore) {
    push @opt_available, ['MS:1001171', 'Mascot:score'];
  }

  if ($pepIdThr) {
    push @opt_available, ['MS:1001371', 'Mascot:identity threshold'];
  }

  if ($pepHomThr) {
    push @opt_available, ['MS:1001370', 'Mascot:homology threshold'];
  }

  if ($pepFrame) {
    # NB: This term was obsoleted in PSI-MS CV for a while but became
    # current again in 3.54.0 due to our request.
    push @opt_available, ['MS:1001024', 'translation frame'];
  }

  if ($pepNumMatch) {
    push @opt_available, ['MS:1001121', 'number of matched peaks'];
  }

  if ($pepScanTitle) {
    push @opt_available, ['MS:1000796', 'spectrum title'];
  }

  for (@opt_available) {
    my ($ms, $name) = @$_;
    $name =~ s/\s/_/g;
    push @column_names, sprintf('opt_global_cv_%s_%s', $ms, $name);
    push @opt_data_names, $ms;
  }

  # Reliability comes after all optional columns.
  my @column_data_names_post = ();

  push @column_data_names_post, qw(
    reliability
  );
  push @column_names, qw(
    reliability
  );

  printTabbed('PEH', @column_names);

  for my $row (@$peptide_rows) {
    my @data = ();
    my %row_data = %$row;

    my $db = $row_data{'database'};
    my $db_entry = $$database_map[$db];
    $row_data{'database'} = $$db_entry{'full_name'};
    $row_data{'database_version'} = $$db_entry{'version'};
    $row_data{'search_engine'} = '[MS, MS:1001207, Mascot, ]';
    if (defined $row_data{'retention_time'}) {
      $row_data{'retention_time_window'} = join('|', $row_data{'retention_time'}, $row_data{'retention_time'});
    }
    $row_data{'mass_to_charge'} = $row_data{'exp_mass_to_charge'};

    # For peptides, the modifications field must contain only variable
    # modifications.
    $row_data{'modifications'} = _format_mztab_peptide_modifications(@{ $row_data{'.varmods'} });

    for (@column_data_names_pre) {
      push @data, $row_data{$_};
    }

    for (@opt_data_names) {
      push @data, $row_data{'opt'}{$_};
    }

    for (@column_data_names_post) {
      push @data, $$row{$_};
    }

    printTabbed('PEP', @data);
  }
}

sub _format_mztab_peptide_modifications {
  my @modifications = @_;

  my @ids = ();

  for (sort { $$a{'idx'} <=> $$b{'idx'} } @modifications) {
    my $id = sprintf('%d-%s', $$_{'idx'}, $$_{'mod_id'});
    if (defined $$_{'neutral_loss'}) {
      $id .= '|' . $$_{'neutral_loss'};
    }
    push @ids, $id;
  }

  if (not @ids) {
    return '0';
  }

  return join(',', @ids);
}

sub printMzTabPSHandPSM {
  my ($peptide_rows, $is_decoy, $is_error_tolerant, $percolate, $database_map,
      $pepDelta, $pepScore, $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch,
      $pepScanTitle) = @_;

  # Order of columns matters.
  my @column_data_names_pre = qw(
    sequence
    PSM_ID
    accession
    unique
    database
    database_version
    search_engine
    search_engine_score[1]
    modifications
    spectra_ref
    retention_time
    charge
    exp_mass_to_charge
    calc_mass_to_charge
    pre
    post
    start
    end
  );

  my @column_names = @column_data_names_pre;
  my @opt_data_names = ();

  my @opt_available = ();

  push @opt_available, ['MS:1001528', 'Mascot query number'];

  if ($percolate == 1) {
    push @opt_available, (
      ['MS:1001491', 'percolator:Q value'],
      ['MS:1001493', 'percolator:PEP'],
    );
  }

  if ($is_decoy) {
    push @opt_available, ['MS:1002217', 'decoy peptide'];
  } elsif ($is_error_tolerant) {
    push @opt_available, ['MS:1002341', 'second-pass peptide identification'];
  }

  if ($pepDelta) {
    push @opt_available, ['MS:1001976', 'delta M'];
  }

  if ($pepScore) {
    push @opt_available, ['MS:1001171', 'Mascot:score'];
  }

  if ($pepIdThr) {
    push @opt_available, ['MS:1001371', 'Mascot:identity threshold'];
  }

  if ($pepHomThr) {
    push @opt_available, ['MS:1001370', 'Mascot:homology threshold'];
  }

  if ($pepFrame) {
    # NB: This term was obsoleted in PSI-MS CV for a while but became
    # current again in 3.54.0 due to our request.
    push @opt_available, ['MS:1001024', 'translation frame'];
  }

  if ($pepNumMatch) {
    push @opt_available, ['MS:1001121', 'number of matched peaks'];
  }

  if ($pepScanTitle) {
    push @opt_available, ['MS:1000796', 'spectrum title'];
  }

  for (@opt_available) {
    my ($ms, $name) = @$_;
    $name =~ s/\s/_/g;
    push @column_names, sprintf('opt_global_cv_%s_%s', $ms, $name);
    push @opt_data_names, $ms;
  }

  # Reliability comes after all optional columns.
  my @column_data_names_post = ();

  push @column_data_names_post, qw(
    reliability
  );
  push @column_names, qw(
    reliability
  );

  printTabbed('PSH', @column_names);

  for my $row (@$peptide_rows) {
    my @data = ();
    my %row_data = %$row;

    my $db = $row_data{'database'};
    my $db_entry = $$database_map[$db];
    $row_data{'database'} = $$db_entry{'full_name'};
    $row_data{'database_version'} = $$db_entry{'version'};
    $row_data{'search_engine'} = '[MS, MS:1001207, Mascot, ]';
    $row_data{'search_engine_score[1]'} = $row_data{'best_search_engine_score[1]'};

    # For PSMs, the modifications field must contain all fixed and variable
    # modifications.
    $row_data{'modifications'} = _format_mztab_peptide_modifications(@{ $row_data{'.fixedmods'} }, @{ $row_data{'.varmods'} });

    for (@column_data_names_pre) {
      push @data, $row_data{$_};
    }

    for (@opt_data_names) {
      push @data, $row_data{'opt'}{$_};
    }

    for (@column_data_names_post) {
      push @data, $$row{$_};
    }

    printTabbed('PSM', @data);
  }
}

sub printMTDsearchParameters {
  my ($software_id, $objParams) = @_;

  for my $i (1 .. $objParams->getNumberOfDatabases) {
    printMTD($software_id, 'DB', '=', $objParams->getDB($i));
  }

  # Parameters not in this list are omitted entirely.
  my @param_names = (
    ['ACCESSION', ''],
    ['CHARGE', ''],
    ['CLE', ''],
    ['DATAURL', ''],
    ['DECOY', -1],
    ['ERRORTOLERANT', ''],
    ['FILENAME', ''],
    ['FORMAT', ''],
    ['FORMVER', ''],
    ['ICAT', '', 0],
    ['INSTRUMENT', ''],
    ['IT_MODS', ''],
    ['MASS', ''],
    ['MODS', ''],
    ['MULTI_SITE_MODS', '', 0],
    ['OVERVIEW', '', 0],
    ['PEP_ISOTOPE_ERROR', -1],
    ['PFA', -1],
    ['PRECURSOR', -1],
    ['QUANTITATION', '', 'none'],
    ['QUE', ''],
    ['REPORT', 0],
    ['REPTYPE', ''],
    ['SEARCH', ''],
    ['SEG', -1],
    ['SUBCLUSTER', -1],
    ['TAXONOMY', ''],
    ['TOL', 0],
    ['TOLU', ''],
  );

  if (not $objResFile->isPMF) {
    push @param_names, (
        ['ITOL', ''],
        ['ITOLU', ''],
    );
  }

  PARAM: for (sort { $$a[0] cmp $$b[0] } @param_names) {
    my ($param_name, @skiplist) = @$_;
    my $f = 'get'.$param_name;
    my $val = $objParams->$f();
    next if not defined $val;
    for (@skiplist) {
      next PARAM if $val eq $_;
    }
    printMTD($software_id, $param_name, '=', $val);
  }

  for (0 .. 12) {
    my $val = $objParams->getUSERField($i);
    next if $val eq '';
    printMTD($software_id, sprintf('USER%02d', $_), '=', $val);
  }

  if (my $str = $objParams->getAllUSERParams()) {
    for (split /[\r\n]+/, $str) {
      printMTD($software_id, $_);
    }
  }

}

sub printMTDformatParameters {
  my ($settings_id, $thisScript, $urlParams) = @_;

  my $args = _cmdline_arguments($thisScript, $urlParams);
  delete $$args{'sessionid'};

  for my $arg (sort keys %$args) {
    my $value = $$args{$arg};
    
    if ($value =~ /\s/) {
      printMTD($settings_id, sprintf("%s='%s'", $arg, $value));
    } else {
      printMTD($settings_id, sprintf("%s=%s", $arg, $value));
    }
  }
}

sub _mk_protein_id { _mk_protein_id_str($_[0]->getDB(), $_[0]->getAccession()) }
sub _mk_protein_id_str { sprintf('%d::%s', $_[0], $_[1]) }

sub collectMzTabProteins {
  my (
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $objMS2Quant, $sigThreshold, $ShowSameSets, $ShowSubSets, $reportType,
    $protDesc, $protTaxStr, $protCover, $protEmPAI, $protpI, $protQuant,
    $query_mod_and_dupe_cache, $filter_duplicates, $sameset_ordering
  ) = @_;

  my $isPMF = $objResFile->isPMF;

  my @protein_rows = ();
  # Peptide map is a mapping from q,p to list of protein data (accession,
  # start, end). We use it later to decide which peptide matches to export
  # and under which accessions to report them.
  my %peptide_map = ();

  for my $i (1 .. $objSummary->getNumberOfHits()) {
    my $objProtein = $objSummary->getHit($i);
    next if $isPMF and $objProtein->isPMFMixture();

    my @sameset = common_subs::compileSameset(
      $objSummary, $objProtein, $ShowSameSets, $$sameset_ordering{$i}
    );

    my $data = collectMzTabProteinData(
      $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
      $objMS2Quant, $sigThreshold, $sameset[0], $reportType, $protDesc,
      $protTaxStr, $protCover, $protEmPAI, $protpI, $query_mod_and_dupe_cache,
      $filter_duplicates, \%peptide_map, 
    );

    $$data{'opt'}{'MS:1001301'} = $i; # rank

    if (not $isPMF) {
        $$data{'opt'}{'MS:1001592'} = 1; # member number
    }

    # Other proteins that match the same set of peptides
    $$data{'ambiguity_members'} = [ map {
      { accession => $_->getAccession(), db => $_->getDB() }
    } @sameset[1 .. $#sameset] ];

    push @protein_rows, $data;

    my %family_superset_map = ();
    $family_superset_map{ _mk_protein_id($sameset[0]) } = $data;

    # We can use the same overall loop over family members and samesets
    # in both kinds of grouping (family and conventional). In the latter
    # case, there are simply no family proteins.
      
    # Other family members
    my $k = 1;
    while (my $objProtein = $objSummary->getNextFamilyProtein($i, $k++)) {
      my @sameset = common_subs::compileSameset(
        $objSummary, $objProtein, $ShowSameSets, $$sameset_ordering{$i},
      );

      my $data = collectMzTabProteinData(
        $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs,
        $objSummary, $objMS2Quant, $sigThreshold, $sameset[0], $reportType,
        $protDesc, $protTaxStr, $protCover, $protEmPAI, $protpI, 
        $query_mod_and_dupe_cache, $filter_duplicates, \%peptide_map, 
      );

      $$data{'opt'}{'MS:1001301'} = $i; # rank

      if (not $isPMF) {
          $$data{'opt'}{'MS:1001592'} = $k; # member number
      }

      $family_superset_map{ _mk_protein_id($sameset[0]) } = $data;

    # And their same-sets
      $$data{'ambiguity_members'} = [ map {
        { accession => $_->getAccession(), db => $_->getDB() }
      } @sameset[1 .. $#sameset] ];

      push @protein_rows, $data;
    }

    # sub-sets and intersections for the entire family
    
    # mzTab 1.0 forbids exporting subset and intersection proteins, so we
    # never do so here. The loop could be of use in future.
    if (0 and $ShowSubSets > 0) {
      my $k = 1;
      while (my $objProtein = $objSummary->getNextSubsetProtein($i, $k++, 1)) {
        # We need the full sameset to filter out proteins from
        # getSimilarProteins().
        my @sameset = common_subs::compileSameset(
          $objSummary, $objProtein, 1, $$sameset_ordering{$i},
        );
        my %is_sameset = map { key_acc($_) => 1 } @sameset;

        my @supersets = _get_supersets($sameset[0], \%is_sameset);
        my @sameset_data = map {
          { accession => $_->getAccession(), db => $_->getDB() }
        } @sameset;

        for (@supersets) {
          my $superset = $family_superset_map{ _mk_protein_id_str(@$_{qw(db acc)}) };
          next if not $superset;

          # The intersection protein appears once under each parent (superset)
          # protein.
          push @{ $$superset{'ambiguity_members'} }, @sameset_data;
        }
      }
    }
  }

  return (\@protein_rows, \%peptide_map);
}

sub _determine_protein_hit_reliability {
  my ($objProtein) = @_;

  my $n = $objProtein->getNumDistinctPeptides(1);

  if ($n >= 2) {
    return 1; # high
  }

  if ($n == 1) {
      return 2; # medium
  }

  return 3; # low
}

sub _determine_peptide_match_reliability {
  my ($objSummary, $objPeptide, $sigThreshold) = @_;

  if ($objPeptide->getIsFromErrorTolerant()) {
      return 3; # no negotiation
  }

  my $score = $objPeptide->getIonsScore();
  my $thr = $objSummary->getPeptideThreshold($objPeptide->getQuery(), 1 / $sigThreshold);

  if ($score > $thr) {
    return 1; # high
  }

  return 3; # low
}

# Mandatory protein columns:
#   accession
#   description (*)
#   taxid (*)
#   species (*)
#   database
#   [database_version]
#   [search_engine]
#   best_search_engine_score[1]
#   search_engine_score[1]_ms_run[i]
#   reliability
#   num_psms_ms_run[k]
#   num_peptides_distinct_ms_run[i]
#   num_peptides_unique_ms_run[i]
#   [ambiguity_members]
#   modifications
#   (uri)
#   (go_terms)
#   protein_coverage (*)
#
# Terms in parentheses are always null. 'database_version' and
# 'search_engine' are set later, as is 'ambiguity_members'.
# Note that best_search_engine_score[] and search_engine_score[]-ms_run[i]
# always have index 1, because we only export Mascot data.
#
# For 'modifications', we need to collect the start and end positions of each
# peptide match and later, once we've looped over peptide matches, construct
# the protein modification list. Also, 'database' is an index, which is mapped
# to the database ID later.
#
# Terms marked with an asterisk (*) may be null, unless the user has opted
# to fetch missing data from ms-getseq.exe and ms-gettaxonomy.exe.
#
# If quantitation is in use, the following optional columns will be
# exported:
#   protein_abundance_assay[k]
#   protein_abundance_study_variable[j]
#   protein_abundance_stdev_study_variable[j]
#   protein_abundance_std_error_study_variable[j]
#
# k is an index over component names in order as given by $objMS2Quant,
# while j is an index over ratio names.
sub collectMzTabProteinData {
  my (
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $objMS2Quant, $sigThreshold, $objProtein, $reportType, $protDesc,
    $protTaxStr, $protCover, $protEmPAI, $protpI, $query_mod_and_dupe_cache,
    $filter_duplicates, $peptide_map, 
  ) = @_;

  my $objResFile = $$ms_run_resfiles[0]{'resfile'};
  my $objParams = $$ms_run_resfiles[0]{'params'};

  my %data = ();

  my $preferredAcc = $data{'accession'} = common_subs::preferredAccession($objProtein->getDB(), $objProtein->getAccession());
  my $dbIdx = $data{'database'} = $objProtein->getDB();

  if ($protDesc) {
    $data{'description'} = common_subs::mustGetProteinDescription($objSummary, $objParams, $shipper, $preferredAcc, $dbIdx);
  }

  if ($protTaxStr) {
    ($data{'taxid'}, $data{'species'}) = common_subs::getProteinTaxonomy($objSummary, $objParams, $preferredAcc, $dbIdx);
  }
  
  # "Best" means the combined file score if there are multiple results files.
  # NB: we should really have a protein-specific CV term here; MS:1001171 is
  # the score for PSMs in PSI-MS CV 3.59 (2014-01-22).
  $data{'best_search_engine_score[1]'} = sprintf('%.0f', $objProtein->getScore());

  # It is rather meaningless to report Mascot protein scores for each ms_run
  # individually.
  for my $i (1 .. $#$ms_run_resfiles) {
    my $summary = $$ms_run_resfiles[$i]{'summary'};
    my $protein = $summary->getProtein($objProtein->getAccession(), $objProtein->getDB());
    if ($protein) {
      $data{'search_engine_score[1]_ms_run['.$i.']'} = sprintf('%.0f', $protein->getScore());

      # "count of the total significant PSMs"
      $data{'num_psms_ms_run['.$i.']'} = $protein->getNumDisplayPeptides(1);

      # "count of the number of different peptide sequences ... above the
      # significance threshold"
      $data{'num_peptides_distinct_ms_run['.$i.']'} = $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE);

      # Uniqueness means:
      # "if ambiguity members have been reported, the count MUST be derived from
      # the number of peptides that can be uniquely mapped to the group of
      # accessions"
      # In terms of Mascot, this means the number of significant PSMs marked
      # unique to the protein hit, as this is the set of matches that defines
      # the sameset proteins (the ambiguity members).
      $data{'num_peptides_unique_ms_run['.$i.']'} = $objProtein->getNumDistinctPeptides(1, $msparser::ms_protein::DPF_SEQUENCE | $msparser::ms_protein::DPF_UNIQUE);
    } else {
      $data{'search_engine_score[1]_ms_run['.$i.']'} = 'null';
      $data{'num_psms_ms_run['.$i.']'} = 'null';
      $data{'num_peptides_distinct_ms_run['.$i.']'} = 'null';
      $data{'num_peptides_unique_ms_run['.$i.']'} = 'null';
    }
  }
  
  $data{'reliability'} = _determine_protein_hit_reliability(
    $objProtein, 
  );

  my @selected_qp = _select_nonduplicate_protein_qp(
    $query_mod_and_dupe_cache, $filter_duplicates, $objResFile, $objParams,
    $objSummary, $objProtein
  );

  my @quant_intensity = ();

  for (@selected_qp) {
    my ($pepnum, $query, $rank) = @$_;
    my $startpos  = $objProtein->getPeptideStart($pepnum);
    my $endpos    = $objProtein->getPeptideEnd($pepnum);
    my $pre       = $objProtein->getPeptideResidueBefore($pepnum);
    my $post      = $objProtein->getPeptideResidueAfter($pepnum);

    push @{ $$peptide_map{$query}{$rank} }, {
      accession => $preferredAcc,
      database => $objProtein->getDB(),
      start => $startpos,
      end => $endpos,
      pre => $pre,
      post => $post,
      pepnum => $pepnum,
    };

    if ($objMS2Quant) {
      my $arr = $objMS2Quant->getComponentIntensities(msparser::ms_peptide_quant_key->new($query, $rank));
      if ($arr and @$arr) {
        # We know that by definition, the component intensities of a PSM
        # can only come from one results file.
        my (undef, undef, $src_resfile_idx) = $objResFile->getSrcQueryAndFileIdForMultiFile($query);
        # We assume component intensities are in the same order as in the
        # quantitation method and that the assays are defined in the same
        # order in mzTabExperimentStructure().
        for my $i (1 .. @$arr) {
          my $assay_idx = $$comp_idx_to_assay[$src_resfile_idx][$i];
          $quant_intensity[$assay_idx] += $$arr[$i-1];
        }
      }
    }
  }

  if ($reportType eq 'concise') {
    my $expect = $objSummary->getProteinExpectationValue($objProtein->getScore());
    $data{'opt'}{'MS:1001172'} = sprintf("%.3g", $expect);
  }

  if ($protEmPAI or $protCover) {
    my $length;
    if (my $mass = $objSummary->getProteinMass($objProtein->getAccession(), $objProtein->getDB())) {
      # approximate but fast
      $length = $mass / 110;
    } else {
      $length = getProteinLen($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
      if ($length) {
      # accurate
      } else {
        my $protMass = mustGetProteinMass($objProtein->getAccession(), $objSummary, $objParams, $objProtein->getFrame(), $objProtein->getDB);
        if ($protMass) {
        # approximate
          $length = $protMass / 110;
        }
      }
    }
    
    if ($protCover) {
      if ($length) {
        $data{'protein_coverage'} = sprintf('%.3f', $objProtein->getCoverage() / $length);
      }
    }

    if ($protEmPAI) {
      my $value = $objSummary->getProteinEmPAI($objProtein->getAccession(), $objProtein->getDB(), $length ? int($length) : -1);
      if ($value >= 0) {
        $data{'opt'}{'MS:1001905'} = sprintf('%.2f', $value);
      }
    }
  }

  if ($protpI) {
      $data{'opt'}{'MS:1000862'} = getProteinpI($objParams, $objProtein->getAccession(), $objProtein->getFrame(), $objProtein->getDB);
  }

  if ($objMS2Quant) {
    for my $assay_idx (1 .. @quant_intensity) {
      # (@quant_intensity is 1-based, so no index error here.)
      if (defined $quant_intensity[$assay_idx]) {
        $data{'protein_abundance_assay['.$assay_idx.']'} = sprintf("%g", $quant_intensity[$assay_idx]);
      } else {
        $data{'protein_abundance_assay['.$assay_idx.']'} = 'null';
      }
    }

    my $ratio_type = lc($objMS2Quant->getQuantitationMethod()->getProteinRatioType());

    for my $sv (@$study_variable_defs) {
      my $ratio = $objMS2Quant->getProteinRatio($objProtein->getAccession(), $objProtein->getDB(), $$sv{'ratio_name'});
      my $idx = $$sv{'idx'};

      if ($ratio->isMissing()) {
        $data{'protein_abundance_study_variable['.$idx.']'} = 'null';
        $data{'protein_abundance_stdev_study_variable['.$idx.']'} = 'null';
        $data{'protein_abundance_std_error_study_variable['.$idx.']'} = 'null';

        $data{'opt'}{'study_variable['.$idx.']_sample_size'} = 'null';
        $data{'opt'}{'study_variable['.$idx.']_is_normal'} = 'null';
        $data{'opt'}{'study_variable['.$idx.']_is_significant'} = 'null';
      } else {
        $data{'protein_abundance_study_variable['.$idx.']'} = sprintf('%.3f', $ratio->getValue());
        $data{'protein_abundance_stdev_study_variable['.$idx.']'} = 'null';
        $data{'protein_abundance_std_error_study_variable['.$idx.']'} = 'null';
        $data{'opt'}{'study_variable['.$idx.']_sample_size'} = $ratio->getSampleSize();

        my $stdev = $ratio->getStandardDeviation();
        my $stderr = $ratio->getStandardError();

        if ($stdev > 0) {
          $data{'protein_abundance_stdev_study_variable['.$idx.']'} = sprintf('%.3f', $stdev);
        } else {
          # Ignore reason why unavailable.
        }

        if ($stderr > 0) {
          $data{'protein_abundance_std_error_study_variable['.$idx.']'} = sprintf('%.3f', $stderr);
        } else {
          # Ignore reason why unavailable.
        }

        if ($ratio->getNormalityPvalue() > 0) {
          $data{'opt'}{'study_variable['.$idx.']_is_normal'} = $ratio->isLogNormal() ? 1 : 0;
          if ($ratio->getHypothesisPvalue() > 0) {
            $data{'opt'}{'study_variable['.$idx.']_is_significant'} = $ratio->isLogNormal() && $ratio->isSignificant() ? 1 : 0;
          }
        }
      }
    }
  }
  
  return \%data;
}


# Mandatory Peptide columns:
#   sequence
#   accession
#   unique
#   database
#   [database_version]
#   [search_engine]
#   best_search_engine_score[1]
#   modifications
#   retention_time
#   retention_time_window
#   charge
#   mass_to_charge
#   search_engine_score[1]_ms_run[k]
#   spectra_ref
#
# Mandatory PSM columns (not already listed above):
#   PSM_ID
#   search_engine_score[1]
#   exp_mass_to_charge
#   calc_mass_to_charge
#   pre
#   post
#   start
#   end
#
# Optional in both:
#   reliability
#   (uri)
#
# Columns in parentheses ('uri') are null. 'database_version'
# and 'search_engine' are filled in later. The mzTab 1.0 specification requires
# that each peptide match in each protein hit be reported, even when the
# peptide is shared between hits, so there may be several nearly identical
# output rows which only differ by the 'accession' column.
#
# To simplify matters, we assume Peptide = PSM. In Quantification mode,
# we output both the Peptides and PSM table; in Identification mode, only
# the PSM table. Both tables use the same source data collected here.
#
# If quantitation is in use, the following optional columns will be
# exported in the Peptides table:
#   peptide_abundance_study_variable[k]
#   (peptide_abundance_stdev_study_variable[k])
#   (peptide_abundance_std_error_study_variable[k])
#   peptide_abundance_assay[j]
#
# k is an index over ratio names in order as given by $objMS2Quant, while
# j is an index over assays (quantitation components). 
# The standard deviation and standard error columns are always null, since we
# don't have those for peptide ratios.
sub collectMzTabPeptides {
  my (
    $ms_run_resfiles, $comp_idx_to_assay, $study_variable_defs, $objSummary,
    $msresFlags2, $objMS2Quant, $sigThreshold, $objUmodConfigFile2, $sessionID,
    $indexIsQueryNumber, $is_decoy, $is_error_tolerant, $pepDelta,
    $pepScore, $pepHomThr, $pepIdThr, $pepFrame, $pepNumMatch, $pepScanTitle,
    $pepQuant, $massDP, $vmMass, $vmString, $percolate, $fixedMods,
    $query_mod_and_dupe_cache, $peptide_map
  ) = @_;

  my $isPMF = $objResFile->isPMF;

  my ($modsByDelta, $full_unimod_file, $query_is_sig);
  
  if (not $isPMF) {
    # Site analysis initialisation.
    $modsByDelta = Util::SiteAnalysis::modifications_by_delta(
      $objUmodConfigFile2, $objParams,
    );
    # $objUmodConfigFile2 could be the reduced file embedded in the results
    # file.
    my $cs = new msparser::ms_connection_settings();
    $cs->setSessionID($sessionID);
    $full_unimod_file = new msparser::ms_umod_configfile(
      "../config/unimod.xml", "../html/xmlns/schema/unimod_2/unimod_2.xsd",
      $cs
    );
  
    $query_is_sig = $msresFlags2 & $msparser::ms_peptidesummary::MSPEPSUM_USE_HOMOLOGY_THRESH
      ? sub { Util::Peptides::above_homology($objSummary, $objSummary->getPeptide($_[0], 1)) }
      : sub { Util::Peptides::above_identity($objSummary, $objSummary->getPeptide($_[0], 1)) }
      ;
  }

  my $export_quant = 0;

  if ($pepQuant and $objMS2Quant) {
    $export_quant = 1;
  }

  # Now collect all peptide matches
  my @peptide_rows = ();
  my %protein_mods = ();

  my %mod_name_lookup = (); # cache for Unimod objects
  my %mod_tally = (); # set of all Unimod and Chemmod identifiers found
  my $has_retention_time = 0;
  my $max_rank_value = $objSummary->getMaxRankValue();

  for my $query (1 .. $objResFile->getNumQueries()) {
    next unless (exists($$peptide_map{$query}) or $isPMF);

    my %query_data = ();
    my %query_data_opt = ();

    $query_data_opt{'MS:1001528'} = $query;
    $query_data{'exp_mass_to_charge'} = sprintf('%.' . $massDP . 'f', $objResFile->getObservedMass($query));

    my $objQuery = new msparser::ms_inputquery($objResFile, $query);
    
    if ($objQuery->getRetentionTimes) {
      $query_data{'retention_time'} = $objQuery->getRetentionTimes();
      ++$has_retention_time;
    }
    
    if ($objQuery->getIT_MODS(1)) {
        #print "            <userParam name=\"IT_MODS\" value=\"" . noXmlTag($objQuery->getIT_MODS(1)) . "\" />\n";
    }
    if ($objQuery->getINSTRUMENT(1)) {
        #print "            <userParam name=\"INSTRUMENT\" value=\"" . noXmlTag($objQuery->getINSTRUMENT(1)) . "\" />\n";
    }
    if ($objQuery->getRULES()) {
        #print "            <userParam name=\"RULES\" value=\"" . $objQuery->getRULES() . "\" />\n";
    }

    if ($pepScanTitle) {
        $query_data_opt{'MS:1000796'} = $objQuery->getStringTitle(1);
    }

    my $siteAnalysisData;
    # NB: Site analysis is not included. The mzTab 1.0 specification provides
    # a mechanism for reporting variable modification site probability, but
    # only one site at a time. Our probabilities are for the whole set of
    # observed modifications.
=pod
    if (not $isPMF) {
      if (&$query_is_sig($query)) {
        $siteAnalysisData = analyse_sites_in_query(
          $objSummary, $objParams, $queryNum, $modsByDelta, $full_unimod_file,
          $objSummary->getMaxRankValue(),
        );
      }
    }
=cut

    for my $rank (1 .. $max_rank_value) {
      my $objPeptide = $objSummary->getPeptide($query, $rank);
      my $pepstr     = $objPeptide->getPeptideStr;
      next if $pepstr eq '';
      next if not exists $$peptide_map{$query}{$rank}; # unassigned

      my $protein_map = $$peptide_map{$query}{$rank};

      my %base_data = %query_data;
      my %base_opt = %query_data_opt;

      $base_data{'PSM_ID'} = 100 * $query + $rank;
      $base_data{'sequence'} = $pepstr;
      $base_data{'unique'} = $objSummary->isPeptideUnique($query, $rank) ? 1 : 0;
  
      for my $i (1 .. $#$ms_run_resfiles) {
          $base_data{'search_engine_score[1]_ms_run['.$i.']'} = 'null';
      }

      my (undef, $src_query_idx, $src_resfile_idx) = $$ms_run_resfiles[0]{'resfile'}->getSrcQueryAndFileIdForMultiFile($query);

      my $scanID = _determine_scan_id($indexIsQueryNumber, $$ms_run_resfiles[0]{'params'}, $objQuery, $src_query_idx);
      $base_data{'spectra_ref'} = sprintf('ms_run[%d]:%s', $src_resfile_idx, $scanID);

      if ($isPMF) {
          $base_data{'best_search_engine_score[1]'} = 'null';
          $base_data{'reliability'} = 'null';
      } else {
          # NB: We export expect value as the search engine score, because the
          # plain score is misleading without the associated significance 
          # thresholds. 

          # Each PSM comes from one file only, and the presence of other files
          # in a combined results file does not affect PSM score. So we only
          # need to report one search_engine_score[].

          if ($is_error_tolerant and $objPeptide->getIsFromErrorTolerant()) {
              # For ET matches, we export null as the search engine score, as
              # the thresholds are not reliable. We do export the actual
              # Mascot score; see below.
          } else {
            $base_data{'best_search_engine_score[1]'} 
              = $base_data{'search_engine_score[1]_ms_run['.$src_resfile_idx.']'} 
              = sprintf("%.2g", $objSummary->getPeptideExpectationValue($objPeptide->getIonsScore(), $query));
          }

          $base_data{'reliability'} = _determine_peptide_match_reliability($objSummary, $objPeptide, $sigThreshold);
      }

      if (not $isPMF) {
          if ($pepScore) {
              $base_opt{'MS:1001171'} = sprintf("%.2g", $objPeptide->getIonsScore());
          }

          if ($is_error_tolerant and $objPeptide->getIsFromErrorTolerant()) {
            # No thresholds.
          } else {
            if ($pepHomThr) {
              $base_opt{'MS:1001370'} = sprintf("%d", $objSummary->getHomologyThreshold($query, 1 / $sigThreshold));
            }
            if ($pepIdThr) {
              $base_opt{'MS:1001371'} = sprintf("%d", $objSummary->getPeptideIdentityThreshold($query, 1 / $sigThreshold));
            }
          }
      }

      $base_data{'calc_mass_to_charge'} = sprintf("%." . $massDP . "f", $objPeptide->getMrCalc());

      if ($pepDelta) {
          $base_opt{'MS:1001976'} = sprintf("%." . $massDP . "f", $objPeptide->getDelta()); # in Daltons
      }

      if ($pepNumMatch) {
          $base_opt{'MS:1001121'} = $objPeptide->getNumIonsMatched();
      }

      my ($fixed_mods, $variable_mods) = _collect_mzTab_modification_data(
        $$ms_run_resfiles[0], $objSummary, $objUmodConfigFile2, $objPeptide,
        exists($$siteAnalysisData{$rank}) ? $$siteAnalysisData{$rank} : undef,
        $vmMass, $vmString, $fixedMods, $query_mod_and_dupe_cache,
        \%mod_name_lookup, \%mod_tally,
      );
  
      # Note: .varmods and .fixedmods are not the final column names. We keep
      # the two types separate here, because the same source data is used for
      # both PSM and peptide rows.
      $base_data{'.varmods'} = $variable_mods;
      $base_data{'.fixedmods'} = $fixed_mods;

      my $pepCharge = $objPeptide->getCharge();
      $pepCharge = 1 if $pepCharge == 0;
  
      $base_data{'charge'} = $pepCharge;
  
      if ($export_quant) {
        my $key = msparser::ms_peptide_quant_key->new($query, $rank);
        for my $sv (@$study_variable_defs) {
          my $ratio = $objMS2Quant->getPeptideRatio($key, $$sv{'ratio_name'});
          my $idx = $$sv{'idx'};

          $base_data{'peptide_abundance_stdev_study_variable['.$idx.']'} = 'null';
          $base_data{'peptide_abundance_std_error_study_variable['.$idx.']'} = 'null';

          if ($ratio->isMissing()) {
            $base_data{'peptide_abundance_study_variable['.$idx.']'} = 'null';
          } else {
            $base_data{'peptide_abundance_study_variable['.$idx.']'} = sprintf('%.3f', $ratio->getValue());
          }
        }
        my $intensities = $objMS2Quant->getComponentIntensities($key);
        # See protein abundance (intensity) reporting in
        # collectMzTabProteinData().
        if ($intensities and @$intensities) {
          for my $i (1 .. @$intensities) {
            my $assay_idx = $$comp_idx_to_assay[$src_resfile_idx][$i];
            if (defined $$intensities[$i-1]) {
              $base_data{'peptide_abundance_assay['.$assay_idx.']'} = sprintf('%.3f', $$intensities[$i-1]);
            } else {
              # No need to set; will be reported 'null'.
            }
          }
        }
      }
  
      if ($percolate == 1) {
        my ($pep, $q) = $objPeptide->getPercolatorScores();
        $base_opt{'MS:1001491'} = $q;
        $base_opt{'MS:1001493'} = $pep;
      }

      if ($is_decoy) {
        $base_opt{'MS:1002217'} = 1;
      } elsif ($is_error_tolerant) {
        $base_opt{'MS:1002341'} = $objPeptide->getIsFromErrorTolerant() ? 1 : 0;
      }

      # And now we repeat once per accession:
  
      for my $protein_context (@$protein_map) {
        my %data = %base_data;
        $data{'opt'} = { %base_opt };

        $data{'accession'} = my $acc = $$protein_context{'accession'};
        $data{'database'} = my $db = $$protein_context{'database'};

        # For the PSM section:
        $data{'start'} = $$protein_context{'start'};
        $data{'end'} = $$protein_context{'end'};
        $data{'pre'} = $$protein_context{'pre'};
        $data{'post'} = $$protein_context{'post'};

        if ($pepFrame) {
          my $objProtein = $objSummary->getProtein($acc, $db);
          my $frame = $objProtein->getPeptideFrame($$protein_context{'pepnum'});
          $data{'opt'}{'MS:1001024'} = $frame > 0 ? $frame : 'null';
        }

        push @peptide_rows, \%data;

        # The mzTab 1.0 specification requires that all modifications
        # (fixed, variable, quantitation) are reported for PSMs, but only
        # variable modifications for proteins and peptides. So, we
        # completely skip processing @$fixed_mods here.

        my $start_idx = $$protein_context{'start'};

        for my $mod (@$variable_mods) {
          next if $$mod{'idx'} == 0 or $$mod{'idx'} == 2 + length($pepstr);
          my $idx = $$mod{'idx'} + $start_idx;
          push @{ $protein_mods{$db}{$acc}{$idx} }, $mod;
        }
      }
    }
  }

  return (\@peptide_rows, \%protein_mods, \%mod_tally, $has_retention_time);
}

sub _collect_mzTab_modification_data {
  my ($ms_run_resfile, $objSummary, $objUmodConfigFile2, $objPeptide,
      $siteAnalysisData, $vmMass, $vmString, $fixedMods, 
      $query_mod_and_dupe_cache, $mod_name_lookup, $mod_tally) = @_;

  my @fixed = ();

  my @pep_str = split //, $objPeptide->getPeptideStr();

  if ($$fixedMods{'n_term'}) {
    push @fixed, {
      idx => 0,
      name_str => $$fixedMods{'n_term'}{'name'},
      delta => $$fixedMods{'n_term'}{'delta'},
    };
  }

  for my $i (1 .. @pep_str) {
    my $residue = $pep_str[$i-1];
    if ($$fixedMods{lc $residue}) {
      push @fixed, {
        idx => $i,
        name_str => $$fixedMods{lc $residue}{'name'},
        delta => $$fixedMods{lc $residue}{'delta'},
      };
    }
  }

  if ($$fixedMods{'c_term'}) {
    push @fixed, {
      idx => scalar(@pep_str),
      name_str => $$fixedMods{'c_term'}{'name'},
      delta => $$fixedMods{'c_term'}{'delta'},
    };
  }

  my @variable = ();

  my @var_mods = split //, $objPeptide->getVarModsStr();
  my @nl = split //, $objPeptide->getPrimaryNlStr();
  my @summed_mods = split //, $objPeptide->getSummedModsStr();
  my $len = 2 + @pep_str;
  my $query = $objPeptide->getQuery();
  my $rank = $objPeptide->getRank();

  my %varmod_index_map = do {
      my @indices = (0..9, 'A'..'W');
      map { $indices[$_] => $_ } 0 .. $#indices
  };

  my %modfile_lookup = ();

  for my $i (0 .. $len) {
    my @mod_ids = ();
    my @neutral_loss = ();

    if (defined $var_mods[$i] and $var_mods[$i] ne '0') {
      my $id = $var_mods[$i];
      push @mod_ids, $id;

      if (defined $nl[$i] and $nl[$i] ne '0') {
        my $nl_id = $varmod_index_map{$nl[$i]};
        my @losses = @{ $$ms_run_resfile{'params'}->getVarModsNeutralLosses($id) };
        if ($nl_id-1 < @losses) {
          my $m = $losses[$nl_id-1];
          if ($m != 0) {
            push @neutral_loss, sprintf('[MS, MS:1001524, fragment neutral loss, %+f]', $m);
          }
        }
      }
    }

    if (defined $summed_mods[$i] and $summed_mods[$i] ne '0') {
      push @mod_ids, $summed_mods[$i];
    }

    for my $j (0 .. $#mod_ids) {
      my $mod_id = $mod_ids[$j];
      my ($mod_name_str, $mod_delta);

      if ($mod_id eq 'X') {
        my $srcSection = $objSummary->getSrcSection($query, $rank);
        my $srcRank = $objSummary->getSrcRank($query, $rank);
        my $et_mods_str = $$ms_run_resfile{'resfile'}->getSectionValueStr($srcSection, "q$query\_p$srcRank\_et_mods");
        if ($et_mods_str) {
          ($mod_delta, undef, $mod_name_str) = split(/,/, $et_mods_str, 3);

          if ($mod_name_str =~ /^NA_/) {
            # We can't export these in mzTab 1.0. The peptide match will be
            # marked error tolerant in an optional column, and that's the best
            # we can do.
            undef $mod_name_str;
          }
        }
      } else {
        $mod_delta = $$vmMass{$mod_id};
        $mod_name_str = $$vmString{$mod_id};
      }

      next if (not $mod_name_str and not $mod_delta);

      push @variable, { 
        idx => $i,
        name_str => $mod_name_str, 
        delta => $mod_delta,
        neutral_loss => $neutral_loss[$j],
      };
    }
  }

  if (not exists $$mod_tally{'fixed'}) {
    $$mod_tally{'fixed'} = {};
  }

  if (not exists $$mod_tally{'variable'}) {
    $$mod_tally{'variable'} = {};
  }

  for ([$$mod_tally{'fixed'}, \@fixed], [$$mod_tally{'variable'}, \@variable]) {
    my ($tally, $mods) = @$_;

    for (@$mods) {
      my $mod_name_str = $$_{'name_str'};
      my $mod_delta = $$_{'delta'};

      my $mod_name;

      if ($mod_name_str) {
        ($mod_name) = $mod_name_str =~ /(.*)\s+\((.*)\)/;
      }
  
      my ($final_id, $tally_id, $position, $site);
  
      if ($mod_name) {
        if (not exists $$mod_name_lookup{$mod_name}) {
          $$mod_name_lookup{$mod_name} = $objUmodConfigFile2->getModificationByName($mod_name);
            # could be undef if the modification doesn't exist in the file
        }
    
        if (defined(my $mod = $$mod_name_lookup{$mod_name})) {
          if ($mod->getRecordID() eq '') {
            # Unimod section from Mascot 2.2 or earlier; use the CHEMMOD
            # syntax.
          } else {
            $final_id = 'UNIMOD:' . $mod->getRecordID();
            $tally_id = sprintf('[UNIMOD, %s, %s,]', $final_id, $mod_name);
          }
  
          ($position, $site) = _unimod_position_and_site($mod, $mod_name_str);
        }
      }
  
      if (not defined $final_id) {
        if ($mod_delta) {
          $final_id = 'CHEMMOD:' . sprintf('%+f', $mod_delta);
          $tally_id = sprintf('[CHEMMOD, %s, %s,]', $final_id, $mod_name ? $mod_name : $mod_name_str);
        } else {
          $_ = undef;
          next;
        }
      }
  
      $_ = {
        idx => $$_{'idx'},
        neutral_loss => $$_{'neutral_loss'},
        mod_id => $final_id,
      };

      if (not exists $$tally{$tally_id}) {
        $$tally{$tally_id} = {
          position => $position,
          site => $site,
        };
      }
    }
  }

  my $local_mods = lookup_query_level_mods($query_mod_and_dupe_cache, $$ms_run_resfile{'resfile'}, $objPeptide, $objUmodConfigFile2);
  my @local_nl = split //, $objPeptide->getLocalModsNlStr();

  if ($local_mods) {
    my $tally = $$mod_tally{'variable'};

    for my $i (keys %$local_mods) {
      my $mod_name_str = $$local_mods{$i}{'name'};
      my ($mod_name) = $mod_name_str =~ /(.*)\s+\((.*)\)/;

      if (not exists $$mod_name_lookup{$mod_name}) {
        $$mod_name_lookup{$mod_name} = $objUmodConfigFile2->getModificationByName($mod_name);
      }

      if (not $$mod_name_lookup{$mod_name}) {
        next;
      }

      my $unimod_obj = $$mod_name_lookup{$mod_name};

      my $unimod_id = $unimod_obj->getRecordID();
      my ($final_id, $tally_id);

      if ($unimod_id eq '') {
        # Mascot 2.2 or earlier.
        $final_id = 'CHEMMOD:' . sprintf('%+f', $$local_mods{$i}{'delta'});
        $tally_id = sprintf('[CHEMMOD, %s, %s,]', $final_id, $mod_name);
      } else {
        $final_id = 'UNIMOD:' . $unimod_id;
        $tally_id = sprintf('[UNIMOD, %s, %s,]', $final_id, $mod_name);
      }

      my $nl_value = undef;

      if (defined($local_nl[$i]) and $local_nl[$i] ne 0) {
        if (not exists $modfile_lookup{$mod_name_str}) {
          my $modfile_mod = msparser::ms_modification->new;
          if ($modfile_mod->getFromUnimod($mod_name_str, $objUmodConfigFile2)) {
            $modfile_lookup{$mod_name_str} = $modfile_mod;
          }
        }

        if ($modfile_lookup{$mod_name_str}) {
          my $nl_id = $varmod_index_map{$local_nl[$i]};
          my $nl_arr = $modfile_lookup{$mod_name_str}->getNeutralLoss($objParams->getMassType());
          if ($nl_arr and $nl_id-1 < @$nl_arr and $$nl_arr[$nl_id-1] > 0) {
            $nl_value = sprintf('[MS, MS:1001524, fragment neutral loss, %+f]', $$nl_arr[$nl_id-1]);
          }
        }
      }

      push @variable, {
        idx => $i,
        neutral_loss => $nl_value,
        mod_id => $final_id,
      };

      if (not exists $$tally{$tally_id}) {
        my ($position, $site) = _unimod_position_and_site($unimod_obj, $mod_name_str);

        $$tally{$tally_id} = {
          position => $position,
          site => $site,
        };
      }
    }
  }

  my @fixed_final = grep { defined($_) } @fixed;
  my @variable_final = sort { $$a{'idx'} <=> $$b{'idx'} } grep { defined($_) } @variable;
  
  return (\@fixed_final, \@variable_final);
}

sub _unimod_position_and_site {
  my ($unimod_obj, $mod_name_str) = @_;

  my $spec_id = $unimod_obj->findSpecGroup($mod_name_str);
  my ($position, @site);

  if ($spec_id != -1) {
    for (0 .. $unimod_obj->getNumberOfSpecificities()-1) {
      my $spec = $unimod_obj->getSpecificity($_);
      if ($spec->getSpecGroup() == $spec_id) {
        if ($spec->havePosition()) {
          # We assume the specification group has the same position for all
          # sites.
          $position = $spec->getPosition();
        }
        if ($spec->haveSite()) {
          push @site, $spec->getSite();
        }
      }
    }
  }

  return ($position, join('', @site));
}


###############################################################################
# &getTaxonomyNameFromId()
# return taxonomy name as string
# $_[0] taxonomy ID
# $_[1] ref to %taxNames (cache)
###############################################################################

sub getTaxonomyNameFromId {

  my ($id, $taxNames_ref) = @_;
  
# First time through, populate hash with some common tax IDs to reduce number
# of times we have to open names.dmp. 
# Also, means we get some information even if names.dmp missing
  unless (scalar(%{ $taxNames_ref })) {
    ${ $taxNames_ref }{'2'} = "Bacteria";
    ${ $taxNames_ref }{'197'} = "Campylobacter jejuni";
    ${ $taxNames_ref }{'358'} = "Agrobacterium tumefaciens";
    ${ $taxNames_ref }{'487'} = "Neisseria meningitidis";
    ${ $taxNames_ref }{'562'} = "Escherichia coli";
    ${ $taxNames_ref }{'590'} = "Salmonella";
    ${ $taxNames_ref }{'1224'} = "Proteobacteria";
    ${ $taxNames_ref }{'1239'} = "Firmicutes";
    ${ $taxNames_ref }{'1313'} = "Streptococcus Pneumoniae";
    ${ $taxNames_ref }{'1423'} = "Bacillus subtilis";
    ${ $taxNames_ref }{'1769'} = "Actinobacteria";
    ${ $taxNames_ref }{'1902'} = "Streptomyces coelicolor";
    ${ $taxNames_ref }{'2093'} = "Mycoplasma";
    ${ $taxNames_ref }{'2157'} = "Archaeobacteria";
    ${ $taxNames_ref }{'2157'} = "Archaeobacteria";
    ${ $taxNames_ref }{'2759'} = "Eukaryota";
    ${ $taxNames_ref }{'3360'} = "Alveolata";
    ${ $taxNames_ref }{'3701'} = "Arabidopsis";
    ${ $taxNames_ref }{'3702'} = "Arabidopsis thaliana";
    ${ $taxNames_ref }{'4530'} = "Oryza sativa";
    ${ $taxNames_ref }{'4754'} = "Pneumocystis carinii";
    ${ $taxNames_ref }{'4896'} = "Schizosaccharomyces pombe";
    ${ $taxNames_ref }{'4932'} = "Saccharomyces Cerevisiae";
    ${ $taxNames_ref }{'5833'} = "Plasmodium falciparum";
    ${ $taxNames_ref }{'6239'} = "Caenorhabditis elegans";
    ${ $taxNames_ref }{'7215'} = "Drosophila";
    ${ $taxNames_ref }{'7711'} = "Chordata";
    ${ $taxNames_ref }{'7898'} = "Actinopterygii";
    ${ $taxNames_ref }{'7955'} = "Danio rerio";
    ${ $taxNames_ref }{'8287'} = "lobe-finned fish and tetrapod clade";
    ${ $taxNames_ref }{'8355'} = "Xenopus laevis";
    ${ $taxNames_ref }{'9443'} = "Primates";
    ${ $taxNames_ref }{'9606'} = "Homo sapiens";
    ${ $taxNames_ref }{'9989'} = "Rodentia";
    ${ $taxNames_ref }{'10088'} = "Mus";
    ${ $taxNames_ref }{'10090'} = "Mus musculus";
    ${ $taxNames_ref }{'10114'} = "Rattus";
    ${ $taxNames_ref }{'10239'} = "Viruses";
    ${ $taxNames_ref }{'11103'} = "Hepatitis C virus";
    ${ $taxNames_ref }{'12908'} = "unclassified";
    ${ $taxNames_ref }{'31033'} = "Takifugu rubripes";
    ${ $taxNames_ref }{'33090'} = "Viridiplantae";
    ${ $taxNames_ref }{'33208'} = "Metazoa";
    ${ $taxNames_ref }{'40674'} = "Mammalia";
    ${ $taxNames_ref }{'45251'} = "Arabidopsis neglecta";
    ${ $taxNames_ref }{'77643'} = "Mycobacterium tuberculosis complex";
    ${ $taxNames_ref }{'117571'} = "bony vertebrates";
  }

# if name not in cache, get from names.dmp and add to cache
  unless (${ $taxNames_ref }{$id}) {
    if (open(TAXFILE, "../taxonomy/names.dmp")) {
      my @matches = grep /^$id\s+/, <TAXFILE>;
      close TAXFILE;
      foreach my $match (@matches) {
        my @fields = split(/\|/, $match);
        if ($fields[3] =~ /\s+scientific name\s+/i) {
          $fields[1] =~ s/^\s+//;
          $fields[1] =~ s/\s+$//;
          ${ $taxNames_ref }{$id} = $fields[1];
          last;
        }
      }
    }
  # if this fails, set cache entry to -1 to prevent further attempts
    unless (${ $taxNames_ref }{$id}) {
      ${ $taxNames_ref }{$id} = "-1";
    }
  }

  if (${ $taxNames_ref }{$id} eq "-1") {
    return "";
  } else {
    return ${ $taxNames_ref }{$id};
  }
      
}

###############################################################################
# &set_option()
# set a non-optional option
# $_[0] parameter name
# $_[1] optional parameter value; 1 by default
###############################################################################

sub set_option {
  
  my $param = shift;
  my $value = shift;
  $value = 1 unless defined $value;
  $urlParams{$param} = $param;
  $thisScript->param(-name=>$param, -value=>$value);
  
}

###############################################################################
# &_percolate()
# run Percolator, if required
###############################################################################
sub _percolate {
 my ($resfile, $use_proteins, $use_rt, $MIME_boundary) = @_;

 my $bin_createpip = join('/', MascotDat->bindir, MascotDat->bin_createpip);
 my $bin_percolator = join('/', MascotDat->bindir, MascotDat->bin_percolator);

 if (not -x $bin_createpip or not -x $bin_percolator) {
   common_subs::fatal("Unable to find Percolator executables", __LINE__, $myName, $idForFatalError, $MIME_boundary);
 }

 my %filenames = MascotResults::percolator_filenames($resfile);

 my $keepalive_pip = undef;
 my $keepalive_perc = undef;

 if (defined $MIME_boundary) {
   $keepalive_pip = "<!-- Creating Percolator input file... (%d s elapsed) -->\n";
   $keepalive_perc = "<!-- Percolating... (%d s elapsed) -->\n";
 }

 run_ms_createpip(
    undef,
    $bin_createpip, 
    $resfile->getFileName,
    $filenames{'pipname'}, 
    $keepalive_pip,
    $use_rt
 );

 run_percolator(
    undef,
    $bin_percolator, 
    MascotResults::percolator_flags($resfile, $use_rt), 
    $keepalive_perc,
 );
}

###############################################################################
# &maybeRedirectToCaching($sub, %extra_params)
# Redirect to nph-cache_families.pl if either the resfile or peptidesummary
# objects need to be cached, or if anything needs percolation. No redirections
# are done if we're running outside CGI.
# $_[0] the subresource to use (sub:resultsfile or sub:peptidesummary)
# $_[1..n] additional parameters overwriting CGI parameters
###############################################################################
sub maybeRedirectToCaching {
  return unless $isCGI;
  my ($sub, %extra_params) = @_;
  my $uri = URI->new_abs(MascotDat->uri_cache_families, Request->cgidir.'/');
  if (%extra_params) {
    # We are free to modify the global CGI state here, because we know we
    # will exit afterwards.
    for (keys %extra_params) {
      $thisScript->param($_, $extra_params{$_});
    }
  }
  # The query string is non-empty, because at least the 'file' parameter
  # has to exist to get this far.
  my $separator = $thisScript->query_string =~ m#\&# ? '&' : ';';
  $uri->query($thisScript->query_string.$separator.'sub='.common_subs::urlEscape($sub).$separator.'c.from='.common_subs::urlEscape($myName));
  print $thisScript->redirect(
    -uri => $uri->as_string,
    -status => '303 See Other',
  );
  exit;
}

sub analyse_sites_in_query {
  my ($cache, $objResFile, $objSummary, $objParams, $queryNum, $modsByDelta,
      $full_unimod_file, $max_p) = @_;

  my @site_analysis_peptides = ();

  for my $rank (1 .. $max_p) {
    my $objPeptide = $objSummary->getPeptide($queryNum, $rank);
    next unless $objPeptide;
    next unless $objPeptide->getAnyMatch();

    push @site_analysis_peptides, $objPeptide;

    if ($objPeptide->getIsFromErrorTolerant()) {
      Util::SiteAnalysis::include_error_tolerant_delta(
        $modsByDelta, $full_unimod_file, $objParams,
        $objSummary->getErrTolModName($queryNum, $rank)
      );
    } elsif ($objPeptide->getLocalModsStr() ne '') {
      Util::SiteAnalysis::include_query_level_delta(
        $modsByDelta, $full_unimod_file, $cache, $objResFile, $objPeptide
      );
    }
  }

  my $siteAnalysisData;
  
  if (@site_analysis_peptides) {
    $siteAnalysisData = Util::SiteAnalysis::analyse_query(
      $cache, $objResFile, $objSummary, $objParams, $modsByDelta,
      \@site_analysis_peptides
    );
  }

  return $siteAnalysisData;
}

